#include "site-stuff.h"

#include <stdio.h>
#include <A++.h>
#include "MHD.h"
#include "ION.h"
#include "ion_variables.h"
#ifdef LC_UNSC
#define ION_POKE ion_poke_
#elif LC_NOUNSC
#define ION_POKE ion_poke
#endif

/// Defeind in ion_fortran_arrays.F
extern "C" void ION_POKE(FLoaT **, int *);
/// Defeind in test-arrays.F
extern "C" void TEST_ARRAYS();

/*
 * Original documentation for this file:
 *
 *  We're only using one processor -- at least initially -- for the
 *  ionosphere. To transfer data, the guard cell part of the
 *  partitioning needs to be consistent. Therefore, we need to slop
 *  back and forth between guard and no guard for the ionospheric
 *  stuff.
 *
 *  We assume that all ion functions operate on no guard
 *  configurations; for communication with the mhd, we need to convert
 *  to MHD partitions and then back when done.
 *
 */

/// Convert back to Ionosphere partition
/**
 *  The ionosphere is currently solved on one processor.  When
 *  transfering data to all other nodes, we need the guard cell part
 *  of the partitioning to be consistent.  So we need to convert
 *  between guard and no guard.
 *
 *  This function converts the electric field E[i,j] and velocity
 *  V[x,y,z] from MHD partitioning to ION partitioning (after
 *  communications with MHD)
 *
 *  @param mhd - MHD info
 *  @param ion - Ionosphere information
 *
 * @see MHD
 * @see ION
 * @see ion2mhd
 */
void mhd2ion( MHD *mhd, ION* ion )
{
#ifdef DEBUG_MODE_ON
  printf("DEBUG:  process (%d) in MHD_ION.C::mhd2ion(...)\n", RTIME.local_proc);
#endif
  
  Range ONE(1,1); Range TWO(1,2); Range THREE(1,3);
  int NJ = mhd->NJ;  int NK = mhd->NK; int vari;
  FLoaT *dummy;
  Range J(1,NJ);   Range K(1,NK);
  int mype = Communication_Manager::My_Process_Number;  
  
  TEST_ARRAYS();
  ion->Rho.partition(ion->ION_3D_GUARD); ion->Rho(TWO,J,K) = mhd->Rho(TWO,J,K);
  ion->Rho.partition(ion->ION_3D); 
  dummy = ion->Rho.getDataPointer(); vari = _i_rho; ION_POKE(&dummy,&vari) ;
  
  ion->Csound.partition(ion->ION_3D_GUARD); ion->Csound(TWO,J,K) = mhd->Csound(TWO,J,K);
  ion->Csound.partition(ion->ION_3D);  
  dummy = ion->Csound.getDataPointer(); vari = _i_csound; ION_POKE(&dummy,&vari) ;
  
  ion->Bx.partition(ion->ION_3D_GUARD); ion->Bx(TWO,J,K) = mhd->Bx(TWO,J,K);
  ion->Bx.partition(ion->ION_3D); 
  dummy = ion->Bx.getDataPointer(); vari = _i_bx;  ION_POKE(&dummy,&vari) ;
  
  ion->By.partition(ion->ION_3D_GUARD); ion->By(TWO,J,K) = mhd->By(TWO,J,K);
  ion->By.partition(ion->ION_3D); 
  dummy = ion->By.getDataPointer(); vari = _i_by;  ION_POKE(&dummy,&vari) ;
  
  ion->Bz.partition(ion->ION_3D_GUARD); ion->Bz(TWO,J,K) = mhd->Bz(TWO,J,K);
  ion->Bz.partition(ion->ION_3D); 
  dummy = ion->Bz.getDataPointer(); vari = _i_bz;  ION_POKE(&dummy,&vari) ;
  
  
#ifdef DEBUG   
  printf("out of mhd2ion %d\n",mype); fflush(stdout);
  Communication_Manager::Sync();
#endif  
}



/// Convert to MHD parition
/** 
 *  The ionosphere is currently solved on one processor.  When
 *  transfering data to all other nodes, we need the guard cell part
 *  of the partitioning to be consistent.  So we need to convert
 *  between guard and no guard.
 *
 *  This function converts the electric field E[i,j] and velocity
 *  V[x,y,z] from ion partitioning to MHD partitioning (for
 *  communications with MHD)
 *
 *  @param mhd - MHD info
 *  @param ion - Ionosphere information
 *
 * @see MHD
 * @see ION
 * @see mhd2ion
 */
void ion2mhd( MHD *mhd, ION* ion ) 
{
#ifdef DEBUG_MODE_ON
  printf("DEBUG:  process (%d) in MHD_ION.C::ion2mhd(...)\n", RTIME.local_proc);
#endif
  
  Range ONE(1,1); Range TWO(1,2); Range THREE(1,3);
  int NJ = mhd->NJ;  int NK = mhd->NK; int vari;
  FLoaT *dummy;
  Range J(1,NJ); Range JP(1,NJ+1);
  Range K(1,NK); Range KP(1,NK+1);    

  // Transform Ei
  ion->Ej_pot.partition(ion->ION_3D_GUARD); 
  mhd->Ej(ONE,J,KP) = ion->Ej_pot(ONE,J,KP);
  ion->Ej_pot.partition(ion->ION_3D); 
  dummy = ion->Ej_pot.getDataPointer(); 
  vari = _i_ej_pot;
  ION_POKE(&dummy,&vari) ;
  
  // Transform Ej
  ion->Ek_pot.partition(ion->ION_3D_GUARD); 
  mhd->Ek(ONE,JP,K) = ion->Ek_pot(ONE,JP,K);
  ion->Ek_pot.partition(ion->ION_3D); 
  dummy = ion->Ek_pot.getDataPointer(); 
  vari = _i_ek_pot;  
  ION_POKE(&dummy,&vari) ;
  
  // Transform Vx
  ion->Vx_pot.partition(ion->ION_3D_GUARD); 
  mhd->Vx(ONE,J,K) = ion->Vx_pot(ONE,J,K);
  ion->Vx_pot.partition(ion->ION_3D); 
  dummy = ion->Vx_pot.getDataPointer(); 
  vari = _i_vx_pot;  
  ION_POKE(&dummy,&vari) ;
  
  // Transform Vy
  ion->Vy_pot.partition(ion->ION_3D_GUARD); 
  mhd->Vy(ONE,J,K) = ion->Vy_pot(ONE,J,K);
  ion->Vy_pot.partition(ion->ION_3D); 
  dummy = ion->Vy_pot.getDataPointer(); 
  vari = _i_vy_pot;  
  ION_POKE(&dummy,&vari) ;
  
  // Transform Vz
  ion->Vz_pot.partition(ion->ION_3D_GUARD); 
  mhd->Vz(ONE,J,K) = ion->Vz_pot(ONE,J,K);
  ion->Vz_pot.partition(ion->ION_3D); 
  dummy = ion->Vz_pot.getDataPointer(); 
  vari = _i_vz_pot;  
  ION_POKE(&dummy,&vari) ;
  
  TEST_ARRAYS();  
}

/// Convert to grid to ionospheric grid
/** 
 *
 *  @param mhd - MHD info
 *  @param ion - Ionosphere information
 *
 * @see MHD
 * @see ION
 */
void grid2ion(MHD* mhd, ION* ion) 
{
#ifdef DEBUG_MODE_ON
  printf("DEBUG:  process (%d) in MHD_ION.C::grid2ion(...)\n", RTIME.local_proc);
#endif

  Range ONE(1,1); Range TWO(1,2); Range THREE(1,3);
  int NJ = mhd->NJ;  int NK = mhd->NK; int vari;
  FLoaT* dummy; 
  Range J(1,NJ); Range JP(1,NJ+1);
  Range K(1,NK); Range KP(1,NK+1);
  
  ion->X.partition(ion->ION_3D_GUARD); 
  ion->X(THREE,JP,KP) = mhd->X(THREE,JP,KP);    
  ion->X.partition(ion->ION_3D);
  dummy = ion->X.getDataPointer(); 
  vari = _i_x;  ION_POKE(&dummy,&vari) ;
  
  ion->Y.partition(ion->ION_3D_GUARD); 
  ion->Y(THREE,JP,KP) = mhd->Y(THREE,JP,KP);    
  ion->Y.partition(ion->ION_3D); 
  dummy = ion->Y.getDataPointer(); 
  vari = _i_y;  ION_POKE(&dummy,&vari) ;
  
  ion->Z.partition(ion->ION_3D_GUARD); 
  ion->Z(THREE,JP,KP) = mhd->Z(THREE,JP,KP);    
  ion->Z.partition(ion->ION_3D); 
  dummy = ion->Z.getDataPointer(); 
  vari = _i_z;  ION_POKE(&dummy,&vari) ;
  
  TEST_ARRAYS();
}


