/**
 * \file CPL_FE_Interface.h
 * \brief CPL_FE_Interface declaration
 *
 * \class CPL_FE_Interface 
 * \brief Defines functions for communication with the MIX coupler (CPL) using File Exchanges
 *
 * \author Peter Schmitt (schmitt at bu.edu) 
 * \since 09-2008
 *
 * This class defines functions for communication between an MHD code
 * (i.e. LFM) and the Ionosphere Coupler/Solver (MIX) code using file
 * exchanges.  Communication occurs through a series of I/O to the
 * following files:
 *
 *   - MHD_status
 *   - MHD_scalars
 *   - MHD_current
 *   - MHD_density
 *   - MHD_soundSpeed
 *
 * If your computing environment supports it, you should use the
 * InterComm version of this class (MHD_IC_InnerBoundaryInterface).
 * The InterComm version communicates with MIX much more efficiently.
 *
 * This code implies that the MHD grid is a logically spherical grid
 * with the symmetry axis aligned with the SM x-axis (LFM definition).
 *
 * \see MHDInnerBoundaryInterface, MHD_IC_InnerBoundaryInterface, MHD_Interface, MHD_FE_Interface
 */
#ifndef CPL_FE_INTERFACE_H
#define CPL_FE_INTERFACE_H

#include "MHDInnerBoundaryInterface.h"

class CPL_FE_Interface : public MHDInnerBoundaryInterface {
public:

  /**\brief Class constructor */
  CPL_FE_Interface(const doubleArray &x, const doubleArray &y, const doubleArray &z, 
		   const int &NI, const int &NJ, const int &NK);

  virtual ~CPL_FE_Interface();

  /**\brief Export the MHD FAC, density and sound speed to the MIX code*/
  void Export(const doubleArray &bx, const doubleArray &by, const doubleArray &bz, 
	      const doubleArray &rho, const doubleArray &cs);
  
  /**\brief Import potential and calculate electric field & velocity at the inner MHD boundary*/
  void Import(doubleArray &eField_j, doubleArray &eField_k, doubleArray &velocity);

  /** \brief Broadcast scalar array (i.e. UT & kill signal) from MHD code. */
  void sendScalars(const doubleArray &scalars);

private:
  /**\name File exchange members */
  //@{

  /**\brief Returns true if MHD is busy */
  bool isMHDOld()  { return ( strcmp( "OLD", readLockFile(MHDLockFile).c_str() ) == 0 ); }
  /**\brief Returns true if MHD is ready and waiting for the MHD */
  bool isMHDNew() { return ( strcmp( "NEW", readLockFile(MHDLockFile).c_str() ) == 0 ); }

  /**\brief Returns true if MIX is busy */
  bool isMIXOld()  { return ( strcmp( "OLD", readLockFile(MIXLockFile).c_str() ) == 0 ); }
  /**\brief Returns true if MIX is ready and waiting for the MHD */
  bool isMIXNew() { return ( strcmp( "NEW", readLockFile(MIXLockFile).c_str() ) == 0 ); }

  /**\brief Set the MHD lock file to "OLD" (aka busy)  */
  bool setMHDOld(void)   { return writeLockFile(MHDLockFile, "OLD"); }
  /**\brief Set the MHD lock file to "NEW" (aka ready) */
  bool setMHDNew(void)  { return writeLockFile(MHDLockFile, "NEW"); }

  /**\brief Set the MIX lock file to "OLD" (aka busy)  */
  bool setMIXOld(void)   { return writeLockFile(MIXLockFile, "OLD"); }
  /**\brief Set the MIX lock file to "NEW" (aka ready) */
  bool setMIXNew(void)  { return writeLockFile(MIXLockFile, "NEW"); }

  /**\brief  Write #status to the MHD lock file */
  bool  writeLockFile(const string &filename, const string &status);
  /**\brief  Read the MIX lock file to check its status */
  string readLockFile(const string &filename);

  /**\brief  Read a 3d array from file into #data*/
  void read3dData(const string &filename, double *data,  
		  const int &ni, const int &nj, const int &nk);
  /**\brief  Write #data to file */
  void write3dData(const string &filename, const double *data,  
		   const int &ni, const int &nj, const int &nk);

  /**\brief filename of MIX lock file. */
  string MIXLockFile;
  /**\brief filename of MHD lock file. */
  string MHDLockFile;
  //@}

};

#endif
