/**
 * \file CPL_IC_Interface.C 
 * \brief CPL_IC_Interface definition.
 */

#include "CPL_IC_Interface.h"

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 05-2005 
 * \param[in] EPS                 IC_EndPointSet used for InterComm communication with the coupler (CPL)
 * \param[in] x                   X coordinates of the MHD grid vertices in cm
 * \param[in] y                   Y coordinates of the MHD grid vertices in cm
 * \param[in] z                   Z coordinates of the MHD grid vertices in cm
 * \param[in] NI  The number of shells in the i-th direction where to calculate the FAC
 * \param[in] NJ  The number of grid cells in the j-th direction
 * \param[in] NK  The number of grid cells in the k-th direction
 *
 * Constructor that initializes the LFM-MIX interface and brings up
 * the InterComm interface.
 *
 * This calls the MHDInnerBoundaryInterface base class constructor,
 * which calculates a lot of grid-related quantities that are needed
 * for future calculations (especially the field-aligned current
 * calculation), but have to be calculated only once. Among these
 * quantities are: cell volumes, cell centers, and face centers as
 * well as magnetic field at the inner MHD boundary (dipole) for the
 * parallel current calculation.
 *
 * \retval The outputs are placed in the global variables declared class-wide: 
 * - #volume - array containing calculated cell volumes
 * - #xCenter - Cell center coordinates
 * - #yCenter
 * - #zCenter
 * - #x_iFaceCenters - Face center coordinates
 * - #y_iFaceCenters etc.
 *
 * These quantities are needed because the calculation is done on a
 * staggered grid. See jpara.tex for details.
 *
 * #bxDot, #byDot, #bzDot are the components of the modified magnetic
 * field multiplying the current density to obtain ionospheric FAC.
 * (see jpara.tex for details)
 *
 * The cell volume is calculated as a triple product of vectors
 * connecting centers of opposing faces.  Cells are offset half-cell
 * size (modified cells are based on original cell centers). See
 * jpara.tex for details.
 *
 * \see MHDInnerBoundaryInterface::MHDInnerBoundaryInterface
 ************************************************************************/
CPL_IC_Interface::CPL_IC_Interface(IC_EndPointSet *EPS,
				   const doubleArray &x, const doubleArray &y, const doubleArray &z,
				   const int &NI, const int &NJ, const int &NK)
  :
  // Base class constructor
  MHDInnerBoundaryInterface(x,y,z, NI,NJ,NK)
{
#ifdef DEBUG_MODE_ON
  cout << "DEBUG:  Inside " << __FILE__ << " " << __FUNCTION__ << endl;
#endif

  epset = EPS;
  cout << "CPL_IC_Interface is here..." << endl;
  
  /* Set up InterComm arrays */
  epset->registerArray("mhd-cpl_fac",current,ic_err);
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("     Current registered. Status",ic_err);
#endif
  CHECK_IC_EPS_STATUS("IC_EndPointSet::registerArray('mhd-cpl_fac')", ic_err);
 
  epset->registerArray("mhd-cpl_density",density,ic_err); 
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("     Density registered. Status",ic_err);
#endif
  CHECK_IC_EPS_STATUS("IC_EndPointSet::registerArray('mhd-cpl_density')", ic_err);

  epset->registerArray("mhd-cpl_sound_speed",soundSpeed,ic_err);
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("     Sound speed registered. Status",ic_err);
#endif
  CHECK_IC_EPS_STATUS("IC_EndPointSet::registerArray('mhd-cpl_sound_speed')", ic_err);

  epset->registerArray("cpl-mhd_potential",potential,ic_err);
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("     Potential registered. Status",ic_err);
#endif
  CHECK_IC_EPS_STATUS("IC_EndPointSet::registerArray('cpl-mhd_potential')", ic_err);

  exchangeAndValidateRION();
} 

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 05-2005 
 *
 * \param[in] bx MHD magnetic field x-component
 * \param[in] by MHD magnetic field y-component
 * \param[in] bz MHD magnetic field z-component
 * \param[in] rho MHD density
 * \param[in] cs MHD sound speed
 *
 * The function is called from the main level program (MHD code). It
 * calls prepareExport to get the FAC, density and sound speed in the
 * second shell. After that InterComm exportArray() functions are
 * called to export current, density and sound speed (precisely in
 * this order) to the MIX code.
 ************************************************************************/
void CPL_IC_Interface::Export(const doubleArray &bx, const doubleArray &by, const doubleArray &bz, 
			      const doubleArray &rho, const doubleArray &cs)
{
#ifdef DEBUG_MODE_ON
  cout << "DEBUG:  Inside " << __FILE__ << " " << __FUNCTION__ << endl;
#endif

  // Calculate the FAC, density and sound speed in the second-shell
  prepareExport(bx,by,bz, rho, cs);

  // Export the data to MIX using InterComm
  epset->exportArray("mhd-cpl_fac",ic_err);
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("     Current sent. Status",ic_err);
#endif
  CHECK_IC_EPS_STATUS("IC_EndPointSet::exportArray('mhd-cpl_fac')", ic_err);

  epset->exportArray("mhd-cpl_density",ic_err);
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("     Density sent. Status",ic_err);
#endif
  CHECK_IC_EPS_STATUS("IC_EndPointSet::exportArray('mhd-cpl_density')", ic_err);

  epset->exportArray("mhd-cpl_sound_speed",ic_err);
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("     Sound speed sent. Status",ic_err);
#endif
  CHECK_IC_EPS_STATUS("IC_EndPointSet::exportArray('mhd-cpl_sound_speed')", ic_err);
}

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 05-2005 
 *
 * \param[out] eField_j   Electric field j-component at the i=1 edge
 * \param[out] eField_k   Electric field k-component at the i=1 edge
 *
 * \param[out] velocity Velocity (x,y,z components) at the center of
 * the first layer of cells
 *
 * The function first calls InterComm importArray() to get the
 * ionospheric potential in the first two shell from the MIX
 * code. Then processImport calculates #eField_j and #eField_k at the
 * first shell (i=1) edges and #velocity is calculated at the centers
 * of the first layer of cells (i=1).
 ************************************************************************/
void CPL_IC_Interface::Import(doubleArray &eField_j, doubleArray &eField_k, doubleArray &velocity)
{
#ifdef DEBUG_MODE_ON
  cout << "DEBUG:  Inside " << __FILE__ << " " << __FUNCTION__ << endl;
#endif

  epset->importArray("cpl-mhd_potential",ic_err);

#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("     Potential for the two first shells received. Status",ic_err);
#endif
  CHECK_IC_EPS_STATUS("IC_EndPointSet::importArray('cpl-mhd_potential')", ic_err);

  // Now do the dirty work: calculate the potential electric field
  // and the velocity at the inner boundary

  processImport(eField_j, eField_k, velocity);
}


/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 06-2007
 * \return   The array of scalars to be sent to the MIX code
 *
 * Use this function to broadcast an array of scalars to the MIX
 * code.The scalars to be sent are defined in the CPL_Scalars
 * enumeration defined in MHDInnerBoundaryInterface.h
 *
 * Note: The scalars array *must* be distributed only on process 0.
 ************************************************************************/
void CPL_IC_Interface::sendScalars(const doubleArray & scalars)
{ 
#ifdef DEBUG_MODE_ON
  cout << "DEBUG:  Inside " << __FILE__ << " " << __FUNCTION__ << endl;
#endif

  int nelems = scalars.elementCount();
  // FIXME:  assert(scalars.elementCount() == Im_Interface::NUMBER_OF_SCALARS)???
  epset->bcastLocalArray("mhd-cpl_scalars",scalars,nelems,ic_err);
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("Scalars sent. Status: ",ic_err);
#endif
  // FIXME: There's a bug where InterComm's bcastLocalArray(...) will
  // set the ic_err status flag to -6 (communication (send/recv)
  // failure).  I have e-mailed Norman Lo on July 17, 2009 about this
  // Temporary fix:  Don't error check! --schmitt<at>ucar<dot>edu

  //  CHECK_IC_EPS_STATUS("IC_EndPointSet::bcastLocalArray(...)", ic_err);
}

/********************************************************************//**
 * Receive RION from the Coupler (MIX) and make sure it matches the
 * value in the LFM.  
 *
 * This *MUST* be called early on before committing arrays.  Make sure
 * to avoid deadlocks by matching the send in MIX.
 ************************************************************************/
void CPL_IC_Interface::exchangeAndValidateRION()
{
#ifdef DEBUG_MODE_ON
  cout << "DEBUG:  Inside " << __FILE__ << " " << __FUNCTION__ << endl;
#endif

  // We must allocate enough size for RION in global memory space, so 1*num_procs.
  int num_procs = Communication_Manager::numberOfProcessors();
  floatArray RION_meters(1*num_procs, Partitioning_Type(num_procs));
  RION_meters = 0;

  epset->recvLocalArray("cpl-mhd_r_i", RION_meters, 1, ic_err);
  CHECK_IC_EPS_STATUS("IC_EndPointSet::recvLocalArray('cpl-mhd_r_i')", ic_err);

  float m_to_cm = 100.0;
  if ( fabs( (RION_meters(0)*m_to_cm) - RION ) > 1e-6 ){
    cerr << __FILE__ << "(" << __LINE__ << ": " << __FUNCTION__ << "\t"
	 << "Radius of ionosphere differs between LFM ( " << RION  << ") "
	 << "and the Coupler, MIX (" << RION_meters(0)*m_to_cm << ").  "
	 << "These values must be identical!" << endl;
    Optimization_Manager::Exit_Virtual_Machine();
  }
}
