/**
 * \file CPL_IC_Interface.h
 * \brief CPL_IC_Interface declaration
 *
 * \class CPL_IC_Interface 
 * \brief Defines functions for communication (using InterComm) with the MIX coupler (CPL).
 *
 * \author Slava Merkin (vgm at bu.edu) 
 * \author Peter Schmitt (schmitt at bu.edu) 
 * \since 08-2005
 *
 * This class defines functions for communication between an MHD code
 * and the Ionosphere Coupler/Solver code using InterComm. The functions visible to
 * the outside world are Export(), Import(), sendScalars(), and
 * contructor CPL_IC_Interface(). 
 *
 * This code implies that the MHD grid is a logically spherical grid
 * with the symmetry axis aligned with the SM x-axis (LFM definition).
 *
 * \see MHDInnerBoundaryInterface
 */
#ifndef CPL_IC_INTERFACE_H
#define CPL_IC_INTERFACE_H

#include "Check_IC_Status.h"
#include "const.h"

// LFM-MIX includes
#include "MHDInnerBoundaryInterface.h"
// 3rd-party includes
#include <IC_EndPointSet.h>

class CPL_IC_Interface : public MHDInnerBoundaryInterface {
public:

  /**\brief Class constructor */
  CPL_IC_Interface(IC_EndPointSet *EPS,
		   const doubleArray &x, const doubleArray &y, const doubleArray &z, 
		   const int &NI, const int &NJ, const int &NK);

  /**\brief Export the MHD FAC, density and sound speed to the MIX code*/
  void Export(const doubleArray &bx, const doubleArray &by, const doubleArray &bz, 
	      const doubleArray &rho, const doubleArray &cs);
  
  /**\brief Import potential and calculate electric field & velocity at the inner MHD boundary*/
  void Import(doubleArray &eField_j, doubleArray &eField_k, doubleArray &velocity);

  /** \brief Broadcast scalar array (i.e. UT & kill signal) from MHD code. */
  void sendScalars(const doubleArray &scalars);

private:
  /**\name InterComm variables */
  //@{
  IC_EndPointSet* epset; //!< InterComm end point set
  int ic_err;
  //@}

  void exchangeAndValidateRION();
};

#endif
