#include "IM_IC_Interface.h"

/********************************************************************//**
 * After this constructor is called, the user *must* commit the arrays
 * via epset->commitArrays(ic_err).
 * 
 * After the arrays are comitted, the user *must* exchange Initialize
 * variables by calling IM_IC_Interface::Initialize(...)!
 *
 * Why do we pass Partitioning_Type mhdPartitioning?  It seems like
 * there's a bug in A++/P++ where Partitioning_Type information is
 * lost when typeArray::convertTo_floatArray() is called.
 *
 * FIXME:  Passing floatArray by reference yields strange results.  Try to 
 *         floatArray::disaplay() and you get a segfault with the error:
 *         "SerialArray is a NULL pointer = 0!"
 *
 ************************************************************************/ 
IM_IC_Interface::IM_IC_Interface(IC_EndPointSet *EPS, 
				 const floatArray &lfmx, const floatArray &lfmy, const floatArray &lfmz, 
				 const Partitioning_Type &mhdPartitioning,
				 const float &inDipoleMoment, const float &inGamma,
				 const DateTime &start, const int &delta_t) :
  epset(EPS), IM_Interface(lfmx, lfmy, lfmz, mhdPartitioning, inDipoleMoment, inGamma, start, delta_t)
{
  cout << "LFMIM_IC_Interface is here..." << endl;

  int num_procs = Communication_Manager::numberOfProcessors();
  
  Partitioning_Type all(num_procs);
  Partitioning_Type p_zero(Range(0,0));

  /************* Get Intermediate Grid size *************/
  // We must allocate enough size for igDims in global memory space, so 3*num_procs.
  intArray igDims(3*num_procs, all);
  igDims = 0;
  epset->recvLocalArray("im-mhd_ig_size", igDims, 3, ic_err);
  CHECK_IC_EPS_STATUS("IC_EndPointSet::recvLocalArray('im-mhd_ig_size')", ic_err);
  ig_dimensions[0] = igDims(0);
  ig_dimensions[1] = igDims(1);
  ig_dimensions[2] = igDims(2);

  /************* Broadcast LFM Grid size *************/
  intArray lfmGridSize(3, p_zero);
  const int ni =  mhd_lfmx.getLength(0)-1;
  lfmGridSize(0) = ni;
  const int nj =  mhd_lfmx.getLength(1)-1;
  lfmGridSize(1) = nj;
  const int nk =  mhd_lfmx.getLength(2)-1;
  lfmGridSize(2) = nk;
  epset->bcastLocalArray("mhd-im_lfm_size", lfmGridSize, 3, ic_err);
  if (Communication_Manager::My_Process_Number == 0)
    CHECK_IC_EPS_STATUS("IC_EndPointSet::bcastLocalArray('im-mhd_lfm_size')", ic_err);

  /************* Broadcast Dipole Moment *************/
  floatArray dipoleMoment(1, p_zero);
  dipoleMoment(0) = dipole_moment;
  epset->bcastLocalArray("mhd-im_dipole_moment", dipoleMoment, 1, ic_err);
  if (Communication_Manager::My_Process_Number == 0)
    CHECK_IC_EPS_STATUS("IC_EndPointSet::bcastLocalArray('mhd-im_dipole_moment')", ic_err);

  /************* Receive IG coordinates from RCM *************/
  // We must allocate enough size for igDims in global memory space, so n*num_procs.
  im_igx.redim(igDims(0)*num_procs);
  im_igx.partition(all);
  im_igx = 0.0;

  im_igy.redim(igDims(1)*num_procs);
  im_igy.partition(all); 
  im_igy = 0.0;

  im_igz.redim(igDims(2)*num_procs);
  im_igz.partition(all);
  im_igz = 0.0;

  epset->recvLocalArray("im-mhd_igx", im_igx, igDims(0), ic_err);
  CHECK_IC_EPS_STATUS("IC_EndPointSet::recvLocalArray('im-mhd_igx')", ic_err);

  epset->recvLocalArray("im-mhd_igy", im_igy, igDims(1), ic_err);
  CHECK_IC_EPS_STATUS("IC_EndPointSet::recvLocalArray('im-mhd_igy')", ic_err);

  epset->recvLocalArray("im-mhd_igz", im_igz, igDims(2), ic_err);
  CHECK_IC_EPS_STATUS("IC_EndPointSet::recvLocalArray('im-mhd_igz')", ic_err);

  //FIXME:  Each of the following is O(n^3) due to P++ trickery!
  //        Should probably serialize this as one triple-nested for... loop.

  const double earthRadiusInM = 6.38e6;
  im_igx /= earthRadiusInM; // convert from meters to Earth Radii
  im_igy /= earthRadiusInM;
  im_igz /= earthRadiusInM;

  /************* Register LFM grid P++ arrays with InterComm *************/
  // Grid variables
  epset->registerArray("mhd-im_lfmx", mhd_lfmxCenter, ic_err);
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("   lfmx registered.  Status", ic_err);
#endif
  CHECK_IC_EPS_STATUS("IC_EndPointSet::registerArray('mhd-im_lfmx')", ic_err);

  epset->registerArray("mhd-im_lfmy", mhd_lfmyCenter, ic_err);
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("   lfmy registered.  Status", ic_err);
#endif
  CHECK_IC_EPS_STATUS("IC_EndPointSet::registerArray('mhd-im_lfmy')", ic_err);

  epset->registerArray("mhd-im_lfmz", mhd_lfmzCenter, ic_err);
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("   lfmz registered.  Status", ic_err);
#endif
  CHECK_IC_EPS_STATUS("IC_EndPointSet::registerArray('mhd-im_lfmz')", ic_err);

  /************* Register Bleed (RCM-to-LFM) P++ arrays (on LFM grid) with InterComm *************/
  epset->registerArray("im-mhd_density", im_density, ic_err);
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("   Density registered.  Status", ic_err);
#endif
  CHECK_IC_EPS_STATUS("IC_EndPointSet::registerArray('im-mhd_density')", ic_err);

  epset->registerArray("im-mhd_pressure", im_pressure, ic_err);
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("   Pressure registered.  Status", ic_err);
#endif
  CHECK_IC_EPS_STATUS("IC_EndPointSet::registerArray('im-mhd_pressure')", ic_err);

  epset->registerArray("im-mhd_rcm_mask", im_rcm_mask, ic_err);
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("   rcm_mask registered.  Status", ic_err); 
#endif
  CHECK_IC_EPS_STATUS("IC_EndPointSet::registerArray('im-mhd_rcm_mask')", ic_err);

  /************* Register Accumulate (LFM-to-RCM) P++ arrays (on Intermediate Grid) with InterComm *************/
  // igSetup will prepare variables on the intermediate grid
  // {mhd-im_density, mhd-im_pressure, mhd-im_bx, mhd-im_by,
  // mhd-im_bz, mhd-im_v}
  igSetup(igDims);

  /************* Arrays are ready to be comitted *************/  
  // Arrays must be comitted outside this class in case additional
  // arrays need to be defined in another interface
  // (i.e. CPL_IC_Interface).

  //Don't forget to call IM_IC_Interface::Initialize() after comitting!
}

/********************************************************************//**
 * Destructor . . . 
 ************************************************************************/ 
IM_IC_Interface::~IM_IC_Interface()
{
  if (mhd_density != NULL){
    delete [] mhd_density;
    mhd_density = NULL;
  }
  if (mhd_pressure != NULL){
    delete [] mhd_pressure;
    mhd_pressure = NULL;
  }
  if (mhd_bx != NULL){
    delete [] mhd_bx;
    mhd_bx = NULL;
  }
  if (mhd_by != NULL){
    delete [] mhd_by;
    mhd_by = NULL;
  }
  if (mhd_bz != NULL){
    delete [] mhd_bz;
    mhd_bz = NULL;
  }
  if (mhd_v != NULL){
    delete [] mhd_v;
    mhd_v = NULL;
  }

  if (igGlobalIndices != NULL){
    delete [] igGlobalIndices;
    igGlobalIndices = NULL;
  }
  if (tasks != NULL){
    delete [] tasks;
    tasks = NULL;
  }
  if (localIndices != NULL){
    delete [] localIndices;
    localIndices = NULL;
  }
  if (globals != NULL){
    delete [] globals;
    globals = NULL;
  }

  IC_Free_desc(desc);
  IC_Free_region(region_set[0]);
}

/********************************************************************//**
 * Setup variables on the intermediate grid {mhd-im_density,
 * mhd-im_pressure, mhd-im_bx, mhd-im_by, mhd-im_bz, mhd-im_v}
 *
 *   - When doing parallel interpolation, use the InterComm ttable
 *   - When doing serial interpolation, register standard P++ array.
 *
 * NOTE: Parallel interpolation is not fully implemented/tested!
 * Development on parallel interpolation stopped at r899 (July 14,
 * 2009). Stick with serial interpolation for the best (only?)
 * results!
 ************************************************************************/ 
void IM_IC_Interface::igSetup(const intArray &igDims)
{
  // Initialize pointers used by parallel interpolation.  
  //
  // Note: These pointers aren't used for serial interp... if we don't
  // set to NULL, *bad* things happen in the destructor!
  mhd_density = NULL;
  mhd_pressure = NULL;
  mhd_bx = NULL;
  mhd_by = NULL;
  mhd_bz = NULL;
  mhd_v = NULL;

  igGlobalIndices = NULL;
  tasks = NULL;
  localIndices = NULL;
  globals = NULL;

  desc = NULL;
  region_set[0] = NULL;
  
#if 0
  // NOTE:  Parallel interpolation is not fully implemented/tested!  Stick with serial.
  igParallelSetup(igDims);
#endif
  igSerialSetup(igDims);
}


/********************************************************************//**
 * Setup intermediate grid variables to use parallel interpolation
 * Uses InterComm translation table to describe the distributed array.
 *
 * WARNING: Parallel interpolation was never fully implemented and
 * active development stopped at r899 (July 14, 2009).  We decided to
 * try serial interpolation, which seems to be working fine.
 ************************************************************************/ 
void IM_IC_Interface::igParallelSetup(const intArray &igDims)
{
  // Configure variables on the intermediate grid
  int igNpoints = igDims(0) * igDims(1) * igDims(2);
  int igNpoints_local;

  // IM_Interface::igParallelInitInterp
  igParallelInitInterp(igNpoints_local, igGlobalIndices);

  mhd_density = new float [igNpoints_local];
  mhd_pressure = new float [igNpoints_local];
  mhd_bx = new float [igNpoints_local];
  mhd_by = new float [igNpoints_local];
  mhd_bz = new float [igNpoints_local];
  mhd_v = new float [igNpoints_local];

  // Create translation table to match local & global indices for
  // arrays distributed across MPI tasks
  tasks        = new int [igNpoints_local];
  localIndices = new int [igNpoints_local];

  for (int i=0; i<igNpoints_local; i++) {
    tasks[i] = Communication_Manager::My_Process_Number; 
    localIndices[i] = i;
  }  

  desc = IC_Create_ttable_desc(igGlobalIndices, localIndices, tasks, igNpoints_local);

  /* Allocate region to handle translation table. */
  globals = new int[igNpoints];
  for (int i=0;i<igNpoints;i++) {
    globals[i]=i;
  }

  region_set[0] = IC_Create_enum_region(globals,igNpoints);

  // Register accumulate regions . . . . . . . . . . . . . . . . . . . . . . . . . 

  // The following are commented out because:
  //  1.  This function is not fully implemented/tested
  //  2.  IC_EndPointSet::registerByConn is not implemented as of InterComm 1.6.5
//   epset->registerByConn("mhd-im_density",  desc, region_set, 1, mhd_density );
//   epset->registerByConn("mhd-im_pressure", desc, region_set, 1, mhd_pressure );
//   epset->registerByConn("mhd-im_bx",       desc, region_set, 1, mhd_bx );
//   epset->registerByConn("mhd-im_by",       desc, region_set, 1, mhd_by );
//   epset->registerByConn("mhd-im_bz",       desc, region_set, 1, mhd_bz );
//   epset->registerByConn("mhd-im_v",        desc, region_set, 1, mhd_v );
}


/********************************************************************//**
 * Setup intermediate grid variables to use Serial interpolation
 * 
 * Uses high-level InterComm to A++/P++ interfaceEndPointSet::RegisterArray 
 * to describe the distributed array.
 ************************************************************************/ 
void IM_IC_Interface::igSerialSetup(const intArray &igDims)
{
  // IM_Interface::igSerialInitInterp
  igSerialInitInterp(igDims);    

  Partitioning_Type P_zero(Range(0,0));

  int ntotal_ig = igDims(0) * igDims(1) * igDims(2);

  fa_mhd_density.partition(P_zero);
  fa_mhd_density.redim( ntotal_ig );
  epset->registerArray("mhd-im_density", fa_mhd_density, ic_err);
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("   mhd-im_density registered.  Status", ic_err);
#endif
  CHECK_IC_EPS_STATUS("IC_EndPointSet::registerArray('mhd-im_density')", ic_err);


  fa_mhd_pressure.partition(P_zero);
  fa_mhd_pressure.redim( ntotal_ig );
  epset->registerArray("mhd-im_pressure", fa_mhd_pressure, ic_err);
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("   mhd-im_pressure registered.  Status", ic_err);
#endif
  CHECK_IC_EPS_STATUS("IC_EndPointSet::registerArray('mhd-im_pressure')", ic_err);


  fa_mhd_bx.partition(P_zero);
  fa_mhd_bx.redim( ntotal_ig );
  epset->registerArray("mhd-im_bx", fa_mhd_bx, ic_err);
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("   mhd-im_bx registered.  Status", ic_err);
#endif
  CHECK_IC_EPS_STATUS("IC_EndPointSet::registerArray('mhd-im_bx')", ic_err);


  fa_mhd_by.partition(P_zero);
  fa_mhd_by.redim( ntotal_ig );
  epset->registerArray("mhd-im_by", fa_mhd_by, ic_err);
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("   mhd-im_by registered.  Status", ic_err);
#endif
  CHECK_IC_EPS_STATUS("IC_EndPointSet::registerArray('mhd-im_by')", ic_err);


  fa_mhd_bz.partition(P_zero);
  fa_mhd_bz.redim( ntotal_ig );
  epset->registerArray("mhd-im_bz", fa_mhd_bz, ic_err);
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("   mhd-im_bz registered.  Status", ic_err);
#endif
  CHECK_IC_EPS_STATUS("IC_EndPointSet::registerArray('mhd-im_bz')", ic_err);


  fa_mhd_v.partition(P_zero);
  fa_mhd_v.redim( ntotal_ig );
  epset->registerArray("mhd-im_v", fa_mhd_v, ic_err);
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("   mhd-im_v registered.  Status", ic_err);
#endif
  CHECK_IC_EPS_STATUS("IC_EndPointSet::registerArray('mhd-im_v')", ic_err);  
}


/********************************************************************//**
 * After constructor is called, the user *must* commit the arrays via
 * epset->commitArrays(ic_err).  
 * 
 * After arrays are comitted, the user *must* exchange Initialize
 * variables (i.e. call this function!)
 ************************************************************************/ 
void IM_IC_Interface::Initialize()
{
  /////////////////////////////////////////////////////////////////////
  // Now that arrays are comitted, we can send the LFM grid to inner magnetosphere.

  epset->exportArray("mhd-im_lfmx", ic_err);
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("   lfmx sent.  Status", ic_err);
#endif
  CHECK_IC_EPS_STATUS("IC_EndPointSet::exportArray('mhd-im_lfmx')", ic_err);

  epset->exportArray("mhd-im_lfmy", ic_err);
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("   lfmy sent.  Status", ic_err);
#endif
  CHECK_IC_EPS_STATUS("IC_EndPointSet::exportArray('mhd-im_lfmy')", ic_err);

  epset->exportArray("mhd-im_lfmz", ic_err);
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("   lfmz sent.  Status", ic_err);
#endif
  CHECK_IC_EPS_STATUS("IC_EndPointSet::exportArray('mhd-im_lfmz')", ic_err);

  /////////////////////////////////////////////////////////////////////
  // Although InterComm requires bleed arrays with ghost cells
  // removed, the rest of the LFM needs the ghost cells put in for two
  // reasons: IM_Interface::Bleed(...) and restart I/O (i.e. saving
  // accumulate & bleed arrays to LFM output files).  Since we've
  // already committed the regions, we can now repartition to add
  // ghost cells.  See IM_Interface::prepareImport() and
  // IM_Interface::processImport() to see how we handle the arrays
  // before/after import.

  im_density.partition(mhdGhostPartitioning);
  im_pressure.partition(mhdGhostPartitioning);
  im_rcm_mask.partition(mhdGhostPartitioning);

  /////////////////////////////////////////////////////////////////////
  // Send the startup date/time to the inner magnetosphere.

  Partitioning_Type p_zero(Range(0,0));
  intArray imScalars(IM_Interface::NUMBER_OF_SCALARS, p_zero);
  imScalars(IM_Interface::KILL_SIGNAL) = 0;                // kill signal
  imScalars(IM_Interface::YEAR) = rcmStartUT.getYear();     // year 
  imScalars(IM_Interface::MONTH) = rcmStartUT.getMonth();   // month
  imScalars(IM_Interface::DAY) = rcmStartUT.getDay();       // day
  imScalars(IM_Interface::HOUR) = rcmStartUT.getHour();     // hour
  imScalars(IM_Interface::MINUTE) = rcmStartUT.getMinute(); // minute
  imScalars(IM_Interface::SECOND) = (int) rcmStartUT.getSecond(); // second
  imScalars(IM_Interface::DELTA_T) = rcm_delta_t;  // number of seconds between LFM-RCM exchange
  imScalars(IM_Interface::LABEL) = 0; // timestep label
  SendScalars("mhd-im_startup_scalars", imScalars);
}

/********************************************************************//**
 * Broadcast an array of scalars to the inner-magnetosphere code.  The
 * scalars to be sent are defined in the IM_Scalars enumeration
 * defined in IM_Interface.h.
 *
 * Two SendScalars calls should happen when coupled with RCM:
 *
 *  1.  IM_IC_Interface::Initialize(...) will send start UT and delta_t
 *  2.  LFM main loop should SendScalars(...) at every LFM-RCM exchange.
 *      FIXME:  Should this move inside of IM_IC_Interface::Export(...)???
 *
 * Note: The scalars array *must* be distributed only on process 0.
 ************************************************************************/ 
void IM_IC_Interface::SendScalars(const char *name, const intArray &scalars)
{
  int nElements = scalars.elementCount();
  // FIXME:  assert(scalars.elementCount() == Im_Interface::NUMBER_OF_SCALARS)???
  epset->bcastLocalArray( (char *) name, scalars, nElements, ic_err);
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("   doubleArray broadcast.  Status", ic_err);
#endif
  // FIXME: There's a bug where InterComm's bcastLocalArray(...) will
  // set the ic_err status flag to -6 (communication (send/recv)
  // failure).  I have e-mailed Norman Lo on July 17, 2009 about this
  // Temporary fix:  Don't error check! --schmitt<at>ucar<dot>edu

  //CHECK_IC_EPS_STATUS("IC_EndPointSet::bcastLocalArray(...)", ic_err);
}

/********************************************************************//**
 * Export variables to Inner Magnetosphere
 * FIXME: Should IM_IC_Interface::Export(...) send time information via
 * IM_IC_Interface::SendScalars(...) ?
 ************************************************************************/ 
void IM_IC_Interface::Export()
{
  prepareExport();

  epset->exportArray("mhd-im_density", ic_err);
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("   density sent.  Status", ic_err);
#endif
  CHECK_IC_EPS_STATUS("IC_EndPointSet::exportArray('mhd-im_density')", ic_err);

  epset->exportArray("mhd-im_pressure", ic_err);
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("   pressure sent.  Status", ic_err);
#endif
  CHECK_IC_EPS_STATUS("IC_EndPointSet::exportArray('mhd-im_pressure')", ic_err);

  epset->exportArray("mhd-im_bx", ic_err);
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("   bx sent.  Status", ic_err);
#endif
  CHECK_IC_EPS_STATUS("IC_EndPointSet::exportArray('mhd-im_bx')", ic_err);

  epset->exportArray("mhd-im_by", ic_err);
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("   by sent.  Status", ic_err);
#endif
  CHECK_IC_EPS_STATUS("IC_EndPointSet::exportArray('mhd-im_by')", ic_err);

  epset->exportArray("mhd-im_bz", ic_err);
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("   bz sent.  Status", ic_err);
#endif
  CHECK_IC_EPS_STATUS("IC_EndPointSet::exportArray('mhd-im_bz')", ic_err);

  epset->exportArray("mhd-im_v", ic_err);   
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("   v sent.  Status", ic_err);
#endif
  CHECK_IC_EPS_STATUS("IC_EndPointSet::exportArray('mhd-im_v')", ic_err);

  processExport();
}

/********************************************************************//**
 * Import variables from Inner Magnetosphere
 ************************************************************************/ 
void IM_IC_Interface::Import()
{
  prepareImport();

  epset->importArray("im-mhd_density", ic_err);
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("   density received.  Status", ic_err);
#endif
  CHECK_IC_EPS_STATUS("IC_EndPointSet::importArray('im-mhd_density')", ic_err);

  epset->importArray("im-mhd_pressure", ic_err);
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("   pressure received.  Status", ic_err);
#endif
  CHECK_IC_EPS_STATUS("IC_EndPointSet::importArray('im-mhd_pressure')", ic_err);

  epset->importArray("im-mhd_rcm_mask", ic_err);
#ifdef DEBUG_MODE_ON
  epset->printErrorMessage("   mask received.  Status", ic_err);
#endif
  CHECK_IC_EPS_STATUS("IC_EndPointSet::importArray('im-mhd_rcm_mask')", ic_err);

  processImport();
}
