#include "IM_Interface.h"

/********************************************************************//**
 * Constructor. 
 *
 * Why do we pass Partitioning_Type mhdPartitioning?  It seems like
 * there's a bug in A++/P++ where Partitioning_Type information is
 * lost when typeArray::convertTo_floatArray() is called.
 ************************************************************************/ 
IM_Interface::IM_Interface(const floatArray &lfmx, const floatArray &lfmy, const floatArray &lfmz, 
			   const Partitioning_Type &mhdPartitioning,
			   const float &inDipoleMoment, const float &inGamma,
			   const DateTime &start, const int &delta_t) :   
  mhd_lfmx(lfmx),    mhd_lfmy(lfmy),     mhd_lfmz(lfmz),
  dipole_moment(inDipoleMoment), gamma(inGamma),
  rcmStartUT(start),  rcm_delta_t(delta_t)
{
  mhdGhostPartitioning = mhdPartitioning;

  const int nip1 = lfmx.getLength(0);
  const int njp1 = lfmx.getLength(1);
  const int nkp1 = lfmx.getLength(2);

  InitializeAccumulateArrays(nip1, njp1, nkp1);
  InitializeBleedArrays(nip1, njp1, nkp1);
  InitializeGridArrays(nip1, njp1, nkp1);
}


/********************************************************************//**
 * Initialize Accumualte arrays:
 *   - redimension to be of length (ni,nj,nk)
 *   - partition using mhdGhostPartitioning to get the proper
 *     partitioning object (among all MHD procs with ghost cells)
 *   - Initialize accum arrays to zero.
 * 
 * Note: This should only be called from the constructor.
 ************************************************************************/ 
void IM_Interface::InitializeAccumulateArrays(const int &nip1, const int &njp1, const int &nkp1)
{
  vector <int> variables;
  variables.insert(variables.end(), AverageVariables::RHO);
  variables.insert(variables.end(), AverageVariables::PRESSURE);
  variables.insert(variables.end(), AverageVariables::BI);
  variables.insert(variables.end(), AverageVariables::BJ);
  variables.insert(variables.end(), AverageVariables::BK);
  variables.insert(variables.end(), AverageVariables::V);
  accum = Average<floatArray>(nip1, njp1, nkp1, 
			      mhdGhostPartitioning,
			      variables);

  accum_bx.redim( nip1, njp1, nkp1 );
  accum_bx.partition(mhdGhostPartitioning);
  accum_bx.setBase(1);
  
  accum_by.redim( nip1, njp1, nkp1 );
  accum_by.partition(mhdGhostPartitioning);
  accum_by.setBase(1);
  
  accum_bz.redim( nip1, njp1, nkp1 );
  accum_bz.partition(mhdGhostPartitioning);
  accum_bz.setBase(1);

  // Initialize accumulate arrays to zero:
  resetAccumulateArrays();

  // Prepare the dipole field values at the cell faces for later
  // subtraction
  INITMAG_DIP();
}

/********************************************************************//**
 * Initialize bleed arrays:
 *   - redimension to be of length (ni,nj,nk)
 *   - partition on all processors with ghost cells removed.
 * 
 * Note: This should only be called from the constructor.
 ************************************************************************/ 
void IM_Interface::InitializeBleedArrays(const int &nip1, const int &njp1, const int &nkp1)
{
  // InterComm does not work with P++ ghost cells.  Initialize with
  // standard partitioning.
  //
  // Note: After we commit the connections, we repartition to include
  // ghost cells.  This is necessary for LFM-RCM restarts: bleeding
  // data into LFM and I/O necessary for restart files requires ghost
  // cell partitioning

  int num_procs = Communication_Manager::numberOfProcessors();  
  Partitioning_Type all(num_procs);

  const int ni = nip1-1;
  const int nj = njp1-1;
  const int nk = nkp1-1;

  im_density.redim( ni, nj, nk );
  im_density.partition(all);
  im_density = 0.0;

  im_pressure.redim( ni, nj, nk );
  im_pressure.partition(all);
  im_pressure = 0.0;

  im_rcm_mask.redim( ni, nj, nk );
  im_rcm_mask.partition(all);
  im_rcm_mask = 0;
}

/********************************************************************//**
 * Initialize LFM grid arrays
 *   - Validate dimensions are right for all grid arrays
 *   - Use 0-based indexing into grid arrays.
 *   - partition on all processors with ghost cells removed.
 *   - Compute cell-centered {x,y,z} coordinates
 * 
 * Note: This should only be called from the constructor.
 ************************************************************************/ 
void IM_Interface::InitializeGridArrays(const int &nip1, const int &njp1, const int &nkp1)
{
  // InterComm requires the base of send/receive variables set to 0.
  // Otherwise it sends shifted arrays to the other side.  A++/P++
  // defaults new objects to base 0, but let's manually set it here
  // just to be careful.
  // 
  // Note: John Lyon calls realArray::setBase(1) all over the place in
  // the MHD code in order to mimic Fortran's 1-based indexing.
  mhd_lfmx.setBase(0);
  mhd_lfmy.setBase(0);
  mhd_lfmz.setBase(0);

  // Make sure that the LFM grid is dimensioned consistently.
  assert(mhd_lfmx.numberOfDimensions() == 3);
  assert(mhd_lfmy.numberOfDimensions() == 3);
  assert(mhd_lfmz.numberOfDimensions() == 3);
  assert((mhd_lfmx.getLength(0) == mhd_lfmy.getLength(0)) && (mhd_lfmy.getLength(0) == mhd_lfmz.getLength(0)));
  assert((mhd_lfmx.getLength(1) == mhd_lfmy.getLength(1)) && (mhd_lfmy.getLength(1) == mhd_lfmz.getLength(1)));
  assert((mhd_lfmx.getLength(2) == mhd_lfmy.getLength(2)) && (mhd_lfmy.getLength(2) == mhd_lfmz.getLength(2)));

  int num_procs = Communication_Manager::numberOfProcessors();
  Partitioning_Type all(num_procs);

  mhd_lfmx.partition(all);
  mhd_lfmy.partition(all);
  mhd_lfmz.partition(all);

  const int ni=nip1-1;
  const int nj=njp1-1;
  const int nk=nkp1-1;

  // Compute cell-centered coordinates
  mhd_lfmxCenter.redim(ni, nj, nk);
  mhd_lfmxCenter.partition(all);
  mhd_lfmyCenter.redim(ni, nj, nk);
  mhd_lfmyCenter.partition(all);
  mhd_lfmzCenter.redim(ni, nj, nk);
  mhd_lfmzCenter.partition(all);
  
  Range I(0,ni-1), J(0,nj-1), K(0,nk-1);

  // This code was adapted from MHDInnerBoundaryInterface.C
  mhd_lfmxCenter(I,J,K) = 0.125*( mhd_lfmx(I,J,K) + mhd_lfmx(I,J,K+1) + mhd_lfmx(I+1,J,K+1) + mhd_lfmx(I+1,J,K) +
				  mhd_lfmx(I,J+1,K) + mhd_lfmx(I,J+1,K+1) + mhd_lfmx(I+1,J+1,K+1) + mhd_lfmx(I+1,J+1,K) );
  
  mhd_lfmyCenter(I,J,K) = 0.125*( mhd_lfmy(I,J,K) + mhd_lfmy(I,J,K+1) + mhd_lfmy(I+1,J,K+1) + mhd_lfmy(I+1,J,K) +
				  mhd_lfmy(I,J+1,K) + mhd_lfmy(I,J+1,K+1) + mhd_lfmy(I+1,J+1,K+1) + mhd_lfmy(I+1,J+1,K) );
  
  mhd_lfmzCenter(I,J,K) = 0.125*( mhd_lfmz(I,J,K) + mhd_lfmz(I,J,K+1) + mhd_lfmz(I+1,J,K+1) + mhd_lfmz(I+1,J,K) +
				  mhd_lfmz(I,J+1,K) + mhd_lfmz(I,J+1,K+1) + mhd_lfmz(I+1,J+1,K+1) + mhd_lfmz(I+1,J+1,K) );

  // Convert LFM grid coordinates at cell-centeres from cm to m
  mhd_lfmxCenter /= 100.0;
  mhd_lfmyCenter /= 100.0;
  mhd_lfmzCenter /= 100.0;

  // r1205 adds in density nudging to IM_Interface::bleed().  This
  // nuding requires the radius of the grid points each time bleed is
  // called.  Although it eats up a tiny bit of memory, lets
  // pre-compute the radius _once_ and partition the array to work
  // with LFM's distributed data.
  mhd_radius.redim(ni,nj,nk);
  mhd_radius.partition(all);
  mhd_radius = sqrt(mhd_lfmxCenter * mhd_lfmxCenter +
		mhd_lfmyCenter * mhd_lfmyCenter + 
		mhd_lfmzCenter * mhd_lfmzCenter);
  mhd_radius.partition(mhdGhostPartitioning);

/////////////////////////////////////////////////////////////////////////
//  // The following code is used to debug the cell-centers computation.
//
//    std::string filename;
//    int mype = Communication_Manager::My_Process_Number; 
//    if (mype == 0){
//      filename = "GRIDOUT.dat";
//    }
//    else{
//      filename = "garbage.dat";
//    }
//
//    ofstream GRIDOUT_out;
//    GRIDOUT_out.open(filename.c_str(), ios::out);
//
//    GRIDOUT_out << setw(5) << ni << "\t" << nj << "\t-999\t-999\t" << nk << endl << "  ";
//    GRIDOUT_out.precision(7);
//
//    const double PI = 3.14159265;
//
//    // Output PHI:
//    for (int k=0; k < nk; k++){
//      // phi = tan^{-1}(z/y)
//      GRIDOUT_out << scientific << setw(13) << atan2(mhd_lfmzCenter(10,10,k),mhd_lfmyCenter(10,10,k))*180.0/PI;
//      if (k % 5 == 4)
//	GRIDOUT_out << endl << "  ";
//      else
//	GRIDOUT_out << " \t";
//    }
//    GRIDOUT_out << endl << "  ";
//    // Output X2:
//    const double earthRadiusInM = 6.38e6;
//    for (int j=0; j < nj; j++){
//      for (int i=0; i < ni; i++){
//	// X2 == x
//	GRIDOUT_out << scientific << setw(13) << mhd_lfmxCenter(i,j,0)/earthRadiusInM;
//	if ((j+i*nj) % 5 == 4)
//	  GRIDOUT_out << endl << "  ";
//	else
//	  GRIDOUT_out << " \t";
//      }
//    }
//    GRIDOUT_out << endl << "  ";
//    for (int j=0; j < nj; j++){
//      for (int i=0; i < ni; i++){
//	GRIDOUT_out << scientific << setw(13) << sqrt( pow(mhd_lfmyCenter(i,j,0),2) + pow(mhd_lfmzCenter(i,j,0),2))/earthRadiusInM;
//	if ((j+i*nj) % 5 == 4)
//	  GRIDOUT_out << endl << "  ";
//	else
//	  GRIDOUT_out << " \t";
//      }
//    }
//    GRIDOUT_out << endl;
//    GRIDOUT_out.close();
//
//    Communication_Manager::Sync();
//
/////////////////////////////////////////////////////////////////////////
}

/********************************************************************//**
 * Overloaded Accumulate to handle DoubleArrays
 *
 * Note special care is taken to ensure appropriate P++ partitioning
 * (i.e. ghost cells) for data arrays.
 *
 * FIXME: This is probably extremely inefficient... How should we
 * handle converting between floatArray and doubleArray?
 ************************************************************************/ 
void IM_Interface::Accumulate(const double &dt, 
			      const doubleArray &density, const doubleArray &sound_speed,
			      const doubleArray &bi, const doubleArray &bj, const doubleArray &bk,
			      const doubleArray &vx, const doubleArray &vy, const doubleArray &vz)
{
  floatArray density_floatArray = density.convertTo_floatArray();
  density_floatArray.partition(mhdGhostPartitioning);

  floatArray sound_speed_floatArray = sound_speed.convertTo_floatArray();
  sound_speed_floatArray.partition(mhdGhostPartitioning);

  floatArray bi_floatArray = bi.convertTo_floatArray();
  bi_floatArray.partition(mhdGhostPartitioning);
  
  floatArray bj_floatArray = bj.convertTo_floatArray();
  bj_floatArray.partition(mhdGhostPartitioning);
  
  floatArray bk_floatArray = bk.convertTo_floatArray();
  bk_floatArray.partition(mhdGhostPartitioning);
  
  floatArray vx_floatArray = vx.convertTo_floatArray();
  vx_floatArray.partition(mhdGhostPartitioning);
  
  floatArray vy_floatArray = vy.convertTo_floatArray();
  vy_floatArray.partition(mhdGhostPartitioning);
  
  floatArray vz_floatArray = vz.convertTo_floatArray();
  vz_floatArray.partition(mhdGhostPartitioning);
  
  Accumulate(dt,
	     density_floatArray, sound_speed_floatArray,
	     bi_floatArray, bj_floatArray, bk_floatArray,
	     vx_floatArray, vy_floatArray, vz_floatArray);
}


/********************************************************************//**
 * Accumulate LFM Grid variables into accumulate variables that will
 * be sent to the RCM:
 *
 *    - density
 *    - sound_speed
 *    - bi
 *    - bj
 *    - bk
 *    - vx,vy,vz => magnitude of vx,yv,vz
 * 
 * Description of accumulate functionality:
 * 
 * This is a brief writeup of the LFM accumulate routine used to
 * average LFM values over a certain time to send to the RCM.  The
 * routine can be found in the file \f[$\mathtt{RCMout.F}$\f] in the
 * subroutine \f[$\mathtt{RCMaccumulate}$\f].  This routine was
 * developed by John Lyon.
 * 
 * Let's assume we want a time average of a quantity \f[$q$\f] (which
 * can be density, pressure, velocity, etc).  Let the LFM timestep at
 * step \f[$N$\f] be defined as \f[$dt_N$\f]. Let's say we have done
 * \f[$N$\f] steps (since the last call to the RCM) in accumulating,
 * the timestep weighed average is then
 * 
 * \f[\begin{equation}
 * <q>_N=\frac{\sum_{i=1}^N q_i dt_i}{\sum_{i=1}^N dt_i}=\frac{\sum_{i=1}^N q_i dt_i}{T_N}
 * \end{equation}\f]
 * 
 * where \f[$T_N=\sum_{i=1}^N dt_i$\f] and is the time elapsed. Now we
 * are at timestep \f[$N+1$\f] and want to compute a new average that
 * takes into account (1) and the new value, so we want to compute
 * 
 * \f[\begin{equation}\setcounter{equation}{2}
 * <q>_{N+1}=\frac{\sum_{i=1}^{N+1} q_i dt_i}{T_{N+1}}
 * \end{equation}\f]
 * 
 * Since we don't have the information for all the steps but we do
 * have (1) we can rewrite (2) as
 * 
 * \f[\begin{equation}\setcounter{equation}{3}
 * <q>_{N+1}=\frac{\sum_{i=1}^N q_i dt_i + q_{N+1}dt_{N+1}}{T_N+dt_{N+1}}
 * =\frac{<q>_N (T_N-dt_{N+1})}{T_{N+1}}+\frac{q_{N+1}dt_{N+1}}{T_{N+1}}
 * \end{equation}\f]
 * 
 * This also saves use from redoing the averaging at each timestep. 
 * 
 * In the code 
 * 
 * \f[\begin{eqnarray*}
 * \verb q_ac  &\equiv & <q>_{N+1} \\
 * \verb tau_rcm &\equiv & T_{N+1} \\
 * \verb saved &\equiv & T_{N+1}-dt \\
 * \verb delta &\equiv & \frac{dt}{T_{N+1}} \\
 * \end{eqnarray*}\f]
 * 
 * where \f[$dt$\f] is the current timestep (\f[$dt_{N+1}$\f]).
 * 
 * ------- ------- ------- ------- ------- ------- ------- ------- -------
 *
 * FIXME: This function can probably be optimized with minimal effort.
 * Note that each A++/P++ operation is O(n^3)!
 ************************************************************************/ 
void IM_Interface::Accumulate(const double &dt, 
			      const floatArray &density, const floatArray &sound_speed,
			      const floatArray &bi, const floatArray &bj, const floatArray &bk,
			      const floatArray &vx, const floatArray &vy, const floatArray &vz)
{
  tau_accumulate += dt;

  // Accumulate. 
  //
  // Note 1: Setting the accumulate variables this way is inefficient.
  // Especially in view of what we've learned about P++ on Kraken.
  // 
  // Note 2: We do not call updateGhostBoundaries here.  P++ should
  // handle the ghost boundaries just fine with this simple
  // arithmetic.  Slava thinks this should be fine, but suggested we
  // call updateGhostBoundaries if we appear to experience
  // underflow/overlow problems.

  accum.setDeltaTime(dt);
  accum.increment(density, AverageVariables::RHO);
  accum.increment(sound_speed*sound_speed*density/gamma, AverageVariables::PRESSURE);
  accum.increment(bi, AverageVariables::BI);
  accum.increment(bj, AverageVariables::BJ);
  accum.increment(bk, AverageVariables::BK);
  accum.increment(sqrt(vx*vx + vy*vy + vz*vz), AverageVariables::V);
}

/********************************************************************//**
 * Overloaded Bleed to handle DoubleArrays
 *
 * See rcm-ut.h for definition of bleed_rate and bleed_type_scheme. 
 *
 * Note special care is taken to ensure appropriate P++ partitioning
 * (i.e. ghost cells) for data arrays.
 *
 * FIXME: This is probably extremely inefficient... How should we
 * handle converting between floatArray and doubleArray?
 ************************************************************************/ 
void IM_Interface::Bleed(const float &bleed_rate, const int &bleed_type_scheme,
			 const double &dt, doubleArray &density, doubleArray &csound,
			 doubleArray &vx, doubleArray &vy, doubleArray &vz)
{
  // Convert to proper precision
  floatArray density_floatArray = density.convertTo_floatArray();
  density_floatArray.partition(mhdGhostPartitioning);

  floatArray csound_floatArray = csound.convertTo_floatArray();
  csound_floatArray.partition(mhdGhostPartitioning);

  floatArray vx_floatArray = vx.convertTo_floatArray();
  vx_floatArray.partition(mhdGhostPartitioning);

  floatArray vy_floatArray = vy.convertTo_floatArray();
  vy_floatArray.partition(mhdGhostPartitioning);

  floatArray vz_floatArray = vz.convertTo_floatArray();
  vz_floatArray.partition(mhdGhostPartitioning);

  // Bleed into temporary floatArrays
  Bleed(bleed_rate, bleed_type_scheme,
	dt, density_floatArray, csound_floatArray,
	vx_floatArray, vy_floatArray, vz_floatArray);

  // Copy data back into doubleArrays
  // FIXME:  Can we do this faster by only making changes where (im_rcm_mask > 0) ?
  for (int i=1; i < density.getLength(0); i++){
    for (int j=1; j < density.getLength(1); j++){
      for (int k=1; k < density.getLength(2); k++){
	density(i,j,k) = density_floatArray(i,j,k);
	csound(i,j,k) = csound_floatArray(i,j,k);
	vx(i,j,k) = vx_floatArray(i,j,k);
	vy(i,j,k) = vy_floatArray(i,j,k);
	vz(i,j,k) = vz_floatArray(i,j,k);
      }
    }
  }
}

/********************************************************************//**
 * Bleed arrays that have just been imported
 *
 * See rcm-ut.h for definition of bleed_rate and bleed_type_scheme. 
 *
 * Note:  Passing in velocities (vx,yv,vz) for testing of momentum fix
 *
 * FIXME: Make sure base of RHS and LHS doesn't cause bugs: InterComm
 * requires that im_density.setBase(0) The LFM uses density.setBase(1)
 ************************************************************************/ 
void IM_Interface::Bleed(const float &bleed_rate, const int &bleed_type_scheme,
			 const double &dt, floatArray &density, floatArray &csound,
			 floatArray &vx, floatArray &vy, floatArray &vz)
{
  // The current implementation was labeled "RCM_bleed_type_scheme == 2" in the OpenMP code.

  // rcm_delta_t = number of seconds between exchange
  const float dt_bleed = float(dt) / float(rcm_delta_t) * bleed_rate;

  // Note the strange conventions:
  //   * Bleed arrays are 0-based indexed and of dimension (ni,nj,nk).
  //   * LFM arrays are 1-based indexed and of dimension (ni+1,nj+1,nk+1),
  //     while meaningful data is only specified up to (ni,nj,nk).
  // So use Range objects to specify where the meaningful data resides:
  const int ni = im_density.getLength(0);
  const int nj = im_density.getLength(1);
  const int nk = im_density.getLength(2);
  Range I(0,ni-1), J(0,nj-1), K(0,nk-1);
  Range Ip1(1,ni), Jp1(1,nj), Kp1(1,nk);

  floatArray densityOld = density;

  where (im_rcm_mask == 1) {
    // FIXME: Here we WOULD use Bill Henshaw's copy function to copy the
    // import arrays into bleed arrays that are partitioned with ghost
    // cells.

    density(Ip1,Jp1,Kp1) += (im_density(I,J,K) - density(Ip1,Jp1,Kp1)) * dt_bleed;

    // pressure     p = 1/GAMMA * rho * c_s^2
    // sound speed  c_s = sqrt( GAMMA * p / rho )
    csound(Ip1,Jp1,Kp1) += (sqrt( gamma * im_pressure(I,J,K) / im_density(I,J,K) ) - csound(Ip1,Jp1,Kp1)) * dt_bleed;
  }
// polar cap density specification  
// set if to 1 to turn this loop on
#if 0
  elsewhere () {
    float scaleHeight = 2.0*6380.e5; // in Re, convert to cm
    float density_0 = 0.05 *1.0e6*1.67e-27/1.0e3; // in gms/cc;

    // Update density as a one-liner::
    //     polar_cap-density = rho_0 * e^(-r/h)
    //     density += (polar_cap_density - density) * dt_bleed; 

    density(Ip1,Jp1,Kp1) += ( density_0 - density(Ip1,Jp1,Kp1) ) 
                             * exp( -mhd_radius(I,J,K)  / scaleHeight ) * dt_bleed;
  }
#endif

#if 0
// set if to 1 to turn this loop on
  //Momentum fix, only update of the density goes up
  floatArray velocityNudge = min( 1, densityOld / density );
  vx *= velocityNudge;
  vy *= velocityNudge;
  vz *= velocityNudge;
#endif  
}


/********************************************************************//**
 * Initialize the intermediate grid global indices for parallel interpolation on local task.
 * WARNING:  This function allocates memory into igGlobalindices!!!!
 *
 * NOTE: Parallel interpolation is not fully implemented/tested!
 * Development on parallel interpolation stopped at r899 (July 14,
 * 2009).
 ************************************************************************/ 
void IM_Interface::igParallelInitInterp(int &igPoints, int *&igGlobalIndices)
{
  cerr << __FILE__ <<  " (Line " << __LINE__ << "):  " << __FUNCTION__ << "(...) Not Fully Implemented."
       << endl << "\tNon-sensical demo/test values set..." << endl;
  
  int mype = Communication_Manager::My_Process_Number; 

  if ( mype == 0) {
    igPoints = 2; 
    igGlobalIndices = new int [igPoints]; 
    igGlobalIndices[0] = 3; 
    igGlobalIndices[1] = 7;
  } else if (mype == 1) {
    igPoints = 3; 
    igGlobalIndices = new int [igPoints]; 
    igGlobalIndices[0] = 8; 
    igGlobalIndices[1] = 12;
    igGlobalIndices[2] = 15;
  } else if (mype == 2) {
    igPoints = 1; 
    igGlobalIndices = new int [igPoints]; 
    igGlobalIndices[0] = 10; 
  } else if (mype == 3) {
    igPoints = 1; 
    igGlobalIndices = new int [igPoints]; 
    igGlobalIndices[0] = 4; 
  } else if (mype == 4) {
    igPoints = 3; 
    igGlobalIndices = new int [igPoints]; 
    igGlobalIndices[0] = 0; 
    igGlobalIndices[1] = 6;
    igGlobalIndices[2] = 9;
  } else if (mype == 5) {
    igPoints = 2; 
    igGlobalIndices = new int [igPoints]; 
    igGlobalIndices[0] = 1; 
    igGlobalIndices[1] = 5;
  } else if (mype == 6) {
    igPoints = 1; 
    igGlobalIndices = new int [igPoints]; 
    igGlobalIndices[0] = 11;
  } else if (mype == 7) {
    igPoints = 3; 
    igGlobalIndices = new int [igPoints]; 
    igGlobalIndices[0] = 2; 
    igGlobalIndices[1] = 13;
    igGlobalIndices[2] = 14;
  }
}


/********************************************************************//**
 * Initialize the intermediate grid global indices for serial interpolation on head node.
 *
 * Requires:
 *   - LFM grid (member vars mhd_lfm{x,y,z}) in units of cm
 *   - RCM Intermediate Grid (member vars im_ig{x,y,z}) in units of Earth Radii.
 * 
 ************************************************************************/ 
void IM_Interface::igSerialInitInterp(const intArray &igDims)
{
  // Move the array to the head node
  Partitioning_Type P_zero(Range(0,0));

  floatArray Xlocal(mhd_lfmx);
  Xlocal.partition(P_zero);

  floatArray Ylocal(mhd_lfmy);
  Ylocal.partition(P_zero);

  floatArray Zlocal(mhd_lfmz);
  Zlocal.partition(P_zero);

  // The rest of this function operates on a single task.
  // However, P++ requires that *all* tasks execute all P++ code
  // (i.e. putting the following within `if mype == 0` would yield
  // errors).
  Range all;
  floatArray X2 = Xlocal(all, all, 0);
  floatArray Y2 = Ylocal(all, all, 0);
  floatArray Phi0 = atan2( Zlocal(10,10,all), Ylocal(10,10,all) );

  // I can't seem to make reshape or resize work (Arghh!!!), so I'll
  // just create a new 1-d array for further passing to Fortran
  // Phi.reshape(mhd->NK+1);   
  floatArray Phi(mhd_lfmz.getLength(2), P_zero);
  float pi = 4.0*atan(1.0); // atan(1) = pi/4
  for (int k=0; k < mhd_lfmz.getLength(2); k++){
    // Make sure to deal with periodic boundary:
    if (Phi0(0,0,k) >= 0){
      Phi(k) = Phi0(0,0,k);
    }
    else{
      Phi(k) = Phi0(0,0,k) + 2*pi;
    }
  }
  
  // Number of cells in LFM grid
  int lfmSize[3] = { mhd_lfmx.getLength(0)-1, mhd_lfmx.getLength(1)-1, mhd_lfmx.getLength(2)-1 };
  
  // Number of grid points in RCM Intermediate Grid
  int igSize[3] =  { igDims(0), igDims(1), igDims(2) };

  // Sync everybody up before we do the 1d interpolation.
  Communication_Manager::Sync();  

  // Calculate interpolation coefficients to be used in
  // IM_Interface::igSerialInterp(...).  This piece of FORTRAN code
  // must be executed only on the head task:
  int mype = Communication_Manager::My_Process_Number; 
  if (mype == 0){
    INTERP1D_INIT_WRAPPER(igSize,  im_igx.getDataPointer(), im_igy.getDataPointer(), im_igz.getDataPointer(),
			  lfmSize, X2.getDataPointer(),     Y2.getDataPointer(),     Phi.getDataPointer());      
  }

  // Must sync after computing interpolation coefficients or bad things will happen.
  Communication_Manager::Sync();  
}



/********************************************************************//**
 * Interpolate values from LFM grid onto the Intermediate Grid.
 * 
 * Note:  accum_b{x,y,z} are populated in IM_Interface::subtractDipole
 * 
 * Note: This should be able to operate in either serial or parallel.
 * However, development on parallel interpolation stopped at
 * r899 (July 14, 2009).
 ************************************************************************/ 
void IM_Interface::igInterp()
{
#if 0
  // Since parallel interpolation is only partially implemented, this should remain commented out!
  igParallelInterp(accum_density,  mhd_density);
  igParallelInterp(accum_pressure, mhd_pressure);
  igParallelInterp(accum_bx,       mhd_bx);
  igParallelInterp(accum_by,       mhd_by);
  igParallelInterp(accum_bz,       mhd_bz);
  igParallelInterp(accum_v,        mhd_v);
#endif

  // Note: Serial interpolation interpolates from the distributed
  // accumulate variables on the LFM grid onto the intermediate grid
  // arrays fa_mhd_* distributed *ON PROCESS 0 ONLY*.

  igSerialInterp(accum.getVariable(AverageVariables::RHO),  fa_mhd_density);
  igSerialInterp(accum.getVariable(AverageVariables::PRESSURE),  fa_mhd_pressure);
  igSerialInterp(accum_bx,       fa_mhd_bx);
  igSerialInterp(accum_by,       fa_mhd_by);
  igSerialInterp(accum_bz,       fa_mhd_bz);
  igSerialInterp(accum.getVariable(AverageVariables::V),  fa_mhd_v);
}


/********************************************************************//**
 * Interpolate values onto the Intermediate Grid on the local task, in parallel.
 *
 * NOTE: Parallel interpolation is not fully implemented/tested!
 * Development on parallel interpolation stopped at r899 (July 14,
 * 2009).
 ************************************************************************/ 
void IM_Interface::igParallelInterp(const floatArray &src, float *&dest)
{
  cerr << __FILE__ <<  " (Line " << __LINE__ << "):  " << __FUNCTION__ << "(...) Not Fully Implemented."
       << endl << "Non-sensical demo/test values set..." << endl;
   
  int mype = Communication_Manager::My_Process_Number; 
   
  if ( mype == 0) {
    dest[0] = 4; 
    dest[1] = 8;
  } else if (mype == 1) {
    dest[0] = 9; 
    dest[1] = 13;
    dest[2] = 16;
  } else if (mype == 2) {
    dest[0] = 11; 
  } else if (mype == 3) {
    dest[0] = 5; 
  } else if (mype == 4) {
    dest[0] = 1; 
    dest[1] = 7;
    dest[2] = 10;
  } else if (mype == 5) {
    dest[0] = 2; 
    dest[1] = 6;
  } else if (mype == 6) {
    dest[0] = 12; 
  } else if (mype == 7) {
    dest[0] = 3; 
    dest[1] = 14;
    dest[2] = 15;
  } 
}


/********************************************************************//**
 * Interpolate values onto the Intermediate Grid as a serial process on the head node
 *
 * Note: dest contains the interpolated values *ON PROCESS 0*.  This
 * function does not redistribute the source data.
 ************************************************************************/ 
void IM_Interface::igSerialInterp(const floatArray &src, floatArray &dest)
{
  // Move the array to the head node
  Partitioning_Type P_zero(Range(0,0));
  floatArray localSrc(src);
  localSrc.partition(P_zero);

  // Interpolate from src (on LFM grid) to dest (on intermediate grid)
  // This uses the interpolation coefficients calculated within
  // IM_IC_Interface::igSerialInitInterp(...).
  int mype = Communication_Manager::My_Process_Number; 

  if (mype == 0){
    int srcSize[3] = { src.getLength(0), src.getLength(1), src.getLength(2) };
    int destSize[3] = { ig_dimensions[0], ig_dimensions[1], ig_dimensions[2] }; 

    INTERP1D_WRAPPER(dest.getDataPointer(),
		     destSize,
		     localSrc.getDataPointer(),
		     srcSize);
  }
  Communication_Manager::Sync();
}


/********************************************************************//**
 * Subtract dipole from b{i,j,k} array & store results to the
 * bx,by,bz parameters.  
 * 
 * When called from IM_Interface & its derived classes, the b{x,y,z}
 * parameters should be the member variables accum_b{x,y,z}.  See
 * prepareExport for an example.
 *
 * \param[out] bx: Accumulate magnetic field with dipole removed on cell centers
 * \param[out] by: Accumulate magnetic field with dipole removed on cell centers
 * \param[out] bz: Accumulate magnetic field with dipole removed on cell centers
 * 
 * \note The subtle "const" after the parameter list: This function
 * does in fact mutate the member variables accum_b{x,y,z} by changing
 * the array pointed to by the P++ getDataPointer!  However, the
 * member variable accum.getVariable(AverageVariables::B{I,J,K}
 * remains untouched: We create a (memory inefficient) copy of
 * b{i,j,k}.  This is necessary because the user may want to look at
 * dipole-subtracted values (i.e. inside MHDManager::dump_step)
 * without mucking with the accumulate rate.
 *
 * ... So if we change accum_b{x,y,z}, why is this a const member
 * function?  Because MHDManager::dump_step writes IM Coupling data
 * from a const pointer to a const IM_Interface object (IM_Interface
 * const * const innerMag) and this was the only way I could make it
 * work!
 *
 * For details on why accum_b{x,y,z} can mutate even though
 * subtractDipole is const, see section 18.10 of "C++ FAQ Lite":
 * http://www.parashift.com/c++-faq-lite/const-correctness.html#faq-18.10
 ************************************************************************/ 
void IM_Interface::subtractDipole() const
{
  // Create a temporary copy of the accumulate arrays that we'll use
  // to subtract the dipole.  This way we don't muck with the real
  // accum.getVariable(AverageVariables::B{I,J,K} arrays.
  floatArray accum_bi_tmp;
  floatArray accum_bj_tmp;
  floatArray accum_bk_tmp;

  const int nip1 = accum.getVariable(AverageVariables::BI).getLength(0);
  const int njp1 = accum.getVariable(AverageVariables::BI).getLength(1);
  const int nkp1 = accum.getVariable(AverageVariables::BI).getLength(2);

  accum_bi_tmp.redim(nip1,njp1,nkp1);
  accum_bj_tmp.redim(nip1,njp1,nkp1);
  accum_bk_tmp.redim(nip1,njp1,nkp1);

  accum_bi_tmp.partition(mhdGhostPartitioning);
  accum_bj_tmp.partition(mhdGhostPartitioning);
  accum_bk_tmp.partition(mhdGhostPartitioning);

  accum_bi_tmp.setBase(0);
  accum_bj_tmp.setBase(0);
  accum_bk_tmp.setBase(0);

  accum_bi_tmp = accum.getVariable(AverageVariables::BI);
  accum_bj_tmp = accum.getVariable(AverageVariables::BJ);
  accum_bk_tmp = accum.getVariable(AverageVariables::BK);

#ifdef NEW_DIPOLE_SUBTRACTION
  // FIXME: This bit of code in the preprocessor flag
  // "NEW_DIPOLE_SUBTRACTION" should be configurable by a variable to
  // the INPUT1.xml file.

  // New dipole subtraction:
  // 1.  Add background field on cell faces:
  ADD_BACKGROUND_FIELD(accum_bi_tmp.getDataPointer(),
		       accum_bj_tmp.getDataPointer(),
		       accum_bk_tmp.getDataPointer());
  // 2.  Convert from cell faces (bi,bj,bk) to cell centers (bx,by,bz)
  BZZ(accum_bx.getDataPointer(),     accum_by.getDataPointer(),     accum_bz.getDataPointer(),
      accum_bi_tmp.getDataPointer(), accum_bj_tmp.getDataPointer(), accum_bk_tmp.getDataPointer());
  // 3.  Subtract Dipole using analytic form of dipole
  SUBTRACT_ANALYTIC_DIPOLE_AT_CELL_CENTERS(accum_bx.getDataPointer(),
					   accum_by.getDataPointer(),
					   accum_bz.getDataPointer());
#else
  // Standard Dipole Subtraction:
  // 1.  Add background field & subtract Dipole on cell faces
  ADD_BACK_AND_SUBTRACT_DIPOLE_AT_CELL_FACES(accum_bi_tmp.getDataPointer(),
					     accum_bj_tmp.getDataPointer(),
					     accum_bk_tmp.getDataPointer());
  // 2.  Convert from cell faces (input bi,bj,bk) to cell centers (output bx,by,bz)
  BZZ(accum_bx.getDataPointer(),     accum_by.getDataPointer(),     accum_bz.getDataPointer(),
      accum_bi_tmp.getDataPointer(), accum_bj_tmp.getDataPointer(), accum_bk_tmp.getDataPointer());
#endif
  // Note: There was some internal debate as to whether or not we need
  // to updateGhostBoundaries() on {accum_bx, accum_by, accum_bz}
  // after removing the dipole from bi,bj,bk.  The code seems to work
  // fine _without_ these calls, but be consider yourself warned, in
  // case there are unanticipated problems in the future!
}


/********************************************************************//**
 * Prepare MHD variables to be exported to Inner Magnetosphere.
 * Interpolate the following variables from the LFM grid onto the
 * Intermediate Grid and convert units:
 *   - Density
 *   - Pressure
 *   - Bx
 *   - By
 *   - Bz
 *   - |v|
 * 
 ************************************************************************/ 
void IM_Interface::prepareExport()
{
  subtractDipole();
/////////////////////////////////////////////////////////////////////////
//   // To debug the interpolation code, it can be helpful to set a
//   // variable on the LFM grid to a nice predictable value
//
//   for (int i=0; i < mhd_lfmx.getLength(0); i++){
//     for (int j=0; j < mhd_lfmx.getLength(1); j++){
//   	  for (int k=0; k < mhd_lfmx.getLength(2); k++){
//   	    accum_bx(i,j,k) = float(i);
//       }
//     }
//   }
/////////////////////////////////////////////////////////////////////////

  igInterp();

  // Convert export variables to Meters Kilograms Seconds (MKS) units
  fa_mhd_density  *= 1.0e3;  // Convert from g/cm^3 to kg/m^3
  fa_mhd_pressure *= 1.0e-1; // Convert from g/(cm s^2) to kg/(m s^2)
  fa_mhd_bx       *= 1.0e-4; // Convert from Gauss to Tesla
  fa_mhd_by       *= 1.0e-4;
  fa_mhd_bz       *= 1.0e-4;
  fa_mhd_v        *= 1.0e-2;  // Convert from cm/s to m/s

/////////////////////////////////////////////////////////////////////////
//   // The following bit of code was used for debugging.  It will write
//   // out the accumulate arrays on both the LFM & Intermediate Grids to
//   // file.  Plotting this data in a program like IDL or MATLAB can
//   // help verify that the interpolation from the LFM grid onto the RCM
//   // Intermediate Grid works properly.
//
//     int ni = mhd_lfmx.getLength(0);
//     int nj = mhd_lfmx.getLength(1);
//     int nk = mhd_lfmx.getLength(2);
//
//     int mype = Communication_Manager::My_Process_Number; 
//
//     std::stringstream filename_ss;
//     filename_ss << "accumulate_lfm_grid_" << mype << ".dat";
//     std::string filename = filename_ss.str();
//
//     ofstream accum_out;
//     accum_out.open(filename.c_str(), ios::out);
//
//     accum_out << ni << " " << nj << " " << nk << endl;
//     accum_out << "i j k clfmx lfmy lfmz density pressure bx by bz v" << endl;
//
//     for (int k=0; k < nk; k++){
//       for (int j=0; j < nj; j++){
//	 for (int i=0; i < ni; i++){
//	   accum_out << i << " " << j << " " << k << " " ;
//
//         // NOTE: accum_{density, pressure, v} are now moved to the
//         // accum class.  Try "accum.getVariable(AverageVariables::{{}}".
//
//	   accum_out << mhd_lfmx(i,j,k) << " " << mhd_lfmy(i,j,k) << " " << mhd_lfmz(i,j,k) << " " 
//		     << accum_density(i,j,k) << " "
//		     << accum_pressure(i,j,k) << " "
//		     << accum_bx(i,j,k) << " "
//		     << accum_by(i,j,k) << " "
//		     << accum_bz(i,j,k) << " "
//		     << accum_v(i,j,k) << endl;
//         }
//       }
//     }
//     accum_out.close();
//
//     Communication_Manager::Sync();
//
//     int num_procs = Communication_Manager::numberOfProcessors();
//     ni = im_igx.getLength(0)/num_procs;
//     nj = im_igy.getLength(0)/num_procs;
//     nk = im_igz.getLength(0)/num_procs;
//
//     if (mype == 0){
//       filename = "intermediate_lfm_grid.dat";
//     }
//     else{
//       filename = "garbage.dat";
//     }
//
//     ofstream ig_out;
//     ig_out.open(filename.c_str(), ios::out); 
//
//     ig_out << ni << " " << nj << " " << nk << endl;
//     ig_out << "i j k igx igy igz density pressure bx by bz v" << endl;
//
//     for (int i=0; i < ni; i++){
//       for (int j=0; j < nj; j++){
//         for (int k=0; k < nk; k++){
//	   int offset = k*ni*nj + j*ni + i;
//
//	   ig_out << i << " " << j << " " << k << " " ;
//	   ig_out << im_igx(i) << " " << im_igy(j) << " " << im_igz(k) << " "
//		  << fa_mhd_density(offset) << " "
//		  << fa_mhd_pressure(offset) << " "
//		  << fa_mhd_bx(offset) << " "
//		  << fa_mhd_by(offset) << " "
//		  << fa_mhd_bz(offset) << " "
//		  << fa_mhd_v(offset) << endl;
//         }
//       }
//     }
//     ig_out.close();
//
//     Communication_Manager::Sync();
//     Optimization_Manager::Exit_Virtual_Machine();
/////////////////////////////////////////////////////////////////////////
}


/********************************************************************//**
 * Post-process after export.  
 *
 * Note: resetting accumulate variables belongs in the driver program,
 * so we can reset _after_ dumping accumulate & bleed arrays to data
 * file.
 ************************************************************************/ 
void IM_Interface::processExport()
{
}


/********************************************************************//**
 * Prepare Inner Magnetosphere variables to be imported.  Must be called 
 * before the IC_Import statements.
 *
 * InterComm gets confused when Importing arrays that have ghost
 * cells.  This function removes ghost cells immediately before
 * importing.
 ************************************************************************/ 
void IM_Interface::prepareImport()
{
  int num_procs = Communication_Manager::numberOfProcessors();
  
  Partitioning_Type all(num_procs);

  im_density.partition(all);
  im_pressure.partition(all);
  im_rcm_mask.partition(all);
}

/********************************************************************//**
 * Process Inner Magnetosphere variables that were just imported,
 * including any unit conversions:
 *   - density
 *   - pressure
 *   - RCM Mask
 ************************************************************************/ 
void IM_Interface::processImport()
{
  // Note: the arrays get a band of garbage values if you do not call
  // typeArray::updateGhostBoundaries().  Make ghost boundaries get
  // updated!

  im_density.partition(mhdGhostPartitioning);
  im_density.updateGhostBoundaries();

  im_pressure.partition(mhdGhostPartitioning);
  im_pressure.updateGhostBoundaries();

  im_rcm_mask.partition(mhdGhostPartitioning);
  im_rcm_mask.updateGhostBoundaries();

  
/////////////////////////////////////////////////////////////////////////
//   // The following bit of code is used for debugging.  It will write
//   // out the accumulate arrays on both the LFM & Intermediate Grids to
//   // file.  Plotting this data in a program like IDL or MATLAB can
//   // help verify that the interpolation from the LFM grid onto the RCM
//   // Intermediate Grid works properly.
     
//     int ni = mhd_lfmx.getLength(0);
//     int nj = mhd_lfmx.getLength(1);
//     int nk = mhd_lfmx.getLength(2);
    
//     int mype = Communication_Manager::My_Process_Number; 

//     std::string filename;

//     if (mype == 0){
//       filename = "bleed_lfm.dat";
//     }
//     else{
//       filename = "garbage.dat";
//     }

//     ofstream bleed_out;
//     bleed_out.open(filename.c_str(), ios::out);

//     bleed_out << ni << " " << nj << " " << nk << endl;
//     bleed_out << "i j k lfmx lfmy lfmz im_density im_pressure im_mask lfm_density" << endl;

//     for (int i=0; i < ni; i++){
//       for (int j=0; j < nj; j++){
//         for (int k=0; k < nk; k++){
//   	bleed_out << i << " " << j << " " << k << " " ;
//   	bleed_out << mhd_lfmx(i,j,k)/6.38e8 << " " 
// 		  << mhd_lfmy(i,j,k)/6.38e8 << " " 
// 		  << mhd_lfmz(i,j,k)/6.38e8 << " " 
//   		  << im_density(i,j,k) << " "
//   		  << im_pressure(i,j,k) << " "
//   		  << im_rcm_mask(i,j,k) << " " 
// 		  << accum_density(i,j,k) << endl;
//         }
//       }
//     }
//     bleed_out.close();

//     Communication_Manager::Sync();
//     Optimization_Manager::Exit_Virtual_Machine();
/////////////////////////////////////////////////////////////////////////
  // Convert Import variables from Meters Kilograms Seconds (MKS) units to LFMs internal units
  im_density  /= 1.0e3;  // Convert from kg/m^3 to g/cm^3
  im_pressure /= 1.0e-1; // Convert from kg/(m s^2) to g/(cm s^2)
}


/********************************************************************//**
 * Reset the accumulate arrays to zero (i.e. at initialization and/or
 * after an Export to Inner Magnetosphere.
 ************************************************************************/ 
void IM_Interface::resetAccumulateArrays()
{
  accum.reset();

  tau_accumulate = 0.0;
}


/********************************************************************//**
 * Destructor . . . 
 ************************************************************************/ 
IM_Interface::~IM_Interface()
{
  SUBTRACT_DIPOLE_CLEANUP();
}
