#ifndef IM_INTERFACE_H
#define IM_INTERFACE_H

// std includes
#include <iostream>
#include <vector>

// 3rd party includes
#define PPLUSPLUS
#include <A++.h>

// LTR-specific includes
#include "Average.h"
#include "DateTime.h"

#ifdef LC_UNSC
#define  INTERP1D_INIT_WRAPPER interp1d_init_wrapper_
#define  INTERP1D_WRAPPER interp1d_wrapper_
#define INITMAG_DIP initmag_dip_
#define ADD_BACKGROUND_FIELD add_background_field_
#define SUBTRACT_ANALYTIC_DIPOLE_AT_CELL_CENTERS subtract_analytic_dipole_at_cell_centers_
#define ADD_BACK_AND_SUBTRACT_DIPOLE_AT_CELL_FACES add_back_and_subtract_dipole_at_cell_faces_
#define SUBTRACT_DIPOLE_CLEANUP subtract_dipole_cleanup_
#define BZZ bzz_
#elif LC_NOUNSC
#define  INTERP1D_INIT_WRAPPER interp1d_init_wrapper
#define  INTERP1D_WRAPPER interp1d_wrapper
#define INITMAG_DIP initmag_dip
#define ADD_BACKGROUND_FIELD add_background_field
#define SUBTRACT_ANALYTIC_DIPOLE_AT_CELL_CENTERS subtract_analytic_dipole_at_cell_centers
#define ADD_BACK_AND_SUBTRACT_DIPOLE_AT_CELL_FACES add_back_and_subtract_dipole_at_cell_faces
#define SUBTRACT_DIPOLE_CLEANUP subtract_dipole_cleanup
#define BZZ bzz
#endif

extern "C" void INTERP1D_INIT_WRAPPER(int *sizes, float *xrect, float *yrect, float *zrect,
                                 int *lfmSizes, float *x2, float *y2, float *phi);
extern "C" void INTERP1D_WRAPPER(float *data_out, int *sizes_out, float *data_in, int *sizes_in);
extern "C" void INITMAG_DIP();
extern "C" void ADD_BACKGROUND_FIELD(float *bi, float *bj, float *bk);
extern "C" void ADD_BACK_AND_SUBTRACT_DIPOLE_AT_CELL_FACES(float *bi, float *bj, float *bk);
extern "C" void SUBTRACT_ANALYTIC_DIPOLE_AT_CELL_CENTERS(float *bx, float *by, float *bz);
extern "C" void SUBTRACT_DIPOLE_CLEANUP();
extern "C" void BZZ(float *, float *, float *, float *, float *, float *);

/********************************************************************//**
 * Abstract base class that serves as a placeholder for variables to
 * be exchanged with the Inner Magnetosphere (e.g. RCM).
 * 
 * Note that actual communication between the codes is handled by
 * classes derived from this, such as IM_IC_Interface.
 *
 * \see IM_IC_Interface
 ************************************************************************/ 
class IM_Interface{
public:
  IM_Interface(const floatArray &lfmx, const floatArray &lfmy, const floatArray &lfmz, 
	       const Partitioning_Type &mhdPartitioning,
	       const float &inDipoleMoment, const float &inGamma,
	       const DateTime &start, const int &delta_t);

  virtual ~IM_Interface();

  virtual void Initialize() = 0;

  virtual void Export() = 0;
  virtual void Import() = 0;

  virtual void SendScalars(const char *name, const intArray &scalars) = 0;

  void Accumulate(const double &dt, 
		  const doubleArray &density, const doubleArray &sound_speed,
		  const doubleArray &bi, const doubleArray &bj, const doubleArray &bk,
		  const doubleArray &vx, const doubleArray &vy, const doubleArray &vz);
  void Accumulate(const double &dt, 
		  const floatArray &density, const floatArray &sound_speed,
		  const floatArray &bi, const floatArray &bj, const floatArray &bk,
		  const floatArray &vx, const floatArray &vy, const floatArray &vz);

  void Bleed(const float &bleed_rate, const int &bleed_type_scheme,
	     const double &dt, doubleArray &density, doubleArray &csound,
	     doubleArray &vx, doubleArray &vy, doubleArray &vz);
  void Bleed(const float &bleed_rate, const int &bleed_type_scheme,
	     const double &dt, floatArray  &density, floatArray  &csound,
	     floatArray &vx, floatArray &vy, floatArray &vz);

  void subtractDipole() const;

  void resetAccumulateArrays();

  /**\name Accessors
   *
   *  MHDManager needs the Accumulate & Bleed arrays in order to write
   *  to file the data necessary for LFM-RCM restarts.  These arrays
   *  are accessed via the following accessor functions.
   */
  //@{
  const DateTime & getCouplingStart() const { return rcmStartUT; }
  const int & getDeltaT() const { return rcm_delta_t; }
  const floatArray & getAccumulateDensity() const { return accum.getVariable(AverageVariables::RHO); }
  const floatArray & getAccumulatePressure() const { return accum.getVariable(AverageVariables::PRESSURE); }
  const floatArray & getAccumulateBi() const { return accum.getVariable(AverageVariables::BI); }
  const floatArray & getAccumulateBj() const { return accum.getVariable(AverageVariables::BJ); }
  const floatArray & getAccumulateBk() const { return accum.getVariable(AverageVariables::BK); }
  const floatArray & getAccumulateBx() const { return accum_bx; }
  const floatArray & getAccumulateBy() const { return accum_by; }
  const floatArray & getAccumulateBz() const { return accum_bz; }
  const floatArray & getAccumulateV() const  { return accum.getVariable(AverageVariables::V); }
  const floatArray & getBleedDensity() const { return im_density; }
  const floatArray & getBleedPressure() const { return im_pressure; }
  const intArray & getBleedMask() const { return im_rcm_mask; }
  //@}
  /**\name Setters
   *
   *  When restarting LFM-RCM, we need both the Accumulate & Bleed
   *  arrays from the previous exchange.  The MHDManager class reads
   *  the Accumulate & Bleed arrays from file and updates the
   *  IM_Interface object.
   */
  //@{
  void setAccumulateDensity(const floatArray &in_density) { accum.setVariable(in_density, AverageVariables::RHO); }
  void setAccumulatePressure(const floatArray &in_pressure) { accum.setVariable(in_pressure, AverageVariables::PRESSURE); }
  void setAccumulateBi(const floatArray &in_bi) { accum.setVariable(in_bi, AverageVariables::BI); }
  void setAccumulateBj(const floatArray &in_bj) { accum.setVariable(in_bj, AverageVariables::BJ); }
  void setAccumulateBk(const floatArray &in_bk) { accum.setVariable(in_bk, AverageVariables::BK); }
  void setAccumulateV(const floatArray &in_v)   { accum.setVariable(in_v,  AverageVariables::V); }
  void setBleedDensity(const floatArray &in_density) { im_density = in_density; }
  void setBleedPressure(const floatArray &in_pressure) { im_pressure = in_pressure; }
  void setBleedMask(const intArray &in_rcm_mask) { im_rcm_mask = in_rcm_mask; }
  //@}


  /** Indices for status scalars to be sent to inner magnetosphere
   *
   * Note: These must match up with the scalars on the receiving side
   * (in Inner Magnetosphere code).
   */
  enum IM_SCALARS {
    KILL_SIGNAL = 0, // 111=kil; 222=last exchange; else=keep running
    YEAR        = 1,
    MONTH       = 2,
    DAY         = 3,
    HOUR        = 4,
    MINUTE      = 5,
    SECOND      = 6,
    DELTA_T     = 7, // number of seconds between LFM-RCM exchange
    LABEL       = 8, // MHD timestep number
    NUMBER_OF_SCALARS = 9
  };
  // Scalars at index KILL_SIGNAL accept the following:
  enum IM_kill_signal{
    KILL_SIGNAL_CONTINUE = 0,
    KILL_SIGNAL_SHUTDOWN = 111,
    KILL_SIGNAL_LAST_EXCHANGE = 222
  };

protected:

  /// Warning:  parallel interpolation not fully implemented/tested!
  void igParallelInitInterp(int &igPoints, int *&igGlobalIndices);
  void igSerialInitInterp(const intArray &igDims);

  void igInterp();
  /// Warning:  parallel interpolation not fully implemented/tested!
  void igParallelInterp(const floatArray &src, float *&dest);
  void igSerialInterp(const floatArray &src, floatArray &dest);

  void prepareExport();
  void processExport();

  void prepareImport();
  void processImport();

  //FIXME:  should these be read from struct/common block mapping?
  float dipole_moment;
  float gamma;

  // Weight for nudging accumulate arrays
  double tau_accumulate;

  /**\name Time Information*/
  //@{
  DateTime rcmStartUT;
  DateTime rcmExchangeUT;
  int rcm_delta_t;
  //@}

  /**\name LFM MHD partitioning type (including ghost cells) */
  //@{
  Partitioning_Type mhdGhostPartitioning;
  //@}

  /**\name Local copies of LFM grid variables */
  //@{
  floatArray mhd_lfmx; // aka "lfmX"
  floatArray mhd_lfmy; // aka "lfmY"
  floatArray mhd_lfmz; // aka "lfmZ"
  //@}

  /**\name coordinates of LFM grid cell-centers */
  //@{
  floatArray mhd_lfmxCenter;  
  floatArray mhd_lfmyCenter;
  floatArray mhd_lfmzCenter;
  //@}
  /**\name radius of LFM grid points (at cell-centers) */
  //@{
  floatArray mhd_radius;
  //@}

  /**\name LFM "Accumulate" Variables */
  //@{
  // FIXME: IC_EndPointSet::registerByConn(...) works only with (float*), not floatArray::GetDataPointer()
  //        So, these are defined for parallel interpolation:
  float *mhd_density;  // AccumRho - pRhoAc
  float *mhd_pressure; // AccumP   - pPAc
  float *mhd_bx;       // AccumBx  - pBxAc
  float *mhd_by;       // AccumBy  - pByAc
  float *mhd_bz;       // AccumBz  - pBzAc
  float *mhd_v;        // AccumV   - pVAc

  /**\name Inner Magnetosphere Intermedate Grid */
  //@{
  //FIXME:  Use ig_dimensions within this class & remove references to intArray &igDims from function parameters.
  int ig_dimensions[3];
  floatArray im_igx;
  floatArray im_igy;
  floatArray im_igz;

  /**\name LFM "Accumulate" Arrays */
  //@{
  // FIXME:  IC_EndPointSet::registerArray(...) works only with floatArray's, not (float*).
  //         So, these are defined for serial interpolation on the head node.
  floatArray fa_mhd_density;
  floatArray fa_mhd_pressure;
  floatArray fa_mhd_bx;
  floatArray fa_mhd_by;
  floatArray fa_mhd_bz;
  floatArray fa_mhd_v;

  Average<floatArray> accum;
  floatArray accum_bx;
  floatArray accum_by;
  floatArray accum_bz;
  //@}

  /**\name Inner Magnetosphere "Bleed" arrays on LFM grid */
  //@{
  floatArray im_density;  // BleedRho  - pRhoBleed
  floatArray im_pressure; // BleedP    - pPBleed
  intArray im_rcm_mask;   //           - pMask
  //@}

private:
  void InitializeAccumulateArrays(const int &ni, const int &nj, const int &nk);
  void InitializeBleedArrays(const int &ni, const int &nj, const int &nk);
  void InitializeGridArrays(const int &ni, const int &nj, const int &nk);
};

#endif
