/**
 * \file MHDInnerBoundaryInterface.h
 * \brief MHDInnerBoundaryInterface declaration
 *
 * \class MHDInnerBoundaryInterface 
 * \brief Abstract base class that defines functions for communication with the MIX
 *
 *\author Slava Merkin (vgm at bu.edu) 
 *\since 08-2005
 *
 * Abstract base class that defines functions for communication
 * between an MHD code and the Ionosphere Coupler/Solver code. 
 *
 * The virtual functions visible to the outside world are 
 *   - MHDInnerBoundaryInterface()
 *   - Export()
 *   - Import(),
 *   - sendScalars()
 * These virtual functions must be defined in a derived class.
 *
 * This code implies that the MHD grid is a logically spherical grid
 * with the symmetry axis aligned with the SM x-axis (LFM definition).
 *
 * \see CPL_IC_Interface.h
 */
#ifndef INTERFACE_H
#define INTERFACE_H


#define PPLUSPLUS  //!< Compile a P++ code rather than A++
#include <A++.h>
#include "const.h"

class MHDInnerBoundaryInterface {
public:

  /**\brief Class constructor */
  MHDInnerBoundaryInterface(const doubleArray &x, const doubleArray &y, const doubleArray &z, 
			    const int &NI, const int &NJ, const int &NK);
  virtual ~MHDInnerBoundaryInterface() {};

  /**\brief Export the MHD FAC, density and sound speed to the MIX code*/
  virtual void Export(const doubleArray &bx, const doubleArray &by, const doubleArray &bz,
		      const doubleArray &rho, const doubleArray &cs) = 0;

  /**\brief Import potential and calculate electric field & velocity at the inner MHD boundary*/
  virtual void Import(doubleArray &eField_j, doubleArray &eField_k, doubleArray &velocity) = 0;

  /** \brief Broadcast scalar array (i.e. UT & kill signal) from MHD code. */
  virtual void sendScalars(const doubleArray &scalars) = 0;

protected:
  /**\brief Calculate FAC, density & sound speed in second shell before */
  void prepareExport(const doubleArray &bx, const doubleArray &by, const doubleArray &bz, 
		     const doubleArray &rho, const doubleArray &cs);

  /**\brief Calculate the potential electric field & the velocity at the inner boundary */
  void processImport(doubleArray &eField_j, doubleArray &eField_k, doubleArray &velocity);


  /**\name MHD grid dimension information */
  //@{
  const int ni;
  const int nj;
  const int nk;
  const int nim1;
  const int nip1;
  const int njm1;
  const int njp1;
  const int nkp1;
  const int nkm1;
  //@}

  doubleArray current, density, soundSpeed, potential;

private:
  /**\name Dipole field functions */
  //@{
  /**\brief X-component of a unit vector in dipole field direction */
  doubleArray bxNorm(const doubleArray &x, const doubleArray &y, const doubleArray &z);

  /**\brief Y-component of a unit vector in dipole field direction */
  doubleArray byNorm(const doubleArray &x, const doubleArray &y, const doubleArray &z);

  /**\brief Z-component of a unit vector in dipole field direction */
  doubleArray bzNorm(const doubleArray &x, const doubleArray &y, const doubleArray &z);

  /**\brief Dipole field magnitude */
  doubleArray bMag(const doubleArray &x, const doubleArray &y, const doubleArray &z);

  /**\brief Ratio of the dipole field magnitude at magnetosphere/ionosphere */
  doubleArray bion2bmag(const doubleArray &x, const doubleArray &y, const doubleArray &z);
  //@}

  /**\name MHD inner boundary condition functions */
  //@{
  /**\brief Get density and sound speed at the second shell vertices */
  doubleArray getVarSecondShell(const doubleArray &var);

  /**\brief Calculate the field-aligned current */
  void getParallelCurrent(const doubleArray &Bx, const doubleArray &By, const doubleArray &Bz);

  /**\brief Smooth the field-aligned current */
  void smoothParallelCurrent();

  /**\brief Calculate the electric field from the potential received
     from MIX */
  void getElectricField(doubleArray &eField_i, doubleArray &eField_j, doubleArray &eField_k);

  /**\brief Calculate the velocity from the electric field at the inner boundary */
  doubleArray getVelocity(const doubleArray &eField_i, const doubleArray &eField_j, const doubleArray &eField_k);
  //@}

  /**\name Vector manipulation functions */
  //@{
  doubleArray dot(const doubleArray &a, const doubleArray &b);
  doubleArray cross(const doubleArray &a, const doubleArray &b);
  doubleArray triple(const doubleArray &a, const doubleArray &b, const doubleArray &c);
  //@}

  /**\name Arrays dimensioned and assigned by the class constructor
   *  and then used by other functions of the class*/
  //@{
  doubleArray volume, xCenter, yCenter, zCenter, bxDot, byDot, bzDot;
  doubleArray x_iFaceCenter, y_iFaceCenter, z_iFaceCenter;
  doubleArray x_jFaceCenter, y_jFaceCenter, z_jFaceCenter;
  doubleArray x_kFaceCenter, y_kFaceCenter, z_kFaceCenter;
  doubleArray di, dj, dk, i_x, i_y, i_z, j_x, j_y, j_z, k_x, k_y, k_z;
  //@}

};

#endif
