#include "Integration.h"
#include <math.h>

/**
 * Integrate FX(x,y,z) over the line (xa,ya,za) to (xb,yb,zb).  This
 * subroutine does Gaussian integration with the first twelve Legendre
 * polynomials as the basis fuctions.  Abromowitz and Stegun page 916.
 */
float gaussianLineIntegral( float (*fx)(const float &x, const float &y, const float &z), 
			    const float &xa, const float &ya, const float &za,
			    const float &xb, const float &yb, const float &zb)
{
  // Positive zeros of 12th order Legendre polynomial
  static const float A[6] = {0.1252334085,  0.3678314989,  0.5873179542,
			     0.7699026741,  0.9041172563,  0.9815606342};
  
  // Gaussian Integration coefficients for a 12th order polynomial
  static const float WT[6] = {0.2491470458,  0.2334925365,  0.2031674267,
			      0.1600783285,  0.1069393259,  0.0471753363};

  const float dx = (xb-xa)/2.0;
  const float dy = (yb-ya)/2.0;
  const float dz = (zb-za)/2.0;
  const float xbar = (xa+xb)/2.0;
  const float ybar = (ya+yb)/2.0;
  const float zbar = (za+zb)/2.0;
  
  float sum = 0;

  for (int k=0; k < 6; k++){
    sum += WT[k] * (
		    fx(xbar+A[k]*dx, ybar+A[k]*dy, zbar+A[k]*dz) +
		    fx(xbar-A[k]*dx, ybar-A[k]*dy, zbar-A[k]*dz)
		    );
  }

  return 0.5*sum;
}


/******************************************************************************/

float vectorPotential::ax(const float &x, const float &y, const float &z)
{
  const float r = sqrt(x*x + y*y + z*z);
  const float s = sqrt( pow((x-XQ_OFF)*AQ,2) + (y*y+z*z)*BQ*BQ );
  if (s < SINNER)
    return ( (constants::GEOQMU/pow(r,2)) * (y/r) );
  else if (s < 1.0)
    return ( (constants::GEOQMU/pow(r,2)) * (y/r)*(1.0-s)/(1.0-SINNER) );
  else
    return 0.0;
}

float vectorPotential::ay(const float &x, const float &y, const float &z)
{
  const float r = sqrt(x*x + y*y + z*z);
  const float s = sqrt( pow((x-XQ_OFF)*AQ,2) + (y*y+z*z)*BQ*BQ );
  if (s < SINNER)
    return ( -(constants::GEOQMU/pow(r,2)) * (x/r) );
  else if (s < 1.0)
    return ( -(constants::GEOQMU/pow(r,2)) * (x/r)*(1.0-s)/(1.0-SINNER) );
  else if (s < SOUTTER) 
    return ( -BXQ*(z*COS_TILT-x*SIN_TILT)*(s-1.0)/(SOUTTER-1.0) );
  else
    return ( -BXQ*(z*COS_TILT-x*SIN_TILT) );
}


/*******************************************************************************
 * For N-S dipoles, return 0.  Otherwise, all 3 vector potentials need
 * to include dipole direction vector.
 ******************************************************************************/
float vectorPotential::az(const float &x, const float &y, const float &z)
{
  return 0.0;
}
