#include "MhdVars.h"

MhdVars::MhdVars(const int &n_i, const int &n_j, const int &n_k) : grid(n_i,n_j,n_k)
{
  rho.redim(grid.nip1,grid.njp1,grid.nkp1);
  rho = 2.0e-25;

  vx.redim(grid.nip1,grid.njp1,grid.nkp1);
  vx = 0.0;
  vy.redim(grid.nip1,grid.njp1,grid.nkp1);
  vy = 0.0;
  vz.redim(grid.nip1,grid.njp1,grid.nkp1);
  vz = 0.0;

  c.redim(grid.nip1,grid.njp1,grid.nkp1);
  c = 1.e5;
  
  bx.redim(grid.nip1,grid.njp1,grid.nkp1);
  bx = 0.0;
  by.redim(grid.nip1,grid.njp1,grid.nkp1);
  by = 0.0;
  bz.redim(grid.nip1,grid.njp1,grid.nkp1);
  bz = 0.0;

  bi.redim(grid.nip1,grid.njp1,grid.nkp1);
  bi = 0.0;
  bj.redim(grid.nip1,grid.njp1,grid.nkp1);
  bj = 0.0;
  bk.redim(grid.nip1,grid.njp1,grid.nkp1);
  bk = 0.0;

  ei.redim(grid.nip1,grid.njp1,grid.nkp1);
  ei = 0.0;
  ej.redim(grid.nip1,grid.njp1,grid.nkp1);
  ej = 0.0;
  ek.redim(grid.nip1,grid.njp1,grid.nkp1);
  ek = 0.0;

  setMagneticFlux();
  setMagneticField();
}

void MhdVars::setMagneticFlux()
{
  floatArray Ai(grid.ni, grid.njp1, grid.nkp1);
  floatArray Aj(grid.nip1, grid.nj, grid.nkp1);
  floatArray Ak(grid.nip1, grid.njp1, grid.nk);

  getVectorPotentials(Ai, Aj, Ak);
  computeMagneticFlux(Ai, Aj, Ak);
}

void MhdVars::getVectorPotentials(floatArray &Ai, floatArray &Aj, floatArray &Ak)
{
  integrateVectorPotentials(Ai, Aj, Ak);
  setVectorPotentialBoundary(Ai, Aj, Ak);
}

void MhdVars::integrateVectorPotentials(floatArray &Ai, floatArray &Aj, floatArray &Ak)
{
  float xijk,yijk,zijk;

  float xip1,yip1,zip1;;
  for (int k=0; k < grid.nkp1; k++){
    for (int j=0; j < grid.njp1; j++){
      for (int i=0; i < grid.ni; i++){
	xijk = grid.x(i,j,k);    xip1 = grid.x(i+1,j,k);
	yijk = grid.y(i,j,k);    yip1 = grid.y(i+1,j,k);
	zijk = grid.z(i,j,k);    zip1 = grid.z(i+1,j,k);

	Ai(i,j,k) = (
		     (xip1-xijk) * gaussianLineIntegral(vectorPotential::ax, xijk,yijk,zijk, xip1,yip1,zip1) +
		     (yip1-yijk) * gaussianLineIntegral(vectorPotential::ay, xijk,yijk,zijk, xip1,yip1,zip1) + 
		     (zip1-zijk) * gaussianLineIntegral(vectorPotential::az, xijk,yijk,zijk, xip1,yip1,zip1)
		     );
      }
    }
  }

  float xjp1,yjp1,zjp1;
  for (int k=0; k < grid.nkp1; k++){
    for (int j=0; j < grid.nj; j++){
      for (int i=0; i < grid.nip1; i++){
	xijk = grid.x(i,j,k);   xjp1 = grid.x(i,j+1,k);
	yijk = grid.y(i,j,k);   yjp1 = grid.y(i,j+1,k);
	zijk = grid.z(i,j,k);   zjp1 = grid.z(i,j+1,k);	              
	
	Aj(i,j,k) = (
		     (xjp1-xijk) * gaussianLineIntegral(vectorPotential::ax, xijk,yijk,zijk, xjp1,yjp1,zjp1) +
		     (yjp1-yijk) * gaussianLineIntegral(vectorPotential::ay, xijk,yijk,zijk, xjp1,yjp1,zjp1) + 
		     (zjp1-zijk) * gaussianLineIntegral(vectorPotential::az, xijk,yijk,zijk, xjp1,yjp1,zjp1) 
		     );
      }
    }
  }

  float xkp1,ykp1,zkp1;
  for (int k=0; k < grid.nk; k++){    
    for (int j=0; j < grid.njp1; j++){
      for (int i=0; i < grid.nip1; i++){
	xijk = grid.x(i,j,k);    xkp1 = grid.x(i,j,k+1);
	yijk = grid.y(i,j,k);    ykp1 = grid.y(i,j,k+1);
	zijk = grid.z(i,j,k);    zkp1 = grid.z(i,j,k+1);	              

	Ak(i,j,k) = (
		     (xkp1-xijk) * gaussianLineIntegral(vectorPotential::ax, xijk,yijk,zijk, xkp1,ykp1,zkp1) +
		     (ykp1-yijk) * gaussianLineIntegral(vectorPotential::ay, xijk,yijk,zijk, xkp1,ykp1,zkp1) + 
		     (zkp1-zijk) * gaussianLineIntegral(vectorPotential::az, xijk,yijk,zijk, xkp1,ykp1,zkp1) 
		     );
      }
    }
  }
}

/*******************************************************************************
 * Applies axis boundary conditions
 ******************************************************************************/
void MhdVars::setVectorPotentialBoundary(floatArray &Ai, floatArray &Aj, floatArray &Ak)
{
  for (int i=0; i < grid.ni; i++){    
    float ei_sum1 = 0.0;
    float ei_sum2 = 0.0;
    for (int k=0; k < grid.nk; k++){
      ei_sum1 += Ai(i, 0, k);
      ei_sum2 += Ai(i, grid.nj, k);
    }
    ei_sum1 /= float(grid.nk);
    ei_sum2 /= float(grid.nk);

    for (int k=0; k < grid.nkp1; k++){
      Ai(i,0,k)  = ei_sum1;
      Ai(i,grid.nj,k) = ei_sum1;
    }
  }
  
  for (int i=0; i < grid.nip1; i++){
    for (int k=0; k < grid.nk; k++){
      Ak(i,0,k)  = 0.0;
      Ak(i,grid.nj,k) = 0.0;
    }
  }
}


/*******************************************************************************
 * Use Stokes theorem to compute magnetic fluxes through the faces
 * (via line integrals of vector potential).
 ******************************************************************************/
void MhdVars::computeMagneticFlux(const floatArray &Ai, const floatArray &Aj, const floatArray &Ak)
{
  for (int k=0; k < grid.nk; k++){
    for (int j=0; j < grid.nj; j++){
      for (int i=0; i < grid.nip1; i++){
	bi(i,j,k) =    Aj(i,j,k) - Aj(i,j,k+1) + Ak(i,j+1,k) - Ak(i,j,k);
      }
    }
  }
  for (int k=0; k < grid.nk; k++){
    for (int j=0; j < grid.njp1; j++){
      for (int i=0; i < grid.ni; i++){
	bj(i,j,k) = -( Ai(i,j,k) - Ai(i,j,k+1) + Ak(i+1,j,k) - Ak(i,j,k) );
      }
    }
  }
  for (int k=0; k < grid.nkp1; k++){
    for (int j=0; j < grid.nj; j++){
      for (int i=0; i < grid.ni; i++){
	bk(i,j,k) =    Ai(i,j,k) - Ai(i,j+1,k) + Aj(i+1,j,k) - Aj(i,j,k);
      }
    }
  }
}

void MhdVars::setMagneticField()
{
  for (int k=0; k < grid.nkp1; k++){
    for (int j=0; j < grid.njp1; j++){
      for (int i=0; i < grid.nip1; i++){
	bx(i,j,k) = Dipole::bx( grid.x(i,j,k), grid.y(i,j,k), grid.z(i,j,k) );
	by(i,j,k) = Dipole::by( grid.x(i,j,k), grid.y(i,j,k), grid.z(i,j,k) );
	bz(i,j,k) = Dipole::bz( grid.x(i,j,k), grid.y(i,j,k), grid.z(i,j,k) );
      }
    }
  }
}
