/* FLoaT, DoubleArray, etc... */
#include "site-stuff.h"
/* "SWIND" namelist */
#include "swind.h"
extern WIND WIND;
/* "CONST" namelist */
#include "help.h"
extern HELPER HELP;
/* "GO" namelist */
#include "info.h"
extern INFO INFO;
/* "ION" namelist */
#include "ionconst.h"
extern IONCONST IONCONST;
/* "RINGS" namelist */
#include "meter.h"
extern METER METER;
/* "GO" namelist */
#include "run-time.h"
extern RTime RTIME;
/* "INPUT-RCM" namelist */
#include "rcm-ut.h"

#include <cassert>
#include <iostream>
#include "tinyxml.h"
#include "DateTime.h"
#include "param.h"

/***********************************************************************/

/// Setup default parameters in the C structs/FORTRAN common blocks
/**
 *  I grabbed these default values from the "NAMELISTS" function in
 *  the old init-fortran.F
 *
 *  @note - We assume that the C++ & FORTRAN compilers will default
 *  all variables to 0... So we only specify non-zero default
 *  variables in this function.
 */
void setDefaultParameterValues(void)
{
  int j;
#ifdef DEBUG_MODE_ON
  std::cout << "DEBUG:  process ( " << RTIME.local_proc << ") in xmlInputReader.C::setDefaultParameterValues(...)\n";
#endif
  /* We'll assume that the compiler defaults everything to 0.0 */
  /* GO */
  ORDER = 8;
  /* RINGS */
  AXFIX = 0.5;
  /* CONST */
  GAMMA = 1.66666666;
  CA = 1.1e8;
  CRNT = 0.3;
  PDMB = 3.0;
  EVMAX = 1.e4;
  BETMAX = 0.9;
  BDIF = 0.0;
  /* SWIND */
  SWFROMFILE = false;
  BX_ZERO = 0.0;
  BY_COEF = 0.0;
  BZ_COEF = 0.0;

  /* ION */
  IONMODEL = true;
  S107 = 100.0;
  PED0 = 5.0;

  PEDMIN = 0.5;
  HALLMIN = 0.0;  
  SIGMA_RATIO = 1.2;
  P_HARDY = 1.19;
  C_HARDY = 2.4;
  H_HARDY = 1.33;
  ITMAX = 400;
  ERRMAX = 2.e-3;
  NLOOP = 10.0;
  LOOPTRIES = 5;

  ITER = 40;

  /* Initialize NRINGS arrays */
  if (N_J < 25) {
    NRINGS = 2;
    for(j=0;j<N_J;j++) NAVER[j]=1;
    NAVER[0]=4;
    NAVER[1]=2;
  } else {
    NRINGS = 5;
    for(j=0;j<N_J;j++) NAVER[j]=1;
    NAVER[0]=8;
    NAVER[1]=4;
    NAVER[2]=4;
    NAVER[3]=2;
    NAVER[4]=2;
  }
}

/***********************************************************************/

/// Convert a string containing {t, true, 1, 1, false, 0} to a boolean
/**
 *  @param data_str - string containing what the boolean should be.
 *  @returns bool - based on the input data_str
 */
bool strToBool(char *data_str)
{
#ifdef DEBUG_MODE_ON
  std::cout << "DEBUG:  process ( " << RTIME.local_proc << ") in xmlInputReader.C::strToBool(...)\n";
#endif
  // Make contents of "data_str" upper case.
  for(int i = 0; data_str[i] != '\0'; i++){
    data_str[i] = toupper(data_str[i]);
  }
  if (  ( strcmp(data_str, "T") == 0) ||
	( strcmp(data_str, ".T.") == 0) ||
	( strcmp(data_str, "TRUE") == 0) ||
	( strcmp(data_str, ".TRUE.") == 0) ||
	( strcmp(data_str, "1") == 0) ){
    return true;
  }
  else if (  ( strcmp(data_str, "F") == 0) ||
	     ( strcmp(data_str, ".F.") == 0) ||
	     ( strcmp(data_str, "FALSE") == 0) ||
	     ( strcmp(data_str, ".FALSE.") == 0) ||
	     ( strcmp(data_str, "0") == 0) ){
    return false;
  }
  else{
    std::cerr << "Error setting \"SWFROMFILE\" in memory."
	      << "\n\t Expected \"true\" or \"false\", but got "
	      << "\"" << data_str << "\"."
	      << "\n\t Defaulting to \"false\".\n";
  }
  return false;
}

/***********************************************************************/

/// Convert whitespace-separated data string to an floating point array
/**
 *  @warning - This function assumes that data_array is already
 *             allocated (and large enough) to handle all the elements
 *             of data_str.
 *
 *  @param data_str - String containing a floating-point data array,
 *                    such as "1 2 3 4 5"
 *
 *  @param data_array - array to store data into (i.e. data_array[0] =
 *                      { 1.0, 2.0, 3.0, 4.0, 5.0 })
 *
 *  @returns int - the number of elements compied from data_str to
 *                 data_array (i.e. the size of data_array).
 */
int strToIntArray(char *data_str, int *data_array)
{
#ifdef DEBUG_MODE_ON
  std::cout << "DEBUG:  process ( " << RTIME.local_proc << ") in xmlInputReader.C::strToFloatArray(...)\n";
#endif
  int nElements = 0;

  char * token;
  token = strtok(data_str, " ");
  if (token == NULL){
    std::cerr << "Error reading data in array containing \"" 
	      << data_str << "\".";
  }
  while (token != NULL){
    data_array[nElements++] = atoi(token);
    token = strtok(NULL, " ");    
  }
  return nElements;
}

/***********************************************************************/

/// Read XML datafile and setup input parameters
/**
 *  Read LFM-para input parameters from an XML input file
 *
 *  @param in_xmlFilename - XML datafile to read
 *  @param in_ni_bin_d - input ni bins
 *  @param in_nj_bin_d - input nj bins
 *  @param out_HDFBASE_IN - prefix to HDF input files
 *  @param out_HDFBASE_OUT - prefix to HDF output files
 *  @param out_lion - not sure what this is....   <<<==== REMOVE & USE NAMELIST
 *  @param out_time_dump_interval -  dump interval (in seconds) for UTIO
 *  @param out_startDateTime - First timestep to load (UTIO)
 *  @param out_stopDateTime - Last timestep to save data for (UTIO)
 *  @param out_ion_exchange_frequency - How many seconds between MHD+ION exchanges (UTIO)
 *  @param out_use_intercomm - bool that determines if we use
 *                             InterComm (true) or File Exchanges
 *                             (false) for communication with MIX.
 *
 *  @fixme - Document how to write an XML input file?
 *  @fixme - Document what variables we set?
 */

bool xmlInputReader(const char * in_xmlFilename, const int in_ni_bin_d, const int in_nj_bin_d,
		    char *out_HDFBASE_IN, char *out_HDFBASE_OUT, 
		    int &out_lion,
		    float &out_time_dump_interval,
		    DateTime &out_startDateTime,
		    DateTime &out_stopDateTime,
		    float &out_ion_exchange_frequency,
		    struct RCMut &innerMagnetosphere,
		    bool &out_use_intercomm
		    )
{
#ifdef DEBUG_MODE_ON
  std::cout << "DEBUG:  process ( " << RTIME.local_proc << ") in xmlInputReader.C::xmlInputReader(...)\n";
#endif
  
  /* Set default parameters */
  NI_BIN = in_ni_bin_d;
  NJ_BIN = in_nj_bin_d;
  out_ion_exchange_frequency = 5.0;

  /* INPUT-RCM */
  // See "rcm-ut.h" for definitions of what these variables represent.
  //FIXME:  document default values on the wiki. 

  // when default year = 9999, the coupling should never begin (in theory)
  innerMagnetosphere.startYear = 9999; 
  innerMagnetosphere.startMonth = 0;
  innerMagnetosphere.startDay = 0;
  innerMagnetosphere.startHour = 0;
  innerMagnetosphere.startMinute = 0;
  innerMagnetosphere.startSecond =  0;

  // 60 second exchange time
  innerMagnetosphere.deltaT = 60;

  // Factor which determines what percentage of the RCM data should be averaged into the LFM
  innerMagnetosphere.bleed_rate  = 0.95;
  //  Type of bleeding
  innerMagnetosphere.RCM_bleed_type_scheme = 2;

  out_use_intercomm = true;
  setDefaultParameterValues();

  /* Local variables */
  int nNAVER = 0;  // number of elements in "NAVER" averaging array
  /* UTIO */
  int start_year = -1, start_month = -1, start_day = -1, start_hour = -1, start_minute = -1;
  float start_second = -1.0;
  int stop_year = -1, stop_month = -1, stop_day = -1, stop_hour = -1, stop_minute = -1;
  float stop_second = -1.0;

  /* Set default outputs */
  if (strcpy(out_HDFBASE_IN, "ChangeName") == NULL){
    std::cerr << "Error setting \"out_HDFBASE_IN\" in memory"
	      << "\n\t while attempting to set default name \"ChangeName\".\n";
    return false;
  }
  if (strcpy(out_HDFBASE_OUT, "ChangeName") == NULL){
    std::cerr << "Error setting \"out_HDFBASE_OUT\" in memory"
	      << "\n\t while attempting to set default name \"ChangeName\".\n";
    return false;
  }
  out_time_dump_interval = -1;
  

  /* Standard I/O */
  NSTART = -1;
  NSTOP = -1;
  NDUMP = -1;

  /* Open the XML file, read the data & set the variables. */
  TiXmlDocument doc(in_xmlFilename);
  assert(doc.LoadFile());

  TiXmlElement *modules = doc.FirstChildElement( "modules" );

  TiXmlNode *name_node, *meta_node, *data_node;
  TiXmlElement *name_element, *meta_element, *data_element;

  char *name_str, *data_str;

  /* Loop over all modules */  
  TiXmlNode *module = 0, *variable = 0;
  for (module = modules->FirstChild( "module" );
       module;
       module = module->NextSibling( "module" ) ){
    
#ifdef DEBUG_MODE_ON    
    std::cout << module->FirstChild("name")->ToElement()->GetText() << "\n";
#endif
    /* Loop over all variables */
    for (variable = module->FirstChild( "variable" );
	 variable;
	 variable = variable->NextSibling( "variable" ) ){

      /* Get the variable name */
      name_node = variable->FirstChild( "name" );
      if (name_node != NULL){
	name_element = name_node->ToElement();
	name_str = (char *) name_element->GetText();
      }

      /* Get the meta data */
      meta_node = variable->FirstChild( "meta" );
      if (meta_node != NULL){
#ifdef DEBUG_MODE_ON
	meta_element = meta_node->ToElement();
	const char *meta_str;
	meta_str = meta_element->GetText();
      std::cout << "\t" << name_str << "\t" << data_str << "\t-\t" << meta_str << "\n";
#endif
      }

      /* Get the actual data */
      data_node = variable->FirstChild( "data" );
      if (data_node != NULL){
	data_element = data_node->ToElement();
	data_str = (char *) data_element->GetText();
      }
      if ( (name_node != NULL) && (data_node != NULL) ){
	// Make contents of "name_str" upper case.
	for(int i = 0; name_str[i] != '\0'; i++){
	  name_str[i] = toupper(name_str[i]);
	}
	/***********
	 * "STEPS" *
	 ***********/
	if (strcmp(name_str, "NSTART") == 0){
	  NSTART = atoi(data_str);
	}
	else if (strcmp(name_str, "NSTOP") == 0){
	  NSTOP = atoi(data_str);
	}
	else if (strcmp(name_str, "NDUMP") == 0){
	  NDUMP = atoi(data_str);
	}
	/*****************
	 * "GO" namelist *
	 *****************/
	else if (strcmp(name_str, "ORDER") == 0){
	  ORDER = atoi(data_str);
	}
	else if (strcmp(name_str, "HDFBASE_IN") == 0){
	  if (strcpy(out_HDFBASE_IN, data_str) == NULL){
	    std::cerr << "Error setting \"out_HDFBASE_IN\" in memory"
		      << "\n\t while attempting to cpy "
		      << "\"" << data_str << "\".\n";
	    return false;
	  }
	}
	else if (strcmp(name_str, "HDFBASE_OUT") == 0){
	  if (strcpy(out_HDFBASE_OUT, data_str) == NULL){
	    std::cerr << "Error setting \"out_HDFBASE_OUT\" in memory"
		      << "\n\t while attempting to cpy "
		      << "\"" << data_str << "\".\n";
	    return false;
	  }
	}
	else if (strcmp(name_str, "RUN_NAME") == 0){
	  if (strcpy(NEW_RUN_NAME, data_str) == NULL){
	    std::cerr << "Error setting \"NEW_RUN_NAME\" in memory"
		      << "\n\t while attempting to cpy "
		      << "\"" << data_str << "\".\n";
	    return false;
	  }
	}
	else if (strcmp(name_str, "USE_INTERCOMM") == 0){
	  out_use_intercomm = strToBool(data_str);
	}
	/********
	 * UTIO *
	 ********/
	else if (strcmp(name_str, "START_YEAR") == 0){ 
	  start_year = atoi(data_str);
	}
	else if (strcmp(name_str, "START_MONTH") == 0){ 
	  start_month = atoi(data_str);
	}
	else if (strcmp(name_str, "START_DAY") == 0){ 
	  start_day = atoi(data_str);
	}
	else if (strcmp(name_str, "START_HOUR") == 0){ 
	  start_hour = atoi(data_str);
	}
	else if (strcmp(name_str, "START_MINUTE") == 0){ 
	  start_minute = atoi(data_str);
	}
	else if (strcmp(name_str, "START_SECOND") == 0){ 
	  start_second = atof(data_str);
	}
	else if (strcmp(name_str, "STOP_YEAR") == 0){ 
	  stop_year = atoi(data_str);
	}	
	else if (strcmp(name_str, "STOP_MONTH") == 0){ 
	  stop_month = atoi(data_str);
	}
	else if (strcmp(name_str, "STOP_DAY") == 0){ 
	  stop_day = atoi(data_str);
	}
	else if (strcmp(name_str, "STOP_HOUR") == 0){ 
	  stop_hour = atoi(data_str);
	}
	else if (strcmp(name_str, "STOP_MINUTE") == 0){ 
	  stop_minute = atoi(data_str);
	}
	else if (strcmp(name_str, "STOP_SECOND") == 0){ 
	  stop_second = atof(data_str);
	}
	else if (strcmp(name_str, "TIME_DUMP_INTERVAL") == 0){ 
	  out_time_dump_interval = atof(data_str);
	}
	else if (strcmp(name_str, "START_DATETIME") == 0){ 
	  int startDT[6];
	  int nStartDT = 0;
	  nStartDT = strToIntArray(data_str, startDT);
	  assert(nStartDT == 6);
	  start_year = startDT[0];
	  start_month = startDT[1];
	  start_day = startDT[2];
	  start_hour = startDT[3];
	  start_minute = startDT[4];
	  start_second = (float) startDT[5];
	}
	else if (strcmp(name_str, "STOP_DATETIME") == 0){ 
	  int stopDT[6];
	  int nStopDT = 0;
	  nStopDT = strToIntArray(data_str, stopDT);
	  assert(nStopDT == 6);	 
	  stop_year = stopDT[0];
	  stop_month = stopDT[1];
	  stop_day = stopDT[2];
	  stop_hour = stopDT[3];
	  stop_minute = stopDT[4];
	  stop_second = (float) stopDT[5];
	}
	/********************
	 * "RINGS" namelist *
	 ********************/
	else if (strcmp(name_str, "NRINGS") == 0){
	  NRINGS = atoi(data_str);
	}
	else if (strcmp(name_str, "NAVER") == 0){
	  nNAVER = strToIntArray(data_str, NAVER);
	}
	else if (strcmp(name_str, "NI_BIN") == 0){
	  NI_BIN = atoi(data_str);
	}
	else if (strcmp(name_str, "NJ_BIN") == 0){
	  NJ_BIN = atoi(data_str);
	}
	else if (strcmp(name_str, "AXFIX") == 0){
	  AXFIX = atof(data_str);
	}
	/********************
	 * "CONST" namelist *
	 ********************/
	else if (strcmp(name_str, "GAMMA") == 0){
	  GAMMA = atof(data_str);
	}
	else if (strcmp(name_str, "CA") == 0){
	  CA = atof(data_str);
	}
	else if (strcmp(name_str, "CRNT") == 0){
	  CRNT = atof(data_str);
	}
	else if (strcmp(name_str, "PDMB") == 0){
	  PDMB = atof(data_str);
	}
	else if (strcmp(name_str, "EVMAX") == 0){
	  EVMAX = atof(data_str);
	}
	else if (strcmp(name_str, "BETMAX") == 0){
	  BETMAX = atof(data_str);
	}
	else if (strcmp(name_str, "BDIF") == 0){
	  BDIF = atof(data_str);
	}
	/********************
	 * "SWIND" namelist *
	 ********************/
	else if (strcmp(name_str, "SWFROMFILE") == 0){
	  SWFROMFILE = strToBool(data_str);
	}
	else if (strcmp(name_str, "UT0") == 0){
	  UT0 = atof(data_str);
	}
	else if (strcmp(name_str, "BX_ZERO") == 0){
	  BX_ZERO = atof(data_str);
	  if (BX_ZERO != 0){
	    //FIXME: refs issues #9, #143, #351
	    std::cout << "WARNING: BX_ZERO is not working properly at the moment.  Forcing BX_ZERO=0.0" << std::endl;
	    std::cerr << "WARNING: BX_ZERO is not working properly at the moment.  Forcing BX_ZERO=0.0" << std::endl;
	  }
	}
	else if (strcmp(name_str, "BY_COEF") == 0){
	  BY_COEF = atof(data_str);
	}
	else if (strcmp(name_str, "BZ_COEF") == 0){
	  BZ_COEF = atof(data_str);
	}
	/******************
	 * "ION" namelist *
	 ******************/
	else if (strcmp(name_str, "S107") == 0){
	  S107 = atof(data_str);
	}
	else if (strcmp(name_str, "PEDMIN") == 0){
	  PEDMIN = atof(data_str);
	}
	else if (strcmp(name_str, "HALLMIN") == 0){
	  HALLMIN = atof(data_str);
	}
	else if (strcmp(name_str, "SIGMA_RATIO") == 0){
	  SIGMA_RATIO = atof(data_str);
	}
	else if (strcmp(name_str, "PED0") == 0){
	  PED0 = atof(data_str);
	}
	else if (strcmp(name_str, "P_HARDY") == 0){
	  P_HARDY = atof(data_str);
	}
	else if (strcmp(name_str, "C_HARDY") == 0){
	  C_HARDY = atof(data_str);
	}
	else if (strcmp(name_str, "H_HARDY") == 0){
	  H_HARDY = atof(data_str);
	}
	else if (strcmp(name_str, "ITER") == 0){
	  ITER = atoi(data_str);
	}
	else if (strcmp(name_str, "ITMAX") == 0){
	  ITMAX = atoi(data_str);
	}
	else if (strcmp(name_str, "ERRMAX") == 0){
	  ERRMAX = atof(data_str);
	}
	else if (strcmp(name_str, "IONMODEL") == 0){
	  IONMODEL = strToBool(data_str);
	}
	else if (strcmp(name_str, "ION_EXCHANGE_FREQUENCY") == 0){
	  out_ion_exchange_frequency = atof(data_str);
	}
	else if (strcmp(name_str, "NLOOP") == 0){
	  NLOOP = atoi(data_str);
	}
	else if (strcmp(name_str, "LOOPTRIES") == 0){
	  LOOPTRIES = atoi(data_str);
	}
	/************************
	 * "INPUT-RCM" namelist *
	 ************************/
	// RCM start date/time (YYYY-MM-DD HH:MM:SS)
	// for LFM-RCM coupling.
	else if (strcmp(name_str, "RCM_START_DATETIME") == 0){ 
	  int imDateTime[6];
	  int nImDateTime = 0;
	  nImDateTime = strToIntArray(data_str, imDateTime);
	  assert(nImDateTime == 6);
	  innerMagnetosphere.startYear   = imDateTime[0];
	  innerMagnetosphere.startMonth  = imDateTime[1];
	  innerMagnetosphere.startDay    = imDateTime[2];
	  innerMagnetosphere.startHour   = imDateTime[3];
	  innerMagnetosphere.startMinute = imDateTime[4];
	  innerMagnetosphere.startSecond = imDateTime[5];
	}
	// Delta T, Restart, Bleed rate, bleed scheme
	else if (strcmp(name_str, "RCM_DELTA_T") == 0){
	  innerMagnetosphere.deltaT = atof(data_str);
	}
	else if (strcmp(name_str, "RCM_BLEED_RATE") == 0){
	  innerMagnetosphere.bleed_rate = atof(data_str);
	}
	else if (strcmp(name_str, "RCM_BLEED_TYPE_SCHEME") == 0){
	  innerMagnetosphere.RCM_bleed_type_scheme = atoi(data_str);
	}
	else {
	  std::cerr << "*** Unknown variable \"" << name_str << "\".\n";
	}
      }
    }
  }

  // Setup the DateTime objects
  if ( ( (start_year < 0) && (start_month < 0) && (start_day < 0) && (start_hour < 0) && (stop_minute < 0) ) ||
       ( (stop_year < 0) && (stop_month < 0) && (stop_day < 0) && (stop_hour < 0) && (stop_minute < 0) ) ){
    out_startDateTime = DateTime(0);
    out_stopDateTime = DateTime(0);
  }
  else{
    if (start_year < 0){
      std::cerr << "Warning: start_year not specified.  Defaulting to start_year=0\n";
      start_year = 0;
    }
    if (start_month < 0){
      std::cerr << "Warning: start_month not specified.  Defaulting to start_month=0\n";
      start_month = 0;
    }
    if (start_day < 0){
      std::cerr << "Warning: start_day not specified.  Defaulting to start_day=0\n";
      start_day = 0;
    }
    if (start_hour < 0){
      std::cerr << "Warning: start_hour not specified.  Defaulting to start_hour=0\n";
      start_hour = 0;
    }
    if (start_minute < 0){
      std::cerr << "Warning: start_minute not specified.  Defaulting to start_minute=0\n";
      start_minute = 0;
    }
    if (start_second < 0){
      std::cerr << "Warning: start_second not specified.  Defaulting to start_second=0\n";
      start_second = 0.0;
    }
    if (stop_year < 0){
      std::cerr << "Warning: stop_year not specified.  Defaulting to stop_year=0\n";
      stop_year = 0;
    }
    if (stop_month < 0){
      std::cerr << "Warning: stop_month not specified.  Defaulting to stop_month=0\n";
      stop_month = 0;
    }
    if (stop_day < 0){
      std::cerr << "Warning: stop_day not specified.  Defaulting to stop_day=0\n";
      stop_day = 0;
    }
    if (stop_hour < 0){
      std::cerr << "Warning: stop_hour not specified.  Defaulting to stop_hour=0\n";
      stop_hour = 0;
    }
    if (stop_minute < 0){
      std::cerr << "Warning: stop_minute not specified.  Defaulting to stop_minute=0\n";
      stop_minute = 0;
    }
    if (stop_second < 0){
      std::cerr << "Warning: stop_second not specified.  Defaulting to stop_second=0\n";
      stop_second = 0.0;
    }

    out_startDateTime = DateTime( start_year, start_month, start_day, start_hour, start_minute, start_second );
    out_stopDateTime = DateTime( stop_year, stop_month, stop_day, stop_hour, stop_minute, stop_second );

    if (out_startDateTime > out_stopDateTime){
      std::cerr << "ERROR:  Start date & time occurs AFTER stop date & time!\n";
      std::cout << "Start: \n\t\t" << out_startDateTime << "\n";
      std::cout << "Stop: \n\t\t" << out_stopDateTime << "\n";
      assert (out_startDateTime < out_stopDateTime);
      return false;
    }
  }

#ifdef DEBUG_MODE_ON
  /* Check everything by printing to stdout. */
  std::cout << std::endl;
  std::cout << "I read the following variables into memory:\n";
  std::cout << "\tGO namelist \n"
	    << "\t\tNSTART = " <<  NSTART << "\n"
	    << "\t\tNSTOP = " << NSTOP << "\n"
	    << "\t\tNDUMP = " << NDUMP << "\n"
	    << "\t\tORDER = " << ORDER << "\n"
	    << "\t\tout_HDFBASE_IN = " << out_HDFBASE_IN << "\n"
	    << "\t\tout_HDFBASE_OUT = " << out_HDFBASE_OUT << "\n"
	    << "\t\tNEW_RUN_NAME = " << NEW_RUN_NAME << "\n";

  std::cout << "\tUTIO\n"
	    << "\t\t" << out_startDateTime << "\n"
	    << "\t\t" << out_stopDateTime << "\n"
	    << "\t\t" << out_time_dump_interval << "\n";

  std::cout << "\tNRINGS namelist \n"
	    << "\t\tNRINGS = " << NRINGS << "\n"
	    << "\t\tNAVER = ";
  for (int i = 0; i < nNAVER; i++)
    std::cout << NAVER[i] << " ";
  std::cout << "\n"
	    << "\t\tNI_BIN = " << "" << NI_BIN << "\n"
	    << "\t\tNJ_BIN = " << "" << NJ_BIN << "\n"
	    << "\t\tAXFIX = " << "" << AXFIX << "\n";

  std::cout << "\tCONST namelist \n"
	    << "\t\tGAMMA = " << GAMMA << "\n"
	    << "\t\tCA = " << CA << "\n"
	    << "\t\tCRNT = " << CRNT << "\n"
	    << "\t\tPDMB = " << PDMB << "\n"
	    << "\t\tEVMAX = " << EVMAX << "\n"
	    << "\t\tBETMAX = " << BETMAX << "\n"
	    << "\t\tBDIF = " << BDIF << "\n";
	  
  std::cout << "\tSWIND namelist\n"
	    << "\t\tSWFROMFILE = " << SWFROMFILE << "\n"
	    << "\t\tUT0 = " << UT0 << "\n"
	    << "\t\tBX_ZERO = " << BX_ZERO << "\n"
	    << "\t\tBY_COEF = " << BY_COEF << "\n"
	    << "\t\tBZ_COEF = " << BZ_COEF << "\n";

  std::cout << "\tION namelist\n"
	    << "\t\tS107 = " << S107 << "\n"
	    << "\t\tPEDMIN = " << PEDMIN << "\n"
	    << "\t\tHALLMIN = " << HALLMIN << "\n"
	    << "\t\tSIGMA_RATIO = " << SIGMA_RATIO << "\n"
	    << "\t\tPED0 = " << PED0 << "\n"
	    << "\t\tP_HARDY = " << P_HARDY << "\n"
	    << "\t\tC_HARDY = " << C_HARDY << "\n"
	    << "\t\tH_HARDY = " << H_HARDY << "\n"
	    << "\t\tITER = " << ITER << "\n"
	    << "\t\tITMAX = " << ITMAX << "\n"
	    << "\t\tERRMAX = " << ERRMAX << "\n"
	    << "\t\tIONMODEL = " << IONMODEL << "\n"
	    << "\t\tNLOOP = " << NLOOP << "\n"
	    << "\t\tLOOPTRIES = " << LOOPTRIES << "\n";  
#endif

  if ( ( (NSTART != NSTOP) && (NSTOP != NDUMP) ) &&
       ( (out_startDateTime.getMJD() > 0.0) && (out_stopDateTime.getMJD() > 0.0) ) ){

    std::cerr << "ERROR reading \"" << in_xmlFilename << "\":\n"
	      << "       Both Timestep information (NSTART/NSTOP/NDUMP) and UTIO information (START_DATETIME & STOP_DATETIME) are specified.\n"
	      << "       You must set EITHER timestep information OR UTIO information, but not both!\n";
    return false;
  }

  /* Set output variables */ 
  // wtf is this???
  //out_lion  = NION;
  return true;
}
