#ifndef _IO_TEST_H_
#define _IO_TEST_H_

// std includes
#include <sys/stat.h>

// 3rd-party includes:
#include <A++.h>

// define resolution info:
#include "param.h"

// Custom includes:
#include "MHD.h"
#include "ION.h"

#include "MHDManager.h"
#include "UTMHD.h"
#include "TimestepMHD.h"

#include "IonIO.h"
#include "UTIonIO.h"
#include "TimestepIonIO.h"

// FIXME:    expand tests to check more variables (especially the grid)
// WARNING:  Several FORTRAN functions are treated as mock objects to reduce the number of dependencies.
//           see mock.F and mock_C++.C for more info. . . 

class TestSuite : public CxxTest::TestSuite
{
private:
  // Does a file exist on the disk?
  // inspired by http://www.physicsforums.com/showthread.php?t=66681
  // FIXME:  "stat" is platform-dependant!  (although it's probably implemented on all Linux platforms)
  bool doesFileExist( char *filename )
  {
    bool status = false;
    struct stat buffer;
    stat( filename, &buffer ) == 0 ? status = true : status = false;   
    return status;
  }

public:

  // FIXME: Should put common functions calls in one of:
  //  * setUp()/tearDown()
  //  * createSuite()/destroySuite()
  //  * setUpWorld()/tearDownWorld()
  // see the CxxTest documentation (4.4.1, 4.7) for more info.

  void testDoesFileNotExist( )
  {
    TS_ASSERT( doesFileExist("blah") == false );
  }
  
  void testReadIon( void )
  {
    /************* Start setUp ***************/
    MHD mhd = MHD(N_I,N_J,N_K,N_ORDER);

    mhd.initial(0,0,0);

    int NION_Shell = 2;
    
    //
    // The following line of code taken from MSPHERE.C on 10/29/2008
    // Sure, it's a great use of the C ternary "?:" operator 
    // but the extra **TWO** lines of code for a standard if(){...} else{}
    // might help with readability!
    // 
    (N_ORDER/2 - 1) > 2  ? NION_Shell = N_ORDER/2-1 : NION_Shell = 2;

    Range ion_procs(0,0);
    ION ion = ION(NION_Shell,N_J,N_K,ion_procs, N_ORDER, NION_I, NION_J);    
    ion.initial(0);

    /************* end setUp ***************/

    MHDManager *magnetosphereManager = new TimestepMHD(100000, 100000, 0);
    IonIO *ionosphere_IO = new TimestepIonIO((TimestepMHD *) magnetosphereManager);

    ionosphere_IO->read_step("TimestepRun",&ion);

    // Check potential in northern hemisphere
    TS_ASSERT_DELTA(ion.Phi_N(10,0),      0.0000, 1e-4);
    TS_ASSERT_DELTA(ion.Phi_N(10,1),      0.0001, 1e-4);
    TS_ASSERT_DELTA(ion.Phi_N(10,2),  -2141.4238, 1e-4);
    TS_ASSERT_DELTA(ion.Phi_N(10,3),  -5148.5122, 1e-4);
    TS_ASSERT_DELTA(ion.Phi_N(10,4),  -9444.2539, 1e-4);
    TS_ASSERT_DELTA(ion.Phi_N(10,5), -15318.5136, 1e-4);
    TS_ASSERT_DELTA(ion.Phi_N(10,6), -22670.7890, 1e-4);
    TS_ASSERT_DELTA(ion.Phi_N(10,7), -29628.7128, 1e-4);
    TS_ASSERT_DELTA(ion.Phi_N(10,8), -31846.3027, 1e-4);
    TS_ASSERT_DELTA(ion.Phi_N(10,9), -25730.5957, 1e-4);

    // Check potential in southern hemisphere
    TS_ASSERT_DELTA(ion.Phi_S(10,0),      0.0000, 1e-4);
    TS_ASSERT_DELTA(ion.Phi_S(10,1),      0.0001, 1e-4);
    TS_ASSERT_DELTA(ion.Phi_S(10,2),  -2151.6125, 1e-4);
    TS_ASSERT_DELTA(ion.Phi_S(10,3),  -5138.8930, 1e-4);
    TS_ASSERT_DELTA(ion.Phi_S(10,4),  -9433.0273, 1e-4);
    TS_ASSERT_DELTA(ion.Phi_S(10,5), -15374.6884, 1e-4);
    TS_ASSERT_DELTA(ion.Phi_S(10,6), -22731.8613, 1e-4);
    TS_ASSERT_DELTA(ion.Phi_S(10,7), -29499.5097, 1e-4);
    TS_ASSERT_DELTA(ion.Phi_S(10,8), -31465.4179, 1e-4);
    TS_ASSERT_DELTA(ion.Phi_S(10,9), -25278.6464, 1e-4);

    ionosphere_IO->dump_step("TimestepRun",&ion);
  }

  void testReadWriteIon( void )
  {
    /************* Start setUp ***************/
    MHD mhd = MHD(N_I,N_J,N_K,N_ORDER);

    mhd.initial(0,0,0);

    int NION_Shell = 2;
    (N_ORDER/2 - 1) > 2  ? NION_Shell = N_ORDER/2-1 : NION_Shell = 2;

    Range ion_procs(0,0);
    ION ion = ION(NION_Shell,N_J,N_K,ion_procs, N_ORDER, NION_I, NION_J);    
    ion.initial(0);

    /************* end setUp ***************/

    MHDManager *magnetosphereManager = new TimestepMHD(100000, 100000, 0);
    IonIO *ionosphere_IO = new TimestepIonIO((TimestepMHD *) magnetosphereManager);

    int status;
    status =  ionosphere_IO->read_step("TimestepRun",&ion);
    TS_ASSERT_EQUALS(status, 0);

    status =  ionosphere_IO->dump_step("TimestepRun",&ion);
    TS_ASSERT_EQUALS(status, 0);

    TS_ASSERT( doesFileExist("TimestepRun_ion_0100000.dmp") );
  }

  void testReadMHD( void )
  {
    /************* Start setUp ***************/
    MHD mhd = MHD(N_I,N_J,N_K,N_ORDER);

    mhd.initial(0,0,0);

    /************* end setUp ***************/

    MHDManager *magnetosphereManager = new TimestepMHD(100000, 100000, 0);

    TS_WARN("TimestepMHD::read_Data() has issues on certain systems.  Skipping test.");
    //int retVal = magnetosphereManager->read_step("TimestepRun", &mhd);

    //TS_ASSERT_EQUALS(retVal, 0);

    /*
    // Check a few values of Bz
    TS_ASSERT_DELTA(1e6*mhd.Bz(10,10,10),     0.0000000, 1e-7);
    TS_ASSERT_DELTA(1e6*mhd.Bz(10,10,11),    31.4336175, 1e-7);
    TS_ASSERT_DELTA(1e6*mhd.Bz(10,10,12),  -134.7623328, 1e-7);
    TS_ASSERT_DELTA(1e6*mhd.Bz(10,10,13),   939.8963885, 1e-7);
    TS_ASSERT_DELTA(1e6*mhd.Bz(10,10,14),     0.0000000, 1e-7);
    TS_ASSERT_DELTA(1e6*mhd.Bz(10,10,15),   -20.8199435, 1e-7);
    TS_ASSERT_DELTA(1e6*mhd.Bz(10,10,16),   -49.9694942, 1e-7);
    TS_ASSERT_DELTA(1e6*mhd.Bz(10,10,17),    44.8179162, 1e-7);
    TS_ASSERT_DELTA(1e6*mhd.Bz(10,10,18), -6673.5111155, 1e-5);
    TS_ASSERT_DELTA(1e6*mhd.Bz(10,10,19),     0.0000000, 1e-7);
    */
  }

  void testReadWriteMHD( void )
  {
    /************* Start setUp ***************/
    MHD mhd = MHD(N_I,N_J,N_K,N_ORDER);

    mhd.initial(0,0,0);

    /************* end setUp ***************/

    MHDManager *magnetosphereManager = new TimestepMHD(100000, 100000, 0);

    int status;
    TS_WARN("TimestepMHD::read_Data() has issues on certain systems.  Skipping test.");
    //status = magnetosphereManager->read_step("TimestepRun", &mhd);
    //TS_ASSERT_EQUALS(status, 0);

    TS_WARN("TimestepMHD::write_Data() has issues on certain systems.  Skipping test.");
    //status = magnetosphereManager->dump_step("TimestepRun",&mhd);
    //TS_ASSERT_EQUALS(status, 0);

    //TS_ASSERT( doesFileExist("TimestepRun_mhd_0100000.dmp") );
  }
};

#endif
