#include "IM_IC_Interface.h"

/********************************************************************//**
 * Exchange RCM model parameters (lat/lon dimensions & R_i) and
 * register exchange variables with InterComm.
 ************************************************************************/ 
IM_IC_Interface::IM_IC_Interface(IC_XJD *XJD) : IM_Interface(XJD) 
{
  // Temporary Receive arrays
  int ia_nLatLon[2] = {1,1};

  // Declare Arrays need by intercomm for descriptors etc
  int iTasks[1] = {0};
  int iaBlocks1d[1][2][1], iaBlocks2d[1][2][2];
  int iLower1d, iUpper1d, iStride1d;
  int iLower2d[2], iUpper2d[2], iStride2d[2];
  int iStatus;

  // Ionosphere Handshake
  iStatus = IC_Recv_local(xjd, "im-cpl_im_size", &ia_nLatLon, 2);
  CHECK_IC_STATUS(xjd, "IC_Recv_local('im-cpl_im_size', ...)", iStatus);

  // RION is defined in const.h
  float RION_meters = RION / 100;
  iStatus = IC_Bcast_local(xjd, "cpl-im_r_i", &RION_meters, 1);
  CHECK_IC_STATUS(xjd, "IC_Recv_local('cpl-im_r_i', ...)", iStatus);
 
  // Register the Latitude Region
  nLat = ia_nLatLon[0];
  colatitudes.redim(nLat);

  iaBlocks1d[0][0][0] = 0;
  iaBlocks1d[0][1][0] = nLat-1;
  iLower1d = 0;
  iUpper1d = nLat-1;
  iStride1d = 1;
  desc_lat = IC_Create_bdecomp_desc(1,&iaBlocks1d[0][0][0],iTasks,1,1);
  region_lat[0]=IC_Create_block_region(1,&iLower1d,&iUpper1d,&iStride1d);
  IC_Register_region(xjd,region_lat,1,"im-cpl_lat",desc_lat,colatitudes.getDataPointer(),&iStatus);
  CHECK_IC_STATUS(xjd, "IC_Register_region('im-cpl_lat', ...)", iStatus);

  // Register the Longitude Region
  nLon = ia_nLatLon[1];
  longitudes.redim(nLon);

  iaBlocks1d[0][0][0] = 0;
  iaBlocks1d[0][1][0] = nLon-1;
  iLower1d = 0;
  iUpper1d = nLon-1;
  iStride1d = 1;
  desc_lon = IC_Create_bdecomp_desc(1,&iaBlocks1d[0][0][0],iTasks,1,1);
  region_lon[0]=IC_Create_block_region(1,&iLower1d,&iUpper1d,&iStride1d);
  IC_Register_region(xjd,region_lon,1,"im-cpl_lon",desc_lon,longitudes.getDataPointer(),&iStatus);
  CHECK_IC_STATUS(xjd, "IC_Register_region('im-cpl_lon', ...)", iStatus);  

  // Dimension data arrays (pot, sigmap, sigmah, avgEng, engFlux, fac)
  dimensionData();

  // Register the 2d exchange variables
  iaBlocks2d[0][0][0] = 0;
  iaBlocks2d[0][0][1] = 0;
  iaBlocks2d[0][1][0] = nLat-1;
  iaBlocks2d[0][1][1] = nLon-1;
  
  iLower2d[0] = 0;
  iLower2d[1] = 0;
  
  iUpper2d[0] = nLat-1;
  iUpper2d[1] = nLon-1;

  iStride2d[0] = 1;
  iStride2d[1] = 1;

  desc_vars = IC_Create_bdecomp_desc(2,&iaBlocks2d[0][0][0],iTasks,1,1);
  region_vars[0]=IC_Create_block_region(2,iLower2d,iUpper2d,iStride2d);

  IC_Register_region(xjd,region_vars,1,"im-cpl_avg_energy",desc_vars,avgEng.getDataPointer(),&iStatus);
  CHECK_IC_STATUS(xjd, "IC_Register_region('im-cpl_avg_energy', ...)", iStatus);

  IC_Register_region(xjd,region_vars,1,"im-cpl_avg_flux",desc_vars,engFlux.getDataPointer(),&iStatus);
  CHECK_IC_STATUS(xjd, "IC_Register_region('im-cpl_avg_flux', ...)", iStatus);

  IC_Register_region(xjd,region_vars,1,"im-cpl_fac",desc_vars,j_fac.getDataPointer(),&iStatus);
  CHECK_IC_STATUS(xjd, "IC_Register_region('im-cpl_fac', ...)", iStatus);

  IC_Register_region(xjd,region_vars,1,"cpl-im_potential",desc_vars,pot.getDataPointer(),&iStatus);
  CHECK_IC_STATUS(xjd, "IC_Register_region('cpl-im_potential', ...)", iStatus);

  IC_Register_region(xjd,region_vars,1,"cpl-im_sigma_p",desc_vars,sigma_p.getDataPointer(),&iStatus);
  CHECK_IC_STATUS(xjd, "IC_Register_region('cpl-im_sigma_p', ...)", iStatus);

  IC_Register_region(xjd,region_vars,1,"cpl-im_sigma_h",desc_vars,sigma_h.getDataPointer(),&iStatus);
  CHECK_IC_STATUS(xjd, "IC_Register_region('cpl-im_sigma_h', ...)", iStatus);

  /************* Arrays are ready to be comitted *************/  
  // Arrays must be comitted outside this class in case additional
  // arrays need to be defined in another interface
  // (i.e. MHD_IC_Interface, ITM_IC_Interface).

  //Don't forget to call IM_IC_Interface::Initialize() after comitting!
}

/********************************************************************//**
 * Destructor.  Free the InterComm descriptors/regions
 ************************************************************************/ 
IM_IC_Interface::~IM_IC_Interface()
{
  IC_Free_desc(desc_lat);
  IC_Free_desc(desc_lon);
  IC_Free_desc(desc_vars);
  IC_Free_region(region_lat[0]);
  IC_Free_region(region_lon[0]);
  IC_Free_region(region_vars[0]);
}

/********************************************************************//**
 * After constructor is called, the user *must* commit the arrays via
 * IC_Commit_region(xjd, status).
 * 
 * After arrays are comitted, the user *must* exchange Initialize
 * variables (i.e. call this function!)
 ************************************************************************/ 
void IM_IC_Interface::Initialize()
{
  int iStatus;
  // Receive Ionosphere Grid from RCM:
  iStatus = IC_Import(xjd, "im-cpl_lat");
  CHECK_IC_STATUS(xjd, "IC_Import('im-cpl_lat', ...)", iStatus);

  iStatus = IC_Import(xjd, "im-cpl_lon");
  CHECK_IC_STATUS(xjd, "IC_Import('im-cpl_lon', ...)", iStatus);

  // Setup RCM grid that will be used for interpolation
  setupIMgrid_2d();
}

/********************************************************************//**
 * Export the Potential, Pedersen & Hall conductances to Inner Magnetosphere
 *
 * prepareExport interpolates the data from the MIX solver grid onto
 * the RCM grid.  See IM_Interface::prepareExport() for more information.
 ************************************************************************/ 
void IM_IC_Interface::Export(const realCompositeGridFunction &mixSolverFunctions) 
{
  int iStatus;

  prepareExport(mixSolverFunctions);

  iStatus = IC_Export(xjd, "cpl-im_potential");
  CHECK_IC_STATUS(xjd, "IC_Export('cpl-im_potential', ...)", iStatus);

  iStatus = IC_Export(xjd, "cpl-im_sigma_p");
  CHECK_IC_STATUS(xjd, "IC_Export('cpl-im_sigma_p', ...)", iStatus);

  iStatus = IC_Export(xjd, "cpl-im_sigma_h");
  CHECK_IC_STATUS(xjd, "IC_Export('cpl-im_sigma_h', ...)", iStatus);
}

/********************************************************************//**
 * Import the following variables from the Inner Magnetosphere:
 *  - Average Energy
 *  - Energy Flux
 *  - Total Field-Aligned density
 * 
 * processImport interpolates from the RCM grid onto the MIX solver grid.
 ************************************************************************/ 
void IM_IC_Interface::Import()
{
  int iStatus;

  iStatus = IC_Import(xjd, "im-cpl_avg_energy");
  CHECK_IC_STATUS(xjd, "IC_Import('im-cpl_avg_energy', ...)", iStatus);

  iStatus = IC_Import(xjd, "im-cpl_avg_flux");
  CHECK_IC_STATUS(xjd, "IC_Import('im-cpl_avg_flux', ...)", iStatus);

  iStatus = IC_Import(xjd, "im-cpl_fac");
  CHECK_IC_STATUS(xjd, "IC_Import('im-cpl_fac', ...)", iStatus);

  processImport();
}
