#include "IM_Interface.h"

/********************************************************************//**
 * Constructor. 
 ************************************************************************/ 
IM_Interface::IM_Interface(IC_XJD *XJD) : IM_Grids() 
{
  xjd = XJD;
}

/********************************************************************//**
 * Dimension the 2-d (nLat, nLon) data exchange arrays
 ************************************************************************/ 
void IM_Interface::dimensionData()
{
  pot.redim(nLat, nLon);
  sigma_h.redim(nLat, nLon);
  sigma_p.redim(nLat, nLon);
  avgEng.redim(nLat, nLon);
  engFlux.redim(nLat, nLon);
  j_fac.redim(nLat, nLon);
}

/********************************************************************//**
 * Get ready to export the potential, pedersen & hall conductances to
 * the Inner Magnetosphere on the (RCM) grid: Intepolate potential,
 * pedersen & hall conductances from MIX grid onto RCM grid and
 * convert units.
 *
 * FIXME: The code currently uses the potential from one hemisphere
 * (North).  Eventually we need a metric to chose which hemisphere to
 * utilize or a method for combining the Northern & Southern
 * hemispheres for more general coupling.
 ************************************************************************/ 
void IM_Interface::prepareExport(const realCompositeGridFunction &mixSolverFunctions)
{
  cerr << __FILE__ <<  " (Line " << __LINE__ << "):  " << __FUNCTION__ << "(...) WARNING:  gets the potential for the Inner Magnetosphere only on the Northern Hemisphere." << endl;

  
/////////////////////////////////////////////////////////////////////////
//   // The following bit of code sets potential == step function
//   // This is useful for validating that interpolation onto RCM grid works.
  
//     MappedGrid *grid = mixSolverFunctions[0].getMappedGrid();
//     Index I1,I2,I3;
//     getIndex((*grid).gridIndexRange(),I1,I2,I3); // Unlike the
  
//     const realArray &x = (*grid).vertex()(I1,I2,I3,0);
//     const realArray &y = (*grid).vertex()(I1,I2,I3,1);
  
//     int ni = x.getLength(0);
//     int nj = x.getLength(1);

//     // Set potential == step function
//     for (int i=0; i < ni; i++){
//       for (int j=0; j < nj; j++){
//         mixSolverFunctions[0](I1,I2,I3,SOLVER_POTENTIAL_NORTH)(i,j) = i;
//       }
//     }
/////////////////////////////////////////////////////////////////////////
  
  //  Interpolate data from MIX grid onto the RCM ionospheric grid:
  interpMixToIm(mixSolverFunctions, SOLVER_POTENTIAL_NORTH, pot);
  interpMixToIm(mixSolverFunctions, SOLVER_SIGMAP_NORTH,    sigma_p);
  interpMixToIm(mixSolverFunctions, SOLVER_SIGMAH_NORTH,    sigma_h);
    
  // Convert export variables to Meters Kilograms Seconds (MKS) units
  // Note:  This is unecessary since MIX variables are already MKS units:
  //
  //pot *= 1.0; // Volts are already MKS.
  //sigma_p *= 1.0; // Siemens are already MKS.
  //sigma_h *= 1.0; // Siemens are already MKS.


/////////////////////////////////////////////////////////////////////////
//   // The following bit of code writes grids & data arrays to a file
//   // for validating interpolation via IDL or Matlab.
  
//     ofstream mix_out;
//     mix_out.open("mix_mix_grid.dat", ios::out);
  
//     realArray raPot = mixSolverFunctions[0](I1,I2,I3,SOLVER_POTENTIAL_NORTH);
//     realArray raPed = mixSolverFunctions[0](I1,I2,I3,SOLVER_SIGMAP_NORTH);
//     realArray raHal = mixSolverFunctions[0](I1,I2,I3,SOLVER_SIGMAH_NORTH);
  
//     int nLatMix = ni;
//     int nLonMix = nj;
//     mix_out << nLatMix << " " << nLonMix << endl;
//     mix_out << "i j x y potential sigma_p sigma_h" << endl;
  
//     for (int i=0; i < nLatMix; i++){
//       for (int j=0; j < nLonMix; j++){
//         mix_out << i << " " << j << " " 
//   	      << x(i,j) << " "
//   	      << y(i,j) << " "
//   	      << raPot(i,j) << " "
//   	      << raPed(i,j) << " "
//   	      << raHal(i,j) << " "
//   	      << endl;
//       }
//     }
//     mix_out.close();
  
//     ofstream rcm_out;
//     rcm_out.open("mix_rcm_grid.dat", ios::out);
  
//     rcm_out << nLat << " " << nLon << endl;
//     rcm_out << "i j x y potential sigma_p sigma_h" << endl;
//     for (int i=0; i < nLat; i++){
//       for (int j=0; j < nLon; j++){
//         rcm_out << i << " " << j << " "
//   	      << IMgrid_2d(i+j*nLat, 0) << " "
//   	      << IMgrid_2d(i+j*nLat, 1) << " "
//   	      << pot(i,j) << " " 
//   	      << sigma_p(i,j) << " " 
//   	      << sigma_h(i,j) << endl;
//       }
//     }
/////////////////////////////////////////////////////////////////////////
}

/********************************************************************//**
 * Interpolate a variable in mixSolverFunctions at index to destination floatArray.
 * (RCM) grid:  Intepolate potential from MIX grid onto RCM grid.
 *
 * Since MIX and RCM both utilize the same coordinate system
 * (spherical grid), we make a simple call to Overture's interpolate
 * function.
 *
 * FIXME: The code currently uses a variable from a single hemisphere
 * (North).  Eventually we need a metric to chose which hemisphere to
 * utilize or a method for combining the Northern & Southern
 * hemispheres for more general coupling.
 ************************************************************************/ 
void IM_Interface::interpMixToIm(const realCompositeGridFunction mixSolverFunctions, 
				 const int &index, floatArray &dest)
{  
  // Get the MIX variable we wish to interpolate. 
  Range R(index, index);
  realCompositeGridFunction MIX_variable;
  MIX_variable.link(mixSolverFunctions, R);

  // Through trial & error, I've discovered that Overture's
  // interpolatePoints(...) requires arrays to be reshaped into 1-d
  // ararys.  Fortunately, floatArray::getDataPointer() returns the
  // same pointer before & after reshaping, so this shouldn't screw up
  // the data exchanges.
  // 
  // Also... Note that Overture's interpolatePoints requires
  // realArrays (not floatArrays).  Must manually copy data into
  // floatArray destination.
  realArray dest_doubleArray(nLat*nLon);

  // Must... Interpolate... Data!
  int r = interpolatePoints(IMgrid_2d, MIX_variable, dest_doubleArray);

  // Copy doubleArray data into destination floatArrray used for
  // communication with Inner Magnetosphere.
  for (int j=0; j < nLon; j++){
    for (int i=0; i < nLat; i++){
      dest(i,j) = dest_doubleArray(i+j*nLat);
    }
  }
}

/********************************************************************//**
 * Process the data imported from the Inner Magnetosphere, including
 * unit conversions of:
 *  - Average Energy
 *  - Energy Flux
 *  - Total field-aligned density
 * 
 * Data has already been imported on the Inner Magnetosphere (RCM)
 * grid.  Need to interpolate onto the MIX grid.
 *
 * FIXME: This is not implemented...  Need to design algorithm that
 * intelligently merges Energy, Flux & FAC from MIX,
 * Ionosphere-Thermosphere-Mesosphere (ITM) and Inner Magnetospher
 * (IM)
 ************************************************************************/ 
void IM_Interface::processImport()
{
  cerr << __FILE__ <<  " (Line " << __LINE__ << "):  " << __FUNCTION__ << "(...) Not Implemented." << endl;

  // Convert Import variables from Meters Kilograms Seconds (MKS) units to MIX's internal units
  avgEng  /= 1.602e-16; // Convert from J to KeV
  engFlux /= 1.602e-12; // Convert from W/m^2 to KeV/(cm^2 s)
  //j_fac *= 1.0; // j_fac has units of A/m^2, which are MKS units
}
