/*!\file ITM_FE_Interface.C
\brief ITM_FE_Interface class definition.
*/
#include "ITM_FE_Interface.h"

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 06-2007
 *
 * \param[in] params Pointer to the Params class
 * \param[in] XJD Pointer to the IC_XJD object
 *
 * The constructor inherits the ITM grid information from
 * ITM_Grids. Then arrays to be transmitted between this code and the
 * ITM model are dimensioned (ITM_Grids::nLat,ITM_Grids::nLon). The
 * InterComm regions are created and prepared for commit.
 ************************************************************************/
ITM_FE_Interface::ITM_FE_Interface(Params *params, IC_XJD *XJD) : ITM_Interface(params, XJD)
{
  cout << "Initializing ITM File Exchange Interface..." << endl;
  
  MIXLockFile = "mhd_status";
  ITMLockFile = "ting_status";

  assert( setMIXOld() );

}

/********************************************************************//**
 * \author Peter Schmitt (schmitt at ucar.edu)
 * \since 9-2008
 *
 * Clean up
 ************************************************************************/
ITM_FE_Interface::~ITM_FE_Interface()
{
  //assert( remove(MIXLockFile.c_str()) == 0 );
  //remove("tin-inputs.txt");
}

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 06-2007
 *
 * \param[in] UT Universal time array (year,month,day,hour,min,sec)
 * \param[in] psi Ionospheric potential
 * \param[in] averageEnergy of precipitating particles
 * \param[in] numberFlux of precipitating particles
 *
 * Prepare & send data to the ITM model. 
 *
 *   - Convert (year,month,day,hour,min,sec) to (year, doy, hour, min, sec) 
 *     and send to the ITM model.
 *    
 *   - Convert the arrays (potential,energy,flux) to lat/lon,
 *     interpolate and IC_export to the ITM model.
 *
 ************************************************************************/
void ITM_FE_Interface::Export(const intArray & UT,
			      realCompositeGridFunction & mhdFunctions)
{
  // Initialize universal time array for use in this function as well as in Import
  intArray YYYYdoyHHmmSS = ut2utdoy(UT);

  // prepare potential, energy & flux
  prepareExport(YYYYdoyHHmmSS, mhdFunctions);

  /***********************************************/
  /* Send potential, energy & flux to ITM model  */
  /***********************************************/
  
  // Wait until TIEGCM has set the MIXstatus to "old"
  
  while ( ! isMIXOld() ){
#ifdef DEBUG_MODE_ON
    std::cout << "  MIX_FE: Waiting for ITM to acknowledge that its ready for potential..." << endl;
#endif
    sleep(4);
  }

  // open the file for writing
  std::ofstream outs("ting-inputs.txt");
  if (outs.bad()){
    std::cerr << "*** Trouble Writting " << "ting-inputs.txt" << "\n";
  }
  
  // write data
  outs << YYYYdoyHHmmSS(1) << "  " << YYYYdoyHHmmSS(2) << "  " << YYYYdoyHHmmSS(3) << "  " << YYYYdoyHHmmSS(4) << endl;
  //outs.precision(4);
  outs.precision(3);
  outs << scientific;
  for (int j = 0; j < nLat; j++){
    for (int i = 0; i < nLon; i++){
      /*
      outs << "  " << leftprec(gridSphericalPositions(j,i,0)*180.0/PI, 0)
	   << "  " << leftprec(gridSphericalPositions(j,i,1)*180.0/PI, 0)
	   << "  " << leftprec(pot(j,i), 0)
	   << "  " << leftprec(eng(j,i), 0)
	   << "  " << leftprec(flux(j,i), 0) << "\n";
      */
      // FORTRAN I/O needs a specific number of items per 12-space column in an ASCII file...
      // So... right-justify & set (minimum) width to 12 spaces.
      outs << right << setw(12) << gridSphericalPositions(j,i,0)*180.0/PI
	   << right << setw(12) << gridSphericalPositions(j,i,1)*180.0/PI
	   << right << setw(12) << pot(j,i)
	   << right << setw(12) << eng(j,i)
	   << right << setw(12) << flux(j,i) << "\n";
    }
  }

  // Make sure to flush the IO to disk.
  outs.flush();
  
  // close the file
  outs.close();

  // Set MHD status to "new" since we've just sent the potential
  assert( setMIXNew() );
  
  /***********************************************/
}


/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 06-2007
 *
 * \param[out] sigmaP_N Pedersen conductance (North) on Solver grid
 * \param[out] sigmaP_S Pedersen conductance (South) on Solver grid
 * \param[out] sigmaH_N Hall conductance (North) on Solver grid
 * \param[out] sigmaH_S Hall conductance (South) on Solver grid
 * \param[out] Unw_N FAC due to neutral wind (North) on Solver grid
 * \param[out] Unw_S FAC due to neutral wind (South) on Solver grid
 *
 * This function receives and transforms data (ped, hall, wfac) from ITM model. 
 *
 *   - Get data arrays (pedersen & hall conductances and FAC) from ITM
 *     model using IC_Import
 *
 *   - Convert arrays (pedersen & hall conductances and FAC) from
 *     lat/lon to geographical angular coordinates. 
 *
 * Results stored in mhdFunctions object.
 ************************************************************************/
void ITM_FE_Interface::Import(realCompositeGridFunction & mhdFunctions)
{
  /***********************************************/
  /*        Get Ped, Hall & wFAC from ITM        */
  /***********************************************/
  //status = IC_Import(xjd,"ped");
  //status = IC_Import(xjd,"hall");
  //status = IC_Import(xjd,"wfac");

  // Wait until ITM data is "new"
  while ( ! isITMNew() ){
#ifdef DEBUG_MODE_ON
    std::cout << "  MIX_FE: Waiting for ITM to export Pedersen & Hall conductances and Field-Aligned Current...." << endl;
#endif
    sleep(4);
  }

#ifdef DEBUG_MODE_ON
  std::cout << "  MIX_FE: ITM exported variables. Read them in." << endl;
#endif

  /***********************************************/

  // open the file for writing
  std::ifstream ins("ting-outputs.txt");
  if (ins.bad()){
    std::cerr << "*** Trouble Reading " << "ting-outputs.txt" << "\n";
  }

  // read data
  double lat_j, lon_i;
  int ndoy, nhr, nmin, nsec;

  ins >> ndoy >> nhr >> nmin >> nsec;

  for (int j = 0; j < nLat; j++){
    for (int i = 0; i < nLon; i++){
      ins >> lat_j >> lon_i >> ped(j,i) >> hall(j,i) >> wfac(j,i);
    }
  }

  // close the file
  ins.close();

  /***********************************************/

  // MIX is done reading.  Tell the ITM to continue:
#ifdef DEBUG_MODE_ON
  std::cout << "  MIX_FE: Pedersen & Hall conductances and Field-Aligned Current read in by MIX.\n";
#endif

  // MIX Is done reading the "new" variables.  Set ITM status to "old"
  assert( setITMOld() );

  /***********************************************/

  // Convert (ped,hall,wfac) to MIX solver coordinates & interpolate.
  // Store results in mhdFunctions
  processImport(mhdFunctions);
}


/************************************************************************
 * Private member functions
 ***********************************************************************/

/********************************************************************//**
 * \author Peter Schmitt (schmitt at ucar.edu)
 * \since 09-2008
 *
 * \param[in] filename - string pointing to path of lock file to write
 * \param[in] status - string that sets the MIX lock file status
 * (e.g. "WAITING" or "WORKING").
 *
 * \param[out] bool - returns true if succesful.
 *
 * Write to code status to the ASCII text file MIXLockFile
 ************************************************************************/
bool ITM_FE_Interface::writeLockFile(const string &filename, const string &status)
{
#ifdef DEBUG_MODE_ON
  cout << "DEBUG: ITM_FE_Interface::writeLockFile(...)" << endl;
#endif
  // open the file for writing
  std::ofstream outs(filename.c_str());
  if (outs.bad()){
    std::cerr << "*** Trouble writting " << filename << "\n";
    return false;
  }

  // write status
  outs << status << "\n";

  // Make sure to flush the IO to disk.
  outs.flush();

  // close the file
  outs.close();
  
  return true;
}

/********************************************************************//**
 * \author Peter Schmitt (schmitt at ucar.edu)
 * \since 09-2008
 *
 * \param[in] filename - path to file
 * \param[out] text contents of file 'filename'.  Should be "NEW" or "OLD".
 *
 * Returns the contents of the ASCII file "filename".
 ************************************************************************/
string ITM_FE_Interface::readLockFile(const string &filename)
{
#ifdef DEBUG_MODE_ON
  cout << "DEBUG: ITM_FE_Interface::readLockFile(...)" << endl;
#endif

  // open the file for reading
  std::ifstream ins(filename.c_str());
  if (ins.bad()){
    std::cerr << "*** Trouble reading " << filename << "\n";
    // FIXME:  Error handling?
    return false;
  }

  // read status
  string statusString;
  ins >> statusString;

  // close the file
  ins.close();

  return statusString;
}
