/**
 * \file ITM_Grids.C
 * \brief ITM_Grids class definition.
 */
#include "ITM_Grids.h"

using namespace std;

//Range all;

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 06-2007
 *
 * \param[in] modelName ITM model name
 *
 * \param[in] gridFileName The name of the file containing the ITM
 * grid. Not used for TING
 *
 * \param[in] gridFileType The type of the ITM grid. Reserved for
 * TIEGCM and others. Not used for TING.
 *
 * The ITM_Grids class turned out to be pretty simple. The constructor
 * is only required to fill in the following variables defining the
 * ITM model grid and used later in ITM_Interface: #nLat, #nLon,
 * #geoPositions, #gridSph. The ITM_Interface is a child of the
 * ITM_Grids class, therefore it inherits all these properties from
 * the latter. The ITM_Class is not called from the ionosphere.C
 * program directly, rather it is initialized by the ITM_Interface
 * class constructor. A more detailed information about the variables
 * listed above is given in their documentation. Here, we will just
 * say that they define the ITM grid completely with the following
 * assumptions implied:
 *
 * - The ITM grid is periodic in the longitudinal direction and the
 * periodic boundary is included in nLon count (this corresponds to
 * gridIndexRange in Overture terms - see, for instance, <A
 * HREF="http://www.llnl.gov/casc/Overture/henshaw/documentation/gf.pdf">Grid
 * Functions for Overture</A>, p. 9.
 *
 * - The #geoPositions array (and any other data associated with the
 * ITM grid) is dimensioned (#nLat,#nLon), latitude being the inner
 * loop coordinate, while longitude being the outer loop coordinate
 * (see, for instance, tingGridPostions()).
 *
 * - Latitude is an angle running up and down from the geographical
 * equator [\f$-\pi/2,\pi/2\f$], negative values corresponding to the
 * southern hemisphere, of course.
 *
 * - Longitude is in the range [\f$-\pi,\pi\f$], with 0 at the noze
 * and \f$\pm\pi\f$ at the tail.
 *
 * The constructor is supposed to call a model specific function,
 * e.g. tingGridPositions(), that fills in the necessary
 * variables. Therefore, in order to couple another ITM model, all one
 * needs to do is to write such a funtion and comply with the
 * requirements listed above.
 ************************************************************************/
ITM_Grids::ITM_Grids(const char *modelName, const char *gridFileName, const int gridFileType)
{
  if ( strcmp(modelName,"TING") == 0 )
    {
      maxLat = 87.5;
      nLat = 36;
      nLon = 73;
      dLat = maxLat*2./(nLat-1);
      dLon = 360./(nLon-1);

      geoPositions.redim(nLat*nLon,3);
      gridSphericalPositions.redim(nLat,nLon,2);
      tingGridPositions();
    }
  else if ( strcmp(modelName,"TIEGCM") == 0 )
    {
      cout << "TIEGCM not yet supported." << endl;
      exit(1);
    }
  else
    {
      cerr << "ERROR: ITM model name not recongnized: \"" << modelName << "\"." << endl;
      exit(1);
    }

  gridSph = generateSphericalGrid();
}

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 06-2007
 *
 * \return Returns #gridSph based on a 2D-mapping in the geographical
 * spherical angular coordinates (latitude,longitude - see contructor
 * documentation)
 * 
 * This is the only grid we need so far. It is used for interpolation
 * from the ITM grid to the Solver grid in ITM_Interface::Import().
 ************************************************************************/
CompositeGrid ITM_Grids::generateSphericalGrid()
{
  DataPointMapping & sphere = *new DataPointMapping;
  sphere.incrementReferenceCount();
  sphere.setDataPoints(gridSphericalPositions,2,2);
  sphere.setIsPeriodic(1,Mapping::functionPeriodic);
   
  CompositeGrid sgrid(2,1);
  sgrid[0].reference(sphere);
  sgrid.update();
  sphere.decrementReferenceCount();

  return sgrid;
}


/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 06-2007
 *
 * \retval #geoPositions are filled in
 * \retval #gridSphericalPositions are filled in
 * 
 * The arrays above are filled in complying with the requirements for
 * the grid ordering described in the class constructor. See the
 * source code for details, it is straightforward.
 ************************************************************************/
void ITM_Grids::tingGridPositions()
{
  // This is the TING grid definition: unlike the normal spherical grid 
  // it goes from -180 to 180 degrees in latitude
  // and from the southern pole (-87.5) to the northern pole (87.5)
  realArray lon(nLon), lat(nLat);
  lon.seqAdd(-180.,dLon);
  lat.seqAdd(-maxLat,dLat);
  lat = lat/180.*PI;
  lon = lon/180.*PI;

  // Coordinates above are defined according to the TING grid definition
  // So, indexing in the arrays that are sent and received from TING corresponds
  // to indexing of above lon and lat array
  // For instance, (using a bogus variable name here) Ting_Pedersen(0,0) corresponds
  // to the value at the TING -87.5 latitutude and -180 longitude
  // meaning (185,180) geomagnetic (colatitude,longitude)

  // We need to calculate GEO cartesian coordinates
  Range NLAT(0,nLat-1);

  realArray x(nLat,nLon), y(nLat,nLon), z(nLat,nLon), theta(nLat,nLon), phi(nLat,nLon);
  for (int k = 0; k < nLon; k++) {
    x(all,k) = cos(lat)*cos(lon(k));
    y(all,k) = cos(lat)*sin(lon(k));
    z(all,k) = sin(lat);

    theta(all,k) = lat;
    phi(all,k) = lon(k);

    geoPositions(NLAT+k*nLat,0) = x(all,k);
    geoPositions(NLAT+k*nLat,1) = y(all,k);
    geoPositions(NLAT+k*nLat,2) = z(all,k);
  }

  gridSphericalPositions(all,all,0) = theta;
  gridSphericalPositions(all,all,1) = phi;
}
