#ifndef ITM_GRID_H
#define ITM_GRID_H

/**
 * \file ITM_Grids.h
 * \brief ITM_Grids class declaration
 * \class ITM_Grids
 * 
 * \brief Ionosphere-thermosphere model grids are defined here
 *
 * \author Slava Merkin (vgm at bu.edu)
 * \since 06-2007
 *
 * This class contains functions and variables that define the ITM
 * grid. At this point it is based on the TING being the ITM model and
 * is rather simple. However, it is not clear whether it should become
 * any more complicated if, say, we replace TING by TIEGCM. The
 * ideology behind this class should allow a very straightforward
 * adaptation of any ITM as long as its grid is a structured one in
 * geographical coordinates. (see the ITM_Grids constructor below).
*/

#include "Overture.h"  
#include "DataPointMapping.h"
#include "SquareMapping.h"

#include "const.h"
#include "ITM_Grids.h"
#include "PlotStuff.h"

#include <iostream>
#include <fstream>

class ITM_Grids  {

 public:
  /**\brief Class constructor*/
  ITM_Grids(const char*, const char*, const int);

 protected:
  /**\name ITM grid definitions to be used by ITM_Interface */
  //@{
  int nLat; //!< Number of grid points in the latitudinal direction
  int nLon; /*!<\brief Number of grid points in the longitudinal
	      direction. Must include the periodic boundary
	      points. See ITM_Grids(). */


  /**The cartesian (x,y,z) geographical positions of the ITM
   * grid. This array should be dimensioned in the ITM_Grids() class
   * constructor as \code geoPositions.redim(nLat*nLon,3); \endcode
   * See also ITM_Grids() and the source code in tingGridPositions()
   * for the details of the array ordering.
   */
  realArray geoPositions;

  /** The 2D geographical spherical angular coorinates of grid points.
   * This array should be dimensioned in the constructor as \code
   * gridSphericalPositions.redim(nLat,nLon,2); \endcode See also
   * ITM_Grids() and the source code in tingGridPositions() for the
   * details of the array ordering. The array is used to contruct
   * #gridSph.
   */
  realArray gridSphericalPositions;

  /** The 2D grid based on #gridSphericalPositions. Used for
   *  interpolation by ITM_Interface::Import()*/
  CompositeGrid gridSph;
  //@}

  /**\brief Creates a spherical grid for ITM_Interface::Import()*/
  CompositeGrid generateSphericalGrid();
  
 private:
  /**\brief Fills in grid positions for the TING model*/
  void tingGridPositions();

  /**\name TING grid specifics */
  //@{
  real maxLat; //!< Maximum latitude in degrees (TING uses a hole around the pole)
  real dLat;   //!< Latitudinal grid step 
  real dLon;   //!< Longitudinal grid step 
  //@}
};

#endif
