/*!\file ITM_IC_Interface.C
\brief ITM_IC_Interface class definition.
*/
#include "ITM_IC_Interface.h"

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 06-2007
 *
 * \param[in] params Pointer to the Params class
 * \param[in] XJD Pointer to the IC_XJD object
 *
 * The constructor inherits the ITM grid information from
 * ITM_Grids. Then arrays to be transmitted between this code and the
 * ITM model are dimensioned (ITM_Grids::nLat,ITM_Grids::nLon). The
 * InterComm regions are created and prepared for commit.
 ************************************************************************/
ITM_IC_Interface::ITM_IC_Interface(Params *params, IC_XJD *XJD) : ITM_Interface(params, XJD)
{
  cout << "Initializing ITM InterComm Interface..." << endl;

  // These are set for a sequential program. If I ever make this code parallel
  // these should be changed (see InterComm manual)
  int tasks[1] = {0};

  int blocks[1][2][2] = {
    {{0,0},{nLat-1,nLon-1}}
  };

  desc = IC_Create_bdecomp_desc(2, &blocks[0][0][0], tasks, 1, 1);
  // IC_ROW_MAJOR = 0                                         ^^^
  // IC_COLUMN_MAJOR = 1                                      ^^^
  
  int lower[2] = {0,0};
  int upper[2] = {nLat-1,nLon-1};
  int stride[2] = {1,1};
  regions[0] = IC_Create_block_region(2,lower,upper,stride);

  IC_Register_region (xjd,regions,1,"cpl-itm_potential",desc,pot.getDataPointer(),&ic_err);
  CHECK_IC_STATUS(xjd, "IC_Register_region('cpl-itm_potential', ...)", ic_err);  

  IC_Register_region (xjd,regions,1,"cpl-itm_energy",desc,eng.getDataPointer(),&ic_err);
  CHECK_IC_STATUS(xjd, "IC_Register_region('cpl-itm_energy', ...)", ic_err);  

  IC_Register_region (xjd,regions,1,"cpl-itm_flux",desc,flux.getDataPointer(),&ic_err);
  CHECK_IC_STATUS(xjd, "IC_Register_region('cpl-itm_flux', ...)", ic_err);  

  IC_Register_region (xjd,regions,1,"itm-cpl_pedersen",desc,ped.getDataPointer(),&ic_err);
  CHECK_IC_STATUS(xjd, "IC_Register_region('itm-cpl_pedersen', ...)", ic_err);  

  IC_Register_region (xjd,regions,1,"itm-cpl_hall",desc,hall.getDataPointer(),&ic_err);
  CHECK_IC_STATUS(xjd, "IC_Register_region('itm-cpl_hall', ...)", ic_err);  

  IC_Register_region (xjd,regions,1,"itm-cpl_wfac",desc,wfac.getDataPointer(),&ic_err);
  CHECK_IC_STATUS(xjd, "IC_Register_region('itm-cpl_wfac', ...)", ic_err);  

}

/********************************************************************//**
 * \author Peter Schmitt (schmitt at ucar.edu)
 * \since 9-2008
 *
 * Clean up InterComm objects
 ************************************************************************/
ITM_IC_Interface::~ITM_IC_Interface()
{
  IC_Free_desc(desc);
  IC_Free_region(regions[0]);
}

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 06-2007
 *
 * \param[in] UT Universal time array (year,month,day,hour,min,sec)
 * \param[in] psi Ionospheric potential
 * \param[in] averageEnergy of precipitating particles
 * \param[in] numberFlux of precipitating particles
 *
 * Prepare & send data to the ITM model. 
 *
 *   - Convert (year,month,day,hour,min,sec) to (year, doy, hour, min, sec) 
 *     and send to the ITM model.
 *    
 *   - Convert the arrays (potential,energy,flux) to lat/lon,
 *     interpolate and IC_export to the ITM model.
 *
 ************************************************************************/
void ITM_IC_Interface::Export(const intArray & UT,
			      realCompositeGridFunction & mhdFunctions)
{
  // Initialize universal time array for use in this function as well as in Import
  utime = ut2utdoy(UT);

  int iStatus;
  Range FOUR(1,4);
  intArray ut(4);
  ut = utime(FOUR);

  // prepare potential, energy & flux
  prepareExport(utime, mhdFunctions);

  /***********************************************/
  /* Send potential, energy & flux to ITM model  */
  /***********************************************/
  iStatus = IC_Export(xjd,"cpl-itm_potential");
  CHECK_IC_STATUS(xjd, "IC_Export('cpl-itm_potential')", iStatus);  

  iStatus = IC_Export(xjd,"cpl-itm_energy");
  CHECK_IC_STATUS(xjd, "IC_Export('cpl-itm_energy')", iStatus);  

  iStatus = IC_Export(xjd,"cpl-itm_flux");
  CHECK_IC_STATUS(xjd, "IC_Export('cpl-itm_flux')", iStatus);  

  /***********************************************/

  /* USE THIS TO PLOT THE INTERPOLATED FUNCTIONS ON THE TING GRID
  realCompositeGridFunction u(gridSph,all,all,all,3);
  Range K(0,nLon-1),J(0,nLat-1),all;
  u[0](J,K,all,0) = pot(all,all);
  u[0](J,K,all,1) = eng(all,all);
  u[0](J,K,all,2) = flux(all,all);
  PlotStuff ps(TRUE,"ionGrid");
  PlotIt::contour(ps,u);
  */  
}


/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 06-2007
 *
 * \param[out] sigmaP_N Pedersen conductance (North) on Solver grid
 * \param[out] sigmaP_S Pedersen conductance (South) on Solver grid
 * \param[out] sigmaH_N Hall conductance (North) on Solver grid
 * \param[out] sigmaH_S Hall conductance (South) on Solver grid
 * \param[out] Unw_N FAC due to neutral wind (North) on Solver grid
 * \param[out] Unw_S FAC due to neutral wind (South) on Solver grid
 *
 * This function receives and transforms data (ped, hall, wfac) from ITM model. 
 *
 *   - Get data arrays (pedersen & hall conductances and FAC) from ITM
 *     model using IC_Import
 *
 *   - Convert arrays (pedersen & hall conductances and FAC) from
 *     lat/lon to geographical angular coordinates. 
 *
 * Results stored in mhdFunctions object.
 ************************************************************************/
void ITM_IC_Interface::Import(realCompositeGridFunction & mhdFunctions)
{
  int iStatus;
  iStatus = IC_Import(xjd,"itm-cpl_pedersen");
  CHECK_IC_STATUS(xjd, "IC_Export('itm-cpl_pedersen')", iStatus);  

  iStatus = IC_Import(xjd,"itm-cpl_hall");
  CHECK_IC_STATUS(xjd, "IC_Export('itm-cpl_hall')", iStatus);  

  iStatus = IC_Import(xjd,"itm-cpl_wfac");
  CHECK_IC_STATUS(xjd, "IC_Export('itm-cpl_wfac')", iStatus);  

  // Convert (ped,hall,wfac) to MIX solver coordinates & interpolate.
  // Store results in mhdFunctions
  processImport(mhdFunctions);

  //  PlotStuff ps(TRUE,"ionGrid");
  //  PlotIt::contour(ps,itmUnw);
  //  PlotIt::contour(ps,Unw_N);
  //  PlotIt::contour(ps,Unw_S);
}
