/*!\file ITM_Interface.C
\brief ITM_Interface class definition.
*/
#include "ITM_Interface.h"

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 06-2007
 *
 * \param[in] params Pointer to the Params class
 * \param[in] XJD Pointer to the IC_XJD object
 *
 * The constructor inherits the ITM grid information from
 * ITM_Grids. Then arrays to be transmitted between this code and the
 * ITM model are dimensioned (ITM_Grids::nLat,ITM_Grids::nLon). The
 * InterComm regions are created and prepared for commit.
 ************************************************************************/
ITM_Interface::ITM_Interface(Params *params, IC_XJD *XJD)
  : ITM_Grids(params->itmModelType,
	      params->itmGridFileName,
	      params->itmGridFileType),
    xjd(XJD),
    pot(nLat,nLon),   // Exports
    eng(nLat,nLon),
    flux(nLat,nLon),
    ped(nLat,nLon),    // Imports
    hall(nLat,nLon),
    wfac(nLat,nLon)
    
{
  cout << "Initializing ITM Interface..."<<endl;
}

/********************************************************************//**
 *  \author Pete Schmitt (schmitt at ucar.edu)
 *  \since March 2008
 *
 * \see ITM_Interface::Export(intArray &UT,realCompositeGridFunction & mhdFunctions)
 *
 * Overloaded Export(...) to handle DateTime objects.  This ultimately
 * makes a call to 
 * ITM_Interface::Export(intArray &UT,realCompositeGridFunction & mhdFunctions)
 ************************************************************************/
void ITM_Interface::Export(DateTime  &UT,
			   realCompositeGridFunction & mhdFunctions)
{
  intArray utime(6);
  utime(0) = UT.getYear();
  utime(1) = UT.getMonth();
  utime(2) = UT.getDay();
  utime(3) = UT.getHours();
  utime(4) = UT.getMinutes();
  utime(5) = UT.getSeconds();

  Export(utime, mhdFunctions);
}

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 06-2007
 *
 * \param[in] UT Universal time array (year,month,day,hour,min,sec)
 * \param[in] psi Ionospheric potential
 * \param[in] averageEnergy of precipitating particles
 * \param[in] numberFlux of precipitating particles
 *
 *   - The ut array (year, doy, hour, min, sec) is used in GEOPACK for
 *     mhdFunctions transforms
 *
 *   - The "mhdFunctions" grid functions (psi,averageEnergy, and
 *     numberFlux) are defined on the entire globe in SM spherical
 *     angular coordinates (see Sover::mapToSMGlobe()).  Convert
 *     Convert to lat/lon and interpolate to the ITM grid by looping
 *     over the grid points.
 *
 *     Note: In the future, we may think of doing a moving grid trick
 *     in the Overture sense (see <A
 *     HREF="http://www.llnl.gov/casc/Overture/henshaw/documentation/primer.pdf">Overture
 *     primer</A>, p. 49).
 ************************************************************************/
void ITM_Interface::prepareExport(intArray &utime, realCompositeGridFunction & mhdFunctions)
{
  int iSMtoGSM = 1,
    iGSMtoSM = -1,
    iGEOtoGSM = 1,
    iGSMtoGEO = -1;
  int  r;

  realArray ptemp(1,2), utemp(1,3);

  real 
    xsm  = 0.,  // WARNING
    ysm  = 0.,  // Data type here is very important!!!
    zsm  = 0.,  // Overture compiles everything in double precision by default.
    xgsm = 0.,  // Therefore, we need to use double precision here as well
    ygsm = 0.,  // Use the 'real' #defined in Overture.
    zgsm = 0.,  // Thus if we choose to recompile Overture without default 
    xgeo = 0.,  // double precision this code is still supposed to work...
    ygeo = 0.,  // However, geopack, needs to be *forced* (via compiler flags) to 
    zgeo = 0.;  // use real*8 to match the Overture default double-preicision.

  // Prepare Geopack coeffs
  RECALC(&utime(0),&utime(1),&utime(2),&utime(3),&utime(4));

  //  The "mhdFunctions" grid functions (psi,averageEnergy,numberFlux)
  //  are defined on the entire globe in SM spherical angular coordinates 
  //  Convert to lat/lon & interpolate to ITM grid.
  for (int k = 0; k < nLon; k++) {
    for (int j = 0; j < nLat; j++) {    
      xgeo = geoPositions(j+k*nLat,0);
      ygeo = geoPositions(j+k*nLat,1);
      zgeo = geoPositions(j+k*nLat,2);

      // Rotate to SM coordinates
      GEOGSW_08(&xgeo,&ygeo,&zgeo, &xgsm,&ygsm,&zgsm, &iGEOtoGSM);
      SMGSW_08(&xsm,&ysm,&zsm, &xgsm,&ygsm,&zgsm, &iGSMtoSM);

      // Convert to lat/lon
      ptemp(0,0) = asin(zsm); // Latitude (-pi/2,pi2) as in TING grid definition
      ptemp(0,1) = atan2(ysm,xsm);

      // Interpolate
      r = interpolatePoints(ptemp,mhdFunctions,utemp); 
      pot(j,k) = utemp(0,SM_GLOBE_POTENTIAL);
      // MJW July 15 2010: Convert from characteristic energy to
      // average energy.  We made need to tweak this down the road to
      // get the fluxes right.
      //eng(j,k) = utemp(0,SM_GLOBE_AVG_ENG);
      eng(j,k) = 0.5*utemp(0,SM_GLOBE_AVG_ENG);
      flux(j,k) = utemp(0,SM_GLOBE_NUM_FLUX);
    }
  }
  // pot now contains the potential for the ITM
  // eng now contains the energy for the ITM
  // flux now contains the flux for the ITM
}

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 06-2007
 *
 * \param[out] sigmaP_N Pedersen conductance (North) on Solver grid
 * \param[out] sigmaP_S Pedersen conductance (South) on Solver grid
 * \param[out] sigmaH_N Hall conductance (North) on Solver grid
 * \param[out] sigmaH_S Hall conductance (South) on Solver grid
 * \param[out] Unw_N FAC due to neutral wind (North) on Solver grid
 * \param[out] Unw_S FAC due to neutral wind (South) on Solver grid
 * 
 *  Transform ITM_Interface member variables (ped,hall,wfac) from ITM
 *  model lat/lon grid to geographic angular coordinates (in the sense
 *  of the ITM_Grids definition), and interpolate.  Store the results
 *  in mhdFunctions.
 ************************************************************************/
void ITM_Interface::processImport(realCompositeGridFunction & mhdFunctions)
{
  // I'm going to inefficiently create grid functions here
  // because i don't know of a way to link them to the ITM grid,
  // which is created when the constructor is called, if I defined them 
  // class-wide using the default constructor
  realCompositeGridFunction itmFunctions(gridSph,all,all,all,3);

  // First fill in the interior of the grid functions with arrays received from ITM
  Index J1,J2,J3;
  getIndex(gridSph[0].gridIndexRange(),J1,J2,J3);
  itmFunctions[0](J1,J2,all,0) = ped;
  itmFunctions[0](J1,J2,all,1) = hall;
  itmFunctions[0](J1,J2,all,2) = wfac;

  int iSMtoGSM = 1,
    iGSMtoSM = -1,
    iGEOtoGSM = 1,
    iGSMtoGEO = -1;
  int  r;

  realArray ptemp(1,2), utemp(1,3);

  real 
    xsm  = 0.,  // Note!!! 
    ysm  = 0.,  // Data type here is very important
    zsm  = 0.,  // Overture compiles everything in double precision
    xgsm = 0.,  // by default (we chose this option in the Makefile)
    ygsm = 0.,  // As a result Geopack uses real*8
    zgsm = 0.,  // Therefore, need to use double precision here as well
    xgeo = 0.,  // We use the 'real' #defined in Overture
    ygeo = 0.,  // Thus if we choose to recompile Overture without default double precision
    zgeo = 0.;  // This code is still supposed to work

  MappedGrid *grid = mhdFunctions[0].getMappedGrid();
  Index I1,I2,I3;
  getIndex((*grid).gridIndexRange(),I1,I2,I3);
  const realArray & x = (*grid).vertex()(I1,I2,I3,0);
  const realArray & y = (*grid).vertex()(I1,I2,I3,1);
  const int nx = x.getBound(0) - x.getBase(0) + 1;
  const int ny = x.getBound(1) - x.getBase(1) + 1;

  //  Loop over the Solver grid points
  for (int k = 0; k < ny; k++) {
    for (int j = 0; j < nx; j++) {

      // Work separately for the two hemispheres
      for (int HEMISPHERE = 0; HEMISPHERE <=1; HEMISPHERE++) {

	xsm = x(j,k);
	if ( HEMISPHERE == NORTH ) {  // SM NORTH
	  ysm = y(j,k);
	  zsm = sqrt(1.-xsm*xsm-ysm*ysm);
	}
	else {              // SM SOUTH
	  ysm = -y(j,k);
	  zsm = -sqrt(1.-xsm*xsm-ysm*ysm);
	}

 	// rotate to GEO coordinates
	// i.e. convert to geographical angular coordinates (in the sense of the ITM_Grids definition)
 	SMGSW_08(&xsm,&ysm,&zsm, &xgsm,&ygsm,&zgsm, &iSMtoGSM);
 	GEOGSW_08(&xgeo,&ygeo,&zgeo, &xgsm,&ygsm,&zgsm, &iGSMtoGEO);

 	ptemp(0,0) = asin(zgeo); // Latitude (-pi/2,pi2) as in TING grid definition
 	ptemp(0,1) = atan2(ygeo,xgeo);

	// Interpolate!
	r = interpolatePoints(ptemp,itmFunctions,utemp); 

	// Notice the cap on the conductances SIGMAP and SIGMAH below.
	// Without this fix, the MIX current acts strangely near the
	// low-latitude boundary near midnight.

	if ( HEMISPHERE == NORTH ) {  // SM NORTH
	  mhdFunctions[0](j,k,all,SOLVER_SIGMAP_NORTH) = utemp(0,0);
	  mhdFunctions[0](j,k,all,SOLVER_SIGMAH_NORTH) = utemp(0,1);
	  mhdFunctions[0](j,k,all,SOLVER_WIND_NORTH)   = utemp(0,2);
	}
	else {               // SM SOUTH
	  mhdFunctions[0](j,k,all,SOLVER_SIGMAP_SOUTH) = utemp(0,0);
	  mhdFunctions[0](j,k,all,SOLVER_SIGMAH_SOUTH) = utemp(0,1);
	  mhdFunctions[0](j,k,all,SOLVER_WIND_SOUTH)   = utemp(0,2);
	}
      } // END SM NORTH/SOUTH LOOP
    }
  }
  // now mhdFunctions should have the following on the Solver grid:
  //   - sigmaP_N & sigmaP_S - North & South Pedersen Conductances
  //   - sigmaH_N & sigmaH_S - North & South Hall Conductances
  //   - Unw_N FACE & Unw_S FAC - North & South neutral wind field-aligned currents
}

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 06-2007
 *
 * \param[in] UT intArray of (Year,Month,Day,Hour,Min,Sec)
 * \return intArray of (Year,DOY,Hour,Min,Sec)
 ************************************************************************/
intArray ITM_Interface::ut2utdoy(const intArray & UT)
{
  time_t rawtime;
  struct tm * timeinfo;
  time(&rawtime);
  timeinfo = localtime( &rawtime );
  timeinfo->tm_year = UT(0) - 1900;
  timeinfo->tm_mon = UT(1) - 1;
  timeinfo->tm_mday = UT(2);
  mktime( timeinfo );

  intArray dummy(5);
  dummy(0) = UT(0); // Year
  dummy(1) = timeinfo->tm_yday + 1; //DOY
  dummy(2) = UT(3);  // Hour
  dummy(3) = UT(4);  // Min
  dummy(4) = UT(5);  // Sec

  return dummy;
}
