#ifndef ITM_INTERFACE_H
#define ITM_INTERFACE_H

/**
 * \file ITM_Interface.h
 * \brief ITM_Interface class declaration
 * \class ITM_Interface
 * 
 * \brief Contains functions for communications with the ITM
 *
 * \author Slava Merkin (vgm at bu.edu)
 * \since 06-2007
 *
 * This class contains functions and variables that establish
 * communication with the ITM model.
*/

// Standard Includes
#include <time.h>

// 3rd-party includes
#include "InterComm.h"
#include "Overture.h"
#include "PlotStuff.h"
#include "interpolatePoints.h"

// Custom includes
#include "Params.h"
#include "ITM_Grids.h"
#include "DateTime.h"
#include "functions.h"

#ifdef LC_UNSC
#define  RECALC recalc_
#define SMGSW_08 smgsw_08_
#define GEOGSW_08 geogsw_08_
#elif LC_NOUNSC
#define RECALC recalc
#define SMGSW_08 smgsw_08
#define GEOGSW_08 geogsw_08
#else
#error "Must define one of (LC_UNSC,LC_NOUNSC) for proper FORTRAN name mangling"
#endif

/**\name GEOPACK functions*/
extern "C" {
  void RECALC(int *, int *, int *, int *, int *);
  void SMGSW_08 (real *, real *, real *, real *, real *, real *, int *);
  void GEOGSW_08 (real *, real *, real *, real *, real *, real *, int *);
} 

class ITM_Interface : public ITM_Grids {

 public:
  /**\brief Class constructor*/
  ITM_Interface(Params *, IC_XJD *);

  /**\brief Class destructor */
  virtual ~ITM_Interface() {};

  /**\brief Import data to ITM*/
  virtual void Import(realCompositeGridFunction &) = 0;

  /**\brief Export data from ITM*/
  virtual void Export(const intArray &, realCompositeGridFunction &) = 0;

  /**\brief Overloaded member to Export data from ITM */
  void Export(DateTime &, realCompositeGridFunction &);

 protected:
  /**\brief Get the transformed Potential, Energy & Flux necessary before export*/
  void prepareExport(intArray &utime, realCompositeGridFunction & mhdFunctions);

  /**\brief Transform the Pedersen & Hall conductances and the Neutral Wind Field Aligned Current after Import. */ 
  void processImport(realCompositeGridFunction & mhdFunctions);

  /**\brief UT to UT & day of year */
  intArray ut2utdoy(const intArray &);

  /**\name InterComm variables */
  IC_XJD *xjd;

  /**\name Universal Time variable */
  intArray utime;

  /**\name Arrays for communications with the ITM model */
  realArray pot, eng, flux, ped, hall, wfac;
};

#endif
