/**
 * \file MHD_FE_Interface.C
 * \brief MHD_FE_Interface definition.
 */

#include "MHD_FE_Interface.h"

using namespace std;

/********************************************************************//**
 * \author Peter Schmitt (schmitt at ucar.edu)
 * \since 09-2008
 * \param[in] params The pointer to the Params object
 * \param[in] XJD The pointer to the IC_XJD object
 *
 * The constructor inherits the MHD grid information from
 * MHD_Interface. Then arrays to be transmitted between this code and the
 * MHD model are dimensioned (MHD_Grids::njp1,MHD_Grids::nkp1). The
 * file exchange lock files are initialized.
 ************************************************************************/
MHD_FE_Interface::MHD_FE_Interface(Params *params, IC_XJD *XJD)
  : MHD_Interface(params, XJD)
{
  // stdout that we're using the file exchange interface
  cout << "Initializing MIX File Exchange Interface..."<<endl;
  ///////////////////////////////////////////////////////////////////////  
  MIXLockFile = "MIX_status";
  MHDLockFile = "MHD_status";

  // Set MIX data to "OLD"
  assert( setMIXOld() );

  ///////////////////////////////////////////////////////////////////////
  // These are set for a sequential program. If I ever make this code parallel
  // these should be changed (see InterComm manual)
  int tasks[1] = {0};

  int njp1 = MHD_Grids::njp1;
  int nk2p1 = MHD_Grids::nk2p1;
  int nkp1 = MHD_Grids::nkp1;

  // Sleep to prevent any buffered I/O issues with file exchanges  
  sleep(5);
}

MHD_FE_Interface::~MHD_FE_Interface()
{
#ifdef DEBUG_MODE_ON
  cout << "  MIX_FE: Destructor called.\n";
#else
  // Leave the files alone in case ITM finishes before MHD.
  //assert( remove(MIXLockFile.c_str()) == 0 );
  //remove("MIX_potential");
#endif
}

/********************************************************************//**
 * \author Peter Schmitt (schmitt at ucar.edu)
 * \since 09-2008
 * \param[out] jpara_N MHD parallel current in the northern hemisphere
 * \param[out] jpara_S MHD parallel current in the southern hemisphere
 * \param[out] rho_N    MHD density in the northern hemisphere
 * \param[out] rho_S    MHD density in the southern hemisphere
 * \param[out] cs_N     MHD sound speed in the northern hemisphere
 * \param[out] cs_S     MHD sound speed in the southern hemisphere
 *
 * The function does three things: 
 *    # Import current, density, and sound speed using file exchanges
 *    # call MapToIonosphere() for each of these variables one time
 *      for each hemisphere
 *    # converts the resulting arrays into grid functions passed to 
 *      Import() as parameters.
 ************************************************************************/
void MHD_FE_Interface::Import(realCompositeGridFunction & mhdFuncs)
{
#ifdef DEBUG_MODE_ON
  cout << "DEBUG: MHD_FE_Interface::Import(...)" << endl;
#endif

  // Wait until MHD data is "new"
  while ( ! isMHDNew() ){
#ifdef DEBUG_MODE_ON
    std::cout << "  MIX_FE: Waiting for MHD to export current, density & soundSpeed...\n" << flush;
#endif
    sleep(4);
  }

#ifdef DEBUG_MODE_ON
  std::cout << "  MIX_FE: MHD exported variables. Read them in.\n" << flush;
#endif
  ///////////////////////////////////////////////////////////////////////
  // Import (read the data from file. . . )

  read3dData("MHD_current", current, njp1, nkp1, 0);
#ifdef DEBUG_MODE_ON
  write3dData("MIX_current", current, njp1, nkp1, 0);
#endif

  read3dData("MHD_density", density, njp1, nkp1, 0);
#ifdef DEBUG_MODE_ON
  write3dData("MIX_density", density, njp1, nkp1, 0);
#endif
  
  read3dData("MHD_soundSpeed", soundSpeed, njp1, nkp1, 0);
#ifdef DEBUG_MODE_ON
  write3dData("MIX_soundSpeed", soundSpeed, njp1, nkp1, 0);
#endif

  ///////////////////////////////////////////////////////////////////////

  // MIX is done reading.  Tell the MHD to continue:
#ifdef DEBUG_MODE_ON
  std::cout << "  MIX_FE: current, density & soundSpeed read in by MIX.\n";
#endif

  // MIX Is done reading the "new" variables.  Set MHD status to "old"
  assert( setMHDOld() );

  ///////////////////////////////////////////////////////////////////////

  // map to the ionosphere & convert arrays into grid functions
  processImport(mhdFuncs);
}

/********************************************************************//**
 * \author Peter Schmitt (schmitt at ucar.edu)
 * \since 09-2008
 *
 * \param[in] psi1_NFunction Northern hemisphere potential grid
 * function to be mapped to the first MHD shell
 *
 * \param[in] psi1_SFunction  Same for the southern hemisphere
 *
 * \param[in] psi2_NFunction Northern hemisphere potential grid
 * functionto be mapped to the second MHD shell
 *
 * \param[in] psi2_SFunction  Same for the southern hemisphere
 *
 * This is the reverse interface for sending the potential obtained by
 * the elliptic solver back to the MHD code. Regardless of whether the
 * MHD code needs it or not, we send back the array containing the
 * potential for the two innermost MHD shells, i.e. dimensioned
 * (2,#njp1,#nkp1).
 *
 * The function first converts the grid functions passed as parameters
 * to plain arrays, then calls MapFromIonosphere(), and finally writes
 * data to the "MIX_potential" file, (i.e. the data is "Exported" to
 * the MHD code.)
 ************************************************************************/
void MHD_FE_Interface::Export(realCompositeGridFunction & mhdFunctions1, 
			      realCompositeGridFunction & mhdFunctions2 )
{
#ifdef DEBUG_MODE_ON
  cout << "DEBUG: MHD_FE_Interface::Export(...)" << endl;
#endif

  prepareExport(mhdFunctions1, mhdFunctions2);

  // Wait until MHD has set the MIXstatus to "old".
  while ( ! isMIXOld() ){
#ifdef DEBUG_MODE_ON
    std::cout << "  MIX_FE: Waiting for MHD to acknowledge that its ready for potential...\n" << flush;
#endif
    sleep(4);
  }  

  ///////////////////////////////////////////////////////////////////////
  // Export the data to MIX using file exchanges.

  write3dData("MIX_potential", potential, 2, njp1, nkp1);

#ifdef DEBUG_MODE_ON
  cout << "  MIX_FE: MIX exported potential to MHD.\n" << flush;
#endif

  // Set MHD status to "new" since we've just sent the potential
  assert( setMIXNew() );
  
#ifdef DEBUG_MODE_ON
  cout << "  MIX_FE: MHD receieved potential.  Moving on.  \n";
#endif
}

/********************************************************************//**
 * \author Peter Schmitt (schmitt at ucar.edu)
 * \since 09-2008
 *
 * Use this function to receive an array of scalars broadcast by the
 * MHD code.  The scalars to be sent are defined in the CPL_Scalars
 * enumeration defined in MHD_Interface.h
 *
 * \return   The array of scalars to be received.
 ************************************************************************/
realArray MHD_FE_Interface::getScalars()
{
#ifdef DEBUG_MODE_ON
  cout << "DEBUG: MHD_FE_Interface::getScalars(...)" << endl;
#endif

  // Wait until MHD data is "new"
  while ( ! isMHDNew() ){
#ifdef DEBUG_MODE_ON
    std::cout << "  MIX_FE: Waiting for MHD to update new scalars..." << endl;
#endif
    sleep(4);
  }
#ifdef DEBUG_MODE_ON
  std::cout << "  MIX_FE: MHD has new scalars ready! Start reading them in." << endl;
#endif

  ///////////////////////////////////////////////////////////////////////

  // Open "MHD_scalars" for reading
  std::ifstream ins("MHD_scalars");
  assert( !ins.bad() );

  // Read scalars
  realArray scalars(MHD_Interface::NUMBER_OF_SCALARS);
  ins >> scalars(MHD_Interface::KILL_SIGNAL);
  ins >> scalars(MHD_Interface::YEAR);
  ins >> scalars(MHD_Interface::MONTH);
  ins >> scalars(MHD_Interface::DAY);
  ins >> scalars(MHD_Interface::HOUR);
  ins >> scalars(MHD_Interface::MINUTE); 
  ins >> scalars(MHD_Interface::SECOND); // Second  (double-precisoin)
  ins >> scalars(MHD_Interface::TILT);  
  ins >> scalars(MHD_Interface::LABEL); // label (mhd time step number)
  
  // Close "MHD_scalars" file handle
  ins.close();

  ///////////////////////////////////////////////////////////////////////

  // MIX Is done reading the "new" scalars.  Set MHD status to "old"
  assert( setMHDOld() );

  return scalars;
}


/************************************************************************
 * Private member functions
 ***********************************************************************/

/********************************************************************//**
 * \author Peter Schmitt (schmitt at ucar.edu)
 * \since 09-2008
 *
 * \param[in] filename - string pointing to path of lock file to write
 * \param[in] status - string that sets the MIX lock file status
 * (e.g. "WAITING" or "WORKING").
 *
 * \param[out] bool - returns true if succesful.
 *
 * Write to code status to the ASCII text file MIXLockFile
 ************************************************************************/
bool MHD_FE_Interface::writeLockFile(const string &filename, const string &status)
{
#ifdef DEBUG_MODE_ON
  cout << "DEBUG: MHD_FE_Interface::writeLockFile(...)" << endl;
#endif
  // open the file for writing
  std::ofstream outs(filename.c_str());
  if (outs.bad()){
    std::cerr << "*** Trouble writting " << filename << "\n";
    return false;
  }

  // write status
  outs << status << "\n";

  // Make sure to flush the IO to disk.
  outs.flush();

  // close the file
  outs.close();
  
  return true;
}

/********************************************************************//**
 * \author Peter Schmitt (schmitt at ucar.edu)
 * \since 09-2008
 *
 * \param[in] filename - path to file
 * \param[out] text contents of file 'filename'.  Should be "NEW" or "OLD".
 *
 * Returns the contents of the ASCII file MHDLockFile.
 ************************************************************************/
string MHD_FE_Interface::readLockFile(const string &filename)
{
#ifdef DEBUG_MODE_ON
  cout << "DEBUG: MHD_FE_Interface::readLockFile(...)" << endl;
#endif

  // open the file for reading
  std::ifstream ins(filename.c_str());
  if (ins.bad()){
    std::cerr << "*** Trouble reading " << filename << "\n";
    // FIXME:  Error handling?
    return false;
  }

  // read status
  string statusString;
  ins >> statusString;

  // close the file
  ins.close();

  return statusString;
}


/********************************************************************//**
 * \author Peter Schmitt (schmitt at ucar.edu)
 * \since 09-2008
 *
 * Read 3d data from #filename into #data
 *
 * \param[in] filename - Points to the file to read data from
 * \param[out] data - 1, 2 or 3d array containing data that will be written to #filename
 * \param[in] ni - number of elements in i direction
 * \param[in] nj - number of elements in j direction (nj and nk should be 0 for 1d data)
 * \param[in] nk - number of elements in k direction (should be 0 for 2d data)
 *
 * \note the file format defined here needs to match that defined in
 * MHD_FE_InnerBoundaryInterface::write3dData
 *
 * \FIXME #data could be a template so we can handle any dimension
 * data?
 ************************************************************************/
void MHD_FE_Interface::read3dData(const string &filename, const doubleArray &data,
				  const int &ni, const int &nj, const int &nk)
{
#ifdef DEBUG_MODE_ON
  cout << "DEBUG: MHD_FE_Interface::read3dData(...)" << endl;
#endif

  // open the file for writing
  std::ifstream ins(filename.c_str());
  if (ins.bad()){
    std::cerr << "*** Trouble Writting " << filename << "\n";
  }

  // read data
  int ni_in, nj_in, nk_in;

  ins >> ni_in >> nj_in >> nk_in;
  assert( ni_in == ni );
  assert( nj_in == nj );
  assert( nk_in == nk );

  for (int k = 0; k < max(nk, 1); k++){
    for (int j = 0; j < max(nj, 1); j++){
      for (int i = 0; i < max(ni, 1); i++){
        ins >> data(i,j,k);
      }
    }
  }

  // close the file
  ins.close();
}

/********************************************************************//**
 * \author Peter Schmitt (schmitt at ucar.edu)
 * \since 09-2008
 *
 * Write 3d array #data to #filename
 *
 * \param[in] filename - filename to write data to
 * \param[in] data - 1, 2 or 3d array containing data that will be written to #filename
 * \param[in] ni - number of elements in i direction
 * \param[in] nj - number of elements in j direction (nj and nk should be 0 for 1d data)
 * \param[in] nk - number of elements in k direction (should be 0 for 2d data)
 *
 * \note The file format defined here needs to match that defined in
 * MHD_FE_Interface::read3dData
 *
 * \FIXME #data could be a template so we can handle any dimension data?
 ************************************************************************/
void MHD_FE_Interface::write3dData(const string &filename, const doubleArray &data,
				   const int &ni, const int &nj, const int &nk)
{
#ifdef DEBUG_MODE_ON
  cout << "DEBUG: MHD_FE_Interface::write3dData(...)" << endl;
#endif

  // open the file for writing
  std::ofstream outs(filename.c_str());
  if (outs.bad()){
    std::cerr << "*** Trouble Writting " << filename << "\n";
  }

  // write data
  outs << ni << "\t" << nj << "\t" << nk << "\n";
  outs.precision(8);
  for (int k = 0; k < max(nk, 1); k++){
    for (int j = 0; j < max(nj, 1); j++){
      for (int i = 0; i < max(ni, 1); i++){
        outs << data(i,j,k) << "\t";
      }
      outs << "\n";
    }
    outs << "\n";
  }

  // Make sure to flush the IO to disk.
  outs.flush();

  // close the file
  outs.close();
}
