#ifndef MHD_FE_INTERFACE_H
#define MHD_FE_INTERFACE_H

/**
 * \file MHD_FE_Interface.h
 * \brief MHD_FE_Interface declaration
 *
 * \class MHD_FE_Interface 
 * \brief The MHD_FE_Interface class utilizes file exchanges (FE) for
 * communication with the MHD model.
 *
 * \author Peter Schmitt (schmitt@ucar.edu)
 * \since 09-2008
 *
 * This class defines functions for communication between an MHD code
 * (i.e. LFM) and the Ionosphere Coupler/Solver (MIX) code using file
 * exchanges.  Communication occurs through a series of I/O to the
 * following files:
 *
 *   - MIX_status
 *   - MIX_potential
 *
 * If your computing environment supports it, you should use the
 * InterComm version of this class (MHD_FE_Interface).
 * The InterComm version communicates with MIX much more efficiently.
 *
 * \see MHD_Interface, MHD_IC_Interface, MHDInnerBoundaryInterface, CPL_FE_Interface
 */

#include <iostream>
#include <fstream>
#include <cstdlib>
#include <string>

#include "MHD_Interface.h"

class MHD_FE_Interface : public MHD_Interface {

 public:
  /** \brief The class constructor. */
  MHD_FE_Interface(Params *params, IC_XJD *XJD);

  virtual ~MHD_FE_Interface();

  /** \brief The function for import from MHD to ionosphere. */
  void Import(realCompositeGridFunction &mhdFuncs);

  /** \brief Sends potential back to MHD. */
  void Export(realCompositeGridFunction &mhdFunctions1, realCompositeGridFunction &mhdFunctions2);
 
  /** \brief Receives a scalar array from the MHD code. Intended to
   * communicate scalars such as UT and kill signal */
  realArray getScalars();

 private:

  /** \name File Exchange members. */
  //@{

  /**\brief Returns true if the MHD is busy */
  bool isMHDOld()  { return ( strcmp( "OLD", readLockFile(MHDLockFile).c_str() ) == 0 ); }
  /**\brief Returns true if the MHD is ready and waiting for MIX */
  bool isMHDNew() { return ( strcmp( "NEW", readLockFile(MHDLockFile).c_str() ) == 0 ); }

  /**\brief Returns true if the MIX is busy */
  bool isMIXOld()  { return ( strcmp( "OLD", readLockFile(MIXLockFile).c_str() ) == 0 ); }
  /**\brief Returns true if the MIX is ready and waiting for MIX */
  bool isMIXNew() { return ( strcmp( "NEW", readLockFile(MIXLockFile).c_str() ) == 0 ); }
  
  /**\brief Set the MHD lock file to "OLD" (aka busy)  */
  bool setMHDOld(void)   { return writeLockFile(MHDLockFile, "OLD"); }
  /**\brief Set the MHD lock file to "NEW" (aka ready) */
  bool setMHDNew(void)  { return writeLockFile(MHDLockFile, "NEW"); }

  /**\brief Set the MIX lock file to "OLD" (aka busy)  */
  bool setMIXOld(void)   { return writeLockFile(MIXLockFile, "OLD"); }
  /**\brief Set the MIX lock file to "NEW" (aka ready) */
  bool setMIXNew(void)  { return writeLockFile(MIXLockFile, "NEW"); }

  /**\brief  Write #status to the MIX lock file */
  bool writeLockFile(const string &filename, const string &status);

  /**\brief  Read the MHD lock file to check its status */
  string readLockFile(const string &filename);

  /**\brief  Read a 3d array from file into #data*/
  void read3dData(const string &filename, const doubleArray &data,
		  const int &ni, const int &nj, const int &nk);
  /**\brief  Write #data to file */
  void write3dData(const string &filename, const doubleArray &data,
		   const int &ni, const int &nj, const int &nk);

  /**\brief filename of MIX lock file. */
  string MIXLockFile;
  /**\brief filename of MHD lock file. */
  string MHDLockFile;
  //@}
};

#endif
