/**
 * \file MHD_Grids.C 
 * \brief MHD_Grids definition.
 */

#include "MHD_Grids.h"

using namespace std;

//Range all;

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 08-22-2005 
 * \param[in] gridFileName    Grid file name.
 * \param[in] fType   Grid file type.
 *
 * The grid file type is one of those defined in the enumerator
 * #gridFileType. The constructor is not called directly at the
 * present although it is possible. Rather, it is called by the child
 * class MHD_Interface constructor, whereas the grid file name and
 * grid file type are defined in the parameter file ionosphere.param.
 *
 * Depending on the value of the #gridFileType the right
 * generateFromFILETYPE() function is called.
 ************************************************************************/
MHD_Grids::MHD_Grids(const char *gridFileName, const int gridFileType)  {
    switch ( gridFileType ) {
    case HDFFILE:
      generateFromHDF(gridFileName);
      break;
    case GRIDOUT: 
      generateFromGRIDOUT(gridFileName);
      break;
    default:
      cout << "The grid file type entered is not supported."<<endl;
      exit(1);
    }
};

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 08-22-2005 
 *
 * \param[in] gridFileName The grid file name.
 *
 * Read the LFM GRIDOUT file and fill in the values of the private
 * grid size variables (#ni, #nip1, #nj, #njp1, #nk, #nkp1, and
 * #nk2p1); then call the function fillInGridArrays() which will fill
 * in the private arrays containing vertex coordinates (#ionArray1,
 * #ionArray2, #ionXYArray1 and #ionXYArray2). These arrays are later used by the
 * public grid generation functions getFirstShellIonSphericalGrid(),
 * getSecondShellIonSphericalGrid(), and
 * getSecondShellIonCartesianGrid(). The ionospheric coordinates are
 * calculated from the MHD coordinates by mapping along the dipole
 * magnetic field lines.
 ************************************************************************/
void MHD_Grids::generateFromGRIDOUT(const char *gridFileName)  {
  cout << "Reading grids from " <<gridFileName<<" ...............";
  ifstream gridFile(gridFileName,ios::in);
  if (!gridFile) {
    cerr << "Could not open the file " <<gridFileName<<endl;
    exit(1);
  }

  int mjin, mjwall;
  gridFile >> nip1 >> njp1 >> mjin >> mjwall >> nkp1;
  ni = nip1 - 1;
  nj = njp1 - 1;
  nk = nkp1 - 1;
  nk2p1 = nk/2+1;

  realArray phi(nkp1);
  for (int k = 0; k < nkp1; k++) {gridFile >> phi(k);}

  realArray x2(nip1,njp1), y2(nip1,njp1);
  for (int j = 0; j < njp1; j++) {
    for (int i = 0; i < nip1; i++) {
      gridFile >> x2(i,j);
    }
  }

  for (int j = 0; j < njp1; j++) {
    for (int i = 0; i < nip1; i++) {
      gridFile >> y2(i,j);
    }
  }
  gridFile.close();

  realArray x(nip1,njp1,nkp1), y(nip1,njp1,nkp1), z(nip1,njp1,nkp1);
  for (int k = 0; k < nkp1; k++) {
    for (int j = 0; j < njp1; j++) {
      for (int i = 0; i < nip1; i++) {
        x(i,j,k) = x2(i,j);
        y(i,j,k) = y2(i,j)*cos(phi(k)/180.*PI);;
        z(i,j,k) = y2(i,j)*sin(phi(k)/180.*PI);
      }
    }
  }

  fillInGridArrays(x,y,z);
  cout << "done"<<endl;

  cout << "Grid dimensions are: " << ni << "x" << nj << "x" << nk << endl;
 }

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 08-22-2005 
 *
 * \param[in] gridFileName The grid file name.
 *
 * \attention THIS FUNCTION IS OBSOLETE SINCE 01-16-2008. THE
 * OLD_STYLE LFM HDF FILES ARE NO LONGER USED
 *
 * Read the LFM HDF file. Then see generateFromGRIDOUT().
 ************************************************************************/
void MHD_Grids::generateFromHDF_old(const char *gridFileName)  {
  cout << "Reading grids from " <<gridFileName<<" ...............\n";
  // Open hdf file
  //  char* fName = const_cast<char*> ( gridFileName.c_str() );  
  int fileID = SDstart(gridFileName,DFACC_READ);
  if (fileID < 0) {
    cout << "Unable to open MHD HDF file for read!\n";
    exit(1);
  }

  // read grid description
  int attrDesc = SDfindattr(fileID,"run_descriptor");
  char gridLabel[STRLEN];
  int status = SDreadattr(fileID,attrDesc,gridLabel);
  cout << "\tGrid info from HDF file:\n\t"<<gridLabel << endl;

  // who wrote the file
  attrDesc = SDfindattr(fileID,"written_by");
  char gridFileOwner[STRLEN];
  status = SDreadattr(fileID,attrDesc,gridFileOwner);
  cout << "\tHDF file written by:\n\t"<< gridFileOwner << endl;

  // read grid info
  int varIndex = SDnametoindex(fileID,"grid_x2");
  int varID = SDselect(fileID, varIndex);
  
  char varName[64];
  int32 varRank, varDims[4],varDataType,varNAttr;  // assume the maximum dimension of arrays = 4
  status = SDgetinfo(varID,varName,&varRank,varDims,&varDataType,&varNAttr);
  if (status < 0) { cout << "SDgetinfo failed!" << endl; exit(1); }

  // get array sizes that are stored as attributes
  attrDesc = SDfindattr(varID,"ni");
  status = SDreadattr(varID,attrDesc,&nip1);
  attrDesc = SDfindattr(varID,"nj");
  status = SDreadattr(varID,attrDesc,&njp1);
  cout << "##\n\tx2, y2 array dimensions: "<<  nip1 << "x" <<njp1 << endl;

  float32 x2c[njp1][nip1],y2c[njp1][nip1];
  int32 start[1],stride[1],edge[1];
  start[0] = 0; stride[0] = 1; edge[0] = nip1*njp1;
  status = SDreaddata(varID,start,stride,edge,x2c);
  floatArray x2(&x2c[0][0],nip1,njp1);
  x2 = x2/RION;

  varIndex = SDnametoindex(fileID,"grid_y2");
  varID = SDselect(fileID, varIndex);
  status = SDreaddata(varID,start,stride,edge,y2c);
  floatArray y2(&y2c[0][0],nip1,njp1);
  y2 = y2/RION;
 
  varIndex = SDnametoindex(fileID,"grid_phi");
  varID = SDselect(fileID, varIndex);
  status = SDgetinfo(varID,varName,&varRank,varDims,&varDataType,&varNAttr);
  attrDesc = SDfindattr(varID,"ni");
  status = SDreadattr(varID,attrDesc,&nkp1);
  cout << "##\n\tphi array dimensions: "<<  nkp1 <<endl;
  float32 phic[nkp1];
  edge[0] = nkp1;
  status = SDreaddata(varID,start,stride,edge,phic);
  floatArray phi(phic,nkp1);

  // fill the grid dimension variables
  ni = nip1 - 1;
  nj = njp1 -1;
  nk = nkp1 - 1;
  nk2p1 = nk/2 + 1;
  fillInGridArrays(x2.convertTo_doubleArray(),
			  y2.convertTo_doubleArray(),
			  phi.convertTo_doubleArray());
  cout << "done"<<endl;
}

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 01-16-2008
 *
 * \param[in] gridFileName The grid file name.
 *
 * Read the LFM HDF file. Then see generateFromGRIDOUT()
 *
 * FIXME: File I/O has high-latency.  Leverage InterComm to have LFM
 * send MIX the grid information.
 *
 * FIXME:  Leverage a generic "Hdf4" class to do grid I/O?  See LFM Startup Generator for more information.
 ************************************************************************/
void MHD_Grids::generateFromHDF(const char *gridFileName)  {
  cout << "Reading grids from " <<gridFileName<<" ...............\n";
  // Open hdf file
  int fileID = SDstart(gridFileName,DFACC_READ);
  if (fileID < 0) {
    cout << "Unable to open MHD HDF file for read!\n";
    exit(1);
  }

  // read grid description
  int attrDesc = SDfindattr(fileID,"file_contents");
  char attrName[MAX_NC_NAME];
  int32 varDataType;
  int32 nCharacters;
  int status = SDattrinfo(fileID, attrDesc, attrName, &varDataType, &nCharacters);
  char *attrStr = new char[nCharacters+1];
  attrStr[nCharacters] = '\0';
  status = SDreadattr(fileID,attrDesc,attrStr);
  if (status < 0){
    cerr << "*** Error in " << __FILE__ << "(L " << __LINE__ << "): " << __FUNCTION__ << "(...)" << endl;
    cerr << "Reading \"" << gridFileName << "\": "
	 << "HDF4 error code: " << status << endl;
    exit(1);
  }  
  cout << "\tHDF file contents \""<<attrStr << "\"" << endl;
  delete []attrStr;
  attrStr = NULL;


  // who wrote the file
  attrDesc = SDfindattr(fileID,"written_by");
  status = SDattrinfo(fileID, attrDesc, attrName, &varDataType, &nCharacters);
  attrStr = new char[nCharacters+1];
  attrStr[nCharacters] = '\0';
  status = SDreadattr(fileID,attrDesc,attrStr);
  if (status < 0){
    cerr << "*** Error in " << __FILE__ << "(L " << __LINE__ << "): " << __FUNCTION__ << "(...)" << endl;
    cerr << "Reading \"" << gridFileName << "\": "
	 << "HDF4 error code: " << status << endl;
    exit(1);
  }  
  cout << "\tHDF file written by \""<< attrStr << "\"" << endl;
  delete []attrStr;
  attrStr = NULL;

  // read grid info
  int varIndex = SDnametoindex(fileID,"X_grid");
  int varID = SDselect(fileID, varIndex);
  
  char varName[64];
  int32 varRank, varDims[3],varNAttr;
  status = SDgetinfo(varID,varName,&varRank,varDims,&varDataType,&varNAttr);
  if (status < 0){
    cerr << "*** Error in " << __FILE__ << "(L " << __LINE__ << "): " << __FUNCTION__ << "(...)" << endl;
    cerr << "Reading \"" << gridFileName << "\": "
	 << "HDF4 error code: " << status << endl;
    exit(1);
  }  


  // Note dimensions are transposed because we are reading into A++
  // floatArray, which uses Fortran column-major ordering
  nip1 = varDims[2];
  njp1 = varDims[1];
  nkp1 = varDims[0];
  cout << "\tGrid dimensions (nip1,njp1,nkp1): ("<<  nip1 << "," 
       << njp1 <<"," << nkp1 << ")" << endl;

  float32 x[nip1][njp1][nkp1],y[nip1][njp1][nkp1],z[nip1][njp1][nkp1];
  int32 start[3];
  start[0] = 0; start[1] = 0; start[2] = 0;
  status = SDreaddata(varID,start,NULL,varDims,x);
  status = SDendaccess(varID);
  floatArray xGrid(&x[0][0][0],nip1,njp1,nkp1);

  varIndex = SDnametoindex(fileID,"Y_grid");
  varID = SDselect(fileID, varIndex);
  status = SDreaddata(varID,start,NULL,varDims,y);
  status = SDendaccess(varID);
  floatArray yGrid(&y[0][0][0],nip1,njp1,nkp1);
 
  varIndex = SDnametoindex(fileID,"Z_grid");
  varID = SDselect(fileID, varIndex);
  status = SDreaddata(varID,start,NULL,varDims,z);
  status = SDendaccess(varID);
  floatArray zGrid(&z[0][0][0],nip1,njp1,nkp1);

  // close the file
  status = SDend(fileID);

  // fill the grid dimension variables
  ni = nip1 - 1;
  nj = njp1 -1;
  nk = nkp1 - 1;
  nk2p1 = nk/2 + 1;

#ifdef OV_USE_DOUBLE
  fillInGridArrays(evaluate(xGrid.convertTo_doubleArray()/RION),
		   evaluate(yGrid.convertTo_doubleArray()/RION),
		   evaluate(zGrid.convertTo_doubleArray()/RION));  
#else
  fillInGridArrays(evaluate(xGrid/RION),
		   evaluate(yGrid/RION),
		   evaluate(zGrid/RION));  
#endif
  cout << "done"<<endl;
}

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 08-22-2005 
 *
 * \param[in] x2 SM x-coordinates of the LFM x2 array (half-slice of the grid)
 * \param[in] y2 SM x-coordinates of the LFM y2 array (half-slice of the grid) 
 * \param[in] phi SM phi-angle (k-direction)
 *
 * This function is supposed to be called from generateFromGRIDOUT()
 * or generateFromHDF() functions that will read the grid dimensions
 * and coordinates from the corresponding files. The coordinate arrays
 * are then passed to this function which fills in values in
 * #ionArray1, #ionArray2, #ionXYArray1 and #ionXYArray2. If another
 * global MHD model needs to be coupled to the MIX, all we need is
 * that #ionXYArray1 and #ionXYArray2 arrays provide a regular
 * grid. Arrays #ionArray1 and #ionArray2 were used before; now they
 * are kept for consistency only.
 ************************************************************************/
void MHD_Grids::fillInGridArrays(const realArray &x, const realArray &y, const realArray &z) {
  Range K2(0,nk/2);

  realArray xFirstShell(1,njp1,nk2p1), yFirstShell(1,njp1,nk2p1), 
    zFirstShell(1,njp1,nk2p1), rFirstShell(1,njp1,nk2p1);
  xFirstShell = x(0,all,K2);
  yFirstShell = y(0,all,K2);
  zFirstShell = z(0,all,K2);
  rFirstShell = sqrt( xFirstShell*xFirstShell + yFirstShell*yFirstShell +zFirstShell*zFirstShell);

  realArray xSecondShell(1,njp1,nk/2+1), ySecondShell(1,njp1,nk/2+1),
               zSecondShell(1,njp1,nk/2+1), rSecondShell(1,njp1,nk/2+1);
  xSecondShell = x(1,all,K2);
  ySecondShell = y(1,all,K2);
  zSecondShell = z(1,all,K2);
  rSecondShell = sqrt( xSecondShell*xSecondShell + 
		       ySecondShell*ySecondShell +
		       zSecondShell*zSecondShell );

  realArray thetaFirstShell = asin( sqrt(xFirstShell*xFirstShell +
					 yFirstShell*yFirstShell) /rFirstShell);
  realArray phiFirstShell     = atan2(yFirstShell,xFirstShell);

  realArray thetaSecondShell = asin( sqrt(xSecondShell*xSecondShell + 
					  ySecondShell*ySecondShell) /rSecondShell);
  realArray phiSecondShell = atan2(ySecondShell,xSecondShell);

  // We create two grids in the ionosphere (x1ion,y1ion) and (x2ion,y2ion).
  // The first one maps to the first shell of the MHD grid.
  // The second one maps to the second shell of the MHD grid.
  // The dimension of the (x2ion,y2ion) grid has 2 additional points in the k-direction.
  // This creates a ring at the lowest latitude, where all physical variables are set to zero.
  // 03-06-07-VGM I am removing this additional ring as of today, because
  // Now i'm solving on a separate spherical grid and don't need it
  realArray x2ion(1,njp1,nk2p1);
  realArray y2ion(1,njp1,nk2p1);

  x2ion = sqrt(1/rSecondShell)*sin(thetaSecondShell)*cos(phiSecondShell);
  y2ion = sqrt(1/rSecondShell)*sin(thetaSecondShell)*sin(phiSecondShell);

  // Reshape the arrays to make them 2-dimensional to avoid scalar indexing
  x2ion.reshape(njp1,nk2p1); y2ion.reshape(njp1,nk2p1);

  ionXYArray2.redim(njp1,nk2p1,2);
  ionXYArray2(all,all,0) = x2ion;
  ionXYArray2(all,all,1) = y2ion;

  // We create a spherical system of coordinates with the symmetry axis along the SM x-axis
  // This is for the (x2ion,y2ion) grid
  realArray thetaIon2(njp1,nk2p1), phiIon2(njp1,nk2p1);
  thetaIon2 = acos(x2ion);
  phiIon2   = atan2(sqrt(1 - x2ion*x2ion - y2ion*y2ion),y2ion);

  // Map the first shell too
  realArray x1ion(1,njp1,nk2p1), y1ion(1,njp1,nk2p1);
  x1ion = sqrt(1/rFirstShell)*sin(thetaFirstShell)*cos(phiFirstShell);
  y1ion = sqrt(1/rFirstShell)*sin(thetaFirstShell)*sin(phiFirstShell);

  // This is for the (x1ion,y1ion) grid
  x1ion.reshape(njp1,nk2p1); y1ion.reshape(njp1,nk2p1);
  ionXYArray1.redim(njp1,nk2p1,2);
  ionXYArray1(all,all,0) = x1ion;
  ionXYArray1(all,all,1) = y1ion;

  realArray thetaIon1(njp1,nk2p1), phiIon1(njp1,nk2p1);
  thetaIon1 = acos(x1ion);
  phiIon1   = atan2(sqrt(1 - x1ion*x1ion - y1ion*y1ion),y1ion);

  // Now create the arrays for both spherical grids (first and second shell mapped)
  ionArray2.redim(njp1,nk2p1,2);
  ionArray2(all,all,0)=thetaIon2(all,all);
  ionArray2(all,all,1)=phiIon2(all,all);

  ionArray1.redim(njp1,nk2p1,2);
  ionArray1(all,all,0)=thetaIon1(all,all);
  ionArray1(all,all,1)=phiIon1(all,all);

}

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 08-22-2005 
 *
 * \return An Overture composite grid (1 member grid) containing the
 * spherical coordinates (in the LFM sense) of the first shell MHD
 * vertices mapped to the ionosphere.
 *
 *  Defines the mapping based on the previously filled #ionArray1 and
 *  the grid based on this mapping. This grid is the return value.
 ************************************************************************/
CompositeGrid MHD_Grids::getFirstShellIonSphericalGrid() const 
{
  DataPointMapping & ionSphericalMapping1 = *new DataPointMapping;
  ionSphericalMapping1.incrementReferenceCount();
  ionSphericalMapping1.setDataPoints(ionArray1,2,2);

  CompositeGrid ionSphericalGrid1(2,1);
  ionSphericalGrid1[0].reference(ionSphericalMapping1);
  ionSphericalGrid1.update();
  ionSphericalMapping1.decrementReferenceCount();
  return ionSphericalGrid1;
};

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 08-22-2005 
 *
 * \return An Overture composite grid (1 member grid) containing the
 * spherical coordinates (in the LFM sense) of the second shell MHD
 * vertices mapped to the ionosphere.
 *
 *  Defines the mapping based on the previously filled #ionArray2 and
 *  the grid based on this mapping. This grid is the return value.
 ************************************************************************/
CompositeGrid MHD_Grids::getSecondShellIonSphericalGrid() const 
{
  DataPointMapping & ionSphericalMapping2 = *new DataPointMapping;
  ionSphericalMapping2.incrementReferenceCount();
  ionSphericalMapping2.setDataPoints(ionArray2,2,2);

  CompositeGrid ionSphericalGrid2(2,1);//.setNumberOfDimensionsAndGrids(2,1);
  ionSphericalGrid2[0].reference(ionSphericalMapping2);
  ionSphericalGrid2.update();
  ionSphericalMapping2.decrementReferenceCount();
  return ionSphericalGrid2;
};

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 11-03-2005 
 *
 * \return An Overture composite grid (1 member grid) containing the
 * spherical coordinates (in the LFM sense) of the second shell MHD
 * vertices mapped to the ionosphere.
 *
 * This function is different from getSecondShellIonSphericalGrid() in
 * that the grid includes ghost points. It has never been used and
 * probably never will.
 *
 * \warning This function has not been tested!
 ************************************************************************/
CompositeGrid MHD_Grids::getSecondShellIonSphericalGridWithGhostLines() const 
{
  DataPointMapping & ionSphericalMapping2WithGhostLines = *new DataPointMapping;
  ionSphericalMapping2WithGhostLines.incrementReferenceCount();
  ionSphericalMapping2WithGhostLines.setDataPoints(ionArray2,2,2,1);

  CompositeGrid ionSphericalGrid2WithGhostLines(2,1);
  ionSphericalGrid2WithGhostLines[0].reference(ionSphericalMapping2WithGhostLines);
  ionSphericalGrid2WithGhostLines.update();
  ionSphericalMapping2WithGhostLines.decrementReferenceCount();
  return ionSphericalGrid2WithGhostLines;
};

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 08-22-2005 
 *
 * \return An Overture composite grid containing the coordinates of the
 * first shell MHD vertices mapped to the ionosphere in cartesian
 * coordinates.
 *
 * Defines the mapping based on the previously filled #ionXYArray1 and
 * the grid based on this mapping. This grid is the return value.
 ************************************************************************/
CompositeGrid MHD_Grids::getFirstShellIonCartesianGrid() const 
{
  /* -VGM- Could not use CompositeGrid for ionCartesianGrid2 because
 for some reason setNumberOfDimensionsAndMHD_Grids does not do anything
 (see the code commented out below. Most probably this grid will be
 used for plotting only (no interpolation) and so it's fine. The
 reason I want to define the grid dimensioning here is because it is
 not nice when one should do it in advance up in ionosphere.C or another
 calling function. 

 UPD. Out of date but still true as of 07-08-05*/

  DataPointMapping & ionCartesianMapping1 = *new DataPointMapping;
  ionCartesianMapping1.incrementReferenceCount();
  ionCartesianMapping1.setDataPoints(ionXYArray1,2,2);

  CompositeGrid ionCartesianGrid1(2,1);
  ionCartesianGrid1[0].reference(ionCartesianMapping1);
  ionCartesianGrid1.update();
  ionCartesianMapping1.decrementReferenceCount();

  return ionCartesianGrid1;
}

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 08-22-2005 
 *
 * \return An Overture composite grid containing the coordinates of the
 * second shell MHD vertices mapped to the ionosphere in cartesian
 * coordinates.
 *
 * Defines the mapping based on the previously filled #ionXYArray2 and
 * the grid based on this mapping. This grid is the return value.
 ************************************************************************/
CompositeGrid MHD_Grids::getSecondShellIonCartesianGrid() const 
{
  DataPointMapping & ionCartesianMapping2 = *new DataPointMapping;
  ionCartesianMapping2.incrementReferenceCount();
  ionCartesianMapping2.setDataPoints(ionXYArray2,2,2);

  CompositeGrid ionCartesianGrid2(2,1);
  ionCartesianGrid2[0].reference(ionCartesianMapping2);
  ionCartesianGrid2.update();
  ionCartesianMapping2.decrementReferenceCount();

  return ionCartesianGrid2;
}

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 08-22-2005 
 *
 * \param[out] Returns njp1 value in the LFM sense
 * \param[out] Returns nk2p1 value in the LFM sense
 *
 * This function can be used to extract dimensions of the ionospheric
 * projection of the MHD grid. It is deprectated and is not being used.
 ************************************************************************/
void MHD_Grids::getDimensions(int & NJ, int & NK) const
{
  NJ = njp1;
  NK = nk2p1;
}


