/**
 * \file MHD_Grids.h
 * \brief MHD_Grids class declaration
 *
 * \class MHD_Grids 
 *
 * \brief Class MHD_Grids defines grids used for coupling as well as
 * functions to access these grids.
 *
 * \author Slava Merkin (vgm at bu.edu) 
 * \since 08-2005
 *
 * Our methodology is that the MHD code that is being coupled to the
 * ionospheric solver has no \em a \em priori knowledge about the
 * ionospheric code nor does it know about the possible ionospheric
 * grids used. Thus, we only require a file containing the two
 * innermost shells of the MHD grid to be provided, and we do the
 * mapping to the ionosphere on our own.
 *
 * The MHD grid is expected to be a general logically-spherical grid
 * with the symmetry axis aligned to the SM x-axis (may be changed
 * later, if we implement an Overture module to interpolate an
 * arbitrary MHD grid to a spherical grid). The restriction above is
 * not really a restriction of the MHD_Grids class, but rather of the
 * MHD_Interface. For MHD_Gris it was true before, when the potential
 * solution was obtained on the MHD ionospheric grid. It was very much
 * LFM centered and required that the MHD grid and its ionospheric
 * projection satisfy the condition above. Now, all we need is that
 * the two inner most shell be given to us, but we don't care about
 * the orientation of the grid axis. In principle, separate readers
 * can be written for different MHD codes and different grid file
 * types as long as they define the grid dimensioning and grid vertex
 * arrays just as functions generateFromGRIDOUT() or generateFromHDF()
 * do at the present time (see below). The arrays #ionXYArray1 and
 * #ionXYArray2 filled in by the function fillInGridArrays() are
 * essential for the right grids to be created. They are what is
 * really necessary to be given to us if another global MHD model is
 * coupled, and they should provide a regular grid in XY
 * coordinates. Arrays #ionArray1 and #ionArray2 are obsolete and are
 * kept for consistency.
*/

#ifndef MHD_GRID_H
#define MHD_GRID_H

#include "Overture.h"  
#include "DataPointMapping.h"
#include "SquareMapping.h"

extern "C" {
  // Include HDF header
#include "mfhdf.h"
}

#include <iostream>
#include <fstream>

#include "const.h"

class MHD_Grids  {

 public:
  /** \brief The enumerator gridFileType is used to distinguish between
   *  different LFM grid storage files: HDF and GRIDOUT. 
   *
   * The last one is provided for convinience but we assume that
   * eventually a general-structure HDF file will be used for storing
   * grids for the LFM and any MHD code in general. Alternatively, a
   * separate reader could be written for a general grid file (see the
   * class description).
  */
  enum gridFileType {HDFFILE,GRIDOUT}; 
  
  /**\brief Class constructor. */
  MHD_Grids(const char*, const int); 

  /**\brief Gets the first shell of the MHD code mapped to the
   * ionosphere in spherical coordiantes (LFM sense). */
  CompositeGrid getFirstShellIonSphericalGrid() const;

  /**\brief Gets the second shell of the MHD code mapped to the
   * ionosphere in spherical coordiantes (LFM sense). */
  CompositeGrid getSecondShellIonSphericalGrid() const; 

  /**\brief Gets the second shell of the MHD code mapped to the
   * ionosphere in spherical coordiantes (LFM sense) including ghost
   * points. */
  CompositeGrid getSecondShellIonSphericalGridWithGhostLines() const;

  /**\brief Gets the first shell of the MHD code mapped to the
     ionosphere ion cartesian coordinates. */
  CompositeGrid getFirstShellIonCartesianGrid() const;

  /**\brief Gets the second shell of the MHD code mapped to the
     ionosphere ion cartesian coordinates. */
  CompositeGrid getSecondShellIonCartesianGrid() const;

  /**\brief Fills in the #ionXYArray1, #ionXYArray2, #ionArray1 and
   * #ionArray2 arrays. It is called from generateFromGRIDOUT() or
   * generateFromHDF()*/
  void fillInGridArrays(const realArray &, const realArray &, const realArray &);

  /**\brief Returns dimensions of the ionospheric projection of MHD
     grid. Deprecated */
  void getDimensions(int &,int  &) const;

 protected:
  /**\name Variables defining the grid. 
   *
   * We use the LFM-style spherical grid definition, i.e. the MHD code
   * grid is logically spherical with the symmetry axis aligned to the
   * SM x-axis. The 'i' index defines the radial direction while the
   * 'j' and 'k' indices define the polar and azimuthal directions,
   * respectively.
  */

  //@{
  int ni;        //!< Number of CELLS in radial direction.
  int nip1;     //!< Number of VERTICES in radial direction (nip1=ni+1).
  int nj;        //!< Number of CELLS in polar direction.
  int njp1;     //!< Number of VERTICES in polar direction (njp1=nj+1).
  int nk;        //!< Number of CELLS in azimuthal direction.
  int nkp1;     //!< Number of VERTICES in azimuthal direction (nkp1=nk+1).
  int nk2p1;   /*!< \brief Number of vertices in azimuthal direction
		* in the ionospheric grid mapped to one hemisphere
		* (nk2p1=nk/2+1). */
  //@}

 private:
  /** \name Arrays for the ionospheric grid vertex coordinate storage.*/
  //@{

  /** The cartesian ionospheric coordinates for the first shell (LFM
   * x1ion-y1ion grid). Should define a regular grid. That's the only
   * requirement. */
  realArray ionXYArray1;

  /** The cartesian ionospheric coordinates for the second shell (LFM
   * x2ion-y2ion grid, but nk2p1 rather than nk2p3 points in k
   * direction). Should define a regular grid. That's the only
   * requirement. */
  realArray ionXYArray2;

  /** The first shell mapped to the ionosphere, in spherical
   * coordinates in the LFM sense. This array is not used any more and
   * is kept for consistency only. */
  realArray ionArray1;

  /** The second shell mapped to the ionosphere, in spherical
   * coordinates in the LFM sense. This array is not used any more and
   * is kept for consistency only. */
  realArray ionArray2;
  //@}

  /**\name Functions for grid array generation.
   *
   * These functions read the MHD grid file of the appropriate file
   * type, process the information stored, and determine the
   * ionospheric grid dimensions and coordinate arrays.
  */
  //@{
  void generateFromGRIDOUT(const char*);  /**< \brief Generates grid
					     dimensions and coordinate
					     arrays from a GRIDOUT
					     file. */

  void generateFromHDF_old(const char*);  /**< \brief Generates grid
					     dimensions and coordinate
					     arrays from a HDF
					     file. This functions was
					     written for the old-style
					     LFM-para hdf file where
					     the grid was stored in
					     x2, y2, and phi
					     arrays. It is now
					     deprecated. The new
					     function
					     generateFromHDF() is used
					     now.*/

  void generateFromHDF(const char*);  /**< \brief Generates grid
					     dimensions and coordinate
					     arrays from a HDF
					     file. */
  //@}
};

#endif
