/**
 * \file MHD_IC_Interface.C
 * \brief MHD_IC_Interface definition.
 */

#include "MHD_IC_Interface.h"

using namespace std;

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 08-23-2005 
 * \param[in] params The pointer to the Params object
 * \param[in] XJD The pointer to the IC_XJD object
 *
 * The constructor inherits the MHD grid information from
 * MHD_Interface. Then arrays to be transmitted between this code and the
 * MHD model are dimensioned (MHD_Grids::njp1,MHD_Grids::nkp1). The
 * InterComm regions are created and prepared for commit.
 ************************************************************************/
MHD_IC_Interface::MHD_IC_Interface(Params *params, IC_XJD *XJD)
  : MHD_Interface(params, XJD)
{
  cout << "Initializing MHD InterComm InterComm Interface..."<<endl;

  // These are set for a sequential program. If I ever make this code parallel
  // these should be changed (see InterComm manual)
  int tasks[1] = {0};

  int njp1 = MHD_Grids::njp1;
  int nkp1 = MHD_Grids::nkp1;

  // IC stuff for the 1-shell arrays
  int blocks1shell[1][2][2] = {
    {{0,0},{njp1-1, nkp1-1}}
  };

  desc1shell = IC_Create_bdecomp_desc(2, &blocks1shell[0][0][0], tasks, 1, 1);
  // IC_ROW_MAJOR = 0                                                     ^^^
  // IC_COLUMN_MAJOR = 1                                                  ^^^
  
  int lower1shell[2] = {0,0};
  int upper1shell[2] = {njp1-1,nkp1-1};
  int stride1shell[2] = {1,1};
  regions1shell[0] = IC_Create_block_region(2,lower1shell,upper1shell,stride1shell);

  // IC stuff for the 2-shell arrays
  int blocks2shell[1][2][3] = {
    {{0,0,0},{1,njp1-1, nkp1-1}}
  };
  desc2shell = IC_Create_bdecomp_desc(3, &blocks2shell[0][0][0], tasks, 1, 1);
  // IC_ROW_MAJOR = 0                                                     ^^^
  // IC_COLUMN_MAJOR = 1                                                  ^^^

  int lower2shell[3] = {0,0,0};
  int upper2shell[3] = {1,njp1-1,nkp1-1};
  int stride2shell[3] = {1,1,1};
  regions2shell[0] = IC_Create_block_region(3,lower2shell,upper2shell,stride2shell);

  IC_Register_region
    (xjd,regions1shell,1,"mhd-cpl_fac",desc1shell,current.getDataPointer(),&ic_err);
  CHECK_IC_STATUS(xjd, "IC_Register_region('mhd-cpl_fac', ...)", ic_err);  
    
  IC_Register_region
    (xjd,regions1shell,1,"mhd-cpl_density",desc1shell,density.getDataPointer(),&ic_err);
  CHECK_IC_STATUS(xjd, "IC_Register_region('mhd-cpl_density', ...)", ic_err);  

  IC_Register_region
    (xjd,regions1shell,1,"mhd-cpl_sound_speed",desc1shell,soundSpeed.getDataPointer(),&ic_err);
  CHECK_IC_STATUS(xjd, "IC_Register_region('mhd-cpl_sound_speed', ...)", ic_err);  

  IC_Register_region
    (xjd,regions2shell,1,"cpl-mhd_potential",desc2shell,potential.getDataPointer(),&ic_err);
  CHECK_IC_STATUS(xjd, "IC_Register_region('mhd-cpl_potential', ...)", ic_err);  

  exchangeRION();
}

/********************************************************************//**
 * \author Pete Schmitt (schmitt at ucar.edu)
 * \since 09-2008
 *
 * Automatic destructor to clean up InterComm objects
 ************************************************************************/
MHD_IC_Interface::~MHD_IC_Interface()
{
  IC_Free_desc(desc1shell);
  IC_Free_desc(desc2shell);
  IC_Free_region(regions1shell[0]);
  IC_Free_region(regions2shell[0]);
}

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 08-23-2005 
 * \param[out] jpara_N MHD parallel current in the northern hemisphere
 * \param[out] jpara_S MHD parallel current in the southern hemisphere
 * \param[out] rho_N    MHD density in the northern hemisphere
 * \param[out] rho_S    MHD density in the southern hemisphere
 * \param[out] cs_N     MHD sound speed in the northern hemisphere
 * \param[out] cs_S     MHD sound speed in the southern hemisphere
 *
 * The function does three things: calls IC_Import for current,
 * density, and sound speed (in exactly that order), then calls
 * MapToIonosphere() for each of these variables one time for each
 * hemisphere, and then converts the resulting arrays into grid
 * functions passed to Import() as parameters.
 ************************************************************************/
void MHD_IC_Interface::Import(realCompositeGridFunction & mhdFuncs)
{
  int status;
  status = IC_Import(xjd,"mhd-cpl_fac");
  CHECK_IC_STATUS(xjd, "IC_Import('mhd-cpl_fac', ...)", status);  

  status = IC_Import(xjd,"mhd-cpl_density");
  CHECK_IC_STATUS(xjd, "IC_Import('mhd-cpl_density', ...)", status);  

  status = IC_Import(xjd,"mhd-cpl_sound_speed");
  CHECK_IC_STATUS(xjd, "IC_Import('mhd-cpl_sound_speed', ...)", status);  

  processImport(mhdFuncs);
}

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 08-2005 
 *
 * \param[in] psi1_NFunction Northern hemisphere potential grid
 * function to be mapped to the first MHD shell
 *
 * \param[in] psi1_SFunction  Same for the southern hemisphere
 *
 * \param[in] psi2_NFunction Northern hemisphere potential grid
 * functionto be mapped to the second MHD shell
 *
 * \param[in] psi2_SFunction  Same for the southern hemisphere
 *
 * This is the reverse interface for sending the potential obtained by
 * the elliptic solver back to the MHD code. Regardless of whether the
 * MHD code needs it or not, we send back the array containing the
 * potential for the two innermost MHD shells, i.e. dimensioned
 * (2,#njp1,#nkp1).
 *
 * The function first converts the grid functions passed as parameters
 * to plain arrays, then calls MapFromIonosphere(), and finally
 * IC_Exports the resulting array to the MHD code.
 ************************************************************************/
void MHD_IC_Interface::Export(realCompositeGridFunction & mhdFunctions1, 
			      realCompositeGridFunction & mhdFunctions2 )
{
  prepareExport(mhdFunctions1, mhdFunctions2);
  int status = IC_Export(xjd,"cpl-mhd_potential");
  CHECK_IC_STATUS(xjd, "IC_Export('cpl-mhd_potential', ...)", status);  
}

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 06-2007
 * \return   The array of scalars to be received.
 *
 * Use this function to receive an array of scalars broadcast by the
 * MHD code. The scalars to be sent are defined in the CPL_Scalars
 * enumeration defined in MHD_Interface.h
 ************************************************************************/
realArray MHD_IC_Interface::getScalars()
{
  realArray scalars(MHD_Interface::NUMBER_OF_SCALARS);
  double *varPtr = scalars.getDataPointer();
  int status = IC_Recv_local(xjd,"mhd-cpl_scalars",varPtr,MHD_Interface::NUMBER_OF_SCALARS);
  CHECK_IC_STATUS(xjd, "IC_Recv_local(..., 'mhd-cpl_scalars', ...)", status);  
  return scalars;
}

void MHD_IC_Interface::exchangeRION()
{
  // RION is defined in const.h
  float RION_meters = RION / 100.0;

  int iStatus = IC_Bcast_local(xjd, "cpl-mhd_r_i", &RION_meters, 1);
  CHECK_IC_STATUS(xjd, "IC_Bcast_local('cpl-mhd_r_i', ...)", iStatus);
}
