/**
 * \file MHD_Interface.C
 * \brief MHD_Interface definition.
 */

#include "MHD_Interface.h"

using namespace std;

//Range all;

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 08-23-2005 
 * \param[in] params The pointer to the Params object
 * \param[in] XJD The pointer to the IC_XJD object
 *
 * The constructor inherits the MHD grid information from
 * MHD_Grids. Then arrays to be transmitted between this code and the
 * MHD model are dimensioned (MHD_Grids::njp1,MHD_Grids::nkp1). 
 ************************************************************************/
MHD_Interface::MHD_Interface(Params *params, IC_XJD *XJD)
  : MHD_Grids(params->mhdGridFileName,params->mhdGridFileType),
    xjd(XJD),
    current(njp1,nkp1),
    density(njp1,nkp1),
    soundSpeed(njp1,nkp1),
    potential(2,njp1,nkp1)
{
  cout << "Initializing MHD Interface..."<<endl;

  forceSymmetricPotential = params->forceSymmetricPotential;
}

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \author Peter Schmitt (schmitt@ucar.edu)
 * 
 * \param[out] jpara_N MHD parallel current in the northern hemisphere
 * \param[out] jpara_S MHD parallel current in the southern hemisphere
 * \param[out] rho_N    MHD density in the northern hemisphere
 * \param[out] rho_S    MHD density in the southern hemisphere
 * \param[out] cs_N     MHD sound speed in the northern hemisphere
 * \param[out] cs_S     MHD sound speed in the southern hemisphere
 *
 * Calls MapToIonosphere() for each current, density and sound speed
 * one time for each hemisphere, and then converts the resulting
 * arrays into grid functions passed to Import() as parameters.
 *
 * The Import(...) method of a MHD_Interface derived class should call
 * "processImport" after the data has been imported.  For example, see
 * MHD_IC_Interface::Import(...)
 ************************************************************************/
void MHD_Interface::processImport(realCompositeGridFunction & mhdFuncs)
{
  realArray j_n,j_s,c_n,c_s,r_n,r_s;

  // Current
  j_n = MapToIonosphere(current,NORTH);
  j_s = MapToIonosphere(current,SOUTH);
  // Density
  r_n = MapToIonosphere(density,NORTH);
  r_s = MapToIonosphere(density,SOUTH);
  // Sound Speed
  c_n = MapToIonosphere(soundSpeed,NORTH);
  c_s = MapToIonosphere(soundSpeed,SOUTH);

  Index I1,I2,I3;
  getIndex(mhdFuncs[0],I1,I2,I3);

  mhdFuncs[0](I1,I2,all,MHD_FAC_NORTH)         = j_n;
  mhdFuncs[0](I1,I2,all,MHD_FAC_SOUTH)         = j_s;
  mhdFuncs[0](I1,I2,all,MHD_DENSITY_NORTH)     = r_n;
  mhdFuncs[0](I1,I2,all,MHD_DENSITY_SOUTH)     = r_s;
  mhdFuncs[0](I1,I2,all,MHD_SOUND_SPEED_NORTH) = c_n;
  mhdFuncs[0](I1,I2,all,MHD_SOUND_SPEED_SOUTH) = c_s;
}

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \author Peter Schmitt (schmitt at bu.edu)
 * \since 08-2005 
 *
 * \param[in] psi1_NFunction Northern hemisphere potential grid
 * function to be mapped to the first MHD shell
 *
 * \param[in] psi1_SFunction  Same for the southern hemisphere
 *
 * \param[in] psi2_NFunction Northern hemisphere potential grid
 * functionto be mapped to the second MHD shell
 *
 * \param[in] psi2_SFunction  Same for the southern hemisphere
 *
 * Prepares MIX data to be exported to the MHD:
 *   - Convert grid functions from input parameters to standard arrays
 *   - Call MapFromIonosphere()
 *
 * This is initialization of the reverse interface for sending the
 * potential obtained by the elliptic solver back to the MHD
 * code. Regardless of whether the MHD code needs it or not, we send
 * back the array containing the potential for the two innermost MHD
 * shells, i.e. dimensioned (2,#njp1,#nkp1).
 *
 * The Export(...) method of a MHD_Interface derived class should call
 * "prepareExport" before the data will be exported.  For example, see
 * MHD_IC_Interface::Export(...)
 ************************************************************************/
void MHD_Interface:: prepareExport(realCompositeGridFunction & mhdFunctions1, 
				   realCompositeGridFunction & mhdFunctions2 )
{
  // Convert grid functions into plain arrays
  realArray psi1_N,psi1_S,psi2_N,psi2_S;
  Index I1,I2,I3;

  getIndex(mhdFunctions1[0],I1,I2,I3,0);
  psi1_N = mhdFunctions1[0](I1,I2,I3,MHD_POTENTIAL_NORTH);
  psi1_S = mhdFunctions1[0](I1,I2,I3,MHD_POTENTIAL_SOUTH);

  getIndex(mhdFunctions2[0],I1,I2,I3,0);
  psi2_N = mhdFunctions2[0](I1,I2,I3,MHD_POTENTIAL_NORTH);
  psi2_S = mhdFunctions2[0](I1,I2,I3,MHD_POTENTIAL_SOUTH);

  // Get the potential in the first two MHD shells
  potential= MapFromIonosphere(psi1_N,psi1_S,psi2_N,psi2_S);
}

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 06-15-2005 
 *
 * \param[in] var An array variable defined on the inner MHD boundary
 * \param[in] HEMISPHERE Either NORTH or SOUTH (see const.h)
 * \return The mapped variable
 *
 * This function maps the field aligned current, density and sound
 * speed from the second shell (or whatever) of the MHD grid to the
 * ionospheric grid defined in class MHD_Grids. The mapping is done by
 * indexing because the ionospheric grid itself is guaranteed to be a
 * dipole mapping of the MHD grid.
 *
 * \note In May-June 2007 this function was significantly
 * simplified. First of all, now all three variables (current,
 * density, and sound speed) are mapped the same way. Before that, the
 * density was massaged at the nose and the tail and the current as
 * well. I have removed these things from here, they don't seem to
 * matter. In case I want to include them, they should be done in the
 * MHD interface rather than in the MIX code.
 *
 * \note Also, note that the field-aligned current in this code is
 * what it is, \f$\vec j \hat b\f$, unlike the LFM code. So, the
 * earthward current is positive in the northern hemisphere, and
 * negative in the southern hemisphere.
 ************************************************************************/
realArray MHD_Interface::MapToIonosphere(const realArray &var, const bool HEMISPHERE)  
{
  int njp1 = MHD_Grids::njp1;
  int nk2p1 = MHD_Grids::nk2p1;
  int nk2 = MHD_Grids::nk/2;

  Range K(0,nk2);
  realArray dummy(njp1,nk2p1);

  /** \note The y-axis is flipped with regard to the conventional LFM
 * ionosphere, so that 
 * \code
 */

 if ( HEMISPHERE == NORTH ) { dummy = var(all,K); }
  else { dummy = var(all,K+nk2); }

 /**\endcode This was done mainly for the reasons of the Solver
  * simplification: this was the triad of axes in the southern
  * hemisphere has the right sense and the Solver can be applied to
  * both hemisphere with just differnt sign of the dip angle
 */

  return dummy;
}

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 08-23-2005 
 * \param[in] varFirstShell_N  The first-shell, northern hemisphere array
 * \param[in] varFirstShell_S  The first-shell, southern hemisphere array
 * \param[in] varSecondShell_N The second-shell, northern hemisphere array
 * \param[in] varSecondShell_S The second-shell, southern hemisphere array
 *
 * \return An array dimensioned (2,#njp1,#nkp1) containing the mapped
 * variable in the first two shells of the MHD grid
 *
 * This function is the counter-part to MapToIonosphere(). The mapping
 * is straightforward, since the arrays are dimensioned the same way
 * on the MHD and ionospheric grids.
 ************************************************************************/
realArray MHD_Interface::MapFromIonosphere(realArray & varFirstShell_N,
					   realArray & varFirstShell_S,
					   realArray & varSecondShell_N,
					   realArray & varSecondShell_S) 
{
  int njp1 = MHD_Grids::njp1;
  int nk2p1 = MHD_Grids::nk2p1;
  int nk2   = MHD_Grids::nk/2;
  int nkp1 = MHD_Grids::nkp1;

  Range K(0,nk2);
  realArray dummy(2,njp1,nkp1); 

  // First shell
  varFirstShell_N.reshape(1,njp1,nk2p1);  // reshape to conform to the geometry of dummy
  varFirstShell_S.reshape(1,njp1,nk2p1);
  dummy(0,all,K)     = varFirstShell_N; // Map from northern hemisphere

  // PJS March 2010: Force symmetric potential when the MIX
  // low-latitude boundary is equatorward of the LFM boundary.
  if (forceSymmetricPotential){
    for (int k=0; k<nk2p1; k++) { 
      varFirstShell_S(0,all,k) = varFirstShell_N(0,all,nk2-k); 
    }
  }
  dummy(0,all,K+nk2) = varFirstShell_S; // Map from southern hemisphere

  // Second shell
  varSecondShell_N.reshape(1,njp1,nk2p1);
  varSecondShell_S.reshape(1,njp1,nk2p1);
  dummy(1,all,K)     = varSecondShell_N;  // Map from northern hemisphere

  // PJS March 2010: Force symmetric potential when the MIX
  // low-latitude boundary is equatorward of the LFM boundary.
  if (forceSymmetricPotential){    
    for (int k=0; k<nk2p1; k++) {
      varSecondShell_S(0,all,k) = varSecondShell_N(0,all,nk2-k); 
    }
  }

  dummy(1,all,K+nk2) = varSecondShell_S;  // Map from southern hemisphere

  return dummy;
}
