#ifndef MHD_INTERFACE_H
#define MHD_INTERFACE_H

/**
 * \file MHD_Interface.h
 * \brief MHD_Interface declaration
 *
 * \class MHD_Interface 
 * \brief Abstract Base Class that does the dirty work in massaging,
 * prior to receiving and sending data from/to the MHD model.
 *
 * \author Slava Merkin (vgm at bu.edu) 
 * \since 08-2005
 *
 * Inherits the ionospheric and magnetospheric grid
 * information from the class MHD_Grids, which is initialized by the
 * MHD_Interface constuctor. 
 * 
 * The derived class MHD_IC_Interface provides functions for
 * communication with the MHD model
 *
 * \see MHD_IC_Interface
 * \see MHD_FE_Interface
 */

// 3rd-party includes
#include "Overture.h"
#include "InterComm.h"

// MIX includes
#include "const.h"
#include "functions.h"
#include "Params.h"
#include "MHD_Grids.h"

class MHD_Interface : public MHD_Grids {

public:
  /** \brief The class constructor. */
  MHD_Interface(Params *, IC_XJD *);

  virtual ~MHD_Interface() {};
  
  /** \brief The function for import from MHD to ionosphere. */
  virtual void Import(realCompositeGridFunction &) = 0;

  /** \brief Sends potential back to MHD. */
  virtual void Export(realCompositeGridFunction &, realCompositeGridFunction &) = 0;
 
  /** \brief Receives a scalar array from the MHD code. Intended to
   * communicate scalars such as UT and kill signal */
  virtual realArray getScalars() = 0;



  /********************************************************************//**
   * Indices for scalars to be received from MHD
   * 
   * Note: These must match up exactly with the scalars on the sending
   * side (in MHD code).
   ************************************************************************/ 
  enum MHD_Scalars{
    KILL_SIGNAL       = 0, // 111=kil; 222=last exchange; else=keep running
    YEAR              = 1,
    MONTH             = 2,
    DAY               = 3,
    HOUR              = 4,
    MINUTE            = 5,
    SECOND            = 6,
    TILT              = 7, // in principle, we can calc this in-place using geopack.  Might as well get it from the MHD code.
    LABEL             = 8, // MHD simulation time step
    EXCHANGE_WITH_IM = 9,
    NUMBER_OF_SCALARS = 10
  };
  // Scalars at index KILL_SIGNAL accept the following indices:
  enum MHD_kill_signal{
    KILL_SIGNAL_CONTINUE = 0,
    KILL_SIGNAL_SHUTDOWN = 111,
    KILL_SIGNAL_LAST_EXCHANGE = 222
  };
  // Scalars at index EXCHANGE_WITH_ITM accept the following indices: 
  enum MHD_exchange_with_IM{
    NO_IM_EXCHANGE     = 0,
    CPL_RECV_FROM_IM   = 1,
    CPL_SEND_TO_IM     = 2,
    DUPLEX_IM_EXCHANGE = 3   // coupler receives/sends from/to inner magnetosphere
  };

protected:
  void processImport(realCompositeGridFunction & mhdFuncs);
  void prepareExport(realCompositeGridFunction & mhdFunctions1, 
		     realCompositeGridFunction & mhdFunctions2 );

  /** \brief Maps the parallel current, density, and sound speed
   * defined on the MHD inner boundary to the northern hemisphere */
  realArray MapToIonosphere(const realArray &, const bool);

  /** \brief Maps the potential (or something else) from the
   * ionosphere to the MHD grid. */
  realArray MapFromIonosphere(realArray &, realArray &,
			      realArray &, realArray &);

  IC_XJD *xjd;

  /**\name Arrays for communications with the MHD model */
  realArray current, density, soundSpeed, potential;

private:
  bool forceSymmetricPotential;
};

#endif
