/**
 * \file Params.C 
 * \brief Params class definition.
 */

#include "Params.h"

using namespace std;

/********************************************************************//**
 * \author Slava Merkin (vgm at bu.edu)
 * \since 06-2007
 *
 * \param[in] Parameter file name
 *
 * The constructor reads and parses the parameter file, e.g. ionosphere.param
 ************************************************************************/
Params::Params(const char* paramFileName)
{
/********************************************************************//**
 * These are the default parameters 
 * \code
 */
  strcpy(localName,"mix");
  strcpy(xjdName,"lfm-ion.xjd");
  strcpy(ionGridFileName,"ion_grid.hdf");

  strcpy(outputFilePrefix,"MIX");
  flushFrequency = 5;

  use_intercomm = true;

  // Default grid is 2x2 spacing down to colatitude of 44 degrees.
  mixGrid.setGrid("2x2_44");
  forceSymmetricPotential = mixGrid.requiresSymmetricPotential();

  strcpy(mhdGridFileName,"/roguehomes/vgm/LFM-para/DUMPMHD.hdf");
  mhdGridFileType = 0; // HDFFILE (1 = GRIDOUT)

  strcpy(itmGridFileName,"apex.dat"); // Not used for TING, reserved for TIEGCM
  itmGridFileType = 0;                // Not used for TING either
  strcpy(itmModelType,"TING");
  itmCouplingStart = DateTime(99999,1,1, 0,0,0);
  itmCouplingDelta = 120.0;

  euvModelType = 0;  // AMIE (1 = MOEN_BREKKE)
  // Alpha, Beta, R (1.03,2.5,5.56e-2) values from VGM testing 12/09
  alpha           = 1.03;
  beta            = 2.5;     
  r               = 5.56e-2; 
  f107            = 200.;
  pedmin          = 2.;
  hallmin         = 1.;
  ped0            = 5.;
  sigma_ratio     = 3.;
  const_sigma     = 0;
/*
 * \endcode
 ************************************************************************/

// Read in parameters from file
    ifstream paramFile(paramFileName,ios::in);
  if (!paramFile) {
    cerr << "Could not open parameter file." << endl;
    exit(1);
  }

  char line[STRLEN],value[STRLEN],varid[STRLEN],mhdGrid[STRLEN],itmGrid[STRLEN],euvModel[STRLEN];

  cout << "\n********** Parameters read from ionosphere.param **********"<<endl;

  while ( paramFile.getline(line,STRLEN) ) {
    strcpy(varid,"");
    strcpy(value,"");

    sscanf(line,"%s %s %*s",varid,value);

    if      ( strcmp(varid,"MHD_GRID_FILE_NAME") == 0 )
      {
	strcpy(mhdGridFileName,value);
	cout << setw(25) << left << "MHD_GRID_FILE_NAME"  << mhdGridFileName << endl;
      }
    else if ( strcmp(varid,"MHD_GRID_FILE_TYPE") == 0 )
      {
	strcpy(mhdGrid,value);
	cout << setw(25) << left << "MHD_GRID_FILE_TYPE" << mhdGrid << endl;
      }
    else if ( strcmp(varid,"XJD_NAME") == 0 )
      {
	strcpy(xjdName,value);
	cout << setw(25) << left << "XJD_NAME" << xjdName << endl;
      }
    else if      ( strcmp(varid,"ITM_GRID_FILE_NAME") == 0 )
      {
	strcpy(itmGridFileName,value);
	cout << setw(25) << left << "ITM_GRID_FILE_NAME" << itmGridFileName << endl;
      }
    else if ( strcmp(varid,"ITM_GRID_FILE_TYPE") == 0 )
      {
	strcpy(itmGrid,value);
	cout << setw(25) << left << "ITM_GRID_FILE_TYPE" << itmGrid << endl;
      }
    else if ( strcmp(varid,"ITM_MODEL_TYPE") == 0 )
      {
	strcpy(itmModelType,value);
	cout << setw(25) << left << "ITM_MODEL_TYPE" << itmModelType << endl;
      }
    else if ( strcmp(varid,"ITM_COUPLING_START") == 0)
      {
	int startDateTime[6];
	assert( strToIntArray(value, startDateTime) == 6 );
	itmCouplingStart = DateTime(startDateTime[0],startDateTime[1],startDateTime[2],
				    startDateTime[3],startDateTime[4],startDateTime[5]);
	cout << setw(25) << left << "ITM_COUPLING_START" <<  itmCouplingStart.getDateTimeString() << endl;
      }
    else if ( strcmp(varid,"ITM_COUPLING_DELTA") == 0)
      {
	itmCouplingDelta = atof ( value );
	cout << setw(25) << left << "ITM_COUPLING_DELTA" << itmCouplingDelta << endl;
      }
    else if ( strcmp(varid,"OUTPUT_FILE_PREFIX") == 0 )
      {
	strcpy(outputFilePrefix,value);
	cout << setw(25) << left << "OUTPUT_FILE_PREFIX" << outputFilePrefix << endl;
      }
    else if ( strcmp(varid,"FLUSH_FREQUENCY") == 0 )
      {
	flushFrequency = atol ( value );
	cout << setw(25) << left  << "FLUSH_FREQUENCY" << flushFrequency << endl;
      }
    else if ( strcmp(varid,"USE_INTERCOMM") == 0 )
      {
	use_intercomm = strToBool( value );
	cout << setw(25) << left  << "USE_INTERCOMM" << use_intercomm << endl;
      }

    else if ( strcmp(varid,"MIX_GRID") == 0){
      string mixGridName(value);
      mixGrid.setGrid(mixGridName);
      forceSymmetricPotential = mixGrid.requiresSymmetricPotential();
      cout << setw(25) << left  << "MIX_GRID" << mixGrid.getGridName() << endl;
    }
    else if ( strcmp(varid,"EUV_MODEL_TYPE") == 0 )
      {
	strcpy(euvModel,value);
	cout << setw(25) << left << "EUV_MODEL_TYPE" << euvModel << endl;
      }
    else if ( strcmp(varid,"ALPHA") == 0 )
      {
	alpha = atof ( value );
	cout << setw(25) << left << "ALPHA" << alpha << endl;
      }
    else if ( strcmp(varid,"BETA") == 0 )
      {
	beta = atof( value );
	cout << setw(25) << left << "BETA" << beta << endl;
      }
    else if ( strcmp(varid,"R") == 0 )
      {
	r = atof( value );
	cout << setw(25) << left << "R" << r << endl;
      }
    else if ( strcmp(varid,"F107") == 0 )
      {
	f107 = atof( value );
	cout << setw(25) << left << "F107" << f107 << endl;
      }
    else if ( strcmp(varid,"PED_MIN") == 0 )
      {
	pedmin = atof( value );
	cout << setw(25) << left << "PED_MIN" << pedmin << endl;
      }
    else if ( strcmp(varid,"HALL_MIN") == 0 )
      {
	hallmin = atof( value );
	cout << setw(25) << left << "HALL_MIN" << hallmin << endl;
      }
    else if ( strcmp(varid,"PED0") == 0 )
      {
	ped0 = atof( value );
	cout << setw(25) << left << "PED0" << ped0 << endl;
      }
    else if ( strcmp(varid,"SIGMA_RATIO") == 0 )
      {
	sigma_ratio = atof( value );
	cout << setw(25) << left << "SIGMA_RATIO" << sigma_ratio << endl;
      }
    else if ( strcmp(varid,"CONST_SIGMA") == 0 )
      {
	const_sigma = atoi( value );
	cout << setw(25) << left  << "CONST_SIGMA" << const_sigma << endl;
      }
  }

  if   ( strcmp(mhdGrid,"HDF") == 0 )
    mhdGridFileType = 0;
  else if ( strcmp(mhdGrid,"GRIDOUT") == 0 )
    mhdGridFileType = 1;

  if   ( strcmp(euvModel,"AMIE") == 0 )
    euvModelType = 0;
  else if ( strcmp(euvModel,"MOEN_BREKKE") == 0 )
    euvModelType = 1;

  cout << "***********************************************************\n" << endl;
}

/*********************  Private Members ************************/

/// Convert a string containing {t, true, 1, 1, false, 0} to a boolean
/**
 *  @param input_str - string containing what the boolean should be.
 *  @returns bool - based on the input data_str
 */
bool Params::strToBool(char *input_str)
{
#ifdef DEBUG_MODE_ON
  std::cout << "DEBUG: in Params.C::strToBool(...)\n";
#endif
  // Make contents of "input_str" upper case.
  for(int i = 0; input_str[i] != '\0'; i++){
    input_str[i] = toupper(input_str[i]);
  }
  if (  ( strcmp(input_str, "T") == 0) ||
        ( strcmp(input_str, ".T.") == 0) ||
        ( strcmp(input_str, "TRUE") == 0) ||
        ( strcmp(input_str, ".TRUE.") == 0) ||
        ( strcmp(input_str, "1") == 0) ){
    return true;
  }
  else if (  ( strcmp(input_str, "F") == 0) ||
             ( strcmp(input_str, ".F.") == 0) ||
             ( strcmp(input_str, "FALSE") == 0) ||
             ( strcmp(input_str, ".FALSE.") == 0) ||
             ( strcmp(input_str, "0") == 0) ){
    return false;
  }
  else{
    std::cerr << "Error setting \"SWFROMFILE\" in memory."
              << "\n\t Expected \"true\" or \"false\", but got "
              << "\"" << input_str << "\"."
              << "\n\t Defaulting to \"false\".\n";
  }
  return false;
}


/// Convert whitespace-separated data string to an floating point array
/**
 *  @warning - This function assumes that data_array is already
 *             allocated (and large enough) to handle all the elements
 *             of input_str.
 *
 *  @param input_str - String containing a floating-point data array,
 *                    such as "1 2 3 4 5"
 *
 *  @param data_array - array to store data into (i.e. data_array[0] =
 *                      { 1.0, 2.0, 3.0, 4.0, 5.0 })
 *
 *  @returns int - the number of elements compied from input_str to
 *                 data_array (i.e. the size of data_array).
 */
int Params::strToIntArray(char *input_str, int *data_array)
{
#ifdef DEBUG_MODE_ON
  std::cout << "DEBUG: in Params.C::strToFloatArray(...)\n";
#endif
  int nElements = 0;

  char * token;
  token = strtok(input_str, "-");
  if (token == NULL){
    std::cerr << "Error reading data in array containing \"" 
	      << input_str << "\".";
  }
  while (token != NULL){
    data_array[nElements++] = atoi(token);
    token = strtok(NULL, "-");    
  }
  return nElements;
}
