/**
 * \file Params.h
 * \brief Params class declaration
 *
 * \class Params
 *
 * \brief Class Params contains the parameters read in from the
 * parameters file (ionosphere.param).
 *
 * \author Slava Merkin (vgm at bu.edu) 
 * \since 07-2007
 *
 * The run-time parameters are packed into this class so that
 * parameter lists in various functions do not get long. Just pass the
 * pointer to this class.
 */

#ifndef PARAMS_H
#define PARAMS_H

#include <iostream>
#include <fstream>
#include <iomanip>

#include "const.h"
#include "DateTime.h"
#include "Grid.h"

void parse_xjd(const char *, char *, char *, int &);

class Params {
 public:
  /**\brief Class constructor */
  Params(const char* paramFileName);

  /**\name InterComm parameters */
  //@{
  char localName[STRLEN]; /*<!\brief The name of the local
			   *  program. Will be parsed from the XJD
			   *  file automatically */
  char xjdName[STRLEN];  //<! The name of the XJD file
  //@}

  /**\name IO parameters */
  //@{
  char ionGridFileName[STRLEN]; /*!< File name to store the MIX grid
				 * in. Obsolete since 01-2008 (grid is
				 * stored along with data in one
				 * file */
  char outputFilePrefix[STRLEN];  //!< Store in the OVERTURE show file or an HDF file
  int  flushFrequency;          /*!< \brief Overture Show file flush frequency 
				 * (<A HREF="http://www.llnl.gov/CASC/Overture/henshaw/documentation/ogshow.pdf">Overture show files</A>) */
  bool use_intercomm;           /*!< \brief Use InterComm for data
				  sharing if true (default).
				  Otherwise, use File Exchanges*/
  //@}

  /**\name MHD model parameters */
  //@{
  char mhdGridFileName[STRLEN]; //!< Where to get the MHD grid from
  int  mhdGridFileType;         //!< Should be one of {HDF,GRIDOUT}
  //@}

  /**\name MIX grid parameters */
  //@{
  Grid mixGrid;
  bool forceSymmetricPotential; /*!< Should MIX force a symmetric potential between hemispheres? */
  //@}

  /**\name ITM model parameters */
  //@{
  char itmGridFileName[STRLEN]; //!< TING and TIEGCM are recognized so far
  int  itmGridFileType;         //!< Not used. Reserved for TIEGCM
  char itmModelType[STRLEN];    //!< Not used. Reserved for TIEGCM
  DateTime itmCouplingStart;
  double itmCouplingDelta;
  //@}

  /**\name ITM model parameters */
  //@{
  int euvModelType; //!<Should be one of (AMIE, MOEN_BREKKE)
  real alpha;     
  real beta;
  real r;
  real f107;
  real pedmin;
  real hallmin;
  real ped0;
  real sigma_ratio;
  int  const_sigma;  //!< Whether to use constant conductance
  //@}

 private:
  bool strToBool(char *input_str);
  int strToIntArray(char *input_str, int *data_array);
};

#endif
