#ifndef __IONOSPHERE_FUNCTIONS_H__
#define __IONOSPHERE_FUNCTIONS_H__

#include "ionosphere.h"
#include "DateTime.h"


/********************************************************************//**
 * Receive data (current, density & sound speed) from  MHD & interpolate to MIX grid
 * 
 * \param[in] mhd
 * \param[inout] secondShellMHDFunctions
 * \param[inout] solverFunctions
 * \param[inout]ionSolver
 *
 * \return elapsed CPU time (in seconds)
 ************************************************************************/
float ionosphere_MHD_to_MIX(MHD_Interface *mhd, 
			    realCompositeGridFunction &secondShellMHDFunctions, realCompositeGridFunction &solverFunctions,
			    Solver &ionSolver)
{
#ifdef DEBUG_MODE_ON
  cout << "MIX/src/ionosphere_functions.h::ionosphere_MHD_to_MIX(...)" << endl;
#endif
  clock_t startTime = clock();

  // ===== Receive data (current, density, sound speed) from MHD code ===================
  mhd->Import(secondShellMHDFunctions);
    
  // ===== Map from LFM cartesian (x2ion,y2ion) onto the solver cartesian ===============
  ionSolver.mapToSolverGrid(secondShellMHDFunctions,solverFunctions);
  return (float) (clock()-startTime)/CLOCKS_PER_SEC;
}


/********************************************************************//**
 * MIX updates the Flux & Energy using the current and density from LFM MHD
 * 
 * \param[inout] sigma
 * \param[in] solverFunctions
 *
 * \return elapsed CPU time (in seconds)
 ************************************************************************/
float ionosphere_get_Flux_and_Energy(const int &exchangeWithIm, IM_Interface *im, Conductance &sigma, realCompositeGridFunction &solverFunctions, const real &tilt)
{
#ifdef DEBUG_MODE_ON
  cout << "MIX/src/ionosphere_functions.h::ionosphere_compute_Flux_and_Energy(...)" << endl;
#endif
  clock_t startTime = clock();

  // MIX (RCM) does not import (export) on the first exchange, as
  // instructed by the LFM via scalar at index
  // MHD_Interface::EXCHANGE_WITH_IM.
  if ( (exchangeWithIm == MHD_Interface::CPL_RECV_FROM_IM) ||
       (exchangeWithIm == MHD_Interface::DUPLEX_IM_EXCHANGE) ){
    //FIXME: What should the import do?  Should it set Flux & Energy?
    //If so, how should we handle data on open field lines?  This is
    //an open question that the LTR team should discuss...
    im->Import();
  }

  // ===== Compute Flux (F) & Energy (E) ================================================
  sigma.precipitation(solverFunctions, tilt, NORTH);
  sigma.precipitation(solverFunctions, -tilt, SOUTH);
  return (float) (clock()-startTime)/CLOCKS_PER_SEC;
}


/********************************************************************//**
 * Update Pedersen & Hall conductances the first iteration of a sequence start/restart
 * 
 * There are three possible outcomes here:
 *   - If this is the first exchange with ITM, & first exchange with
 *     MHD: use MIX to update conductances
 *   - If this is the first exchange with ITM, & we've exchanged with
 *     MHD: do nothing (i.e. use the LFM-MIX conductance model until
 *     we've receieved at least once from ITM model).
 *   - first sequence restart exchange: read conductance profile from file
 * 
 * \note This function assumes that MIX is coupled to an ITM model
 * \note This function assumes that this is the first exchange for a run or sequence
 * \note Should only be called via conductances_update(...)
 * 
 * \param[in] params
 * \param[out] itmExchangeCount
 * \param[in] currentUT
 * \param[in] DXHDF
 * \param[in] tilt
 * \param[inout] solverFunctions
 *
 * \return elapsed CPU time (in seconds)
 ************************************************************************/
void conductances_update_sequence_begin(const Params &params, const unsigned long int &mhdExchangeCount, unsigned long int &itmExchangeCount, 
					DateTime &currentUT, DateTime &itmCouplingNext, IO &DXHDF, real &tilt, 
					realCompositeGridFunction &solverFunctions, Conductance &sigma)
{
#ifdef DEBUG_MODE_ON
  cout << "MIX/src/ionosphere_functions.h::conductances_update_sequence_begin(...)" << endl;
#endif
  // How many exchanges have we done across all restarts for a run?
  // Get the ITM coupling delta in terms of MJD:  divide by  86,400 = 60*60*24
  itmExchangeCount = floor( (currentUT.getMJD() - params.itmCouplingStart.getMJD()) / (params.itmCouplingDelta/86400));
      
  // Has MIX exchanged with the ITM before?
  if (itmExchangeCount > 0){
    cout << "***MIX-ITM has already exchanged " << itmExchangeCount << " times." << endl;	  
    // ... ITM-MIX has exchanged before 
    // ... i.e.  this is a restart run!
    //
    // Read Pedersen & Hall conductances from old MIX file:
    //
    // For now, we assume that MIX saves output on every //
    // MIX-ITM exchange...  So a the previous restart    //
    // file should be at                                 //
    //       nextCouplingTime - itmCouplingDelta         //
    itmCouplingNext = params.itmCouplingStart;
    itmCouplingNext.incrementSeconds( (itmExchangeCount) * params.itmCouplingDelta );
    unsigned long int previousTimeStep;
    // The "previous" MIX file should be named:
    //       "[run-name]_mix_[itmCouplingNext.getDateTimeString].hdf"
#ifdef DX_OUT
    cout << "Reading Conductances from restart file at " << itmCouplingNext.getDateTimeString() << endl;

    DXHDF.ReadVariable(itmCouplingNext, solverFunctions, SOLVER_SIGMAP_NORTH);
    DXHDF.ReadVariable(itmCouplingNext, solverFunctions, SOLVER_SIGMAP_SOUTH);

    DXHDF.ReadVariable(itmCouplingNext, solverFunctions, SOLVER_SIGMAH_NORTH);
    DXHDF.ReadVariable(itmCouplingNext, solverFunctions, SOLVER_SIGMAH_SOUTH);

    DXHDF.ReadVariable(itmCouplingNext, solverFunctions, SOLVER_WIND_NORTH);
    DXHDF.ReadVariable(itmCouplingNext, solverFunctions, SOLVER_WIND_SOUTH);
    
#else
#error ERROR: MIX-ITM needs to read data from previous MIX file, but MIX was not compiled with -DDX_OUT !
#endif

    // Pre-pend MIX-TIEGCM exchange stdout to designate this is a restarting from "MIX-TIEGCM exchange # ..."
    // i.e. this will eventually output:
    //      "Restart MIX-TIEGCM exchange number [#]; UT: YYYY-MM-DDTHH:MM:SS"
    cout << "Restart ";
  }
  else{
    // ... ITM-MIX has never exchanged & itmExchangeCount == 0.

    // If this is the first LFM-MIX exchange, update the conductances
    if (mhdExchangeCount == 1){
      sigma.getTotal(solverFunctions, tilt, NORTH);
      sigma.getTotal(solverFunctions, -tilt, SOUTH);
      sigma.applyCap(solverFunctions);
    }
    // ... otherwise, use the Pedersen (sigma_p) & Hall (sigma_h)
    // conductances from the previous LFM-MIX exchange.

    // Pre-pend MIX-TIEGCM exchange stdout to designate the first exchange
    // i.e. this will eventually output:
    //      "First MIX-TIEGCM exchange number 1; UT: YYYY-MM-DDTHH:MM:SS"
    cout << "First ";
  }
  itmExchangeCount++;
}


/********************************************************************//**
 * Update the Pedersen & Hall conductances.  
 * 
 * There are three possible outcomes here:
 *   - first exchange with ITM:, do nothing (i.e. use the LFM-MIX
 *     conductance model until we've receieved at least once from ITM
 *     model).
 *   - first sequence restart exchange: read conductance profile from file
 *   - not the first exchange: import data from ITM model
 * 
 * \note Assumes that MIX is coupled to an ITM model & it's time to exchange data.
 * \note Should only be called via ionosphere_get_Conductances(...) 
 *
 * \param[in] params
 * \param[out] itmExchangeCount
 * \param[in] currentUT
 * \param[in] DXHDF
 * \param[in] tilt
 * \param[inout] solverFunctions
 * \param[in] itm
 *
 ************************************************************************/
void conductances_update(const Params &params, const unsigned long int &mhdExchangeCount, unsigned long int &itmExchangeCount, 
			 DateTime &currentUT, DateTime &itmCouplingNext, IO &DXHDF, real &tilt, 
			 realCompositeGridFunction &solverFunctions, ITM_Interface *itm, Conductance &sigma)
{
#ifdef DEBUG_MODE_ON
  cout << "MIX/src/ionosphere_functions.h::conductances_update(...)" << endl;
#endif
  // We are coupled to ITM!
    
  // Is this the first ITM exchange for this run?
  if (itmExchangeCount == 0) {	
    conductances_update_sequence_begin(params, mhdExchangeCount, itmExchangeCount, 
				       currentUT, itmCouplingNext, DXHDF, tilt, solverFunctions, sigma);
  }
  else{
    // itmExchange > 1 if we're here... Meaning that the ITM should have :
    //    1.  Received F,E,phi at step (itmExchange-1)
    //    2.  Time-stepped itmCouplingDelta seconds
    //    3.  Exported (sigma_p, sigma_h, J_w) to MIX for step (itmExchange)
    // 
    // Get Pedersen & Hall conductances and Field-Aligned Current from ITM:
    cout << "Importing from ITM at " << currentUT.getDateTimeString() 
	 << " (" << "unknown" << ")" << endl;
    itm->Import(solverFunctions);
    sigma.applyCap(solverFunctions);
  }
	
  // stdout MIX-TIEGCM exchange number & UT:
  cout << "MIX-TIEGCM exchange number " << itmExchangeCount << "; "
       << "UT: " <<  currentUT.getDateTimeString() << endl;
}


/********************************************************************//**
 * Update the Pedersen & Hall conductances.  
 * 
 * If we're not coupled to ITM, update conductances with the MIX conductance model
 * If coupled to ITM & it's time to exchange: get from ITM via conductances_update(...)
 *
 * \param[in] params
 * \param[out] itmExchangeCount
 * \param[in] currentUT
 * \param[in] itmCouplingNext
 * \param[out] sigma
 * \param[inout] solverFunctions
 * \param[in] DXHDF
 * \param[in] tilt
 * \param[in] itm
 *
 * \return elapsed CPU time (in seconds)
 ************************************************************************/
float ionosphere_get_Conductances(const Params &params, const unsigned long int &mhdExchangeCount, unsigned long int &itmExchangeCount, 
				  DateTime &currentUT, DateTime &itmCouplingNext,
				  Conductance &sigma, realCompositeGridFunction &solverFunctions, 
				  IO &DXHDF, real &tilt, ITM_Interface *itm)
{
#ifdef DEBUG_MODE_ON
  cout << "MIX/src/ionosphere_functions.h::ionosphere_get_Conductances(...)" << endl;
#endif
  clock_t startTime = clock();

  // ===== Compute Pedersen (sigma_p) & Hall (sigma_h) conductances =====================
#ifdef DEBUG_MODE_ON
  cout << "*** MHD UT(i): " << currentUT.getDateTimeString() << endl
       << "*** ITM UT(0): " << params.itmCouplingStart.getDateTimeString() << endl
       << "*** ITM_UT(k): " << itmCouplingNext.getDateTimeString() << endl;
#endif
  if ( currentUT < params.itmCouplingStart ){
    // MIX computes new Pedersen & Hall conductances since we haven't coupled to MIX (yet).
    sigma.getTotal(solverFunctions, tilt, NORTH);
    sigma.getTotal(solverFunctions, -tilt, SOUTH);
    sigma.applyCap(solverFunctions);
  }
  else if ( currentUT >= itmCouplingNext ){
    // MIX imports conductances from:
    //   * TIEGCM if it's time to 
    //   * file   if this is the first iteration of a restart
    conductances_update(params, mhdExchangeCount, itmExchangeCount, currentUT, 
			itmCouplingNext, DXHDF, tilt, solverFunctions, itm, sigma);
  }
  else{
    // MIX uses conductances from the previous ITM exchange
    cout << "*** Using Pedersen (sigma_p) & Hall (sigma_h) conductances from previous MIX-ITM exchange." << endl;
  }

  return (float) (clock()-startTime)/CLOCKS_PER_SEC;
}


/********************************************************************//**
 * Compute Potential (phi) using MIX solver (conductances & current ==>> update potential)
 * 
 * \param[inout] solverFunctions
 * \param[inout] ionSolver
 *
 * \return elapsed CPU time (in seconds)
 ************************************************************************/
float ionosphere_compute_Potential(Solver &ionSolver, realCompositeGridFunction &solverFunctions)
{
#ifdef DEBUG_MODE_ON
  cout << "MIX/src/ionosphere_functions.h::ionosphere_compute_Potential(...)" << endl;
#endif
  clock_t startTime = clock();
  ionSolver.solve(solverFunctions,NORTH);
  ionSolver.solve(solverFunctions,SOUTH);
  return (float) (clock()-startTime)/CLOCKS_PER_SEC;
}


/********************************************************************//**
 * Interpolate & send the Potential from  MIX to MHD
 * 
 * \param[in] mhd
 * \param[inout] solverFunctions
 * \param[inout] firstShellMHDFunctions
 * \param[inout] secondShellMHDFunctions
 *
 * \return elapsed CPU time (in seconds)
 ************************************************************************/
float ionosphere_MIX_to_MHD(MHD_Interface *mhd, Solver &ionSolver, realCompositeGridFunction &solverFunctions,
			    realCompositeGridFunction &firstShellMHDFunctions, 
			    realCompositeGridFunction &secondShellMHDFunctions)
{
#ifdef DEBUG_MODE_ON
  cout << "MIX/src/ionosphere_functions.h::ionosphere_MIX_to_MHD(...)" << endl;
#endif
  clock_t startTime = clock();
  // ===== Map from solver cartesian onto the MHD cartesian (x2ion,y2ion in LFM terms) ==
  ionSolver.mapFromSolverGrid(solverFunctions,secondShellMHDFunctions);
  ionSolver.mapFromSolverGrid(solverFunctions,firstShellMHDFunctions);      
  
  // ===== Send data (potential) to MHD code ============================================
  mhd->Export(firstShellMHDFunctions,secondShellMHDFunctions);

  return (float)(clock()-startTime)/CLOCKS_PER_SEC;
}


/********************************************************************//**
 * Interpolate & send the Potential, Energy & Flux from MIX to ITM
 * 
 * \param[in] mhd
 * \param[inout] solverFunctions
 * \param[inout] firstShellMHDFunctions
 * \param[inout] secondShellMHDFunctions
 *
 * \return elapsed CPU time (in seconds)
 ************************************************************************/
float ionosphere_MIX_to_ITM(const Params &params, const bool &isLastStep, unsigned long int &itmExchangeCount, 
			    DateTime &currentUT, DateTime &itmCouplingNext, ITM_Interface *itm, Solver &ionSolver, 
			    realCompositeGridFunction &solverFunctions, realCompositeGridFunction &smGlobeFunctions,
			    IO &DXHDF, real &tilt)
{
#ifdef DEBUG_MODE_ON
  cout << "MIX/src/ionosphere_functions.h::ionosphere_MIX_to_ITM(...)" << endl;
#endif
  clock_t startTime = clock();

  // Dont send to ITM if this is the last timestep of a sequence.
  if ( isLastStep ){
    return 0.0;
  }
  // ===== Send data (potential, Flux & Energy) to ITM code =============================
  else if (currentUT >= itmCouplingNext){	

    // Interpolate from solver grid to ITM lat/lon grid:
    ionSolver.mapToSMGlobe(solverFunctions,smGlobeFunctions);

    // Send ITM the Potential, avg. Energy & avg. Flux
    // itm->Export(currentUT,smGlobeFunctions);
    cout << "Exporting to ITM at " << currentUT.getDateTimeString() 
	 << " (" << itmCouplingNext.getDateTimeString() << ")" << endl;
    itm->Export(itmCouplingNext, smGlobeFunctions);

    // Increment next ITM exchange time:
    itmCouplingNext = params.itmCouplingStart;
    // FIXME:  do we really want to increment by epsilon?
    itmCouplingNext.incrementSeconds(itmExchangeCount * params.itmCouplingDelta);
    itmExchangeCount++;

    ////////////
    // FIXME: //
    ///////////////////////////////////////////////////////
    // Do we want to write the MIX datafile or trust that//
    // the user will setup the exchange rate & dump      //
    // frequency properly?                               //
    // For now, trust the user...                        //
    ///////////////////////////////////////////////////////

    ////////////
    // FIXME: // 
    ////////////////////////////////////////////////////////
    // MIX crashes when IO::Write(...) tries to overwrite //
    // existing files.
    ///////////////////////////////////////////////////////
    //DXHDF.Write(tilt, currentUT, mhdTimeStep, solverFunctions);

    return (float)(clock()-startTime)/CLOCKS_PER_SEC;
  }
  else {
    return 0.0;
  }
}


/********************************************************************//**
 * Interpolate & send the Potential from MIX to the Inner Magnetosphere
 * 
 * \param[in] exchangeWithIm
 * \param[inout] im
 * \param[inout] solverFunctions
 *
 * \return elapsed CPU time (in seconds)
 ************************************************************************/
float ionosphere_MIX_to_IM(const int &exchangeWithIm, IM_Interface *im, realCompositeGridFunction &solverFunctions)
{
#ifdef DEBUG_MODE_ON
  cout << "MIX/src/" << __FILE__ << "::" << __FUNCTION__ << endl;
#endif
  clock_t startTime = clock();

  // MIX (RCM) does not export (import) on the last exchange, as
  // instructed by the LFM via scalar at index
  // MHD_Interface::EXCHANGE_WITH_IM.
  if ( (exchangeWithIm == MHD_Interface::CPL_SEND_TO_IM) || 
       (exchangeWithIm == MHD_Interface::DUPLEX_IM_EXCHANGE) ){
    // interpolation occurs within IM_Interface::Export(...)

    //FIXME: What should the Export send to the RCM?  Potential on the
    //Northern hemisphere?  Southern?  This is an open question that
    //the LTR team should discuss...
    im->Export(solverFunctions);
  }

  return (float)(clock()-startTime)/CLOCKS_PER_SEC;
} 


/********************************************************************//**
 * Receive MHD scalars that describe the state of LFM-MIX exchange
 * 
 * \param[in] mhd
 * \param[out] currentDateTime
 * \param[out] tilt
 * \param[out] mhdTimeStep
 * \param[out] isLastStep
 * \param[out] exchangeWithIm - Is this the first/intermediate/last exchange with RCM?  See MHD_Interface::MHD_EXCHANGE_WITH_ITM for valid integer values.
 * \param[inout] imExchangeCount - Number of exchanges have occured between MIX & RCM
 *
 * \return elapsed CPU time (in seconds)
 ************************************************************************/
bool receiveMHDscalars(MHD_Interface *mhd,  DateTime &currentDateTime, real &tilt, unsigned long int &mhdTimestep, bool &isLastStep, int &exchangeWithIm, unsigned long int &imExchangeCount)
{
#ifdef DEBUG_MODE_ON
  cout << "MIX/src/ionosphere_functions.h::receiveMHDscalars(...)" << endl;
#endif
  realArray scalars = mhd->getScalars();
  isLastStep = false; // must initialize to false!!!

  if ( fabs(scalars(MHD_Interface::KILL_SIGNAL) - MHD_Interface::KILL_SIGNAL_LAST_EXCHANGE) < 1e-5 ){
    cout << "Last LFM-MIX exchange" << endl;
    isLastStep = true;
  }   
  else if ( fabs(scalars(MHD_Interface::KILL_SIGNAL) - MHD_Interface::KILL_SIGNAL_SHUTDOWN) < 1e-5 ){  
    cout << "MIX received kill signal from the MHD code" << endl;
    //  exit the code
    return false;
  }
  
  currentDateTime = DateTime(scalars(MHD_Interface::YEAR),
			     scalars(MHD_Interface::MONTH),
			     scalars(MHD_Interface::DAY), 
			     scalars(MHD_Interface::HOUR), 
			     scalars(MHD_Interface::MINUTE), 
			     (double) scalars(MHD_Interface::SECOND));
  tilt = scalars(MHD_Interface::TILT);
  mhdTimestep = (unsigned long int)(scalars(MHD_Interface::LABEL));

  // See MHD_Interface::MHD_EXCHANGE_WITH_ITM for valid integer values
  exchangeWithIm = (int) scalars(MHD_Interface::EXCHANGE_WITH_IM);

  // Output to stdout to let the user know what's going on.
  if (exchangeWithIm == MHD_Interface::CPL_SEND_TO_IM){
    std::cout << "First exchange with RCM  (#" << ++imExchangeCount << ") "
	      << currentDateTime.getDateTimeString() << std::endl;
  }
  else if (exchangeWithIm == MHD_Interface::CPL_RECV_FROM_IM){
    std::cout << "Last exchange with RCM  (#" << ++imExchangeCount << ") "
	      << currentDateTime.getDateTimeString() << std::endl;
  }
  else if (exchangeWithIm == MHD_Interface::DUPLEX_IM_EXCHANGE) {
    std::cout << "MIX-RCM exchange  (#" << ++imExchangeCount << ") " 
	      << currentDateTime.getDateTimeString() << std::endl;
  }

  // If we made it this far, MIX should keep running.
  return true;
}

#endif
