#ifndef IO_TEST_H_
#define IO_TEST_H_

// Test
#include <cxxtest/TestSuite.h>

// 3rd-party
#include "Overture.h"
#include "DataPointMapping.h"

// Common 
#include "DateTime.h"

// MIX
#include "const.h"
#include "functions.h"
#include "IO.h"
#include "Params.h"

class IOTest : public CxxTest::TestSuite
{
private:
  realCompositeGridFunction solverFunctions;
  // Grid information:
  int nLat, nLon;
  realArray x,y;
  Params *input_parameters;
  CompositeGrid sgrid;

public:

  /******************************************************************/

  // setUpWorld is executed once at startup of testing process
  // (note: setUpWorld is in the documentation, but didn't work... define a constructor)
  IOTest(){
    // Setup (x,y) lat/lon grid:
    nLat = NLAT;
    nLon = 360.0/(SOLVER_LON_GRID_STEP+1.0);

    x.redim(nLat,nLon);
    y.redim(nLat,nLon);

    realArray lon(nLon), lat(nLat);
    for (int i = 0; i < nLat; i++) {
      lat(i) = LATITUDE[i];
    }
    lon.seqAdd(0.,SOLVER_LON_GRID_STEP);
    
    lon = lon/180.*PI;
    lat = lat/180.*PI;
    
    for (int k = 0; k < nLon; k++) {
      x(all,k)=sin(lat)*cos(lon(k));
      y(all,k)=sin(lat)*sin(lon(k));
    }

    // Setup input Param object
    input_parameters = new Params("IO_test.param");
  }

  /******************************************************************/

  // tearDownWorld is executed once at cleanup, like a destructor
  // (note: tearDownWorld is in the documentation, but didn't work... define a destructor)
  ~IOTest(){
    delete input_parameters;
  }

  /******************************************************************/

  // setUp() ( tearDown() ) are exectuted before ( after ) each test
  void setUp(){
    realArray ionArray(nLat,nLon,2);
    ionArray(all,all,0)=x;
    ionArray(all,all,1)=y;
    
    DataPointMapping & sphere = *new DataPointMapping;
    sphere.incrementReferenceCount();
    sphere.setDataPoints(ionArray,2,2);
    sphere.setIsPeriodic(1,Mapping::functionPeriodic);
    
    sgrid = CompositeGrid(2,1);
    sgrid[0].reference(sphere);
    sgrid.update();
    sphere.decrementReferenceCount();

    solverFunctions = realCompositeGridFunction(sgrid,all,all,all,18);
    solverFunctions.setName("Solver grid functions");
    solverFunctions.setName("Potential North [V]",SOLVER_POTENTIAL_NORTH);         // MIX: phi
    solverFunctions.setName("Potential South [V]",SOLVER_POTENTIAL_SOUTH);         //      phi
    solverFunctions.setName("FAC North [A/m^2]",SOLVER_FAC_NORTH);                 // ITM: J_w
    solverFunctions.setName("FAC South [A/m^2]",SOLVER_FAC_SOUTH);                 //      J_w
    solverFunctions.setName("Sound speed North [cm/s]",SOLVER_SOUND_SPEED_NORTH);  // MHD: c_s
    solverFunctions.setName("Sound speed South [cm/s]",SOLVER_SOUND_SPEED_SOUTH);  //      c_s
    solverFunctions.setName("Density North [g/cm^3]",SOLVER_DENSITY_NORTH);        // MHD: rho
    solverFunctions.setName("Density South [g/cm^3]",SOLVER_DENSITY_SOUTH);        //      rho
    solverFunctions.setName("Pedersen conductance North [S]",SOLVER_SIGMAP_NORTH); // ITM: sigma_p
    solverFunctions.setName("Pedersen conductance South [S]",SOLVER_SIGMAP_SOUTH); //      sigma_p
    solverFunctions.setName("Hall conductance North [S]",SOLVER_SIGMAH_NORTH);     // ITM: sigma_h
    solverFunctions.setName("Hall conductance South [S]",SOLVER_SIGMAH_SOUTH);     //      sigma_h
    solverFunctions.setName("Average energy North [keV]",SOLVER_AVG_ENG_NORTH);    // MIX: E
    solverFunctions.setName("Average energy South [keV]",SOLVER_AVG_ENG_SOUTH);    //      E
    solverFunctions.setName("Number flux North [1/cm^2 s]",SOLVER_NUM_FLUX_NORTH); // MIX: F
    solverFunctions.setName("Number flux South [1/cm^2 s]",SOLVER_NUM_FLUX_SOUTH); //      F
    solverFunctions.setName("Neutral wind speed North [m/s]",SOLVER_WIND_NORTH);   //
    solverFunctions.setName("Neutral wind speed South [m/s]",SOLVER_WIND_SOUTH);   //    
  }

  /******************************************************************/

  void setAllData(realCompositeGridFunction &toFunc, bool setToZero = false)
  {
    setData(toFunc, SOLVER_POTENTIAL_NORTH, setToZero);
    setData(toFunc, SOLVER_POTENTIAL_SOUTH, setToZero);
    setData(toFunc, SOLVER_FAC_NORTH, setToZero);
    setData(toFunc, SOLVER_FAC_SOUTH, setToZero);
    setData(toFunc, SOLVER_SOUND_SPEED_NORTH, setToZero);
    setData(toFunc, SOLVER_SOUND_SPEED_SOUTH, setToZero);
    setData(toFunc, SOLVER_DENSITY_NORTH, setToZero);
    setData(toFunc, SOLVER_DENSITY_SOUTH, setToZero);
    setData(toFunc, SOLVER_SIGMAP_NORTH, setToZero);
    setData(toFunc, SOLVER_SIGMAP_SOUTH, setToZero);
    setData(toFunc, SOLVER_SIGMAH_NORTH, setToZero);
    setData(toFunc, SOLVER_SIGMAH_SOUTH, setToZero);
    setData(toFunc, SOLVER_AVG_ENG_NORTH, setToZero);
    setData(toFunc, SOLVER_AVG_ENG_SOUTH, setToZero);
    setData(toFunc, SOLVER_NUM_FLUX_NORTH, setToZero);
    setData(toFunc, SOLVER_NUM_FLUX_SOUTH, setToZero);
    setData(toFunc, SOLVER_WIND_NORTH, setToZero);
    setData(toFunc, SOLVER_WIND_SOUTH, setToZero);
  }

  // Set default values for a realCompositeGridFunction
  void setData(realCompositeGridFunction &toFunc, int variable, bool setToZero = false)
  {
    //cout << "setting " << variable << endl;
    MappedGrid *grid = toFunc[0].getMappedGrid();
    Index I1,I2,I3;
    getIndex((*grid).gridIndexRange(),I1,I2,I3);
    const realArray & x = (*grid).vertex()(I1,I2,I3,0);
    const realArray & y = (*grid).vertex()(I1,I2,I3,1);
    const int nx = x.getBound(0) - x.getBase(0) + 1;
    const int ny = x.getBound(1) - x.getBase(1) + 1;

    if (setToZero){
      for (int j = 0; j < ny; j++) {
	for (int i = 0; i < nx; i++) {
	  toFunc[0](i,j,all,variable) = real(0.0);
	}
      }
    }
    else{
      for (int j = 0; j < ny; j++) {
	for (int i = 0; i < nx; i++) {
	  toFunc[0](i,j,all,variable) = real(i*1000 + j);
	}
      }
    }
    }

  /******************************************************************/
  
  void test_WarningMessage()
  {
    TS_WARN( "Params class sends a bunch of junk to stdout." );
    TS_WARN( "TODO: Clean up Params class stdout..." );
  }

  /******************************************************************/

  void test_resolution()
  {
    TS_ASSERT_EQUALS(nLat, 27);
    TS_ASSERT_EQUALS(nLon, 120);
  }  

  /******************************************************************/

  /***********
   * WARNING *
   *******************************************************************
   * Order is important here... test_write must be called before the *
   * tests that read data.                                           *
   *******************************************************************/
  void test_write()
  {   
    setAllData(solverFunctions, false);

    IO DXHDF(input_parameters);
    double tilt = -1;
    DateTime currentUT(2002,3,21,0,0,0.0);
    long mhdTimeStep = 0;    
    //TS_ASSERT_THROWS_NOTHING( DXHDF.Write(tilt, currentUT, mhdTimeStep, solverFunctions) );
    DXHDF.Write(tilt, currentUT, mhdTimeStep, solverFunctions);
  }

  /******************************************************************/

  // Read *all* variables & store varNorth and varSouth in north & south.
  void readData(const int varNorth, doubleArray &north, const int varSouth, doubleArray &south)
  {
    setAllData(solverFunctions, true);

    IO DXHDF(input_parameters);
    double tilt;
    DateTime currentUT(2002,3,21,0,0,0.0);
    long mhdTimeStep;    

    DXHDF.Read(tilt, currentUT, mhdTimeStep, solverFunctions);

    // Get grid index ranges:
    MappedGrid *grid = solverFunctions[0].getMappedGrid();
    Index I1,I2,I3;
    // include the periodic points here (gridIndexRange) to make it compliant with the grid storage
    getIndex((*grid).gridIndexRange(),I1,I2,I3);

    //solverFunctions[0](I1,I2,I3).display("---------");

    north = solverFunctions[0](I1,I2,I3, varNorth);
    south = solverFunctions[0](I1,I2,I3, varSouth);
  }

  /******************************************************************/

  // Read *one* variable & store in north & south.
  void readVariable(const int varNorth, doubleArray &north, const int varSouth, doubleArray &south)
  {
    IO DXHDF(input_parameters);
    double tilt;
    DateTime currentUT(2002,3,21,0,0,0.0);
    long mhdTimeStep;    

    // Get grid index ranges:
    MappedGrid *grid = solverFunctions[0].getMappedGrid();
    Index I1,I2,I3;
    // include the periodic points here (gridIndexRange) to make it compliant with the grid storage
    getIndex((*grid).gridIndexRange(),I1,I2,I3);

    DXHDF.ReadVariable(currentUT, solverFunctions, varNorth);
    north = solverFunctions[0](I1,I2,I3, varNorth);

    DXHDF.ReadVariable(currentUT, solverFunctions, varSouth);
    south = solverFunctions[0](I1,I2,I3, varSouth);
  }

  void checkVariable(realArray var, const int ni, const int nj)
  {
    int nErrors = 0;
    for (int j = 0; j < nj; j++) {
      for (int i = 0; i < ni; i++) {
	if  ( fabs( var(i,j) - real(i*1000+j) ) > 1e-5 )
	  nErrors++;
	//TS_ASSERT_DELTA( var(i,j), real(i*1000 + j), 1e-5 );
      }
    }    
    std::cout << nErrors << std::endl;
  }

  /******************************************************************/

  void test_checkPotential()
  {
    realArray north, south;

    readData(SOLVER_POTENTIAL_NORTH, north,
	     SOLVER_POTENTIAL_SOUTH, south);

    // Check for valid variables
    for (int j = 0; j < nLon; j++) {
      for (int i = 0; i < nLat; i++) {
	TS_ASSERT_DELTA( north(i,j), real(i*1000 + j), 1e-5 );
	TS_ASSERT_DELTA( south(i,j), real(i*1000 + j), 1e-5 );
      }
    }
    checkVariable(north, nLon, nLat);
    //    checkVariable(south, nLon, nLat);

    realArray north2, south2;
    readVariable(SOLVER_POTENTIAL_NORTH, north2,
		 SOLVER_POTENTIAL_SOUTH, south2);
    checkVariable(north2, nLon, nLat);
    //checkVariable(south2, nLon, nLat);

  }

  /******************************************************************/

  void test_checkFac()
  {
    realArray north, south;

    readData(SOLVER_FAC_NORTH, north,
	     SOLVER_FAC_SOUTH, south);

    // Check for valid variables
    for (int j = 0; j < nLon; j++) {
      for (int i = 0; i < nLat; i++) {
	TS_ASSERT_DELTA( north(i,j), real(i*1000 + j), 1e-5 );
	TS_ASSERT_DELTA( south(i,j), real(i*1000 + j), 1e-5 );
      }
    }
  }

  /******************************************************************/

  void test_checkSigmap()
  {
    realArray north, south;

    readData(SOLVER_SIGMAP_NORTH, north,
	     SOLVER_SIGMAP_SOUTH, south);

    // Check for valid variables
    for (int j = 0; j < nLon; j++) {
      for (int i = 0; i < nLat; i++) {
	TS_ASSERT_DELTA( north(i,j), real(i*1000 + j), 1e-5 );
	TS_ASSERT_DELTA( south(i,j), real(i*1000 + j), 1e-5 );
      }
    }
  }

  /******************************************************************/

  void test_checkSigmah()
  {
    realArray north, south;

    readData(SOLVER_SIGMAH_NORTH, north,
	     SOLVER_SIGMAH_SOUTH, south);

    // Check for valid variables
    for (int j = 0; j < nLon; j++) {
      for (int i = 0; i < nLat; i++) {
	TS_ASSERT_DELTA( north(i,j), real(i*1000 + j), 1e-5 );
	TS_ASSERT_DELTA( south(i,j), real(i*1000 + j), 1e-5 );
      }
    }
  }

  /******************************************************************/

  void test_checkAvgEng()
  {
    realArray north, south;

    readData(SOLVER_AVG_ENG_NORTH, north,
	     SOLVER_AVG_ENG_SOUTH, south);

    // Check for valid variables
    for (int j = 0; j < nLon; j++) {
      for (int i = 0; i < nLat; i++) {
	TS_ASSERT_DELTA( north(i,j), real(i*1000 + j), 1e-5 );
	TS_ASSERT_DELTA( south(i,j), real(i*1000 + j), 1e-5 );
      }
    }
  }

  /******************************************************************/

  void test_checkNumFlux()
  {
    realArray north, south;

    readData(SOLVER_NUM_FLUX_NORTH, north,
	     SOLVER_NUM_FLUX_SOUTH, south);

    // Check for valid variables
    for (int j = 0; j < nLon; j++) {
      for (int i = 0; i < nLat; i++) {
	TS_ASSERT_DELTA( north(i,j), real(i*1000 + j), 1e-5 );
	TS_ASSERT_DELTA( south(i,j), real(i*1000 + j), 1e-5 );
      }
    }
  }

  /******************************************************************/

  void test_checkWind()
  {
    realArray north, south;

    readData(SOLVER_WIND_NORTH, north,
	     SOLVER_WIND_SOUTH, south);

    // Check for valid variables
    for (int j = 0; j < nLon; j++) {
      for (int i = 0; i < nLat; i++) {
	TS_ASSERT_DELTA( north(i,j), real(i*1000 + j), 1e-5 );
	TS_ASSERT_DELTA( south(i,j), real(i*1000 + j), 1e-5 );
      }
    }
  }

  /******************************************************************/


};
#endif /* IO_TEST_H_ */
