#ifndef INTERPOLATEPOINTS_TEST_H_
#define INTERPOLATEPOINTS_TEST_H_

#include <cxxtest/TestSuite.h>

#include "Overture.h"
#include "interpolatePoints.h"
#include "DataPointMapping.h"

static const Range all;

class OvertureSingleton : CxxTest::GlobalFixture
{
public:
  virtual bool setUpWorld()
  {
    cout << "---Initialize Overture---" << endl;
    int argc;
    char ** argv;
    Overture::start(argc,argv); 
    return true;
  }
  virtual bool tearDownWorld() 
  {
    cout << "---Shutting down Overture---" << endl;
    Overture::finish();
    return true;
  }
};

static OvertureSingleton myOverture;

class Interpolate2d : public CxxTest::TestSuite
{
  int nLon, nLat;
  realArray longitude, latitude;
  realArray gridPositions;
  
  CompositeGrid latLonGrid;
  realCompositeGridFunction functionValues;

public:
  void setUp()
  {
    // Create evenly-spaced lat/lon grid
    nLon = 3;
    nLat = 3;

    latitude.redim(nLat);
    longitude.redim(nLon);
    
    realArray lat(nLat);
    lat.seqAdd(0.0, 1.0);

    realArray lon(nLon);
    lon.seqAdd(0.0, 1.0); 

    gridPositions.redim(nLat, nLon, 2);

    for (int j=0; j < nLon; j++){
      for (int i=0; i < nLat; i++){
	gridPositions(i,j,0) = lat(i);
	gridPositions(i,j,1) = lon(j);	
      }
    }
    
    // Instantiate Overture data structure for lat/lon grid.
    DataPointMapping &gridMap = *new DataPointMapping;
    gridMap.incrementReferenceCount();
    gridMap.setDataPoints(gridPositions, 2, 2);

    // A CompositeGrid is a class that holds an overlapping grid.
    latLonGrid = CompositeGrid(2, 1);
    latLonGrid[0].reference(gridMap);
    latLonGrid.update();
    gridMap.decrementReferenceCount();

    // Create f(lat, lon) realCompositeGridFunction

    //functionValues = realCompositeGridFunction(latLonGrid, all, all, all, 1);
    functionValues = realCompositeGridFunction(latLonGrid);
    functionValues.setName("Test Function Values");
    functionValues.setName("f(lat, lon) = lat", 0);

    functionValues[0](Range(0,nLat-1),
		      Range(0,nLon-1),
		      Range(0,0),
		      0) = 2.0*gridPositions(Range(0,nLat-1),
					     Range(0,nLon-1),
					     Range(0,0)
					     );
					 
    //gridPositions.display("Lat/Lon Grid");
    //cout << endl;
    //functionValues.display("Function Values");
    //cout << endl;

    ////////////////////////////////////////////////////////////////////////////
    // ready to interpolate. . . 
    ////////////////////////////////////////////////////////////////////////////
  }

  void tearDown()
  {
    //FIXME:  Need to delete DataPointMapping object somehow?
  }

  void testTruth()
  {
    ////////////////////////////////////////////////////////////////////////////
    //FIXME I could only get interpolatePoints to work with
    //      (nLat*nLon, 1) data... I had problems with (nLat, nLon, 1)
    //      data...  Strange, since Bill Henshaw said:
    //
    //         "interpolatePoints does not work in 1D :( There are not
    //         too many complex geometry problems in 1D so not all of
    //         Overture treats this dimension properly.".
    ////////////////////////////////////////////////////////////////////////////

    realArray interpolatedValues(nLat * nLon, 1);
    realArray interpolatedLat(nLat);
    interpolatedLat.seqAdd(0.5, 0.5);
    realArray interpolatedLon(nLon);
    interpolatedLon.seqAdd(0.5, 0.5);
    realArray positionsToInterpolate(nLat * nLon, 2);
    for (int j=0; j < nLon; j++){
      for (int i=0; i < nLat; i++){
	//positionsToInterpolate(i,j,0) = interpolatedLat(i);
	//positionsToInterpolate(i,j,1) = interpolatedLon(j);
	positionsToInterpolate(i+j*nLat, 0) = interpolatedLat(i);
	positionsToInterpolate(i+j*nLat, 1) = interpolatedLon(j);
      }
    }

    //positionsToInterpolate.display("Interpolation grid");
    //cout << endl;
    //interpolatedValues.display("Before interpolation");
    //cout << endl;

    int r = interpolatePoints(positionsToInterpolate, functionValues, interpolatedValues);
    interpolatedValues.reshape(nLat, nLon);
    positionsToInterpolate.reshape(nLat, nLon, 2);

    //interpolatedValues.display("After interpolation");           

    for (int j=0; j < nLon; j++){
      for (int i=0; i < nLat; i++){
	TS_ASSERT_DELTA( interpolatedValues(i,j), 2.0*positionsToInterpolate(i, j, 0), 1e-5 );
      }
    }
  }
};


#endif /*INTERPOLATEPOINTS_TEST_H_*/
