program rcm_lfm  
  use Rcm_mod_subs
  use intercomm_module
  use intermediate_grid, only: setupIg, tearDownIg
  use ionosphere_intermediate, only : setupIon, tearDownIon
  use constants, ONLY : radius_earth_m, radius_iono_m
  use rice_housekeeping_module

  implicit none
  integer(iprec) :: dayOfYear 
  CHARACTER(LEN=3) :: months(12)=(/'JAN','FEB','MAR','APR','MAY','JUN','JUL','AUG','SEP','OCT', 'NOV','DEC'/)

  integer(iprec) :: ierr   !> Error code...
  integer(iprec) :: system !> This code makes a call to the
                           !! non-standard SYSTEM function.  Must
                           !! define as integer to avoid compile error.

  double precision :: time  !> Modified Julian Date + decimal fraction of
                            !! day of current sim time (according to LFM)
  double precision :: time0 !> Modfieid Julian Date + decimal fraction of
                            !! day when coupling starts
  double precision :: dt    !> Interval between RCM exchanges in seconds

  integer(iprec) :: timercm = 0 !> Time (in seconds) 

  integer(iprec) :: itimei !> RCM(...) param:  start time
  integer(iprec) :: itimef !> RCM(...) param:  end time
  integer(iprec) :: itimef_old = -1
  integer(iprec) :: irdr = 1 !> RCM(...) param:  record # to read in
  integer(iprec) :: irdw = 1 !> RCM(...) param:  record # to write out
  integer(iprec) :: idt   !> RCM(...) param:  basic time step in program
  integer(iprec) :: idt1  !> RCM(...) param:  time step for
                          !! changing disk & write records
  integer(iprec) :: idt2  !> RCM(...) param:  time step for
                          !! writting formatted output

  integer(iprec) :: rec = 1 !> Record number that torcm/tomhd use to
                        !! import/export data with coupled models.

  

  real(rprec) :: t1, t2  !> Used for performance timing


  !> Model coupling variables
  integer :: iaScalars(NUMBER_OF_SCALARS) !> Scalars received from LFM via ReceiveScalars
  integer :: exchangeNum = 0
  logical :: isFirstExchange
  logical :: isLastExchange

  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


  IsCoupledExternally = .TRUE.  ! switch RCM to "coupled" mode before doing anything else

  write (*,'(TR1,A,L7)') 'Welcome to the RCM, IsCoupledExternally=', IsCoupledExternally

  CALL Read_rcm_lfm_params
  CALL Rcm (itimei, itimef, irdr, irdw, idt, idt1, idt2,icontrol=0)


  ! Setup Intermediate Grid
  call setupIg()                   ! here radius_earth_m is invoked first time
  call allocate_conversion_arrays (isize,jsize,kcsize)


  ! Set up RCM ionospheric grid:
  call Grid_torcm (75.0, 15.0, 0.0, radius_earth_m, radius_iono_m)  ! set up RCM ionospheric grid here


  ! Setup Ionosphere intermediate Grid by equating it to the RCM grid, without angular overlap:
  call setupIon()


  ! Setup initial RCM plasma edges
  call plasma_rcm_setup

  CALL Rcm (itimei, itimef, irdr, irdw, idt, idt1, idt2, icontrol=1)
  CALL Rcm (itimei, itimef, irdr, irdw, idt, idt1, idt2, icontrol=2)


  ! Initialize InterComm (must be called after setupIg & setupIon).
  call Initialize()

  ! Initialize() receives the date & time to start coupling
  ! Get day of year for converting to MJD.  
  !FIXME:  Do we really n eed day of year to get modified julian date?
  call date_doy(0, &
       iaUtRcmStart(1), &
       months(iaUtRcmStart(2)), &
       iaUtRcmStart(2), &
       iaUtRcmStart(3), &
       dayOfYear)       
  ! Convert to modified Julian Date (MJD) since RCM requires MJD's.
  call ut2mjd(time0, &    ! Modified Julian Date/time (MJD)
       iaUtRcmStart(1), & ! Year
       dayOfYear, &       ! Day of Year
       iaUtRcmStart(2), & ! Month
       iaUtRcmStart(3), & ! Day
       iaUtRcmStart(4), & ! Hour
       iaUtRcmStart(5), & ! Minute
       DBLE(iaUtRcmStart(6)))    ! Second
  dt = iRcmDelta

  couplingTimeLoop: DO  ! exit the loop only by stopping

     !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     ! Receive & process scalars
     !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

     call ReceiveScalars("mhd-im_scalars", iaScalars)

     ! Is there any more work to do?
     if (iaScalars(KILL_SIGNAL) == KILL_SIGNAL_SHUTDOWN) then
        exit couplingTimeLoop
     end if

     ! Is this the first exchange? (Import only)
     isFirstExchange = &
          (exchangeNum == 0) &
          .and. (iaScalars(KILL_SIGNAL) /= KILL_SIGNAL_SHUTDOWN) &
          .and. (iaScalars(KILL_SIGNAL) /= KILL_SIGNAL_LAST_EXCHANGE)
     ! Is this the last exchange? (Export only)
     isLastExchange = (iaScalars(KILL_SIGNAL) == KILL_SIGNAL_LAST_EXCHANGE)

     ! Tell the user what's going on via stdout message:
     if (isFirstExchange) then
        write(*,*) "First LTR exchange #", exchangeNum
     else if (isLastExchange) then
        write(*,*) "Last LTR exchange #", exchangeNum
     else
        write(*,*) "Intermediate LTR exchange # ", exchangeNum
     end if

     ! Get the current Modified Julian Date (MJD) from time array
     ! First get the day of year (required for calculating MJD)
     !FIXME:  Do we really n eed day of year to get modified julian date?
     call date_doy(0, &
          iaScalars(YEAR), &
          months(iaScalars(MONTH)), &
          iaScalars(MONTH), &
          iaScalars(DAY), &
          dayOfYear)       
     ! Convert to modified Julian Date (MJD) since RCM requires MJD's.
     call ut2mjd(time, & ! Modified Julian Date/time (MJD)
          iaScalars(YEAR), &
          dayOfYear, &  ! Day of year (unecessary)
          iaScalars(MONTH), &
          iaScalars(DAY), &
          iaScalars(HOUR), &
          iaScalars(MINUTE), &
          DBLE(iaScalars(SECOND))) 

     ! Kind of redundant to set delta_t, but maybe we'll use variable time steps in the future?
     dt = iaScalars(DELTA_T)
     ! Convert dt from # of seconds into % of a day:
     dt = dt / 86400.0 ! 86400 = 60*60*24 = # of seconds in day


     if ( isFirstExchange .or. (.not. isLastExchange) ) then
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        ! Determine exchange times...
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        ! set the record number; Note: nint seems to fix some rounding errors.
        rec = nint(2.0*(time-time0)/dt + 1)
        timercm = nint((time-time0)*86400.0)  ! RCM elapsed time in seconds
        WRITE (*,'(//)')
        write (*,*) ' time[mjd]=',time,'  time0[mjd]=',time0, '  Delta_t[s]=',(time-time0)*86400.0
        write (*,*) ' ____T_rcm[s] =', timercm,'  rec=',rec,'  T_lfm[mjd]=',time
        WRITE (*,'(//)')

        if (itimef_old < 0 )then
           itimei = timercm
        else
           itimei = itimef
        end if

        itimef = nint((time-time0+dt)*86400.0D00)

        idt = Idt_overwrite ! RCM internal time step in seconds

        ! now round to to fit the correct number rcm timesteps
        itimef = itimei + idt *((itimef-itimei)/idt)
        itimef_old = itimef

        ! Frequency (in seconds) to change disk & write records
        idt1 = itimef - itimei
        ! Frequency (in seconds) to write formatted output
        idt2 = idt1

        IF (isFirstExchange .AND. itimei == 0) then 
           !special case. Tell RCM to start from rec=1, but most initial
           ! conditions will actually be done in TORCM
           irdr = 1
        ELSE
           irdr = rec - 1
        END IF

        irdw = rec

        if (isFirstExchange) then ! Set RCM initial conditions on plasma:
           call rcm (itimei, itimef, irdr, irdw, idt, idt1, idt2, icontrol=3)
        end if


        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        ! Import data from LFM/MIX
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        call cpu_time(t1)
        write(6,*)' calling torcm with rec =',rec,' itimei=',itimei
        call print_date_time(6)
        call torcm(rec,itimei,ierr)
        call Ionosphere_toRCM
        if (ierr > 0 )then
           stop ' error in torcm '
        endif
        exchangeNum = exchangeNum + 1
        call cpu_time(t2)

        write(*,*)' torcm cpu time= ',t2-t1,' seconds'

        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        ! Advance RCM
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        call cpu_time(t1)
        write(*,*)' call rcm at itimei =',itimei,' to itimef =',itimef,' dt=',dt
        call print_date_time(6)
        call rcm (itimei, itimef, irdr, irdw, idt, idt1, idt2, icontrol=4)
        rec = rec + 1
        call cpu_time(t2)
        write(*,*)'   rcm cpu time= ',t2-t1,' seconds'
        call print_date_time(6)

     end if


!     if (( .not. isFirstExchange) .or. isLastExchange) then
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        ! Export data to LFM/MIX
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        write(6,*)' calling tomhd with rec =',rec
        call cpu_time(t1)
        ! Export magnetosphere variables in intermediate_grid module to LFM MHD.
        call Tomhd (rec, ierr)
        ! Export Ionosphere variables in ionosphere_intermediate
        ! module to the MIX ionosphere coupler/solver.
        call ExportIonosphere()    

        call cpu_time(t2)
        call print_date_time(6)
        write(*,*)' tomhd cpu time= ',t2-t1,' seconds'
        if (ierr > 0 )then
           stop ' error in tomhd '
        end if
!    end if

     !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  end do couplingTimeLoop

  call rcm (itimei,itimef,irdr,irdw,idt,idt1,idt2,icontrol=5)
  call Finalize()    ! Matches Initialize() above
  call tearDownIon() ! Matches setupIon() above
  call tearDownIg()  ! Matches setupIg() above.

end program rcm_lfm

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      
      subroutine min2hr(time,hours)
! returns a string of hr:min:sec from an input in minutes
! 2/04 frt
      USE Rcm_mod_subs, ONLY : iprec,rprec
      implicit none
      real(rprec) :: time
      real(rprec) :: time_hours
      real(rprec) :: time_minutes
      real(rprec) :: time_seconds
      character (len=*) :: hours
      character (len=3) :: char_time_hours
      character (len=2) :: char_time_minutes,char_time_seconds
! time is in minutes

      time_hours = floor(time/60.)
      time_minutes = floor(time - 60*time_hours)
      time_seconds = 60*(time -floor(time))

      write(char_time_hours,'(i3.3)')int(time_hours)
      write(*,*)' char_time_hours =',char_time_hours
!     if(int(time_hours) < 10)then
!             char_time_hours = '00'//adjustr(char_time_hours)
!     endif
!     write(*,*)' char_time_hours =',char_time_hours
      write(char_time_minutes,'(i2.2)')int(time_minutes)
      write(char_time_seconds,'(i2.2)')int(time_seconds)
!     write(*,*)' char_time_seconds =',char_time_seconds
       
      hours = adjustr(char_time_hours) //':'//char_time_minutes//':' &
      //adjustl(char_time_seconds)

      return
      end subroutine min2hr

      SUBROUTINE Plasma_rcm_setup
      USE Rcm_mod_subs
!_________________________________________________________________
!
! purpose: to setup the intial rcm plasma edges, need to be run once
! before the first call to the rcm
! input data
! following wolf doc dated 4/14/95.  let eta=5.9e21 wb**(-1)
!                                        alam=5680 ev (re/nt)^2/3e
! moved inner edge to iner boundary of rcm grid
!_________________________________________________________________
!
      IMPLICIT NONE
      INTEGER(iprec) :: ierr, n, nn, nx, irdw, k
      REAL(rprec) :: almdel(kcsize), almmin(kcsize), almmax(kcsize)
!
      IF (ksize /= kcsize) STOP 'DESIGNED FOR KSIZE=KCSIZE'
!
      CALL Read_alam (kcsize, alamc, ikflavc, fudgec, almdel, &
                      almmax, almmin, iesize, ierr)
      IF (ierr /= 0) STOP 'ERROR EXITING READ_ALAM'
      etac (:) = 1.5E+15
!
!
!     Set edges info to be the same as grid-based channels but with
!     zero densities:
!
      alam = alamc
      ikflav = ikflavc
      fudge = fudgec
      eta = 0.0
!
      n=0
      DO k = 1, ksize
         mpoint(k) = n + 1
         nn=0
         DO nx=2, (jsize-3), 2
            n = n + 1
            nn=nn+1
            bi(n) = REAL(isize) -10.0
            bj(n) = REAL (nx)
            itrack(n)=n
            etab(n)= eta(k)
         END DO
         npoint(k)=nn
      END DO
      mpoint (ksize+1)=n+1
!
!
      CALL Write_plasma ()
!
      irdw = 1
      CALL Write_array (rcmdir//'rcmbi_inp', irdw, label, ARRAY_1D=bi, SETUP=.TRUE.)
      CALL Write_array (rcmdir//'rcmbj_inp', irdw, label, ARRAY_1D=bj, SETUP=.TRUE.)
      CALL Write_array (rcmdir//'rcmitrack_inp',irdw,label, ARRAY_1D=itrack,SETUP=.TRUE.)
      CALL Write_array (rcmdir//'rcmetab_inp',irdw, label, ARRAY_1D=etab, SETUP=.TRUE.)
      CALL Write_array (rcmdir//'rcmmpoint_inp',irdw, label, ARRAY_1D=mpoint,SETUP=.TRUE.)
      CALL Write_array (rcmdir//'rcmnpoint_inp',irdw, label, ARRAY_1D=npoint,SETUP=.TRUE.)
!
      RETURN
      END SUBROUTINE Plasma_rcm_setup

      subroutine print_date_time(LUN)
! prints date and time, 6/08, frt (from rcm)      
      USE Rcm_mod_subs, ONLY : iprec
      IMPLICIT NONE
      CHARACTER(LEN=8) :: real_date
      CHARACTER(LEN=10) ::real_time
      INTEGER(iprec), INTENT(IN) :: LUN

      CALL Date_and_time (real_date, real_time)
      WRITE (LUN,'(A11,A4,A1,A2,A1,A2, A8,A2,A1,A2,A1,A2)') &
              '  Today is ', real_date(1:4), '/', &
                             real_date(5:6), '/', &
                             real_date(7:8), & 
                  ' Time: ', real_time(1:2), ':', &
                             real_time(3:4), ':', &
                             real_time(5:6)
      return
      end subroutine print_date_time


