#ifndef GEOPACK_TEST_H
#define GEOPACK_TEST_H

#include <cxxtest/TestSuite.h>

// 'real' is defined in "OvertureTypes.h" as either a flot or a
// double...  MIX requires OV_USE_DOUBLE is enabled...  
//
// Cannot include "Overture.h" due to STD C++ build errors with
// CxxTest ..  so hard-code real = 8-byte real = double
typedef double real;

#ifdef LC_UNSC
#define RECALC recalc_
#define SMGSW_08 smgsw_08_
#define GEOGSW_08 geogsw_08_
#elif LC_NOUNSC
#define RECALC recalc
#define SMGSW_08 smgsw_08
#define GEOGSW_08 geogsw_08
#else
#error "Must define one of (LC_UNSC,LC_NOUNSC) for proper FORTRAN name mangling"
#endif

/**\name GEOPACK functions*/
extern "C" void RECALC(int *, int *, int *, int *, int *);
extern "C" void SMGSW_08 (real *, real *, real *, real *, real *, real *, int *);
extern "C" void GEOGSW_08(real *, real *, real *, real *, real *, real *, int *);

class GeopackTest : public CxxTest::TestSuite
{
public:

  int iSMtoGSM, iGSMtoSM, iGEOtoGSM, iGSMtoGEO;

  void setUp(){
    iSMtoGSM = 1;
    iGSMtoSM = -1;
    iGEOtoGSM = 1;
    iGSMtoGEO = -1;
    
    int year = 2002;
    int doy  = 80;
    int hour = 0;
    int min  = 0;
    int sec  = 0;   
    
    // YOU ABSOLUTELY MUST (!!!) call "RECALC" before calling SMGSW_08 &
    // GEOGSW_08 in order to calculate/initialize fit parameters
    RECALC(&year, &doy, &hour, &min, &sec);
  }
  
  void testPrintWarning( void )
  {
    TS_WARN("If geopack_test errors, make sure 'GEOPACK_FFLAGS' sets 8-byte reals.");
    TS_WARN("TODO: Write this warning as a try{...} catch(...){...} exception.");
  }

  void testGeoToGSM( void )
  {
    real xgeo,ygeo,zgeo;
    real xgsm,ygsm,zgsm;
    real xsm, ysm, zsm;

    xgeo = -0.0436194;
    ygeo = -5.34183e-18;
    zgeo = -0.999048;

    // Rotate to SM coordinates
    GEOGSW_08(&xgeo,&ygeo,&zgeo, &xgsm,&ygsm,&zgsm, &iGEOtoGSM);

    TS_ASSERT_DELTA(xgsm, 0.0421836, 1e-6);
    TS_ASSERT_DELTA(ygsm, 0.171217, 1e-6);
    TS_ASSERT_DELTA(zgsm, -0.98433, 1e-6);
  }

  void testGSMToSM( void )
  {
    real xgeo,ygeo,zgeo;
    real xgsm,ygsm,zgsm;
    real xsm, ysm, zsm;

    xgeo = -0.0436194;
    ygeo = -5.34183e-18;
    zgeo = -0.999048;

    // Rotate to SM coordinates
    GEOGSW_08(&xgeo,&ygeo,&zgeo, &xgsm,&ygsm,&zgsm, &iGEOtoGSM);
    SMGSW_08(&xsm,&ysm,&zsm, &xgsm,&ygsm,&zgsm, &iGSMtoSM);    

    TS_ASSERT_DELTA(xsm, -0.00674672, 1e-6);
    TS_ASSERT_DELTA(ysm,  0.171217, 1e-6);
    TS_ASSERT_DELTA(zsm, -0.98521, 1e-6);
  }

  void testConvertToLatLon( void )
  {
    real xgeo,ygeo,zgeo;
    real xgsm,ygsm,zgsm;
    real xsm, ysm, zsm;
    real lat, lon;

    xgeo = -0.0436194;
    ygeo = -5.34183e-18;
    zgeo = -0.999048;

    // Rotate to SM coordinates
    GEOGSW_08(&xgeo,&ygeo,&zgeo, &xgsm,&ygsm,&zgsm, &iGEOtoGSM);
    SMGSW_08(&xsm,&ysm,&zsm, &xgsm,&ygsm,&zgsm, &iGSMtoSM);
    
    // Convert to lat/lon
    lat = asin(zsm); // Latitude (-pi/2,pi2) as in TING grid definition
    lon = atan2(ysm,xsm);    

    TS_ASSERT_DELTA(lat, -1.3986, 1e-5);
    TS_ASSERT_DELTA(lon,  1.61018, 1e-5);
  }
};

#endif /* GEOPACK_TEST_H */
