;+------------------------------------------------------------------------
; This package of IDL functions facilitates reading data and metadata from
; Common Data Format (CDF) files.  While CDF provides the all the benefits
; of a portable, self-documenting scientific data format, reading them is
; not always a simple matter.  To make it simple, I have created this IDL
; package so that all of the data and metadata from multiple variables can 
; be read from multiple CDF files ... in one single, simple command.  The 
; function is called 'READ_MYCDF' and it returns an anonymous structure of 
; the form:
;
;       structure_name.variable_name.attribute_name.attribute_value
;
; From this structure, all data and metadata for the requested variables
; is easily accessed.
;
; AUTHOR:
;       Richard Burley, NASA/GSFC/Code 632.0, Feb 13, 1996
;       burley@nssdca.gsfc.nasa.gov    (301)286-2864
; 
; NOTES:
;
; Two additional 'attributes' will be included in the sub-structure for 
; each variable.  The first is the 'VARNAME' field.  Because IDL structure
; tags are always uppercase, and because CDF variable names are case sen-
; sitive, a case sensitive copy of the variable name is created.  The second
; 'attribute' to be added is the 'DAT' field.  The 'DAT' field will contain
; the actual data values read from the CDF for the the variable.  If more
; than one CDF file is read, the data will be concatenated to the data from
; any previously read CDF's.
;
; This package will look for and utilize certain special attributes required
; by the International Solar Terrestrial Physics Key Parameters Generation
; Software Standards and Guidelines.  The existance of these attributes is
; not required for the operation of this software, but will enhance its
; usefullness, primarily by reading variables that will be needed for proper
; utilization of the data, even though you may not have asked for them 
; explicitly.
;
; This package was tested under IDL version 4.0.1b.  This package was tested
; on CDF's up to version 2.5 and on both r-variables and z-variables.
;
; 1996, NASA/Goddard Space Flight Center
; This software may be used, copied, or redistributed as long as it is not
; sold and this copyright notice is reproduced on each copy made.  This
; routine is provided as is without any express or implied warranties
; whatsoever.
;-------------------------------------------------------------------------



;+------------------------------------------------------------------------
; NAME: BREAK_MYSTRING
; PURPOSE: 
;       Convert a string into a string array given a delimiting character 
; CALLING SEQUENCE:
;       out = break_mystring(instring)
; INPUTS:
;       instring = input text string
; KEYWORD PARAMETERS:
;       delimiter = character to parse by.  Default = ' '
; OUTPUTS:
;       out = string array
; AUTHOR:
;       Richard Burley, NASA/GSFC/Code 632.0, Feb 13, 1996
;       burley@nssdca.gsfc.nasa.gov    (301)286-2864
; MODIFICATION HISTORY:
;-------------------------------------------------------------------------
FUNCTION break_mySTRING, s, DELIMITER=delimiter
; Validate the input parameters
s_size=size(s) & n_size=n_elements(s_size)
if (s_size(n_size - 2) ne 7) then begin
  print,'ERROR>Argument to break_mySTRING must be of type string'
  return,-1
endif
if s eq '' then return, [ '' ]
if n_Elements(delimiter) eq 0 then delimiter = ''
; dissect the string
byte_delim = Byte( delimiter ) ; convert string to byte delimiter
result = Where( Byte(s) eq byte_delim(0), count ) ; count occurences
result = StrArr( count+1 ) & pos = -1
if (count gt 0) then begin
  for i=0, count-1 do begin
    oldpos = pos + 1
    pos = StrPos(s, delimiter, oldpos)
    result(i) = StrMid(s, oldpos, pos-oldpos)
  endfor
endif
pos = pos + 1
result(count) = StrMid( s, pos, StrLen(s) - pos )
return, result
end


;+------------------------------------------------------------------------
; NAME: DECODE_CDFEPOCH
; PURPOSE: 
;	Create a yyyy/mm/dd hh:mm:ss string given a CDF Epoch time
; CALLING SEQUENCE:
;       out = decode_cdfepoch(e)
; INPUTS:
;       e = CDF Epoch timetag (i.e. DOUBLE, millisecs from 0 A.D.)
; KEYWORD PARAMETERS:
; OUTPUTS: out = string in the format 'yyyy/mm/dd hh:mm:ss'
; AUTHOR:
;       Richard Burley, NASA/GSFC/Code 632.0, Feb 13, 1996
;       burley@nssdca.gsfc.nasa.gov    (301)286-2864
; MODIFICATION HISTORY:
;-------------------------------------------------------------------------
FUNCTION decode_CDFEPOCH, etime
; Create a yyyy/mm/dd hh:mm:ss string given a CDF Epoch time
CDF_EPOCH,etime,Yr,Mo,Day,Hr,Min,Sec,Mil,/BREAKDOWN_EPOCH
estr1 =         strtrim(string(Yr ,FORMAT='(I4.4)'),2) + '/'
estr1 = estr1 + strtrim(string(Mo ,FORMAT='(I2.2)'),2) + '/'
estr1 = estr1 + strtrim(string(Day,FORMAT='(I2.2)'),2) + ' '
estr1 = estr1 + strtrim(string(Hr ,FORMAT='(I2.2)'),2) + ':'
estr1 = estr1 + strtrim(string(Min,FORMAT='(I2.2)'),2) + ':'
estr1 = estr1 + strtrim(string(Sec,FORMAT='(I2.2)'),2)
return,estr1
end


;+------------------------------------------------------------------------
; NAME: ENCODE_CDFEPOCH
; PURPOSE: 
;	Convert a 'yyyy/mm/dd hh:mm:ss' string into CDF epoch time
; CALLING SEQUENCE:
;	e = encode_cdfepoch(instring)
; INPUTS:
;       instring = string in the form: 'yyyy/mm/dd hh:mm:ss'
; KEYWORD PARAMETERS:
; OUTPUTS:
;       e = CDF epoch timetag (i.e. DOUBLE, millisecs since 0 A.D.)
; AUTHOR:
;       Richard Burley, NASA/GSFC/Code 632.0, Feb 13, 1996
;       burley@nssdca.gsfc.nasa.gov    (301)286-2864
; MODIFICATION HISTORY:
;-------------------------------------------------------------------------
FUNCTION encode_CDFEPOCH, estr
; Convert the yyyy/mm/dd hh:mm:ss string into a CDF Epoch time
estr = strtrim(estr,2) ; trim any leading and trailing whitespace
; Define punctuation marks to be searched for
Punctuation = [44B,45B,46B,47B,58B,59B] ; (i.e. ",-./:;")
a=0 & reads,estr,a,FORMAT='(I)' ; read the year field from the string
s = strpos(estr,'/',0) & estr = strmid(estr,s+1,20) ; reposition after '/'
b=0 & reads,estr,b,FORMAT='(I)' ; read the month field form the string
s = strpos(estr,'/',0) & estr = strmid(estr,s+1,20) ; reposition after '/'
c=0 & reads,estr,c,FORMAT='(I)' ; read the day field from the string
s = strpos(estr,' ',0) & estr = strmid(estr,s+1,20) ; reposition after ' '
d=0 & reads,estr,d,FORMAT='(I)' ; read the hour field from the string
s = strpos(estr,':',0) & estr = strmid(estr,s+1,20) ; reposition after ':'
e=0 & reads,estr,e,FORMAT='(I)' ; read the minute field from the string
s = strpos(estr,':',0) & estr = strmid(estr,s+1,20) ; reposition after ':'
f=0 & reads,estr,f,FORMAT='(I)' ; read the sec field from the string
; Perform TBD validation
CDF_EPOCH,etime,a,b,c,d,e,f,/COMPUTE_EPOCH
return,etime
end


;+------------------------------------------------------------------------
; NAME: AMI_ISTPPTR
; PURPOSE:
;       Return true(1) or false(0) depending on whether or not the
;       given attribute name qualifies as an ISTP pointer-class attribute.
; CALLING SEQUENCE:
;	out = amI_ISTPptr(attribute_name)
; INPUTS:
;	attribute_name = name of a CDF attribute as a string
; KEYWORD PARAMETERS:
; OUTPUTS:
;       True(1) or False(0)
; AUTHOR:
;       Richard Burley, NASA/GSFC/Code 632.0, Feb 13, 1996
;       burley@nssdca.gsfc.nasa.gov    (301)286-2864
; MODIFICATION HISTORY:
;-------------------------------------------------------------------------
FUNCTION amI_ISTPptr, aname
if (aname eq 'UNIT_PTR')        then return,1
if (aname eq 'FORMAT_PTR')      then return,1
if (aname eq 'DELTA_PLUS_VAR')  then return,1
if (aname eq 'DELTA_MINUS_VAR') then return,1
len = strlen(aname) & pos = strpos(aname,'LABL_PTR_')
if ((len gt 9)AND(pos eq 0)) then begin ; label pointer found
  ON_IOERROR,escape ; return false if non-digit found
  for j=0,(len-10) do begin ; check one character at a time
    r = strmid(aname,(9+j),1) & READS,r,v,FORMAT='(I1)'
  endfor
  return,1 ; remaining characters in label pointer are valid
endif
pos = strpos(aname,'OFFSET_')
if ((len gt 7)AND(pos eq 0)) then begin ; label pointer found
  ON_IOERROR,escape ; return false if non-digit found
  for j=0,(len-8) do begin ; check one character at a time
    r = strmid(aname,(7+j),1) & READS,r,v,FORMAT='(I1)'
  endfor
  return,1 ; remaining characters in offset pointer are valid
endif
escape: return,0
end

;+------------------------------------------------------------------------
; NAME: PARSE_DISPLAY_TYPE
; PURPOSE: 
;	Parse and examine the input string.  It should be the value of the
;	CDF attribute 'DISPLAY_TYPE'.  Return an array of variable names
;       that it 'points' to.
; CALLING SEQUENCE:
;	out = parse_display_type(instring)
; INPUTS:
;       instring = string, value of a CDF attribute called 'DISPLAY_TYPE'
; KEYWORD PARAMETERS:
; OUTPUTS:
;       out = string array, names of other variables required for display
; NOTES: This routine expects to find 'DISPLAY_TYPE' values looking like:
;        PLOT_TYPE>x=vname,y=vname ...
;        PLOT_TYPE>y=vname,z=vname(*,1),z=vname(*,2) ...
; AUTHOR:
;       Richard Burley, NASA/GSFC/Code 632.0, Feb 13, 1996
;       burley@nssdca.gsfc.nasa.gov    (301)286-2864
; MODIFICATION HISTORY:
;-------------------------------------------------------------------------
FUNCTION parse_DISPLAY_TYPE, instring

num_vnames = 0L & spos = 0L & i=0L ; initialize
a = break_mySTRING(instring,DELIMITER='>') ; break string into components
if n_elements(a) le 1 then return,-1 ; no '>' following plot type
if a(1) eq '' then return,-1 ; no info past '>' to parse
b = break_mystring(a(1),delimiter=',') ; Parse the string into substrings
for i=0,n_elements(b)-1 do begin ; examine each substring
  s = strpos(b(i),'=') & p = strpos(b(i),'(') ; find '=' and '(' signs
  if s ne -1 then begin ; probably a properly formated DISPLAY_TYPE vattr
    s = s + 1 ; point to first character past the '=' sign
    if p ne -1 then vname = strmid(b(i),s,(p-s)) $ ; extract vname
    else vname = strmid(b(i),s,(strlen(b(i))-s))   ; extract vname
    if num_vnames eq 0 then begin
      variable_names = vname & num_vnames = num_vnames + 1
    endif else begin
      index = where(variable_names eq vname,wc)
      if wc eq 0 then begin
        variable_names = [variable_names,vname] ; add to existing list
        num_vnames = num_vnames + 1 ; keep count of number of names
      endif
    endelse
  endif
endfor
if num_vnames eq 0 then return,-1 $
else return,variable_names
end


;+------------------------------------------------------------------------
; NAME: follow_myDEPENDS
; PURPOSE: 
;	Search the metadata anonymous structure for ISTP 'DEPEND' attributes.
;       If and when found, add the variable name that it points to to the
;       vnames array if it is not already present, and increase the size
;       of the dhids and mhids arrays.
; CALLING SEQUENCE:
;       follow_myDEPENDS, metadata, vnames, dhids, mhids
; INPUTS:
;       metadata = anonymous structure holding attribute values
;       vnames   = string array of the names of variables already processed
;       vvarys   = string array of the record variance for each variable
;       dhids    = array of data handle id's
;       mhids    = array of metadata handle id's
; KEYWORD PARAMETERS:
; OUTPUTS:
;       dhids    = array of data handle id's
;       mhids    = array of metadata handle id's
; AUTHOR:
;       Richard Burley, NASA/GSFC/Code 632.0, Feb 13, 1996
;       burley@nssdca.gsfc.nasa.gov    (301)286-2864
; MODIFICATION HISTORY:
;-------------------------------------------------------------------------
PRO follow_myDEPENDS, metadata, vnames, vvarys, ctypes, dhids, mhids
tnames = tag_names(metadata)
for i=0,n_elements(tnames)-1 do begin
  ; Determine if the current tagname is a legal ISTP-style DISPLAY_TYPE vattr
  if (tnames(i) eq 'DISPLAY_TYPE') then begin
    dvnames = parse_DISPLAY_TYPE(metadata.(i)) & dvs = size(dvnames)
    if (dvs(n_elements(dvs)-2) eq 7) then begin ; variable names found
      for j=0,n_elements(dvnames)-1 do begin
        a = where(vnames eq dvnames(j),count) ; search vnames array
        if ((dvnames(j) ne '')AND(count eq 0)) then begin
          ; add the display variable name all array parameters
          n = n_elements(vnames)
          newn = strarr(n+1) & newd = lonarr(n+1) & newm = lonarr(n+1)
          newv = strarr(n+1) & newr = lonarr(n+1) & newc = strarr(n+1)
          newn(0:(n-1)) = vnames(0:(n-1))
          newv(0:(n-1)) = vvarys(0:(n-1))
          newc(0:(n-1)) = ctypes(0:(n-1))
          newd(0:(n-1)) = dhids(0:(n-1))
          newm(0:(n-1)) = mhids(0:(n-1))
          newn(n)=dvnames(j) & newv(n)='' & newc(n)='' & newd(n)=0 & newm(n)=0
          vnames=newn & vvarys=newv & ctypes=newc & dhids=newd & mhids=newm
        endif
      endfor
    endif
  endif

  ; Determine if the current tagname is a legal ISTP-style depend attribute
  len = strlen(tnames(i)) & pos = strpos(tnames(i),'DEPEND_')
  if ((len gt 7)AND(pos eq 0)) then begin ; DEPEND found, check remainder
    ON_IOERROR, escape ; return false if non-digit found
    for j=0,(len-8) do begin ; check one character at a time
      r = strmid(tnames(i),(7+j),1) & READS,r,v,FORMAT='(I1)'
    endfor
    dvname = metadata.(i) ; depend attribute FOUND
    a = where(vnames eq dvname,count)    ; search vnames array
    if ((dvname ne '')AND(count eq 0)) then begin
      ; add the depend variable name to all array parameters
      n = n_elements(vnames)
      newn = strarr(n+1) & newd = lonarr(n+1) & newm = lonarr(n+1)
      newv = strarr(n+1) & newr = lonarr(n+1) & newc = strarr(n+1)
      newn(0:(n-1)) = vnames(0:(n-1))
      newv(0:(n-1)) = vvarys(0:(n-1))
      newc(0:(n-1)) = ctypes(0:(n-1))
      newd(0:(n-1)) = dhids(0:(n-1))
      newm(0:(n-1)) = mhids(0:(n-1))
      newn(n) = dvname & newv(n)='' & newc(n)='' & newd(n)=0 & newm(n)=0
      vnames = newn & vvarys=newv & ctypes=newc & dhids=newd & mhids=newm
    endif
  endif
  escape: ; Current tag name is not a depend attribute
endfor
end


;-------------------------------------------------------------------------
; NAME: CDFTYPE_T0_MYIDLTYPE
; PURPOSE: 
;	Convert from CDF type number to IDL type number
; CALLING SEQUENCE:
;       out = CDFtype_to_myIDLtype(in)
; INPUTS:
;       in = integer, CDF type number
; KEYWORD PARAMETERS:
; OUTPUTS:
;       out = integer, IDL type number
; AUTHOR:
;       Richard Burley, NASA/GSFC/Code 632.0, Feb 13, 1996
;       burley@nssdca.gsfc.nasa.gov    (301)286-2864
; MODIFICATION HISTORY:
;-------------------------------------------------------------------------
FUNCTION CDFtype_to_myIDLtype,cdftype
case cdftype of
  22L : idltype = 5 ; CDF_REAL8
  45L : idltype = 5 ; CDF_DOUBLE
  31L : idltype = 5 ; CDF_EPOCH
  21L : idltype = 4 ; CDF_REAL4
  44L : idltype = 4 ; CDF_FLOAT
  4L  : idltype = 3 ; CDF_INT4
  14L : idltype = 3 ; CDF_UINT4
  2L  : idltype = 2 ; CDF_INT2
  12L : idltype = 2 ; CDF_UINT2
  51L : idltype = 7 ; CDF_CHAR
  52L : idltype = 1 ; CDF_UCHAR
  1L  : idltype = 1 ; CDF_INT1
  11L : idltype = 1 ; CDF_UINT1
  41L : idltype = 1 ; CDF_BYTE
  else: idltype = 0 ; undefined
endcase
return,idltype
end


;+------------------------------------------------------------------------
; NAME: APPEND_MYDATA
; PURPOSE: 
; 	Append the 'new' data to the 'old' data using array concatenation.
; CALLING SEQUENCE:
;       out = append_mydata(new,old)
; INPUTS:
;       new = data to be appended to the old data
;       old = older data that new data is to be appended to
; KEYWORD PARAMETERS:
; OUTPUTS:
;       out = product of concatenating the old and new data arrays
; NOTES:
; 	Special case check: if old data was from either a skeleton CDF or from
; 	a CDF with only a single record, then the last dimension was dropped 
;	during the process of saving/retrieving the data from a handle.  
;	Must compare the dimensionality of the new and old data to determine 
;	if this drop has occured, and if so, reform the old data to include 
;       the extra dimension so that the data can be appended.
; AUTHOR:
;       Richard Burley, NASA/GSFC/Code 632.0, Feb 13, 1996
;       burley@nssdca.gsfc.nasa.gov    (301)286-2864
; MODIFICATION HISTORY:
;-------------------------------------------------------------------------
FUNCTION append_myDATA, new, old
a = size(new) & b = size(old) & data=0L
if (a(0) ne b(0)) then begin ; dimension mismatch - reform old data
  print,'WARNING> Dimension mismatch detected, attempting to reform...' ;DEBUG
  case b(0) of
    0    : ; no special actions needed
    1    : old = reform(old,b(1),1)
    2    : old = reform(old,b(1),b(2),1)
    3    : old = reform(old,b(1),b(2),b(3),1)
    else : print,'ERROR>Cannot reform single-record variable with > 3 dims'
  endcase
  b = size(old) ; recompute b after it has been reformed
endif

; append the new data to the old data
case a(0) of
  0:begin
    data = [old,new]
    end
  1:begin
    data = [old,new]
    end
  2:begin
    data = make_array(a(1),a(2)+b(2),type=a(3))
    data(*,0:b(2)-1)=old(*,*) & data(*,b(2):*)=new(*,*)
    end
  3:begin
    data = make_array(a(1),a(2),a(3)+b(3),type=a(4))
    data(*,*,0:b(3)-1)=old(*,*,*) & data(*,*,b(3):*)=new(*,*,*)
    end
  else: print,'ERROR>Cannot append arrays with > 3 dimensions yet'
endcase
new = 0L & old = 0L ; free up unneeded memory
return,data
end


;+------------------------------------------------------------------------
; NAME: READ_MYVARIABLE
; PURPOSE: 
;	Return the data for the requested variable.
; CALLING SEQUENCE:
;       out = read_myvariable(vname, CDFid, vary, dtype, recs)
; INPUTS:
;       vname = string, name of variable to be read from the CDF
;       CDFid = integer, id or already opened CDF file.
; KEYWORD PARAMETERS:
; OUTPUTS:
;       out = all data from the CDF for the variable being read
;       vary = True(1) or False(0) is variable record-varying
;       dtype= string, CDF data type
;       recs = integer, number of data records
; AUTHOR:
;       Richard Burley, NASA/GSFC/Code 632.0, Feb 13, 1996
;       burley@nssdca.gsfc.nasa.gov    (301)286-2864
; MODIFICATION HISTORY:
;-------------------------------------------------------------------------
FUNCTION read_myVARIABLE, vname, CDFid,vary,dtype,recs

; Get needed information about the cdf and variable
cinfo = cdf_inquire(CDFid) ; inquire about the cdf
vinfo = cdf_varinq(CDFid,vname) ; inquire about the variable
cdf_control,CDFid,VAR=vname,GET_VAR_INFO=vinfo2 ; inquire more about the var
vnum = cdf_varnum(CDFid,vname,zflag) ; determine if r-variable or z-variable
rcount = cinfo.maxrec+1 & if (vinfo.RECVAR eq 'NOVARY') then rcount=1L
vary = vinfo.RECVAR & dtype = vinfo.DATATYPE & recs = vinfo2.MAXREC

; Read the CDF for the data for the requested variable
if (zflag eq 1) then begin ; read the z-variable
  cdf_control,CDFid,VAR=vname,GET_VAR_INFO=zvinfo
  if zvinfo.MAXREC eq -1 then begin
    print,'WARNING:',vname,' has ZERO records!' & return,0
  endif else cdf_varget,CDFid,vname,dat,REC_COUNT=zvinfo.MAXREC+1
endif else begin ; read the r-variable
  dims = total(vinfo.dimvar) & dimc = vinfo.dimvar * cinfo.dim
  dimw = where(dimc eq 0,dcnt) & if (dcnt ne 0) then dimc(dimw)=1
  CDF_varget,CDFid,vname,dat,COUNT=dimc,REC_COUNT=rcount
  ds = size(dat) ; get size info to determine if dat is scalar or not
  if (ds(0) ne 0) then dat = reform(dat) ; eliminate extraneous dims
endelse

; Correct for fact that IDL retrieves character data as bytes
if vinfo.DATATYPE eq 'CDF_CHAR' then begin ; IDL retrieves as bytes
  ds = size(dat) ; get dimensions of dat for single char special case
  if (ds(0) gt 1) then begin 
    dat = string(dat) & dat = reform(dat) ; eliminate extraneous dims
  endif else begin ; process each element of array
    d2=strarr(ds(1)) & for i=0,ds(1)-1 do d2(i)=string(dat(i))
    dat = d2
  endelse
endif

; If this variable is a record-varying variable, but this CDF only happens
; to have one record, then we must add the extra dimension onto the end
; for proper appending to take place when other CDF's are read
if ((vinfo.RECVAR eq 'VARY')AND(rcount eq 1L)) then begin
  ; print,'WARNING>Reforming single-record variable' ;DEBUG
  ds = size(dat) ; get dimensions of dat
  case ds(0) of
    0    : rcount = 1L ; do nothing
    1    : dat = reform(dat,ds(1),1)
    2    : dat = reform(dat,ds(1),ds(2),1)
    3    : dat = reform(dat,ds(1),ds(2),ds(3),1)
    else : print,'ERROR>Cannot reform single-record variable with > 3 dims'
  endcase
endif
; Return the data read from the CDF
return,dat
end


;+------------------------------------------------------------------------
; NAME: READ_MYATTRIBUTE
; PURPOSE: 
;	Return the value of the requested attribute for the requested variable.
; CALLING SEQUENCE:
;       out = read_myattribute(vname,anum,CDFid)
; INPUTS:
;       vname = string, name of variable whose attribute is being read
;       anum = integer, number of attribute being read
;       CDFid = integer, id of already opened CDF file.
; KEYWORD PARAMETERS:
; OUTPUTS:
;       out = anonymous structure holding both the name of the attribute
;             and the value of the attribute
; AUTHOR:
;       Richard Burley, NASA/GSFC/Code 632.0, Feb 13, 1996
;       burley@nssdca.gsfc.nasa.gov    (301)286-2864
; MODIFICATION HISTORY:
;-------------------------------------------------------------------------
FUNCTION read_myATTRIBUTE, vname, anum, CDFid

cdf_attinq,CDFid,anum,aname,ascope,maxe,maxze ; inquire about the attribute
aname = strtrim(aname,2) ; trim any possible leading or trailing blanks
attval='' & astruct=create_struct(aname,attval) ; initialize anonymous structure
CATCH,error_status & if error_status ne 0 then return,astruct
if (ascope eq 'GLOBAL_SCOPE')OR(ascope eq 'GLOBAL_SCOPE_ASSUMED') then begin
  cdf_attget,CDFid,anum,0,aval & attval = aval ; get the global attribute
  for entry=1,maxe do begin ; get remaining entrys if any
    if (entry eq 1) then begin ; create array to hold the data
      asize = size(aval) & dtype = asize(n_elements(asize)-2)
      attval = make_array((maxe+1),TYPE=dtype) & attval(0) = aval
    endif
    cdf_attget,CDFid,anum,entry,aval
    attval(entry) = aval
  endfor
  asize = size(attval) & nea = n_elements(asize)
  if (asize(nea-2) eq 7) then begin
    if asize(0) eq 0 then attval = strtrim(attval,2) $
    else for i=0,asize(1)-1 do attval(i) = strtrim(attval(i),2)
  endif
  astruct = create_struct(aname,attval)
endif else begin ; 'VARIABLE_SCOPE' or 'VARIABLE_SCOPE_ASSUMED'
  cdf_attget,CDFid,anum,vname,aval & attval = aval ; read variable attribute
  if (amI_ISTPptr(aname) eq 1) then begin ; check for pointer-type attribute
    attval = read_myVARIABLE(attval,CDFid,vary,ctype,recs)
  endif
  asize = size(attval) & nea = n_elements(asize)
  if (asize(nea-2) eq 7) then begin
    if asize(0) eq 0 then attval = strtrim(attval,2) $
    else for i=0,asize(1)-1 do attval(i) = strtrim(attval(i),2)
  endif
  astruct = create_struct(aname,attval)
endelse
return,astruct
end


;+------------------------------------------------------------------------
; NAME: READ_MYMETADATA
; PURPOSE: 
;	To read all of the attribute values for the requested variable, and
;       to return this information as an anonymous structure.
; CALLING SEQUENCE:
;       metadata = read_mymetadata(vname,CDFid)
; INPUTS:
;       vname = string, name of variable whose metadata is being read
;       CDFid = integer, id of already opened CDF file
; KEYWORD PARAMETERS:
; OUTPUTS:
;       metadata = anonymous structure whose tags are the attribute names
;                  and whose fields are the corresponding attribute values.
; AUTHOR:
;       Richard Burley, NASA/GSFC/Code 632.0, Feb 13, 1996
;       burley@nssdca.gsfc.nasa.gov    (301)286-2864
; MODIFICATION HISTORY:
;-------------------------------------------------------------------------
FUNCTION read_myMETADATA, vname, CDFid

cinfo = cdf_inquire(CDFid) ; inquire about the cdf to get #attributes
; Create initial data structure to hold all of the metadata information
METADATA = create_struct('varname',vname)
; Extract all metadata information for the all attributes
for anum=0,cinfo.natts-1 do begin
  astruct = 0 ; initialize astruct
  ; Get the name and value of the next attribute for vname
  astruct = read_myATTRIBUTE(vname,anum,CDFid)
  METADATA = create_struct(METADATA,astruct)
endfor ; for each attribute
return,METADATA
end


;+------------------------------------------------------------------------
; NAME: GET_ALLVARNAMES
; PURPOSE: 
; 	To return a string array containing the names of all of the
;	variables in the given CDF file.
; CALLING SEQUENCE:
;       vnames = get_allvarnames()
; INPUTS:
; KEYWORD PARAMETERS:
;	CNAME = string, name of a CDF file to be opened and read
;	CDFid = integer, id of an already opened CDF file
;       VAR_TYPE = string, only return the names for variables who have an
;                  attribute called 'VAR_TYPE' and whose value matches the
;                  value given by this keyword.  (ex. VAR_TYPE='data')
; OUTPUTS:
;       vnames = string array of variable names
; AUTHOR:
;       Richard Burley, NASA/GSFC/Code 632.0, Feb 13, 1996
;       burley@nssdca.gsfc.nasa.gov    (301)286-2864
; MODIFICATION HISTORY:
;-------------------------------------------------------------------------
FUNCTION get_allvarnames, CNAME=CNAME, CDFid=CDFid, VAR_TYPE=VAR_TYPE

; validate keyword combination and open cdf if needed
if keyword_set(CNAME) AND keyword_set(CDFid) then return,0 ; invalid
if keyword_set(CNAME) then CDFindex = CDF_OPEN(CNAME) ; open the cdf
if keyword_set(CDFid) then CDFindex = CDFid ; save the cdf file number

; determine the number of variables 
cinfo = CDF_INQUIRE(CDFindex) ; inquire about number of variables
numvars = cinfo.nvars + cinfo.nzvars & vnames=strarr(numvars)

; Get the name of every variable
for i=0,numvars-1 do begin
  zflag=0 & CATCH, Error_status ; establish error handler for zvars
  if Error_status ne 0 then begin
    if zflag ne 1 then zflag = 1 $
    else begin print,!ERR_STRING & stop & endelse
  endif
  vinfo = CDF_VARINQ(CDFindex,i,ZVARIABLE=zflag) ; inquire about ith var
  if keyword_set(VAR_TYPE) then begin ; only get VAR_TYPE='data'
    metadata = read_myMETADATA(vinfo.name,CDFindex)
    tags = tag_names(metadata) & temp = where(tags eq 'VAR_TYPE',count)
    if count ne 0 then begin
      if metadata.VAR_TYPE eq VAR_TYPE then vnames(i) = vinfo.name
    endif
  endif else begin ; do not utilize metadata for variable determination
    if vinfo.recvar eq 'VARY' then vnames(i) = vinfo.name
  endelse
endfor
if keyword_set(CNAME) then CDF_CLOSE,CDFindex
; strip blank vnames from the array
v = where(vnames ne "",vc)
if vc gt 0 then begin
  temp = strarr(vc) & temp = vnames(v) & vnames = temp
endif
return,vnames
end


;+------------------------------------------------------------------------
; NAME: READ_MYCDF
; PURPOSE: 
;	Read all data and metadata for given variables, from given CDF
;       files, and return all information in a single anonymous structure
;       of the form: 
;          structure_name.variable_name.attribute_name.attribute_value
;
; CALLING SEQUENCE:
;       out = read_mycdf(vnames,cnames)
; INPUTS:
;       vnames = string, array of variable names or a single string of
;                names separated by a comma.  (ex. 'Epoch,Magfld,Bmax')
;       cnames = string, array of CDF filenames or a single string of
;                names separated by a comma.
; KEYWORD PARAMETERS:
;	ALL = If set, get data and metadata for ALL variables in the CDFs.
;       NODATASTRUCT = If set, instead of returning the data for each variable
;                   in the 'DAT' attribute field, create a 'HANDLE' field
;                   and set it to the handle id of a data handle which
;                   holds the data for each variable.
;       NOQUIET = If set, do NOT set the !QUIET system variable before
;                 reading the cdf file(s).
;       DEBUG = If set, print out some progress information during reading.
; OUTPUTS:
;       out = anonymous structure holding all data and metadata for the
;             requested variables. -1 is returned if an error occurs.
; AUTHOR:
;       Richard Burley, NASA/GSFC/Code 632.0, Feb 13, 1996
;       burley@nssdca.gsfc.nasa.gov    (301)286-2864
; MODIFICATION HISTORY:
;-------------------------------------------------------------------------

FUNCTION read_myCDF, vnames, cnames, ALL=ALL,NODATASTRUCT=NODATASTRUCT, $
                                     NOQUIET=NOQUIET,DEBUG=DEBUG

; establish exception handler to trap errors from all sources.
CATCH,error_status
if (error_status ne 0) then begin
  print,!ERR_string  ; output description of error
  print,'Error Index=',error_status
  return,-1 ; return failure flag
endif

; Validate cnames parameter, remove .cdf extensions if present
s = size(cnames) & ns = n_elements(s)
if (s(ns-2) eq 7) then begin
  if (s(0) eq 0) then cnames = break_mySTRING(cnames,DELIMITER=',')
  for i=0,n_elements(cnames)-1 do begin
    j=strpos(cnames(i),'.cdf') & if (j eq -1) then j=strpos(cnames(i),'.CDF')
    if (j ne -1) then cnames(i) = strmid(cnames(i),0,j)
  endfor
endif else begin
  print,'ERROR>CDF filenames must be given as strings.' & return,-1
endelse
if keyword_set(DEBUG) then print,'Number of CDFs to read=',n_elements(cnames)

; If the ALL keyword is set then get names of all data variables
if keyword_set(ALL) then vnames = get_allvarnames(CNAME=cnames(0))

; Validate vnames parameter.  May be a strarr or single string
s = size(vnames) & ns = n_elements(s)
if (s(ns-2) eq 7) then begin
  if (s(0) eq 0) then vnames = break_mySTRING(vnames,DELIMITER=',')
  dhids = lonarr(n_elements(vnames)) ; create array of handle ids for data
  mhids = lonarr(n_elements(vnames)) ; create array of handle ids for metadata
endif else begin
  print,'ERROR>variable names must be given as strings.' & return,-1
endelse
vvarys = strarr(n_elements(vnames)) ; initialize variable variance array
cdftyp = strarr(n_elements(vnames)) ; initialize cdftype array

quiet_flag = !quiet ; save current state of quiet flag
if not keyword_set(NOQUIET) then !quiet=1 ; turn off annoying cdf messages

; get the data and metadata for all variables from all cdfs
for cx=0,n_elements(cnames)-1 do begin
  ; Open the CDF and inquire about global information about file
  if keyword_set(DEBUG) then print,'Opening the file ',cnames(cx)
  CDFid = cdf_open(cnames(cx)) & cinfo = cdf_inquire(CDFid)

  ; Read all of the selected variables from the open CDF
  vx = 0 & REPEAT begin

    ; Read the data for the variable unless it is known to be non varying
    if (vvarys(vx) ne 'NOVARY') then begin
      if keyword_set(DEBUG) then print,'Reading data for ',vnames(vx)
      data = read_myVARIABLE(vnames(vx),CDFid,vary,dtype,recs) ; read the data
      vvarys(vx) = vary  ; save the record variance flag
      cdftyp(vx) = dtype ; save the cdf data type
    endif

    ; Process the metadata of the current variable
    if (cx eq 0) then begin ; first cdf in dataset
      vvarys(vx) = vary ; set value in variable variance array
      if keyword_set(DEBUG) then print,'Reading metadata for ',vnames(vx)
      metadata = read_myMETADATA(vnames(vx),CDFid) ; read variable metadata
      mhids(vx) = HANDLE_CREATE() & HANDLE_VALUE, mhids(vx), metadata, /SET
      ; Check metadata for ISTP depend attr's, modify other arrays accordingly
      follow_myDEPENDS, metadata, vnames, vvarys, cdftyp, dhids, mhids
    endif

    ; Process the data of the current variable
    if (recs gt -1) then begin ; current cdf is not a skeleton
      if (dhids(vx) eq 0) then begin ; create handle to hold data
        dhids(vx) = HANDLE_CREATE() & HANDLE_VALUE, dhids(vx), data, /SET
      endif else begin ; only append record varying data
        if (vvarys(vx) eq 'VARY') then begin ; append the data to previous data
          HANDLE_VALUE, dhids(vx), olddata    ; get data from previous cdf's
          data = append_myDATA(data,olddata)  ; append new data to old data
          HANDLE_VALUE, dhids(vx), data ,/SET ; save back into handle
        endif
      endelse
    endif

    vx = vx + 1 ; increment variable name index
  ENDREP UNTIL (vx eq n_elements(vnames))
  cdf_close,CDFid ; close the open cdf
endfor ; loop thru all cdfs

if keyword_set(DEBUG) then print,'Assembling Anonymous Structure'
; It is possible that some of the variable names may be padded with blanks
; This will likely cause problems later, so trim any blanks around vnames.
for i=0,n_elements(vnames)-1 do vnames(i) = strtrim(vnames(i),2)


; Retrieve the data and metadata from first handle, and append them
; together to create a data structure to be output from this function.
HANDLE_VALUE, mhids(0), metadata, /NO_COPY  & HANDLE_FREE,mhids(0)
if keyword_set(NODATASTRUCT) then begin
  ; Rather than place the actual data into the megastructure, create
  ; a data handle structure field and put the data handle id in it.
  mytype = create_struct('cdftype',cdftyp(0)) ; create .cdftype structure
  mydata = create_struct('handle',dhids(0))   ; create .handle structure
  mydata = create_struct(mytype,mydata)       ; append the structures
  mydata = create_struct(metadata,mydata)     ; append the metadata
  burley = create_struct(vnames(0),mydata)    ; create initial structure
endif else begin
  ; Place the actual data into the large data structure.  This requires
  ; moving data and can take a long time with large image data arrays.
  if (dhids(0) ne 0) then begin ; data exists, retrieve it
    HANDLE_VALUE, dhids(0), data, /NO_COPY    & HANDLE_FREE,dhids(0)
  endif else data = ''
  ds = size(data) & if (ds(0) ne 0) then data = reform(data) ; special case
  mytype = create_struct('cdftype',cdftyp(0)) ; create .cdftype structure
  mydata = create_struct('dat',data)          ; create .dat structure
  mydata = create_struct(mytype,mydata)       ; append the structures
  mydata = create_struct(metadata,mydata)     ; append the metadata
  burley = create_struct(vnames(0),mydata)    ; create initial structure
endelse

; If more than one variable is being processed, then retrieve the data
; and metadata from the handles, and append them into an anonymous struct
; and append these structures into a single anonymous struct for output.
for vx = 1,n_elements(vnames)-1 do begin ; retrieve and append
  HANDLE_VALUE, mhids(vx), metadata, /NO_COPY  & HANDLE_FREE,mhids(vx)
  if keyword_set(NODATASTRUCT) then begin
    ; Rather than place the actual data into the megastructure, create
    ; a data handle structure field and put the data handle id in it.
    mytype = create_struct('cdftype',cdftyp(vx)) ; create .cdftype structure
    mydata = create_struct('handle',dhids(vx))   ; create .handle structure
    mydata = create_struct(mytype,mydata)        ; append the structures
    mydata = create_struct(metadata,mydata)      ; append the metadata
    rick   = create_struct(vnames(vx),mydata)    ; create new structure
    burley = create_struct(burley,rick)          ; create initial structure
  endif else begin
    if (dhids(vx) ne 0) then begin ; data exists, retrieve it
      HANDLE_VALUE, dhids(vx), data, /NO_COPY    & HANDLE_FREE,dhids(vx)
    endif else data = ''
    ds = size(data) & if (ds(0) ne 0) then data = reform(data) ; special case
    mytype = create_struct('cdftype',cdftyp(vx)) ; create .cdftype structure
    mydata = create_struct('dat',data)           ; create .dat structure
    mydata = create_struct(mytype,mydata)        ; append the structures
    mydata = create_struct(metadata,mydata)      ; append the metadata
    rick   = create_struct(vnames(vx),mydata)    ; create new structure
    burley = create_struct(burley,rick)          ; append the structures
  endelse
endfor

!quiet = quiet_flag ; restore
; Return successfull
Return, burley
end

