import datetime
import getpass
import platform

import yaml

import Release
import share
import Models
import Environments

class Configuration(object):
    """
    Configuration Class deals with makeItSo.config files:
      1.  Write config file out
      2.  Read config file & instantiate makeItSo objects.

    FIXME: Overload __init__ so it like a function rather than a class
           e.g. (name, shared, models, env) = Configuration.init(filename)
    """
    
    def __init__(self):
        None

    ############################################################################

    def readConfigurationFile(self, filename):
        """
        Read coupled model configuration from file and return properly
        configured script generator objects
        
        Returns:
          coupled_name: Name of the coupled model (e.g. 'LFM-MIX')
          shared:  Shared Model object
          models:  A list of Model objects (e.g. [LFM, MIX, XJD]
          environment:  Environment object (e.g. SH, LSF or SGE)
        """

        yaml_file = open(filename)
        configDict = yaml.safe_load(yaml_file)
        yaml_file.close()
                
        (coupledName, shared, models) = self.__setModelConfiguration(configDict)
        environment = self.__setEnvironmentConfiguration(filename, configDict, coupledName, shared.is_Coupled_Via_InterComm(coupledName))

        return (coupledName, shared, models, environment)

    def __setModelConfiguration(self, configDict):
        """
        Parse configDict and prepare the objects necessary for
        makeItSo to provide meaningful output.
        """
        (coupledName, modelNames) = self.__getModelNames(configDict)
        (shared, models) = self.__instantiateModels(coupledName, modelNames)        
        self.__setupModel(shared, configDict[coupledName]["Models"])

        for model in models:
            self.__setupModel(model, configDict[coupledName]["Models"])
    
        return (coupledName, shared, models)

    def __getModelNames(self, configDict):
        """
        Returns the coupled model name (including the models that are
        coupled) as defined in configDict.

        Validates parts of configDict most likely to have fatal problems
        """
        # Validate that there's only one model type.
        if len(configDict.keys()) != 1:
            raise Exception, ("Error: configDict should define only one " +
                              "coupled model type, but it defines " +
                              str(len(configDict.keys())) + ": " +
                              str(configDict.keys())
                              )

        coupledName = configDict.keys()[0]

        # Validate parts of configDict that are most likely to have fatal errors.    
        if not configDict[coupledName].has_key("About"):
            raise Exception, ("Error: configDict must define an \"About\" " +
                              "section. Is the YAML config file valid?"
                              )
        if not configDict[coupledName].has_key("Models"):
            raise Exception, ("Error: configDict must define one or more " +
                              "\"Model\" for coupling type \"" + coupledName + "\"."
                              )

        # Make sure there are parameters defined for each model coupled:
        modelNames = []
        for model in configDict[coupledName]["Models"]:
            if (len(model.keys()) != 1):
                raise Exception, ("Error:  configDict should define only one "+
                                  "model, but appears to define " +
                                  str(len(model.keys())) + ":  " +
                                  str(model.keys())
                                  )
            modelNames.append( model.keys()[0] )

        return (coupledName, modelNames)


    def __instantiateModels(self, coupledName, modelNames):
        """
        Instantiate & return all Model objects components (modelNames)
        for the coupled model for (coupledName).

        Validates that all the models are present.

        Parameters:
          coupledName: Name of coupled model
          modelNames: list of model names that are the coulped
        
        Returns:
          shared model:
          List of Models:
        """

        # Validate that all models are defined for the coupled model:
        models_dict = Models.helpers.getModelCoupling()
        validated = False
        coupled_model_dict = None
        for coupled in models_dict:
            if coupled.keys()[0] == coupledName:
                # Hit!  Make sure each component is present.
                for modelName in coupled[coupledName]:
                    if (modelNames.count(modelName) != 1) & Release.verboseDebug:
                        print ("WARNING: Could not find Model \"" + modelName +
                               "\" in the list of models defined in the config " +
                               "file (" + str(modelNames) + ").")
                validated = True
                coupled_model_dict = coupled
                break
            
        if not validated:
            raise Exception, ("Error validating coupled model \"" +
                              coupledName + "\" (" +
                              str(modelNames) + ")."
                              )

        # Instantiate all models from the models.yml coupled model
        # definition, not the 'modelNames' input list.  Why do this?
        # In case the a model is absent from the configDict for
        # whatever reason.
        #
        # For example, during development & testing, the "XJD" Model
        # Object might not be present in the config file, since it has
        # no user configurable options.
        coupled_model_dict[coupledName].remove("Shared")
        shared = eval("Models.Shared")()

        models = []
        for modelName in coupled_model_dict[coupledName]:
            models.append( eval("Models." + modelName)() )
        
        return (shared, models)


    def __setEnvironmentConfiguration(self, filename, configDict, coupledName, isInterCommCoupled):
        """
        Parse configDict for Environment input. If Environment is not
        specified, query user (currently hardcoded via stdin) for
        environment information.
        """    
        (environmentName) = self.__getEnvironmentName(configDict)
        env = None
        if environmentName:
            (env) = self.__instantiateEnvironment(environmentName)
            self.__setupModel(env, configDict[coupledName]["Environment"])
        else:
            print 'WARNING: Did not find environment in config file "%s". Querying for input' % filename
            env = Environments.inputs.stdin(False, False, isInterCommCoupled)            
        
        return env


    def __getEnvironmentName(self, configDict):
        """
        Returns the name of the selected environment as defined in
        configDict.  Otherwise returns None (and calling function
        should then somehow prepare an environment).
        """
        coupledName = configDict.keys()[0]
        
        if not configDict[coupledName].has_key("Environment"):
            return None

        if (len(configDict[coupledName]["Environment"][0].keys()) != 1):
            raise Exception, ("Error:  Should only have one environment " +
                              "defined, but appears to define " +
                              str(len(configDict[coupledName]["Environment"].keys())) +
                              ":  " + str(configDict[coupledName]["Environment"].keys())
                              )
        return ( configDict[coupledName]["Environment"][0].keys()[0] )


    def __instantiateEnvironment(self, environmentName):
        """
        Instantiate & return the Environment object corresponding to
        the input string 'environmentName'.  return (0)
        """
        env = Environments.Environment()
        isEnvironmentFound = False
        for envIndex, environment in enumerate(env.environments):
            if environmentName == environment["__metadata__"]["environment_name"]:                
                env.selectEnvironment( envIndex )
                isEnvironmentFound = True
                break

        if not isEnvironmentFound:
            raise Exception, ("Error:  Could not instantiate Environment \"" +
                              environmentName + "\".  is this a valid Environment type?")
        
        return env



    def __setupModel(self, model, configDict):
        """
        Using the configuration defined in configDict, setup the model
        object.

        Parameters:
          model: Either a Model (Shared, LFM, MIX, etc.) or an
            Environment (sh, LSF, SGE, etc.)
          configDict: Configuration read in via YAML to a Python
            dictionary/list object.
        """
        # Get the model name
        if model.metadata.has_key("model_name"):
            descriptor = model.metadata["model_name"]
        elif model.metadata.has_key("environment_name"):
            descriptor = model.metadata["environment_name"]
        else:
            raise Exception, ("Did not understand which Model you want me to " +
                              "configure.  Does the [Model_Name].yml file " +
                              "have a __metadata__ section that describes " +
                              "basic information for the model?"
                              )
        # Find dictionary
        dictIndex = -1
        for i, config in enumerate(configDict):
            if config.has_key(descriptor):
                dictIndex = i
                break
        # Some models (e.g. XJD) aren't written to configDict.  For
        # these models, we dont need to do any setup.
        if dictIndex == -1:
            return

        self.__setParametersByType(model, configDict[i][descriptor])

    def __setParametersByType(self, model, configDict):
        """
        Set model object parameters using the nested dictionary.

        Parameters:
          model: EIther a Model (Shared, LFM, MIX, etc.) or an
            Environment (sh, LSF, SGE, etc.)
          configDict: A nested dictionary of one or more parameter
            types, with one or more variable lists for each type.
        """
        # Dictionary of Parameter Types (e.g. the LFM has several:
        # "Constants", "Solar Wind", "Ionosphere")
        for parameterType in configDict.keys():            

            # List of parameters (e.g. the LFM "Constants" type has
            # several parameters: "CA", "GAMMA", "CRNT", ....)
            for parameter in configDict[parameterType]:

                assert(len(parameter.keys()) == 1)
                paramName = parameter.keys()[0]
                
                varDict = model.get_variable_dict(parameterType, paramName)
                if varDict.has_key("format_regex"):
                    regex = varDict["format_regex"]
                else:
                    # When there is no regex defined, match anything.
                    regex = ".*"
                    if Release.verboseDebug:
                        print ("WARNING:  No \"format_regex\" for variable \"" +
                               parameterType + ": " + paramName + "\".  Matching anything.")

                # Individual variable can be stored as a dictionary.
                # This allows us to specify conditions of the
                # variable.  Here's an example:
                # 
                #   - AURORA: 
                #       status: disabled
                #       value: 1
                #
                # Here the variable "aurora" is disabled, but has a
                # value of "1" if it were "enabled".
                if type(parameter[paramName]) == dict:
                    keys = parameter[paramName].keys()
                    
                    if 'status' in keys:
                        varDict['status'] = parameter[paramName]['status']
                        # Select the default value in case status==specified & value omitted.
                        paramValueStr = varDict['value']
                            
                    if 'value' in keys:
                        paramValueStr = str(parameter[paramName]['value'])
                        
                    if ('status' not in keys) & ('value' not in keys):
                        raise Exception, ('Error:  Invalid format for variable "' +
                                          paramName + '".  ' +
                                          'Expected "valid" and "status" flags for dictionary, but none were found.')
                else:
                    paramValueStr = str(parameter[paramName])
                    # Enable variables that aren't explicitly disabled!
                    if 'status' in varDict:
                        if varDict['status'] == 'disabled':
                            varDict['status'] = 'enabled'
                        else:
                            if Release.verboseDebug:
                                print 'WARNING: Found status for "' + paramName + '". Not enabling because it is set to "' + varDict['status'] + '".'
                        
                # Check regular expression to ensure the variable has a valid value.
                if not share.matches_regex(regex, paramValueStr):
                    format_human = varDict["format_human"]
                    
                    raise Exception, ("Error:  Invalid format for Variable " +
                                      "\"" + paramName +  ".  "
                                      "Got: \"" + str(parameter[paramName]) + "\"." +
                                      "Expected: " + format_human + "\".  " +
                                      "RegEx: \"" + regex + "\".")

                # type cast variable appropriately:
                if  varDict.has_key('type'):
                    paramValueStr = share.typeCast(paramValueStr, varDict['type'])
                
                varDict["value"] = paramValueStr

    ############################################################################

    def writeConfigurationFile(self, filename, coupled_name, shared, models, environment,
                               includeOptional=False, includeExpert=False):
        """
        Write the model configuration to a YAML file.

        Parameters:
          filename:  File to write
          coupled_name:  Name of the coupled model (e.g. 'LFM-MIX', 'CMIT', etc.)
          shared:  Shared Model object
          models:  list of Model objects (e.g. for LFM-MIX this might be LFM, MIX, XJD)
          environment:  Compute Environment object
          includeOptional:  Should we write out all optional variables?
          includeExpert:  Should we write out all optional variables?
        """
        # Get the model configruation as a Python object.
        config = self.__getModelConfiguration(coupled_name, shared, models, environment,
                                              includeOptional, includeExpert)

        yaml_string =  yaml.safe_dump(config, default_flow_style=False)
        
        # Write to file
        yaml_file = open(filename, "w")
        yaml_file.write(yaml_string)
        yaml_file.close()
    

    def __getModelConfiguration(self, coupled_name, shared, models, environment,
                                includeOptional, includeExpert):
        """
        Get a Python dictionary representing the makeItSo configuration

        Parameters:
          coupled_name:  Stand-alone (e.g. 'TIEGCM') or coupled model
            (e.g. 'CMIT') name.
          shared: Parameters shared amongst one (or more) model(s)
          models: List of Model objects that define parameters for
            each (of the coupled) model(s)
          environment: Environment object that defines parameters for
            the compute environment.
        """
        #
        # About
        #
        configDict = {}
        configDict[coupled_name]={}
        configDict[coupled_name]["About"] = []
        README = """
        This configuration file can be used for repeated runs of the
        script generator with minor changes.  For example, if you wish
        to do a parameter study, you could run makeItSo.py with manual
        stdin input once, then tweak a variable defined in this
        configuration file slightly and pass the modified file to
        makeItSo.py.  Execute makeItSo.py --help for details on
        running makeItSo.py from a config file.
        """
        # Cleanup README string for output to YAML file.
        README=README.replace("\n        "," ")
        configDict[coupled_name]["About"].append({"README": README })
        configDict[coupled_name]["About"].append({"Created at": datetime.datetime.now()})
        #FIXME:  Append real version number:
        configDict[coupled_name]["About"].append({"Created by": "Make It So, v " + Release.version})
        configDict[coupled_name]["About"].append({"Configured by": getpass.getuser()})
        configDict[coupled_name]["About"].append({"Configured on": platform.node()})

        #
        # Models & Compute Environment
        #
        configDict[coupled_name]["Models"] = []
        configDict[coupled_name]["Environment"] = []
        userConfiguration = [shared] + models + [environment]
        
        for config in userConfiguration:
            if config.metadata.has_key("model_name"):
                descriptor = config.metadata["model_name"]
                isModel = True
                isEnvironment = False
            elif config.metadata.has_key("environment_name"):
                descriptor = config.metadata["environment_name"]
                isModel = False
                isEnvironment = True
            else:
                raise Exception, "ERROR:  Do not understand metadata for model."

            # Create & append model/environment to the dictionary
            
            newConfigDict = self.__getModelDictionary(coupled_name, config,
                                                      includeOptional, includeExpert)

            # Add non-empty dictionaries of parameters
            if len(newConfigDict.keys()) > 0:
                if isModel:
                    configDict[coupled_name]["Models"].append({descriptor:newConfigDict})
                if isEnvironment:
                    configDict[coupled_name]["Environment"].append({descriptor:newConfigDict})

        return configDict

    def __getModelDictionary(self, coupled_name, model, includeOptional=False, includeExpert=False):
        """
        Returns a dictionary of modified model parameters sorted by
        parameter type.

        Note: if no parameters were changed from the default, returns
        an empty dictionary.

        Parameters:
          model: makeItSo Model object
        """
        modelDict = {}

        for parameterType in model.parameters.keys():
            # Don't add duplicate parameter types (in theory that should never happen anyway)
            if not modelDict.has_key(parameterType):
                listOfParameters = self.__getParametersList(coupled_name, model, parameterType, includeOptional, includeExpert)
                # Add non-empty lists of parameters
                if len(listOfParameters) > 0:
                    modelDict[parameterType] = listOfParameters

        return modelDict

    def __getParametersList(self, coupled_name, model, parameterType, includeOptional=False, includeExpert=False):
        """
        Returns list of changed parameters that are a part of
        model.parameters[parameterType].

        Note: If no parameters were changed from default, returns an empty list.

        Parameters:
          model: makeItSo Model object
          parameterType: String containing parameter type name
          includeOptional: Append all (expert) variables
          includeExpert: Append all (optional) variables
        """
        parametersList = []

        for parameterIndex, parameter in enumerate(model.parameters[parameterType]):
            paramName = parameter.keys()[0]
            # Some configurations (i.e. XJD) do not have a "value".
            if not parameter[paramName].has_key("value"):
                continue

            # Ignore variables that do not belong to this coupled
            # model type.  (e.g. LFM's S107 doesn't make sense when
            # running LFM-MIX)
            if parameter[paramName].has_key("models"):
                modelsList = str.split( parameter[paramName]["models"].replace(" ", ""), ",")
                if (modelsList.count(coupled_name) == 0):
                    continue

            try:
                defaultParameter = model.parameters_defaults[parameterType][parameterIndex][paramName]
            except:
                # It's possible that the parameters_defaults list got
                # shuffled around.  If we had a hard time finding a
                # defaultParameter, let's search through the list for
                # the right index:
                for idx,variable in enumerate(model.parameters_defaults[parameterType]):
                    if paramName == variable.keys()[0]:
                        break
                defaultParameter = model.parameters_defaults[parameterType][idx][paramName]

            paramDict = self.__getParameterDictionary(paramName,
                                                      parameter[paramName], defaultParameter,
                                                      includeOptional, includeExpert)

            # Make sure we got a variable. Then append!
            if paramDict.keys():
                parametersList.append(paramDict)

        return parametersList


    def __getParameterDictionary(self, paramName, param, defaultParam, includeOptional=False, includeExpert=True):
        """
        Obtain dictionary for a given parameter.  For example:
        
           - AURORA: 
               status: disabled
               value: 1

        Parameters:
          paramName:  Name of parameter (ie. 'AURORA')
          param:  Dictionary of parameter (see YAML file)
          defaultParam:  Default dictionary of parameter
          includeOptional: Append all (expert) variables
          includeExpert: Append all (optional) variables

        Returns: Dictionary for given parameter.  See unit tests for example.
        """
        
        # Do not write shared/automated parameters.
        if ( (param['user_input'] == 'shared_model_parameter') |
             (param['user_input'] == 'automated') ):
            return {}
   
        # Write any variables that
        # 1. have changed
        # 2. are required
        # 3. are optional (if requested)
        # 4. are expert (if requested)
        if ( (param != defaultParam) |
             ( (param['user_input'] == 'required') ) |
             ( (param['user_input'] == 'optional') & (includeOptional | includeExpert) ) |
             ( (param['user_input'] == 'expert_only') & (includeExpert) ) ):                        
            # Certain variables can be disabled, enabled or
            # required.  When a variable is set to "disabled",
            # store status in config file.  Here's an example:
            # 
            #   - AURORA: 
            #       status: disabled
            #       value: 1
            #
            # Here the variable "aurora" is disabled, but has a
            # value of "1" if it were "enabled".
    
            # Also note the goofy if ... else ... else logic is
            # required. Can't do "if (a & b)" if "b" doesn't exist!  Must
            # split it up into multiple conditionals.
            if 'status' not in defaultParam:
                return { paramName: param['value'] }
            else:
                if (param['status'] in ['enabled', 'mandatory']):
                    return { paramName: param['value'] }
                elif (param['status'] == 'disabled'):
                    return { paramName: {'value': param['value'], 'status': param['status']} }
                else:
                    raise Exception, 'Error:  Status for "' + paramName + '" is set to an invalid value: "' + param['status'] + '".'

        # If we made it this far, 'param' is unecessary.
        return {}

import unittest
import tempfile
import os
import copy

class TestConfiguration(unittest.TestCase):
    def setUp(self):
        self.Shared = Models.Shared()

        self.LFM = Models.LFM()
        self.MIX = Models.MIX()
        self.XJD = Models.XJD()

        self.env = Environments.Environment()
        for index, environment in enumerate(self.env.environments):
            if (environment["__metadata__"]["environment_name"] == "Bourne Shell (sh)"):
                self.env.selectEnvironment( int(index) )
                break
        if self.env.selected_environment == -1:
            raise Exception, "Error:  No environment found/selected!"

        self.config = Configuration()

        (fileId, self.filename) = tempfile.mkstemp()
        os.close(fileId)

        # Make a temporary file.
        self.config.writeConfigurationFile(self.filename, "LFM-MIX", self.Shared, [self.LFM, self.MIX, self.XJD], self.env)

    def tearDown(self):
        os.remove(self.filename)

    def test_write_file_when_parameters_are_changed(self):
        self.Shared.get_variable_dict("SHARED","RUN_IDENTIFIER")["value"] = "purpleMonkeyDishwasher"
        self.LFM.get_variable_dict("Solar Wind","BX_ZERO")["value"] = 2.71828183
        self.MIX.get_variable_dict("LOCAL CONDUCTANCE MODEL PARAMETERS","F107")["value"] = 3.141592
        self.env.get_variable_dict("environment", "mpirun")["value"] = "purpleMonkeyDishwasher"

        self.config.writeConfigurationFile(self.filename, "LFM-MIX", self.Shared, [self.LFM, self.MIX, self.XJD], self.env)

    def test_readConfigurationFile(self):
        (name, shared, models, env) = self.config.readConfigurationFile(self.filename)

    def test_getParameterDictionary(self):
        paramName='F107'
        param = { 'description': 'F10.7 cm daily solar flux',
                  'status': 'disabled',
                  'type': 'float',
                  'units': 'unknown',
                  'user_input': 'expert_only',
                  'value': '100.0',
                  'format_regex': '"^[+-]?\\ *(\\d+(\\.\\d*)?|\\.\\d+)([eE][+-]?\\d+)?$"',
                  'format_human': '"Must be a number."'}
        defaultParam = copy.deepcopy(param)

        varDict=self.config._Configuration__getParameterDictionary(paramName, param, defaultParam, False, False)
        self.assertEqual(varDict, {})
        
        param['value'] = 200.0
        varDict=self.config._Configuration__getParameterDictionary(paramName, param, defaultParam, False, False)
        self.assertEqual('F107', varDict.keys()[0])
        self.assertTrue(varDict['F107'].has_key('status'))
        self.assertEqual(varDict['F107']['status'],'disabled')
        self.assertTrue(varDict['F107'].has_key('value'))
        self.assertEqual(varDict['F107']['value'], 200.0)

        param['status'] = 'enabled'
        varDict=self.config._Configuration__getParameterDictionary(paramName, param, defaultParam, False, False)
        self.assertEqual('F107', varDict.keys()[0])
        self.assertEqual(varDict['F107'],200.0)
