import copy
import datetime
import os
import re
import unittest

from MakeItSo import yaml

class Environment(object):
    """
    Set the job scripting environment
    """
    def __init__(self):
        """
        Queries the 'Environments' directory for all YAML description
        files & reads the results into the 'self.environments' array.

        Note: You must call selectEnvironment before calling most
        member functions.
        """
        
        yaml_filenames = []
        
        # Get list of queueing environments:
        dirname = os.path.dirname(__file__)
        files = os.listdir(dirname)
        for filename in files:
            if filename.find(".yml", len(filename)-4,len(filename)) > 0:
                yaml_filenames.append(os.path.join(dirname, filename))

        self.environments = []
        # Read all environment specs from YAML files.
        for filename in yaml_filenames:
            yaml_file = open(filename)
            self.environments.append( yaml.safe_load(yaml_file) )
            yaml_file.close()

        self.selected_environment = -1
        self.metadata = None
        self.parameters_defaults = None
        self.parameters = None

    def selectEnvironment(self, selection):
        """
        Select a particular environment & set helper member variables
        internal to the Environment class.
        """
        self.selected_environment = int(selection)
        assert(self.__is_valid_environment_selection)

        self.parameters = self.environments[self.selected_environment]
        self.metadata = self.parameters.pop("__metadata__")
        self.parameters_defaults = copy.deepcopy(self.parameters)

    def __is_valid_environment_selection(self):
        return ( (self.selected_environment >= 0) & (self.selected_environment < len(self.environments)) )

    def get_variable_dict(self, variable_type, variable_name):
        """
        Search the parameters dictionary for `variable_name`, which is
        of `variable_type`.

        This function uses a 'soft copy' of (akin to a C pointer to)
        the variable, and hence it's the proper way to get and set
        variables.
        """
        assert(self.__is_valid_environment_selection)
        variable_dict = None

        for variable in self.parameters[variable_type]:
            if variable_name == variable.keys()[0]:
                variable_dict = variable[variable_name]
                break

        if variable_dict == None:
            raise Exception, "Error: [%s][%s] not found in the selected environment" % (variable_type, variable_name)

        return variable_dict

    def get_num_cpus(self, cpuString):
        """
        Return the integer number of processors to use.

        FIXME: This currently the 'n_processors' string to determines
               how many CPU's should be used.  In the future, I could
               add a __metadata__ flag like 'nCpusPerNode' and use
               this to determine how many nodes & cpus should be used
        """
        cpuString = str(cpuString)
        
        if re.match('^\d+$', cpuString.strip()):
            return( int( cpuString.strip() ) )
        elif re.match('^nodes=\d+:ppn=\d+',cpuString.strip()):
            matchTuple = re.findall('^nodes=(\d+):ppn=(\d+)$', cpuString.strip())
            assert( len(matchTuple) == 1 )

            nNodes = int( matchTuple[0][0].strip() )
            cpusPerNode = int( matchTuple[0][1].strip() )
            return nNodes * cpusPerNode
        else:
            raise Exception, 'Error:  Did not understand "n_processors" value of %s.' % (cpuString)                      
        

    def get_run_command(self, is_mpi, executable, flags, stdin, stdout_and_stderr):
        """
        Get the command used to run an executable.

        For example usage, see 'test_get_run_instructions'.

        Parameters:
          is_mpi:  Boolean to determine whether this particular command needs to be run via MPI.
          executable:  Name of the executable
          flags:  flags for executable
          stdin:  file to stdin redirect
          stdout_and_stderr:  stdout/stderr redirects.

        Returns:
          string: Complete Unix-style run command.
        """
        assert(self.__is_valid_environment_selection)

        # If everything is Null or empty, assume we dont run an executable.
        if ( ((executable == "") | (executable == None)) &
             ((flags == "") | (flags == None)) &
             ((stdin == "") | (stdin == None)) &
             ((stdout_and_stderr == "") | (stdout_and_stderr == None)) ):
            return ""
        
        # Sanitize input
        if is_mpi:            
            mpirun  = self.get_variable_dict("environment", "mpirun")["value"] + " "
            # Some machines dont require "-np X", so dont print it if "mpi np flag" is empty.
            if (self.get_variable_dict("environment", "mpi np flag")["value"] != ""):
                mpirun += self.get_variable_dict("environment", "mpi np flag")["value"] + " "
                mpirun += str(self.get_num_cpus(self.get_variable_dict("flags", "n_processors")["value"])) + " "
        else:
            mpirun = ""

        if (flags == "") | (flags == None):
            flags = ""
            
        if (stdin == "") | (stdin == None):
            stdin = "/dev/null"

        if (stdout_and_stderr == "") | (stdout_and_stderr == "None"):
            stdout_and_stderr = "/dev/null"
        
        if self.metadata["shell"] == "sh":
            redirect_in  = " < "
            redirect_out = " > "
        else:
            raise Exception, "Error:  Environment Shell misunderstood." + str(self.metadata)

        # sh requires stdout/stderr redirect have "2>&1" at the end of
        # the line to redirect both stderr and stdout.
        return (mpirun + "./" + executable + " " + flags + redirect_in + stdin + redirect_out + stdout_and_stderr + " 2>&1")


    def get_script_header(self, run_identifier, iteration, has_InterComm):
        """
        Returns a string containing the header of a job script,
        including any queueing envrinoment-specific settings.

        WARNING:  Must input the scripting parameter values first!
        """
        assert(self.__is_valid_environment_selection)

        outs = ["#!/bin/sh"]

        for variable in self.parameters["flags"]:
            variable_name = variable.keys()[0]
            variable_dict = variable[variable_name]

            if variable_name == "job_name":
                variable_dict["value"] = "%s-%03d" % (run_identifier, iteration)
            elif variable_name == "output_file":
                variable_dict["value"] = "%s-%03d.output" % (run_identifier, iteration)
            elif variable_name == "error_file":
                variable_dict["value"] = "%s-%03d.error" % (run_identifier, iteration)
                
            if variable_dict.has_key("flag"):
                if (variable_dict["flag"].count("None")) == 0:
                    parameter = ( str(self.metadata["definition"]) + " " +
                                  str(variable_dict["flag"]) + " " +
                                  str(variable_dict["value"]) )
                    outs.append( parameter )

        outs.append( "" )
        outs.append( "# Auto-generated by makeItSo (%s)" % (datetime.datetime.now()).strftime("%a, %d %b %Y %H:%M:%S") )

        outs.append( "echo \"%s-%03d started at \" > %s-%03d.status" % (run_identifier, iteration, run_identifier, iteration) )
        outs.append( "echo `date` >> %s-%03d.status" % (run_identifier, iteration) )
        outs.append( "echo \"`whoami`@`hostname`\" >> %s-%03d.status" % (run_identifier, iteration) )
        outs.append( "echo `uname -a` >> %s-%03d.status" % (run_identifier, iteration) )
        outs.append( "echo \"Run directory: `pwd`\" >> %s-%03d.status" % (run_identifier, iteration) )
        #FIXME:  Populate this via a PBS_JOBID or LSB_JOBID to environment config files.
        outs.append( "echo \"Job ID: ${PBS_JOBID} ${LSB_JOBID}\" >> %s-%03d.status" % (run_identifier, iteration) )        
        outs.append( "" )

        outs.append( "# Setup envirnoment" )
        if (self.get_variable_dict("environment", "Append to LD_LIBRARY_PATH")["value"] != ""):
            outs.append( "export LD_LIBRARY_PATH=%s:${LD_LIBRARY_PATH}" % ( self.get_variable_dict("environment", "Append to LD_LIBRARY_PATH")["value"] ) )

        if has_InterComm:
            outs.append( self.get_pvm_startup() )
        outs.append( "" )

        outs.append( "# Record hosts we will use for computation" )
        outs.append( "echo \"HOSTS:\" >> %s-%03d.status" % (run_identifier, iteration) )
        outs.append( "cat ${MY_HOSTS} >> %s-%03d.status" % (run_identifier, iteration) )
        outs.append( "" )        
        
        return( "\n".join(outs) )


    def get_pvm_startup(self):
        """
        Return a string containing the code necessary to configure & start the PVM daemon.
        """
        assert(self.__is_valid_environment_selection)

        outs = []        

        # Set standard environment variables
        #FIXME:  Replace this with
        #        `$PVM_ROOT/lib/pvmgetarch`
        outs.append( "export PVM_ARCH=%s" % ( self.get_variable_dict("environment", "PVM_ARCH")["value"] ) )
        outs.append( "export PVM_ROOT=%s" % ( self.get_variable_dict("environment", "PVM_ROOT")["value"] ) )
        outs.append( "export PVM_BIN=${PVM_ROOT}/bin" )
        outs.append( "export PVM_RSH=%s" % ( self.get_variable_dict("environment", "PVM_RSH")["value"] ) )

        # Some environments require PVM_VMID.
        if (self.get_variable_dict("environment", "PVM_VMID")["value"] != ""):
            outs.append( "export PVM_VMID=%s" % ( self.get_variable_dict("environment", "PVM_VMID")["value"] ) )

        outs.append( "export MY_HOSTS=pvm_hostfile" )

        # Create a .pvmprofile that defines all PVM variables
        outs.append( "" )
        outs.append( "rm -f ~/.pvmprofile" )
        outs.append( "env | grep PVM_ > ~/.pvmprofile" )
        outs.append( "" )

        # Create hostfile
        outs.append( "# Create file containing _unique_ host names.  Note that there are two possible sources of available hosts" )
        outs.append( "if [ -f \"%s\" ]; then" % self.get_variable_dict("environment", "HOSTS")["value"] )
        outs.append( "    #   1.  Host File, containing one host per file" )
        outs.append( "    sort -k 1,1 -u %s | awk '{print $1}' | perl -pe \"s/^([a-zA-Z0-9\-\_]+)/\\1/\" > ${MY_HOSTS}" % (self.get_variable_dict("environment", "HOSTS")["value"]) )        
        outs.append( "else" )
        outs.append( "    #   2.  Environment variable: space-delimited list (e.g. $HOSTS='one one two two three three')" )
        outs.append( "    echo %s|awk '{for(i=1;i<NF;i+=1) b[$i]=1; for(i in b)print i}' > ${MY_HOSTS}" % (self.get_variable_dict("environment", "HOSTS")["value"]) )
        outs.append( "fi" )
        outs.append( "" )        

        # Remove /tmp/aaa* on all nodes before starting PVM (required according to sghosh@ucar.edu)
        outs.append( "for h in `cat ${MY_HOSTS}`; do " )
        #outs.append( "  %s ${h} \"find /tmp -user $LOGNAME -exec rm -fr {} \; 2>/dev/null\" " % (self.get_variable_dict("environment", "PVM_RSH")["value"])))
        outs.append( "  %s ${h} \"rm -f /tmp/aaa* 2> /dev/null\" " % (self.get_variable_dict("environment", "PVM_RSH")["value"]))
        outs.append( "done" )
        outs.append( "" )

        # Start PVM & Add Nodes.        
        outs.append( "# Start PVM & add nodes" )
        outs.append( "printf \"%s\\n%s\\n\" conf quit|${PVM_ROOT}/lib/pvm ${MY_HOSTS}" )
        outs.append( "wait" )
        outs.append( "sleep 2" )

        return( "\n".join(outs) )
        

    def get_sh_script_footer(self, run_identifier, iteration, has_InterComm):
        """
        Returns a string containing the footer of a job script.
        """
        assert(self.__is_valid_environment_selection)

        current_name = "%s-%03d" % (run_identifier, iteration)
        next_name = "%s-%03d" % (run_identifier, iteration+1)
        current_status_file = "%s.status" % (current_name)
        next_status_file = "%s.status" % (next_name)
        
        file_suffix = self.metadata["output_suffix"]
        submit_via =  self.metadata["submit_via"]

        outs = []
        if has_InterComm:
            outs.append( "wait" )
            outs.append( "# Exit PVM daemon" )
            outs.append( "echo \"reset\" | $PVM_ROOT/lib/pvm" )
            outs.append( "echo \"halt\" | $PVM_ROOT/lib/pvm" )
            outs.append( "" )

        outs.append( "# record the ending time of this portion of the run" )        
        outs.append( "echo \"%s ended at\" >> %s" % (current_name, current_status_file) )
        outs.append( "echo `date` >> %s" % (current_status_file) )
                
        outs.append( "" )
        outs.append( "# submit the next job if there's more to do" )
        outs.append( "if [ -f \"%s.%s\" ]; then" % (next_name, file_suffix) )
        outs.append( "  echo \"Submitting %s.%s\"" % (next_name, file_suffix) )
        outs.append( "  %s %s.%s" % (submit_via, next_name, file_suffix) )
        outs.append( "fi" )
        outs.append( "" )

        # I oringally appended this to kill rogue PVM daemons running
        # on Bluefire, preventing the job from exiting cleanly.
        # It didn't fix the problem & seems to have broken running
        # jobs in the queue on Rice's STIC cluster.  Commenting this out for now:
        #outs.append( "# Kill any rogue processes still lying around before exiting." )        
        #outs.append( "wait" )
        #outs.append( "kill -9 -1" )
        outs.append( "" )

        return( "\n".join(outs) )

class TestEnvironment(unittest.TestCase):

    def setUp(self):
        self.env = Environment()
        # find & select bourne shell (sh) environment
        for index, environment in enumerate(self.env.environments):
            if (environment["__metadata__"]["environment_name"] == "Bourne Shell (sh)"):
                self.env.selectEnvironment( int(index) )
                break
        if self.env.selected_environment == -1:
            raise Exception, "Error:  No environment found/selected!"            

    def test_get_num_cpus(self):
        nCpus = self.env.get_num_cpus(8)
        self.assertEqual(nCpus, 8)

    def test_get_num_cpus_TRM(self):
        """
        Torque Resource Manager (TRM) requires a different format for nCpus:
                   -lnodes=3:ppn:8
        This will utilize 3 nodes and 8 CPUs per node, so 24 CPUs total.
        """
        nCpus = self.env.get_num_cpus('nodes=3:ppn=8')
        self.assertEqual(nCpus, 24)

        nCpus = self.env.get_num_cpus('nodes=2:ppn=12')
        self.assertEqual(nCpus, 24)

        nCpus = self.env.get_num_cpus('nodes=18:ppn=8')
        self.assertEqual(nCpus, 144)        


    def test_get_run_command(self):
        # LFM, LFM-MIX and LFM-CMIT:
        run_string = self.env.get_run_command(True, "LFM-MIX", "cmit.xjd", None, "LFM-MIX-002.out")
        self.assertEqual(run_string, "mpirun -np 8 ./LFM-MIX cmit.xjd < /dev/null > LFM-MIX-002.out 2>&1")

        run_string = self.env.get_run_command(True, "LFM-MIX", "lfm-mix.xjd", None, "LFM-MIX-020.out")
        self.assertEqual(run_string, "mpirun -np 8 ./LFM-MIX lfm-mix.xjd < /dev/null > LFM-MIX-020.out 2>&1")

        run_string = self.env.get_run_command(True, "LFM", None, None, "LFM-MIX-200.out")
        self.assertEqual(run_string, "mpirun -np 8 ./LFM  < /dev/null > LFM-MIX-200.out 2>&1")

        # MIX
        run_string = self.env.get_run_command(False, "ionosphere", "cmit.xjd", None, "IONOSPHERE-002.out")
        self.assertEqual(run_string, "./ionosphere cmit.xjd < /dev/null > IONOSPHERE-002.out 2>&1")

        run_string = self.env.get_run_command(False, "ionosphere", "lfm-mix.xjd", None, "IONOSPHERE-300.out")
        self.assertEqual(run_string, "./ionosphere lfm-mix.xjd < /dev/null > IONOSPHERE-300.out 2>&1")

        # TIEGCM
        run_string = self.env.get_run_command(False, "tiegcm", None, "tiegcm1.9-002.inp", "TIEGCM-002.out")
        self.assertEqual(run_string, "./tiegcm  < tiegcm1.9-002.inp > TIEGCM-002.out 2>&1")
        
        # XJD
        run_string = self.env.get_run_command(False, "", "", "", "")
        self.assertEqual(run_string, "")

if __name__ == '__main__':
    import doctest
    doctest.testmod()
    unittest.main()
