from Model import Model
from MakeItSo import share
from MakeItSo import Release

import datetime
import os
import re
import unittest
import xml.dom.minidom

##########################################################################################

class LFM(Model):
    """ LFM Model Type """

    def __init__(self):
        filename = os.path.join(os.path.dirname(__file__), "LFM.yml")
        Model.__init__(self, filename)

        # Make sure the YAML file is for the LFM:
        assert( self.metadata["model_name"] == "LFM" )
        assert( self.metadata["output_prefix"] == "INPUT1" )
        assert( self.metadata["output_suffix"] == "xml" )

    def set_shared_parameters(self, coupled_model_name, shared_model, env=None):
        """
        Note: Required function (standardized parameters & return variables)        
        """
        self.set_shared_model_parameters(shared_model.get_variable_dict("SHARED", "RUN_IDENTIFIER")["value"],
                                         shared_model.get_variable_dict("SHARED", "LFM_MIX_EXCHANGE_FREQUENCY")["value"],
                                         shared_model.get_variable_dict("SHARED", "TIME_DUMP_INTERVAL")["value"],
                                         shared_model.get_variable_dict("SHARED", "USE_INTERCOMM")["value"])        

        # Set the rcm coupling start time.
        if (coupled_model_name == "LFM-RCM"):
            self.get_variable_dict("INPUT-RCM","RCM_START_DATETIME")["value"] = shared_model.get_variable_dict("SHARED", "START_YMDHMS")["value"]


    def set_shared_model_parameters(self, run_prefix, lfm_mix_exchange_frequency, dumpInterval, use_intercomm):
        """
        Setup internal structure of generic LFM shared parameters

        This sets four parameters defined in LFM.yml:
          * HDFBASE_IN, HDFBASE_OUT, RUN_NAME: strings used for naming LFM input/output files
          * TIME_DUMP_INTERVAL:  Number of seconds between file dumps
          * USE_INTERCOMM:  do we use InterComm (True) or adhoc (False) coupling?
        """
        
        self.get_variable_dict("GO","HDFBASE_IN")["value"] = run_prefix
        self.get_variable_dict("GO","HDFBASE_OUT")["value"] = run_prefix
        self.get_variable_dict("GO","RUN_NAME")["value"] = run_prefix
        self.get_variable_dict("GO", "USE_INTERCOMM")["value"] = use_intercomm
        self.get_variable_dict("UTIO","TIME_DUMP_INTERVAL")["value"] = str(float(dumpInterval))

        self.get_variable_dict("IONOSPHERE", "ION_EXCHANGE_FREQUENCY")["value"] = str(float(lfm_mix_exchange_frequency))

    def __calc_tzero(self, spinupDDHHMMSS_str):
        """
        Returns floating point TZERO: length of preconditioning
        interval in seconds)

        Parameters:
          spinupDDHHMMSS_str: String of integers 'DD HH MM SS'
            representing number of days (DD), hours (HH), minutes (MM)
            and seconds (SS) to use for the spinup interval.  Typical
            value is 50 minutes (0 0 50 0)

        returns:
          TZERO: Length of preconditioning interval in seconds.
            During this interval, the solar wind has a velocity of
            400km/s, a density of 5.8 #/cc and a sound speed of 35 km/s.
            No IMF is imposed on the magnetosphere.  These conditions
            helps the magnetosphere take shape within the computational
            domain.  For all runs started from scratch, this period must
            be conducted.
        """
        
        assert(len(spinupDDHHMMSS_str.strip().split()) == 4)
        (dd,hh,mm,ss) = [float(s) for s in spinupDDHHMMSS_str.strip().split()]

        return  (ss + 60.0*mm + 60.0*60.0*hh + 60.0*60.0*24.0*dd)    

        
    def set_time_information(self, start_datetime, stop_datetime):
        """
        Setup the internal structure of the LFM parameters to generate
        the correct set of scripts.

        This function sets two parameters defined in LFM.yml:
          * START_DATETIME: Whitespace delimited integer array of YYYY
              MM DD HH MM SS for the first timestep on the current
              segment.
          * STOP_DATETIME:  Whitespace delimited integer array of YYYY
              MM DD HH MM SS for the final timestep on the current
              segment.

        start_datetime: datetime object
        stop_datetime: datetime object

        Note: Required function (standardized parameters & return variables)
        """

        timeString = []
        timeString.append( str(int(start_datetime.year)) )
        timeString.append( str(int(start_datetime.month)) )
        timeString.append( str(int(start_datetime.day)) )
        timeString.append( str(int(start_datetime.hour)) )
        timeString.append( str(int(start_datetime.minute)) )
        timeString.append( str(float(start_datetime.second)) )
            
        self.get_variable_dict("UTIO","START_DATETIME")["value"] = " ".join(timeString)

        timeString = []
        timeString.append( str(int(stop_datetime.year)) )
        timeString.append( str(int(stop_datetime.month)) )
        timeString.append( str(int(stop_datetime.day)) )
        timeString.append( str(int(stop_datetime.hour)) )
        timeString.append( str(int(stop_datetime.minute)) )
        timeString.append( str(float(stop_datetime.second)) )

        self.get_variable_dict("UTIO","STOP_DATETIME")["value"]  = " ".join(timeString)

    def get_spinup_coupled_model_name(self, coupled_model_name):
        spinup_coupled_model_name = "LFM"

        if ( (coupled_model_name == "CMIT") |
             (coupled_model_name == "LFM-MIX") ):
            spinup_coupled_model_name = "LFM-MIX"
        elif ( coupled_model_name == "LFM-RCM"):
            spinup_coupled_model_name = "LFM-RCM"

        return spinup_coupled_model_name


    def get_model_setup(self, model_name, has_spinup, run_identifier, start_datetime):
        shScript = ['# LFM']

        # Copy/soft link binary
        execBaseName = self.get_run_instructions(model_name, 0)[1]
        execName = execBaseName + '-' + self.__resolution_string()
        
        shScript.append('cp ${LTRROOT}/build/bin/%s .' % (execName) )
        shScript.append('ln -s %s %s' % (execName, execBaseName) )
        shScript.append('')

        # Create startupfile, if necessary
        if has_spinup:
            nixnjxnk_np = self.__resolution_string()
            (ni,nj,nk,np) = re.match('^(\d+)x(\d+)x(\d+)-(\d+)$', nixnjxnk_np).groups()

            pipeInput  = 'printf "'
            pipeInput += ni + ' ' + nj + ' ' + nk + '\\n'
            pipeInput += 'F\\n'
            pipeInput += run_identifier + '\\n'
            pipeInput += 'T\\n'

            pipeInput += start_datetime.strftime('%Y %m %d %H %M %S') + '\\n'

            # TZERO (Spinup time in seconds) is documented in __calc_tzero.
            pipeInput += str( self.__calc_tzero(self.get_variable_dict("Required Base Settings","spinup")["value"]) ) + '\\n'

            # Run identifier:
            pipeInput += 'Automatically created by MakeItSo version ' + Release.version + '\\n'
            pipeInput += '"'

            shScript.append(pipeInput + ' | ${LTRROOT}/build/bin/mkLfmStartup > /dev/null')
            shScript.append('')

        if model_name != 'LFM':
            shScript.append('rm %s_ion_%sZ.hdf' % (run_identifier, start_datetime.strftime('%Y-%m-%dT%H-%M-%S')) )

        shScript.append('')
        
        # Model-specific tasks
        instructions = ['\tLFM:']
        instructions.append(share.word_wrap('Obtain solar wind data in SM coordinates.  File must be named (or soft-linked) "SW-SM-DAT".', prefix='\t   '))
        instructions.append('\t   See the documentation for more information:')
        instructions.append('\t   https://wiki.ucar.edu/display/LTR/Solar+Wind')

        return ('\n'.join(shScript), '\n'.join(instructions))

    
    def get_iteration_setup(self, model_name, iteration):
        """
        Setup environment to run a model, ends with newline when not empty
        """
        return ( "ln -sf INPUT1-%03d.xml INPUT1.xml\n" % (iteration) )

    def get_iteration_teardown(self, model_name, iteration):
        """
        Teardown environment to run a model, ends with newline when not empty

        Note:  this is /bin/sh specific!
        """
        outs = []
        outs.append( "for i in *.dmp; do ")
        outs.append( "  mv \"$i\" \"`basename $i .dmp`.hdf\"" )
        outs.append( "done" )
        outs.append( "")

        return( "\n".join(outs) )

    def get_run_instructions(self, model_name, iteration):
        """
        Get the executable, inputs & outputs needed to run the model.

        For example usage, see 'test_get_run_instructions'.

        Note: Required function (standardized parameters & return variables)

        Returns:
          boolean:  True when parallel
          executable:  Name of executable to run
          flags:  any command-line arguments
          stdin:  where to redirect stdin input from
          stdout_and_stderr:  Where to redirect stdout/stderr to?
        """
        executable = ""
        flags = ""
        stdin = "/dev/null"
        stdout_and_stderr = ""
        
        if (model_name == "LFM-MIX"):
            executable = "LFM-MIX"
            flags = "lfm-mix.xjd"
            stdout_and_stderr = "LFM-MIX-%03d.out" % (iteration)            
        elif (model_name == "CMIT"):
            executable = "LFM-MIX"
            flags = "cmit.xjd"
            stdout_and_stderr = "LFM-MIX-%03d.out" % (iteration)
        elif (model_name == "LFM-RCM"):
            executable = "LFM-MIX"
            flags = "lfm-mix-rcm.xjd"
            stdout_and_stderr = "LFM-MIX-%03d.out" % (iteration)            
        elif (model_name == "LFM"):
            executable = "LFM"
            stdout_and_stderr = "LFM-%03d.out" % (iteration)
        else:
            raise Exception, "ERROR:  Did not understand model \"" + model_name + "\""

        return (True, executable, flags, stdin, stdout_and_stderr)

    def get_input_file(self):
        """
        Get contents of LFM INPUT1.xml file.

        Returns:
          str_script: String of job script

        FIXME: If no module variables are added, the XML still gets an
               empty <module>...</module> tag...

        Note: Required function (standardized parameters & return variables)
        """
        # Create the minidom document
        doc = xml.dom.minidom.Document()

        # <modules> base element
        modules = doc.createElement("modules")
        doc.appendChild(modules)

        # Create modules:
        variable_types_sorted = self.parameters.keys()
        variable_types_sorted.reverse()
        for name in variable_types_sorted:
            if name != 'Required Base Settings':
                modules.appendChild(self.__get_module_as_xml_string(doc, name))
        
        return doc.toprettyxml(indent="  ", encoding="utf-8")


    def __resolution_string(self):
        """
        Returns a string 'NIxNJxNK-NP' encoding resolution (NI,NJ,NK)
        and number of prcocesses (NP).      
        """
        resolution = self.get_variable_dict('Required Base Settings', 'Resolution')['value']

        if (resolution == 'single'):
            return '53x24x32-8'
        elif (resolution == 'double'):
            return '53x48x64-24'
        elif (resolution == 'quad'):
            return '106x96x128-144'
        else:
            return resolution

    def __get_module_as_xml_string(self, xml_doc, name):
        """
        Returns string containing one <module>...</module> containing
        all variables (<variable>...</variable>) belonging to that
        module.

        XML syntax returned:
        <module>
          <name>...</name>
          <variable>...</variable>
          <variable>...</variable>
          ...
          <variable>...</variable>
        </module>
        """
        variables = self.parameters[name]
        # Create <module> tag
        xml_module = xml_doc.createElement("module")
        xml_name = xml_doc.createElement("name")
        xml_name.appendChild(xml_doc.createTextNode(name))
        xml_module.appendChild(xml_name)
        
        # Add <variable> tags
        for i, variable in enumerate(variables):
            # Get a dictionary of elements for the current variable
            variable_name = variable.keys()[0]
            variable_dict = variable[variable_name]
            # Only output variables that have changed from the default.
            if (variable_dict["value"] != self.parameters_defaults[name][i][variable_name]["value"]):
            
                xml_variable = self.__get_variable_as_xml_string(xml_doc,
                                                                   variable_name,
                                                                   variable_dict["description"],
                                                                   variable_dict["value"],
                                                                   variable_dict["type"])
                xml_module.appendChild(xml_variable)
            
        return xml_module


    def __get_variable_as_xml_string(self, doc, name, meta, data, format):
        """
        Returns a string containing one variable.

        XML syntax returned:
        <variable>
          <name>...</name>
          <meta>...</meta>
          <data>...</data>
          <format>...</format>
        </variable>

        Note:  Should only be called by self.__get_module_as_xml_string(...)
        """
        
        # Variable
        variable = doc.createElement("variable")

        varName = doc.createElement("name")
        varName.appendChild(doc.createTextNode(name))
        variable.appendChild(varName)

        varMeta = doc.createElement("meta")
        varMeta.appendChild(doc.createTextNode(meta))
        variable.appendChild(varMeta)

        varData = doc.createElement("data")
        varData.appendChild(doc.createTextNode(str(data)))
        variable.appendChild(varData)

        varFormat = doc.createElement("format")
        varFormat.appendChild(doc.createTextNode(format))
        variable.appendChild(varFormat)

        return variable

class TestLFM(unittest.TestCase):
    def setUp(self):
        self.LFM_params = LFM()

    def test_set_shared_model_paraeters(self):
        self.LFM_params.set_shared_model_parameters("TestCase", 10.0, 300.0, True)
        # exchange with MIX every 10 seconds.
        # dump every 300 seconds (5 minutes)

        self.assertEqual(self.LFM_params.get_variable_dict("GO","HDFBASE_IN")["value"], "TestCase")
        self.assertEqual(self.LFM_params.get_variable_dict("GO","HDFBASE_OUT")["value"], "TestCase")
        self.assertEqual(self.LFM_params.get_variable_dict("GO","RUN_NAME")["value"], "TestCase")
        self.assertEqual(self.LFM_params.get_variable_dict("GO","USE_INTERCOMM")["value"], True)
        self.assertEqual(self.LFM_params.get_variable_dict("UTIO","TIME_DUMP_INTERVAL")["value"], "300.0")
        self.assertEqual(self.LFM_params.get_variable_dict("IONOSPHERE","ION_EXCHANGE_FREQUENCY")["value"], "10.0")
        
    def test__calc_tzero(self):
        self.assertAlmostEqual(3000.0, self.LFM_params._LFM__calc_tzero("0 0 50 0"))

        # 86400 + 7200 + 1800 + 20 = 95420
        self.assertAlmostEqual(95420.0, self.LFM_params._LFM__calc_tzero("1 2 30 20"))

        self.assertRaises(AssertionError, self.LFM_params._LFM__calc_tzero, "0 0 50")


    def test_set_time_information(self):
        run_prefix = "TestCase"
        segment_start_time = datetime.datetime(2009,1,1,0,0,0) # start on the new year!
        segment_stop_time = datetime.datetime(2009,1,19,0,0,0) # start on MLK day

        self.LFM_params.set_time_information(segment_start_time, segment_stop_time)

        self.assertEqual(self.LFM_params.get_variable_dict("UTIO","START_DATETIME")["value"], "2009 1 1 0 0 0.0")
        self.assertEqual(self.LFM_params.get_variable_dict("UTIO","STOP_DATETIME")["value"], "2009 1 19 0 0 0.0")

    def test_get_spinup_coupled_model_name(self):        
        self.assertEqual(self.LFM_params.get_spinup_coupled_model_name("CMIT"), "LFM-MIX")
        self.assertEqual(self.LFM_params.get_spinup_coupled_model_name("LFM-MIX"), "LFM-MIX")
        self.assertEqual(self.LFM_params.get_spinup_coupled_model_name("LFM"), "LFM")

    def test_get_model_setup(self):
        
        (shScript, instructions) = self.LFM_params.get_model_setup('LFM-MIX',
                                                                   has_spinup=True,
                                                                   run_identifier='test',
                                                                   start_datetime=datetime.datetime(year=2000,month=3,day=14,hour=3,minute=23,second=int(53.60496)))
        # start date/time is pi day (and time! 0.14159265 percentage of the day => 3:23:53.60496)

        expectedScript="""# LFM\ncp ${LTRROOT}/build/bin/LFM-MIX-53x24x32-8 .\nln -s LFM-MIX-53x24x32-8 LFM-MIX\n\nprintf "53 24 32\\nF\\ntest\\nT\\n2000 03 14 03 23 53\\n3000.0\\nAutomatically created by MakeItSo version 2.1.5.svn.r1993\\n" | ${LTRROOT}/build/bin/mkLfmStartup > /dev/null\n\nrm test_ion_2000-03-14T03-23-53Z.hdf\n"""
        share.unittestAssertBigStrings(self, shScript, expectedScript)
        expectedInstructions="""\tLFM:\n\t   Obtain solar wind data in SM coordinates.  File must be named (or soft-linked)\n\t   "SW-SM-DAT".\n\t   See the documentation for more information:\n\t   https://wiki.ucar.edu/display/LTR/Solar+Wind"""
        share.unittestAssertBigStrings(self, instructions, expectedInstructions)        

    def test_get_iteration_setup(self):
        self.assertEqual("ln -sf INPUT1-000.xml INPUT1.xml\n", self.LFM_params.get_iteration_setup("Test", 0))
        self.assertEqual("ln -sf INPUT1-010.xml INPUT1.xml\n", self.LFM_params.get_iteration_setup("Test", 10))
        self.assertEqual("ln -sf INPUT1-123.xml INPUT1.xml\n", self.LFM_params.get_iteration_setup("Test", 123))
                        
    def test_get_iteration_teardown(self):
        self.assertEqual("for i in *.dmp; do \n  mv \"$i\" \"`basename $i .dmp`.hdf\"\ndone\n", self.LFM_params.get_iteration_teardown("Test", 0))

    def test_get_run_instructions(self):
        (isParallel, executable, flags, stdin, stdout_and_stderr) = self.LFM_params.get_run_instructions("CMIT", 2)
        self.assertEqual(isParallel, True)
        self.assertEqual(executable, "LFM-MIX")
        self.assertEqual(flags, "cmit.xjd")
        self.assertEqual(stdin, "/dev/null")
        self.assertEqual(stdout_and_stderr, "LFM-MIX-002.out")        

        (isParallel, executable, flags, stdin, stdout_and_stderr) = self.LFM_params.get_run_instructions("LFM-MIX", 2)
        self.assertEqual(executable, "LFM-MIX")
        self.assertEqual(flags, "lfm-mix.xjd")
        self.assertEqual(stdin, "/dev/null")
        self.assertEqual(stdout_and_stderr, "LFM-MIX-002.out")

        (isParallel, executable, flags, stdin, stdout_and_stderr) = self.LFM_params.get_run_instructions("LFM-RCM", 3)
        self.assertEqual(executable, "LFM-MIX")
        self.assertEqual(flags, "lfm-mix-rcm.xjd")
        self.assertEqual(stdin, "/dev/null")
        self.assertEqual(stdout_and_stderr, "LFM-MIX-003.out")

        (isParallel, executable, flags, stdin, stdout_and_stderr) = self.LFM_params.get_run_instructions("LFM", 4)
        self.assertEqual(executable, "LFM")
        self.assertEqual(flags, "")
        self.assertEqual(stdin, "/dev/null")
        self.assertEqual(stdout_and_stderr, "LFM-004.out")

        self.assertRaises(Exception,     self.LFM_params.get_run_instructions, "supercalafragalisticexpialadoshus", 5)

    def test__resolution_string(self):
        self.assertEqual(self.LFM_params._LFM__resolution_string(), '53x24x32-8')
        self.LFM_params.get_variable_dict('Required Base Settings', 'Resolution')['value'] = 'double'
        self.assertEqual(self.LFM_params._LFM__resolution_string(), '53x48x64-24')
        self.LFM_params.get_variable_dict('Required Base Settings', 'Resolution')['value'] = 'quad'
        self.assertEqual(self.LFM_params._LFM__resolution_string(), '106x96x128-144')
        self.LFM_params.get_variable_dict('Required Base Settings', 'Resolution')['value'] = '123x456x789-0'
        self.assertEqual(self.LFM_params._LFM__resolution_string(), '123x456x789-0')

if __name__ == '__main__':
    import doctest
    doctest.testmod()
    unittest.main()

