from Model import Model

import datetime
import os
import unittest

#FIXME:  When coupled via CMIT, F107 flux should be input once and shared between both MIX and TIEGCM

##########################################################################################

class MIX(Model):
    """ MIX Model Type """
    def __init__(self):
        filename = os.path.join(os.path.dirname(__file__), "MIX.yml")
        Model.__init__(self, filename)
        # Make sure the YAML file is for MIX:
        
        assert( self.metadata["model_name"] == "MIX" )
        assert( self.metadata["output_prefix"] == "MIX" )
        assert( self.metadata["output_suffix"] == "param" )        

    def get_model_setup(self, model_name, has_spinup, run_identifier, start_datetime):
        shScript = ['# MIX']
        execName = self.get_run_instructions(model_name, 0)[1]
        shScript.append('cp ${LTRROOT}/build/bin/%s .' % (execName))
        shScript.append('')

        instructions = []

        return ('\n'.join(shScript), '\n'.join(instructions))

    def get_iteration_setup(self, model_name, iteration):
        """
        Setup environment to run a model, ends with newline when not empty
        """
        if (model_name == "LFM-MIX"):
            self.get_variable_dict("INTERCOMM PARAMETERS", "XJD_NAME")["value"] = "lfm-mix.xjd"
        elif (model_name == "CMIT"):
            self.get_variable_dict("INTERCOMM PARAMETERS", "XJD_NAME")["value"] = "cmit.xjd"
        elif (model_name == "LFM-RCM"):
            self.get_variable_dict("INTERCOMM PARAMETERS", "XJD_NAME")["value"] = "lfm-mix-rcm.xjd"
        else:
            raise Exception, "Do not understand model_name \"" + model_name + "\""
            
        return ""

    def get_iteration_teardown(self, model_name, iteration):
        """
        Teardown environment to run a model, ends with newline when not empty
        """
        return ""

    def get_run_instructions(self, model_name, iteration):
        """
        Get the executable, inputs & outputs needed to run the model.

        For example usage, see 'test_get_run_instructions'.
        
        Note: Required function (standardized parameters & return variables)        
        """
        executable = ""
        flags = ""
        stdin = "/dev/null"
        stdout_and_stderr = ""
        
        if ( (model_name == "LFM-MIX") | (model_name == "CMIT") | (model_name == "LFM-RCM") ):
            executable = "MIX"
            flags = "MIX-%03d.param" % (iteration)
            stdout_and_stderr = "MIX-%03d.out" % (iteration)            
        else:
            raise Exception, "ERROR:  Did not understand model \"" + model_name + "\""

        return (False, executable, flags, stdin, stdout_and_stderr)
        

    def get_input_file(self):
        """
        Get contents of MIX parameter file

        Returns:
          str_script: String of job script

        Note: Required function (standardized parameters & return variables)
        """
        ret = []
        variable_types_sorted = self.parameters.keys()
        variable_types_sorted.reverse()
        for name in variable_types_sorted:
            ret.append( "" )
            ret.append( "============= " + name + " ==========================" )
            
            for variable in self.parameters[name]:
                # Get a dictionary of elements of the variable
                variable_name = variable.keys()[0]
                variable_dict = variable[variable_name]

                ret.append( variable_name + "\t\t" + str(variable_dict["value"]) )
                                
            ret.append( "=============================================================" )

        return( "\n".join(ret) )

    def set_shared_parameters(self, coupled_model_name, shared_model, env=None):
        """
        Note: Required function (standardized parameters & return variables)
        """        
        self.set_shared_model_parameters(shared_model.get_variable_dict("SHARED", "RUN_IDENTIFIER")["value"],
                                         shared_model.get_variable_dict("SHARED", "LFM_MIX_EXCHANGE_FREQUENCY")["value"],
                                         shared_model.get_variable_dict("SHARED", "TIME_DUMP_INTERVAL")["value"],
                                         shared_model.get_variable_dict("SHARED", "USE_INTERCOMM")["value"])

        # Set the rcm coupling start time.
        if (coupled_model_name == "CMIT"):
            # Shared parameter sets whitespace delimited date/time string.  MIX requires dash delimited!
            
            # Whitespace delimited (2008 1 1 0 0 0):
            dateTime = shared_model.get_variable_dict("SHARED", "START_YMDHMS")["value"]
            # Dash delimited (2008-1-1-0-0-0):
            dashDateTime = "-".join(dateTime.split())
            
            self.get_variable_dict("ITM MODEL PARAMETERS","ITM_COUPLING_START")["value"] = dashDateTime
        

    def set_shared_model_parameters(self, run_prefix, exchange_frequency, dump_interval, use_intercomm):
        """
        Setup internal structure of generic MIX shared parameters

        This sets two parameters defined in MIX.yml:
          * OUTPUT_FILE_PREFIX: string to prepend to each MIX output file.
          * FLUSH_FREQUENCY:  Number of LFM-MIX exchanges between file dumps (set to match the 'dumpInterval').
          * USE_INTERCOMM: Is coupling Adhoc or via InterComm?

        Parameters:
          run_prefix: string to prepend to LFM output files.
          exchange_frequency: Number of seconds between exchanges with LFM
          dump_interval: Number of seconds between file dumps
          use_intercomm: do we use InterComm (True) or adhoc (False) coupling?
        """
        self.get_variable_dict("INTERCOMM PARAMETERS","USE_INTERCOMM")["value"] = use_intercomm

        self.get_variable_dict("IO PARAMETERS","OUTPUT_FILE_PREFIX")["value"] = run_prefix
        self.get_variable_dict("IO PARAMETERS","FLUSH_FREQUENCY")["value"] = int( float(dump_interval) / float(exchange_frequency) )

    def set_time_information(self, start_datetime, stop_datetime):
        """
        Set start time (i.e. to point to the correct MHD_GRID_FILE_NAME).

        This sets one parameters defined in MIX.yml:
          * MHD_GRID_FILE_NAME:  to something like 'CMIT_mhd_YYYY-MM-DDTHH-MM-SSZ.hdf'

        Parameters:
          start_datetime: datetime object specifying the starting
            year-month-day-hour-minute-second for this segment.
          stop_datetime:  Unused.  Stop time determined by LFM kill signal.

        Note: Required function (standardized parameters & return variables)
        """
        # MHD_GRID_FILE_NAME:
        filename = []
        filename.append( self.get_variable_dict("IO PARAMETERS","OUTPUT_FILE_PREFIX")["value"] )
        filename.append( "_mhd_" )
        filename.append( "%04d-%02d-%02d" % (start_datetime.year, start_datetime.month, start_datetime.day) )
        filename.append( "T%02d-%02d-%02d" % (start_datetime.hour, start_datetime.minute, start_datetime.second) )
        filename.append( "Z.hdf" )

        self.get_variable_dict("MHD MODEL PARAMETERS","MHD_GRID_FILE_NAME")["value"] = "".join(filename)

class TestMIX(unittest.TestCase):
    def setUp(self):
        self.MIX_params = MIX()

    def test_set_shared_parameters(self):
        import MakeItSo.Models.Shared as Shared
        sharedParameters = Shared()

        self.assertEqual( self.MIX_params.get_variable_dict("ITM MODEL PARAMETERS","ITM_COUPLING_START")["value"], '99999-5-19-16-0-0' )
        self.MIX_params.set_shared_parameters("CMIT", sharedParameters)
        self.assertEqual( self.MIX_params.get_variable_dict("ITM MODEL PARAMETERS","ITM_COUPLING_START")["value"], '2000-1-1-0-0-0' )                                   

    def test_set_shared_model_parameters(self):
        self.MIX_params.set_shared_model_parameters("TestCase", 10.0, 60.0, True)
        # exchange with LFM every 10 seconds.
        # dump every 60 seconds

        self.assertEqual(self.MIX_params.get_variable_dict("INTERCOMM PARAMETERS","USE_INTERCOMM")["value"], True)

        self.assertEqual(self.MIX_params.get_variable_dict("IO PARAMETERS","OUTPUT_FILE_PREFIX")["value"], "TestCase")
        self.assertEqual(self.MIX_params.get_variable_dict("IO PARAMETERS","FLUSH_FREQUENCY")["value"], 6)
        

    def test_set_time_information(self):
        run_prefix = "TestCase"
        segment_start_time = datetime.datetime(2009,1,1,0,0,0) # start on the new year!

        self.MIX_params.set_time_information(segment_start_time, None)

        self.assertEqual(self.MIX_params.get_variable_dict("MHD MODEL PARAMETERS","MHD_GRID_FILE_NAME")["value"], "ChangeName_mhd_2009-01-01T00-00-00Z.hdf")

    def test_get_model_setup(self):
        (shScript, instructions) = self.MIX_params.get_model_setup('LFM-MIX',
                                                                   has_spinup=True,
                                                                   run_identifier='test',
                                                                   start_datetime=datetime.datetime(year=2000,month=3,day=14,hour=3,minute=23,second=int(53.60496)))
        # start date/time is pi day (and time! 0.14159265 percentage of the day => 3:23:53.60496)

        self.assertEqual(shScript,"""# MIX\ncp ${LTRROOT}/build/bin/MIX .\n""")
        self.assertEqual(instructions,'')

    def test_get_iteration_setup(self):
        self.assertEqual("", self.MIX_params.get_iteration_setup("LFM-MIX", 0))
        self.assertEqual(self.MIX_params.get_variable_dict("INTERCOMM PARAMETERS", "XJD_NAME")["value"], "lfm-mix.xjd")

        self.assertEqual("", self.MIX_params.get_iteration_setup("CMIT", 0))
        self.assertEqual(self.MIX_params.get_variable_dict("INTERCOMM PARAMETERS", "XJD_NAME")["value"], "cmit.xjd")
        
        self.assertEqual("", self.MIX_params.get_iteration_setup("LFM-RCM", 0))
        self.assertEqual(self.MIX_params.get_variable_dict("INTERCOMM PARAMETERS", "XJD_NAME")["value"], "lfm-mix-rcm.xjd")

        self.assertRaises(Exception,      self.MIX_params.get_iteration_setup, "Purple-Monkey-Dishwasher", 0 )


    def test_get_iteration_teardown(self):
        self.assertEqual("", self.MIX_params.get_iteration_teardown("Test", 0))

    def test_get_run_instructions(self):
        (isParallel, executable, flags, stdin, stdout_and_stderr) = self.MIX_params.get_run_instructions("CMIT", 2)
        self.assertEqual(isParallel, False)
        self.assertEqual(executable, "MIX")
        self.assertEqual(flags, "MIX-002.param")
        self.assertEqual(stdin, "/dev/null")
        self.assertEqual(stdout_and_stderr, "MIX-002.out")

        (isParallel, executable, flags, stdin, stdout_and_stderr) = self.MIX_params.get_run_instructions("LFM-MIX", 200)
        self.assertEqual(executable, "MIX")
        self.assertEqual(flags, "MIX-200.param")
        self.assertEqual(stdin, "/dev/null")
        self.assertEqual(stdout_and_stderr, "MIX-200.out")

if __name__ == '__main__':
    import doctest
    doctest.testmod()
    unittest.main()
