"""
Individual model parameters for LFM, MIX, TIEGCM, and any variables shared between the models.
"""

# Standard import
import copy
import datetime
import unittest

# 3rd-party imports
from MakeItSo import yaml
from MakeItSo import share
import helpers

##########################################################################################

class Model(object):
    """ Abstract base class for code models """

    def __init__(self, filename):
        """
        Open the YAML file (filename) and read the parameters
        dictionary.  Pop off the __metadata__ key/value pair.

        This init function should be explicitly called from any
        derived class.
        """
        yaml_file = open(filename)
        self.parameters = yaml.safe_load(yaml_file)
        yaml_file.close()

        # pop of the metadata element.  This makes it easier to loop over keys later on.
        self.metadata = self.parameters.pop("__metadata__")

        self.__typecastParameters()

        self.parameters_defaults = copy.deepcopy(self.parameters)

    def __typecastParameters(self):
        """
        yaml.safe_load attempts to guess the data type of each
        node.  Let's explicitly type-cast each variable based
        on the 'type' specified in each variable's dictionary.
        """
        var_types_sorted = self.parameters.keys()
        var_types_sorted.reverse()
        for name in var_types_sorted:
            for i,variable in enumerate(self.parameters[name]):                
                varName = variable.keys()[0]
                varDict = variable[varName]

                # Convert to the prescribed data type if variable has a 'value' and 'type'.
                if varDict.has_key('value'):
                    if varDict.has_key('type'):
                        self.parameters[name][i][varName]['value'] = share.typeCast(str(varDict['value']), varDict['type'])
                    else:
                        if Release.verboseDebug:
                            print('WARNING:  Variable "' + varName + '" does not define a data type.')
        

    # Derived classes must implement the following functions:
    def get_model_setup(self, model_name, has_spinup, run_identifier, start_datetime):
        raise NotImplementedError
    def get_iteration_setup(self, model_name, iteration):
        raise NotImplementedError
    def get_iteration_teardown(self, model_name, iteration):
        raise NotImplementedError
    def get_run_instructions(self, model_name, iteration):
        raise NotImplementedError
    def get_input_file(self):
        raise NotImplementedError
    def set_shared_parameters(self, shared_model, env=None):
        raise NotImplementedError
    def set_time_information(self, start_datetime, stop_datetime):
        raise NotImplementedError

    def get_variable_index(self, variable_type, variable_name):
        """
        Returns the index into self.parameters['variable_type']
        correspndoing to the 'variable_name' dictionary.  Returns -1
        if the variable isn't found.
        
        self.parameters is a dictionary.  Each key is a variable_type.
        The variable_type value is a list of variables.
        """
        for idx,variable in enumerate(self.parameters[variable_type]):
            if variable_name == variable.keys()[0]:
                return idx

        raise LookupError, 'Could not find variable "' + variable_name + '" of type "' + variable_type + '"."'
            
                            
    def get_variable_dict(self, variable_type, variable_name):        
        """
        Search the parameters dictionary for `variable_name`, which is
        of `variable_type`.

        This function uses a 'soft copy' of (akin to a C pointer to)
        the variable, and hence this function can be used to both get
        and set variables.
        """

        varIdx = self.get_variable_index(variable_type,variable_name)

        return self.parameters[variable_type][varIdx][variable_name]
        
        
    def get_spinup_times(self, eventStart, hoursPerSegment):
        """
        Get segments required for model spinup (ending at 'eventStart' datetime).
        
        Parameters:
          eventStart: datetime requested by user to start model
            execution for a particular event.          
          hoursPerSegment: Maximum number of simulation hours to
            compute for a given segment. If the model computes
            reasonably fast, set this to sys.maxint.
          
        Returns:
          spinup_start:  list of datetime objects corresponding to start date & time for model spinup segments.
          spinup_stop:  list of datetime objects corresponding to stop date & time for model spinup segments.
          spinup_dt:  List of float objects corresponding to the time dump interval for the spinup
          
        FIXME:  There's a peculiar error here:  Given certain CMIT inputs:

                 - START_YMDHMS: 1996 5 19 16 0 0
                 - STOP_YMDHMS: 1996 5 20 6 0 0
                 - TIME_DUMP_INTERVAL: 120.0                 

                this function returns a spinup_dt where at least one
                element of the list are datetime.timedelta objects
                rather than a floating-point number of seconds.  I
                suspect this happens because it jumps from returning
                number of seconds to either number of minutes or
                hours.
        """
        spinup_start = []
        spinup_stop  = []
        spinup_dt    = []

        if not self.parameters.has_key('Required Base Settings'):
            return ( spinup_start, spinup_stop, spinup_dt )

        try:
            time_str = self.get_variable_dict('Required Base Settings', 'spinup')['value']
        except LookupError:
            return ( spinup_start, spinup_stop, spinup_dt )
        
        spinupTimeDelta = datetime.timedelta( days = int(time_str.split()[0]),
                                              hours = int(time_str.split()[1]),
                                              minutes = int(time_str.split()[2]),
                                              seconds = int(time_str.split()[3]) )

        spinupStart = eventStart - spinupTimeDelta
        spinupStop = eventStart


        #NOTE: we set dt=0.0 because we will calculate the dump
        #interval based on the number and length of segments.        
        (spinup_start, spinup_stop) = helpers.getSegmentTimes(spinupStart.strftime('%Y %m %d %H %M %S'),
                                                              spinupStop.strftime('%Y %m %d %H %M %S'),
                                                              0.0, hoursPerSegment)

        # Figure out dt based on segments calculated
        for i in range(len(spinup_start)):
            dt = spinup_stop[i] - spinup_start[i]
            spinup_dt.append(dt.days*24.0*60.0*60.0 + dt.seconds)

        return ( spinup_start, spinup_stop, spinup_dt )

    def get_spinup_coupled_models(self):
        coupled_models = []
        if self.metadata.has_key("spinup_couple_with"):
            models = self.metadata["spinup_couple_with"].split(",")
            for model_name in models:
                coupled_models.append(model_name.strip())

        return coupled_models
