import MakeItSo.share
from Model import Model

import datetime
import math
import os
import unittest

#FIXME: spinup period should use mpi with 8 processors
#FIXME: When coupled via CMIT F 10.7 should be shared between MIX and TIEGCM

##########################################################################################

class TIEGCM(Model):
    """ TIEGCM Model  Type """
    def __init__(self):
        filename = os.path.join(os.path.dirname(__file__), "TIEGCM.yml")
        Model.__init__(self, filename)
        # Make sure the YAML file is for the TIEGCM:
        
        assert( self.metadata["model_name"] == "TIEGCM" )
        assert( self.metadata["output_prefix"] == "tiegcm1.9" )
        assert( self.metadata["output_suffix"] == "inp" )

        # These are used when "Required Base Settings" parameter
        # "FILENAME_CONVENTION" is set to old-school integer-based
        # filenames.
        # Member variables are used to generate OUTPUT and SECOUT values.
        self.__pTimeIdx = [0,0]
        self.__sTimeIdx = [0,0]

        # TIEGCM Continuation Runs require that the "SOURCE" and
        # "SOURCE_START" paramaters are set for the first iteration.
        # Subsequent iterations grab the start time & input file from
        # "START" and the first file in "OUTPUT".  This flag
        # determines if SOURCE and SOURCE_START need to be set.
        self.__isFirstIteration = True
        self.__isSourceRemoved = False
        self.__isSourceStartRemoved = False

        self.__sourceFilename = None


    def __get_filename_from_date(self, prefix, date):
        """
        Get the filename for TIEGCM input/output of the form:
        
           prefix_tie_YYYY-MM-DDThh-mm-ss.nc
        
        Note: We dont name the file using minutes or seconds, since
        each TIEGCM NetCDF history file may contain multiple timesteps
        """
        filename = []
        filename.append( prefix + "_tie_" )
        filename.append( "%04d-%02d-%02dT%02d-%02d-%02d" % (date.year, date.month, date.day,
                                                            date.hour, date.minute, date.second) )
        filename.append( ".nc" )
        return( "".join(filename) )
                                                 

    def __get_filename_from_dates(self, prefix, startDate, stopDate):
        """
        Get the filename for TIEGCM input/output of the form:

           prefix_tie_YYYY-MM-DDThh-mm-ss_YYYY-MM-DDThh-mm-ss.nc

        Note: We dont name the file using minutes or seconds, since
        each TIEGCM NetCDF history file may contain multiple timesteps
        per file.

        Function parametes:
          prefix: string
          date: datetime object
        """
        filename = []
        filename.append( prefix + "_tie_" )
        filename.append( "%04d-%02d-%02dT%02d-%02d-%02d" % (startDate.year, startDate.month, startDate.day,
                                                            startDate.hour, startDate.minute, startDate.second) )
        filename.append( "_" )
        filename.append( "%04d-%02d-%02dT%02d-%02d-%02d" % (stopDate.year, stopDate.month, stopDate.day,
                                                            stopDate.hour, stopDate.minute, stopDate.second) )
        filename.append( ".nc" )
        return( "".join(filename) )

    def __get_filename_from_iteration(self, prefix, iteration, type='secondary'):
        """
        Get the filename for TIEGCM input/output of the form:
          
           prefix_p000.nc
           prefix_s000.nc
           
        Function parameters:
          prefix: string
          iteration: integer
          isPrimary: boolean (only options are primary (True) or secondary (False)          
        """
        filename = []
        filename.append(prefix)
        filename.append('-')
        if type == 'primary':
            filename.append('p')
        elif type == 'secondary':
            filename.append('s')
        else:
            raise Exception, 'ERROR:  __get_filename_from_iteration expects type to be either "secondary" or "primary"!'
        filename.append("%03d" % iteration)
        filename.append(".nc")
        return ("".join(filename) )

    def __seconds_to_days_hours_minutes(self, seconds):
        """ How many days, hours and minutes are in float(seconds)?"""
        nSeconds = float(seconds)
        nDays = 0
        while (nSeconds >= 60.0*60.0*24.0):
            nSeconds -= 60.0*60.0*24.0
            nDays += 1
        nHours = 0
        while ( nSeconds >= 60.0*60.0 ):
            nSeconds -= 60.0*60.0
            nHours += 1
        nMinutes = 0
        while ( nSeconds >= 60.0):
            nSeconds -= 60.0
            nMinutes += 1

        return (nDays, nHours, nMinutes)

    def __getSource(self, startTime):
        """
        Returns SOURCE and SOURCE_START strings for a run-from-scratch via
          - 'Required Base Settings':
           - 'SOLAR FORCING': minimum, medium or maximum
        start date & time (December Solstice, June Solstice or Spring Equinox)

        Example SOURCE values are:
        
           TGCM.tiegcm1.94.pcntr_decsol_smax.nc
           TGCM.tiegcm1.94.pcntr_decsol_smin.nc
           TGCM.tiegcm1.94.pcntr_junsol_smax.nc
           TGCM.tiegcm1.94.pcntr_junsol_smin.nc
           TGCM.tiegcm1.94.pcntr_mareqx_smax.nc
           TGCM.tiegcm1.94.pcntr_mareqx_smin.nc
           TGCM.tiegcm1.94.pcntr_sepeqx_smax.nc
           TGCM.tiegcm1.94.pcntr_sepeqx_smin.nc

        Returns tuple containing:
          - source: filename
          - source_start: day-of-year, hour-of-day, minute-of-day
            mapping to source filename.

        See "test_getSource" unittest for examples.

        FIXME: This will probably need to be updated with future
        releases of TIEGCM... the version number is encoded directly
        to the filename!
        """
        tiegcm193 = {'basename': 'TGCM.tiegcm1.93.pcntr_',
                     'seasons': {'march':     {'seasonStr': 'eqnx', 'dateStr': '80,0,0'},
                                 'june':      {'seasonStr': 'jsol', 'dateStr': '172,0,0'},
                                 'september': {'seasonStr': 'eqnx', 'dateStr': '80,0,0'},
                                 'december':  {'seasonStr': 'dsol', 'dateStr': '355,0,0'}
                                  }
                      }
        tiegcm194 = {'basename': 'TGCM.tiegcm1.94.pcntr_',
                     'seasons': {'march':     {'seasonStr': 'mareqx', 'dateStr': '80,0,0'},
                                 'june':      {'seasonStr': 'junsol', 'dateStr': '172,0,0'},
                                 'september': {'seasonStr': 'sepeqx', 'dateStr': '264,0,0'},
                                 'december':  {'seasonStr': 'decsol', 'dateStr': '355,0,0'}
                                  }
                      }

        tiegcmVer = tiegcm194
        
        ### Solar Forcing
        forcing = self.get_variable_dict("Required Base Settings", "SOLAR FORCING")["value"].lower()
        if ( (forcing == 'min') | (forcing == 'minimum') ):
            forcing = 'smin'
        elif ( (forcing == 'med') | (forcing == 'medium') ):
            forcing = 'smin'
        elif ( (forcing == 'max') | (forcing == 'maximum') ):
            forcing = 'smax'
        else:
            raise Exception, 'Did not understand solar forcing.  Expected "minimum", "medium" or "maximum".  Got "%s"' % (forcing)

        ### Season
        dayOfYear = int( startTime.strftime('%j') )
        # Each season is roughly 91 days long.  So dates within 46 (91
        # days / 2) belong to that season.
        if ( (34 <= dayOfYear) & (dayOfYear < 126) ):
            # March Equinox
            month = 'march'
        elif ( (126 <= dayOfYear) & (dayOfYear < 218) ):
            # June Solstice
            month = 'june'
        elif ( (218 <= dayOfYear) & (dayOfYear < 309) ):
            # September Equinox
            month = 'september'
        elif ( (309 <= dayOfYear) | (dayOfYear < 34) ):
            # December Solstice
            month = 'december'
        else:
            raise Exception, 'Did not understand Day of Year for this run.  Expected a value in [0,366].  Got %d' % (dayOfYear)

        # Construct the "SOURCE" filename
        source = '%s%s_%s.nc' % (tiegcmVer['basename'],
                                 tiegcmVer['seasons'][month]['seasonStr'],
                                 forcing)
        # Get the date string (day-of-year,hour,minute).
        source_start = tiegcmVer['seasons'][month]['dateStr']

        return (source,source_start)

    def __getTiegcmTimeTuples(self, startDateTime, stopDateTime):
        """
        Returns string 'DOY,HH,MM' where DOY = day of year; HH=hour; MM=minute

        >>> t=TIEGCM()
        >>> t._TIEGCM__getTiegcmTimeTuples(datetime.datetime(2012,6,12,14,0,0), datetime.datetime(2012,6,12,16,0,0))
        ('164,14,0','164,16,0')

        # not a leap year
        >>> t=TIEGCM()
        >>> t._TIEGCM__getTiegcmTimeTuples(datetime.datetime(2001,12,31,0,0,0), datetime.datetime(2002,1,1,0,0,0)
        ('365,0,0','366,0,0')

        # leap year
        >>> t=TIEGCM()
        >>> t._TIEGCM__getTiegcmTimeTuples(datetime.datetime(2000,12,31,0,0,0), datetime.datetime(2001,1,1,0,0,0)
        ('366,0,0','367,0,0')
        """
        startTimeTuple  = str( startDateTime.timetuple().tm_yday ) + ","
        startTimeTuple += str( startDateTime.hour ) + ","
        startTimeTuple += str( startDateTime.minute )

        stopTimeTuple  = str( stopDateTime.timetuple().tm_yday ) + ","
        stopTimeTuple += str( stopDateTime.hour ) + ","
        stopTimeTuple += str( stopDateTime.minute )

        if stopTimeTuple == '1,0,0':
            stopTimeTuple=str( (stopDateTime-datetime.timedelta(1)).timetuple().tm_yday+1 ) + ',0,0'

        return (startTimeTuple, stopTimeTuple)

    def __setPrimaryHistoryParameters(self, prefix, start, stop):
        """
        Sets "PRIMARY HISTORY FILE PARAMETERS":
          - START (DayOfYear,Hour,Minute)          
          - STOP (DayOfYear,Hour,Minute)
          - HIST: Frequency to save primary history files
          - MXHIST_PRIM (Integer)
          - OUTPUT: list of filenames ('prefix_tie_YYYY-MM-DDThh.nc','prefix_tie_YYYY-MM-DDThh.nc',... or "'prefix_p001.nc','to','prefix_p020.nc','by','1'")

        Note: OUTPUT needs to set filenames so the first file in the
              list was created as the last file from a previous
              iteration.
        
        FIXME: Some primary history parameters are hard-coded.  For
               example, one primary history is saved at the beginning & end
               iteration in a sequence.
        """
        #
        # Primary history output parameters:
        # Note that we have a single primary history file at the start & stop of segment.
        #
        (start_string, stop_string) = self.__getTiegcmTimeTuples(start,stop)
        self.get_variable_dict("PRIMARY HISTORY FILE PARAMETERS","START")["value"] = start_string
        self.get_variable_dict("PRIMARY HISTORY FILE PARAMETERS","STOP")["value"] = stop_string

        primary_filenames = ''
        if int(self.get_variable_dict('Required Base Settings','FILENAME_CONVENTION')['value']) == 1:
            ####################################################################
            # time-based file names: One file per segment
            ####################################################################
            dt = stop - start
            (nDays, nHours, nMinutes) = self.__seconds_to_days_hours_minutes( dt.days*24.0*(60.0*60.0) + dt.seconds )
            hist_string  = str(nDays) + "," + str(nHours) + "," + str(nMinutes)
            self.get_variable_dict("PRIMARY HISTORY FILE PARAMETERS","HIST")["value"] = hist_string

            primary_filenames += "'" + self.__get_filename_from_date(prefix, start) + "',"
            primary_filenames += "'" + self.__get_filename_from_date(prefix, stop) + "'"
            self.get_variable_dict("PRIMARY HISTORY FILE PARAMETERS","MXHIST_PRIM")["value"] = int(1)
        else:
            raise NotImplementedError, "Integer-filenames is disabled.  We haven't fully tested it yet."
            ####################################################################
            # Integer-based filenames (aka old-school style): One file per day
            ####################################################################
            self.__pTimeIdx[0] = self.__pTimeIdx[1]

            ####################################################################
            #FIXME: Start hard-coded values
            # Hard-code primary history time cadence.  Choose the
            # lesser of daily primary histories or matching up with
            # the sequence start/stop.

            dt = stop - start
            (nDays, nHours, nMinutes) = self.__seconds_to_days_hours_minutes( dt.days*24.0*(60.0*60.0) + dt.seconds )

            if ( (nHours != 0) &  (nMinutes != 0) ):
                hist_string = "1,0,0"
            else:
                hist_string  = str(nDays) + "," + str(nHours) + "," + str(nMinutes)

            self.get_variable_dict("PRIMARY HISTORY FILE PARAMETERS","HIST")["value"] = hist_string
            self.get_variable_dict("PRIMARY HISTORY FILE PARAMETERS","MXHIST_PRIM")["value"] = "10"
            #FIXME: End hard-coded values
            ####################################################################

            elapsedTime = (stop-start)
            elapsedTime = elapsedTime.days * 24.0*60.0*60.0 + elapsedTime.seconds

            dt_str = self.get_variable_dict("PRIMARY HISTORY FILE PARAMETERS","HIST")["value"]
            dt_day = float(dt_str.split(",")[0])
            dt_hour = float(dt_str.split(",")[1])
            dt_minute = float(dt_str.split(",")[2])
            
            n_total_histories = elapsedTime / (dt_day*24.0*60.0*60.0 + dt_hour*60.0*60.0 + dt_minute*60.0)
            n_histories_per_file = int(self.get_variable_dict("PRIMARY HISTORY FILE PARAMETERS","MXHIST_PRIM")["value"])

            self.__pTimeIdx[1] += n_total_histories / n_histories_per_file
            self.__pTimeIdx[1] = math.floor(self.__pTimeIdx[1])
    
            if (self.__pTimeIdx[0] == self.__pTimeIdx[1]):
                primary_filenames = "'" + self.__get_filename_from_iteration(prefix, self.__pTimeIdx[0], 'primary') + "'"
            else:
                primary_filenames  = "'" 
                primary_filenames += self.__get_filename_from_iteration(prefix, self.__pTimeIdx[0], 'primary')
                primary_filenames += "','to','"
                primary_filenames += self.__get_filename_from_iteration(prefix, self.__pTimeIdx[1], 'primary')
                primary_filenames += "','by','1'"
            self.__pTimeIdx[1] += 1

        self.get_variable_dict("PRIMARY HISTORY FILE PARAMETERS","OUTPUT")["value"] = primary_filenames
        

    def __setSecondaryHistoryParameters(self, prefix, startDate, stopDate):
        """
        Sets "SECONDARY HISTORY FILE PARAMETERS":
          - SECSTART (DayOfYear,Hour,Minute)
          - SECSTOP (DayOfYear,Hour,Minute)
          - MXHIST_SECH (Integer)
          - SECOUT list of filenames ('prefix_sech_tie_YYYY-MM-DDThh.nc','prefix_sech_tie_YYYY-MM-DDThh.nc',... or or "'prefix_s001.nc','to','prefix_s020.nc','by','1'"))

        Thanks to a 'feature' with TIEGCM, the first TIEGCM timestep
        at 'start' cannot be saved to a secondary history file.  So we
        save the next time value at start + dt.
        """

        # Dump file time cadence:
        dt_str = self.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","SECHIST")["value"]
        dt_day = float(dt_str.split(",")[0])
        dt_hour = float(dt_str.split(",")[1])
        dt_minute = float(dt_str.split(",")[2])

        # SECSTART & STOP
        # Data is not set for the first secondary history, so let's skip it (ie. "+timedelta(...)").
        start = startDate + datetime.timedelta(days=dt_day, hours=dt_hour, minutes=dt_minute)
        stop = stopDate
        (start_string, stop_string)  = self.__getTiegcmTimeTuples(start, stop)
        self.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","SECSTART")["value"] = start_string
        self.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","SECSTOP")["value"] = stop_string

        if (start > stop):
            errorString  = 'Calculated start time "'
            errorString += start.strftime("%a, %d %b %Y %H:%M:%S")
            errorString += '" exceeds requested stop time "'
            errorString += stop.strftime("%a, %d %b %Y %H:%M:%S")
            errorString += '" Likely cause: TIME_DUMP_INTERVAL="' + dt_str + '"'
            errorString += ' exceeds simulation run time for one or more time segments in a simulation.'

            raise Exception, errorString

        #
        # Compute a list of output history filenames
        #
        secondary_filenames = ""
        if int(self.get_variable_dict('Required Base Settings','FILENAME_CONVENTION')['value']) == 1:
            ####################################################################
            # time-based file names    
            ####################################################################

            if (start == stop):
                # filename should look something like: 'CMITs_sech_tie_YYYY-MM-DDTHH-MM-SS.nc'                
                secondary_filenames ="\'" + self.__get_filename_from_date(prefix+'_sech', start) + "\'"
            else:
                # filename should look something like: 'CMITs_sech_tie_YYYY-MM-DDTHH-MM-SS_YYYY-MM-DDTHH-MM-SS.nc'
                secondary_filenames = "\'" + self.__get_filename_from_dates(prefix+'_sech', start, stop) +"\'"

            dt_segment = stopDate-startDate
            dt_sechist = datetime.timedelta( days=int(dt_day), hours=int(dt_hour), minutes=int(dt_minute) )

            n = math.floor( (dt_segment.days*24*60*60 + dt_segment.seconds) / (dt_sechist.days*24*60*60 + dt_sechist.seconds) )            
            self.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","MXHIST_SECH")["value"] = int(n)
        else:            
            raise NotImplementedError, "Integer-filenames is disabled.  We haven't fully tested it yet."
            ####################################################################
            # Integer-based filenames (aka old-school style)
            ####################################################################
            self.__sTimeIdx[0] = self.__sTimeIdx[1]
            
            elapsedTime = (stop-start)
            elapsedTime = elapsedTime.days * 24.0*60.0*60.0 + elapsedTime.seconds
            n_total_histories = elapsedTime / (dt_day*24.0*60.0*60.0 + dt_hour*60.0*60.0 + dt_minute*60.0)
            n_histories_per_file = int(self.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","MXHIST_SECH")["value"])
            
            self.__sTimeIdx[1] += n_total_histories / n_histories_per_file
            self.__sTimeIdx[1] = math.floor(self.__sTimeIdx[1])
    
            if (self.__sTimeIdx[0] == self.__sTimeIdx[1]):
                secondary_filenames = "'" + self.__get_filename_from_iteration(prefix, self.__sTimeIdx[0], 'secondary') + "'"
            else:
                secondary_filenames  = "'" 
                secondary_filenames += self.__get_filename_from_iteration(prefix, self.__sTimeIdx[0], 'secondary')
                secondary_filenames += "','to','"
                secondary_filenames += self.__get_filename_from_iteration(prefix, self.__sTimeIdx[1], 'secondary')
                secondary_filenames += "','by','1'"
            self.__sTimeIdx[1] += 1

        self.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","SECOUT")["value"] = secondary_filenames
        

    def get_spinup_times(self, eventStart, hoursPerSegment):
        """
        Overloaded Model::get_spinup_times(...) that ensures the
        initial starting time is 00:00:00 for any given
        Year/Month/Day.

        Note: this routine adds spinup times when crossing a year boundary.

        Parameters:
          eventStart:  datetime used for starting a TIE-GCM run
          hoursPerSegment: ignored & overwritten to a unreasonably
            large number (65536=2^16=2730 days).  See
            Model::get_spinup_times for details.

        Returns:
          spinup_start:  list of datetime objects corresponding to start date & time for model spinup segments.
          spinup_stop:  list of datetime objects corresponding to stop date & time for model spinup segments.
        """
        (spinup_start, spinup_stop, spinup_dt) = Model.get_spinup_times(self, eventStart, 65536)

        # Should output once per day:
        spinup_dt.pop()
        spinup_dt.append(24.0*60.0*60.0)
        
        # Initial datafile must start at hour=0, minutes=0
        if ( (spinup_start[0].hour != 0) |
             (spinup_start[0].minute != 0) |
             (spinup_start[0].second != 0) ):
            
            spinup_stop.reverse()
            spinup_stop.append(spinup_start[0])
            spinup_stop.reverse()

            tmp_start = spinup_start[0]
            
            spinup_start.reverse()
            spinup_start.append(datetime.datetime(year=tmp_start.year,
                                                  month=tmp_start.month,
                                                  day=tmp_start.day))
            spinup_start.reverse()
        
            spinup_dt.append(datetime.timedelta(hours=1))
            spinup_dt.reverse()

        # Does start/stop time cross a year boundary?
        startYear = spinup_start[0].year
        endYear = spinup_stop[len(spinup_stop)-1].year
        if startYear < endYear:
            jan1=datetime.datetime(endYear, 1, 1, 0, 0, 0)

            for i,d in enumerate(spinup_stop):
                if (d.year == endYear):
                    spinup_start.insert(i+1, jan1)
                    spinup_stop.insert(i, jan1)
                    spinup_dt.append(24.0*60.0*60.0)                    
                    break

        return (spinup_start, spinup_stop, spinup_dt)

    def get_spinup_coupled_model_name(self, coupled_model_name):
        """
        Get the coupled model name required for a spinup iteration.
        For CMIT, this should be stand-alone TIEGCM.
        """
        return "TIEGCM"

    def get_model_setup(self, model_name, has_spinup, run_identifier, start_datetime):
        """
        Create script & documentation for the user on how to finish
        setup & start to execute the run.
        
        parameters:
          model_name:  Name of (coupled) model (ie. TIEGCM, CMIT, LFM-MIX, etc)
          has_spinup:  Does this run include a model spinup (ie. 20 days for TIEGCM)
          run_identifier: string corresponding to this run.
          start_datetime: First time step of the run.

        Returns: 
          string shScript: Shell script that configures this model.
          string instructions: Specific instructions or notes on running this model.
          
        FIXME:  In the future, provide support for 2.5 or 5 degree resolution.
        FIXME:  In the future, allow for MPI or serial tiegcm.
        """
        shScript = ['# TIEGCM']

        # Copy the executable:
        if (model_name == 'TIEGCM'):
            shScript.append('cp ${LTRROOT}/build/bin/TIEGCM .')
        elif (model_name == 'CMIT'):
            shScript.append('cp ${LTRROOT}/build/bin/TIEGCM .')
            shScript.append('cp ${LTRROOT}/build/bin/TIEGCM-MIX .')
        else:
            raise Exception, "ERROR:  Did not understand model \"" + model_name + "\""            

        # Startup file configuration
        instructions = ['\tTIEGCM:']

        if has_spinup:            
            (self.__sourceFilename, source_start) = self.__getSource(start_datetime)
            instructions.append('\t   Guessed initial TIEGCM SOURCE file "%s".' % self.__sourceFilename)
            instructions.append('\t   Make sure you have downloaded the TIEGCM data directory and set the $TGCMDATA environment variable.')
            shScript.append('\n\n')
            shScript.append('# TIEGCM startup file')
            shScript.append('cp ${TGCMDATA}/%s .' % self.__sourceFilename)
        else:
            instructions.append('\t   It looks like you are running without a spinup interval.  You must obtain a TIEGCM primary history to start the model with.')
            shScript.append('')
            shScript.append('echo "TIEGCM warning: No spinup interval selected. Make sure to copy a startup file for TIEGCM!"')
            shScript.append('')
            
        instructions.append('\t   See the documentation for more information.')
        instructions.append('\t   https://wiki.ucar.edu/display/LTR/TGCMDATA')

        return ('\n'.join(shScript), '\n'.join(instructions))

    def get_iteration_setup(self, model_name, iteration):
        """
        Setup environment to run a model, ends with newline when not empty
        """
        return ""

    def get_iteration_teardown(self, model_name, iteration):
        """
        Teardown environment to run a model, ends with newline when not empty
        """

        # After the first iteration, we need to pop SOURCE and
        # SOURCE_START to do a proper continuation run.
        if not self.__isFirstIteration:
            if not self.__isSourceRemoved:
                sourceIndex = self.get_variable_index("RUN INFORMATION", "SOURCE")
                self.parameters["RUN INFORMATION"].pop(sourceIndex)
                self.__isSourceRemoved = True

            if not self.__isSourceStartRemoved:
                source_startIndex = self.get_variable_index("RUN INFORMATION", "SOURCE_START")
                self.parameters["RUN INFORMATION"].pop(source_startIndex)
                self.__isSourceStartRemoved = True
        
        return ""

    def get_run_instructions(self, model_name, iteration):
        """
        Get the executable, inputs & outputs needed to run the model.

        For example usage, see 'test_get_run_instructions'.

        Note: Required function (standardized parameters & return variables)
        """
        executable = ""
        flags = ""
        stdin = "/dev/null"
        stdout_and_stderr = ""
        
        if (model_name == "TIEGCM"):
            executable = "TIEGCM"
            stdin = "tiegcm1.9-%03d.inp" % (iteration)
            stdout_and_stderr = "TIEGCM-%03d.out" % (iteration)
        elif (model_name == "CMIT"):
            executable = "TIEGCM-MIX"
            stdin = "tiegcm1.9-%03d.inp" % (iteration)
            stdout_and_stderr = "TIEGCM-MIX-%03d.out" % (iteration)
        else:
            raise Exception, "ERROR:  Did not understand model \"" + model_name + "\""

        return (False, executable, flags, stdin, stdout_and_stderr)


    def get_input_file(self):
        """
        Get contents of TIEGCM input namelist

        Returns:
          str_script: String of job script

        Note: Required function (standardized parameters & return variables)
        """
        ret = []
        ret.append( "&tgcm_input" )
        ret.append( "; Auto-generated at <" + str(datetime.datetime.now()) + ">" )
        variable_types_sorted = self.parameters.keys()
        variable_types_sorted.reverse()
        for name in variable_types_sorted:

            if name == 'Required Base Settings':
                continue
            
            ret.append( "" )
            ret.append( "; " + name )
            for variable in self.parameters[name]:
                variable_name = variable.keys()[0]
                variable_dict = variable[variable_name]

                variableString = " " + variable_name + " = "
                # Strings must be surrounded by single quotes
                if ( (variable_dict["type"] == "string") &
                     (str(variable_dict["value"]).count("'") < 2) ):
                    variableString += "'" + str(variable_dict["value"]) + "'"
                else:
                    variableString += str(variable_dict["value"])
                    
                if not variable_dict.has_key('status'):
                    ret.append( variableString )
                elif (variable_dict['status'] != 'disabled'):
                    # Can be "disabled", "enabled" or "mandatory"
                    ret.append( variableString )

        ret.append( "/" )
        ret.append( "" )

        return "\n".join(ret)
                                        
    def set_shared_parameters(self, coupled_model_name, shared_model, env=None):
        """
        Note: Required function (standardized parameters & return variables)        
        """

        if (coupled_model_name == 'TIEGCM'):
            self.set_shared_model_parameters(shared_model.get_variable_dict("SHARED", "RUN_IDENTIFIER")["value"],
                                             shared_model.get_variable_dict("SHARED", "TIME_DUMP_INTERVAL")["value"])
        elif (coupled_model_name == 'CMIT'):
            self.set_shared_model_parameters(shared_model.get_variable_dict("SHARED", "RUN_IDENTIFIER")["value"],
                                             self.get_variable_dict("Required Base Settings", "CMIT TIME DUMP INTERVAL")["value"])
        else:
            raise Exception, "ERROR:  Did not understand model \"" + model_name + "\""            
            


    def set_shared_model_parameters(self,label, dump_interval):
        """
        Setup internal structure of generic TIEGCM shared parameters

        This sets some shared parameters defined in TIEGCM.yml:
         * LABEL
         * SECHIST: Frequency to save secondary history files

        Function Parameters:
          label: string
          dump_interval: floating-point number of seconds between dumps
        """
        self.get_variable_dict("RUN INFORMATION","LABEL")["value"] = "'" + label + "'"
    
        (nDays, nHours, nMinutes) = self.__seconds_to_days_hours_minutes(dump_interval)
        self.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","SECHIST")["value"] = str(nDays) + "," + str(nHours) + "," + str(nMinutes)

    def set_time_information(self, start, stop):
        """
        This function sets the following TIEGCM parameters based on function input parameters.
          * START_YEAR (4-digit integer) 
          * START_DAY (day of year)
          * SOURCE & SOURCE_START (only for the first iteration!)
            * SOURCE  (ie. 'TGCM.tiegcm1.94.pcntr_decsol_smin.nc')
            * SOURCE_START (ie. '355,0,0' == 'DayOfYear,Hour,Minute')
          * START (DayOfYear,Hour,Minute)
          * STOP (DayOfYear,Hour,Minute)
          * OUTPUT: list of filenames ('prefix_tie_YYYY-MM-DDThh.nc','prefix_tie_YYYY-MM-DDThh.nc',... or "'prefix_p001.nc','to','prefix_p020.nc','by','1'")
          * MXHIST_PRIM (integer)
          * SECSTART (DayOfYear,Hour,Minute)
          * SECSTOP (DayOfYear,Hour,Minute)
          * SECOUT list of filenames ('prefix_sech_tie_YYYY-MM-DDThh.nc','prefix_sech_tie_YYYY-MM-DDThh.nc',... or or "'prefix_s001.nc','to','prefix_s020.nc','by','1'"))
          * MXHIST_SECH (integer)

        Note: This function contains a considerable amount of
        business logic to determine the primary & secondary history
        file names.

        Function input parameters:
          start: segment start time
          stop: segment stop time

        Note: Required function (standardized parameters & return variables)

        FIXME:  This could be heavily refactored to keep it DRY. Especially the bits with TIEGCM filename setting (OUTPUT, SECOUT, etc.).
        FIXME:  TIEGCM primary history filenames (OUTPUT) are hard-coded by MXHIST_PRIM=10 when the "old school" file setting is chosen.
        """
        
        # Year & Day of Year to start the run
        self.get_variable_dict("RUN INFORMATION","START_YEAR")["value"] = start.year
        self.get_variable_dict("RUN INFORMATION","START_DAY")["value"] = start.timetuple().tm_yday

        # First iteration in a sequence requires a source primary
        # history file obtained via the TGCMDATA distribution.
        if self.__isFirstIteration:
            (self.__sourceFilename, source_start) = self.__getSource(start)
            self.get_variable_dict("RUN INFORMATION","SOURCE")["value"] = "'%s'" % self.__sourceFilename
            self.get_variable_dict("RUN INFORMATION","SOURCE_START")["value"] = source_start
            self.__isFirstIteration = False

        # Set the primary & secondary history file parameters.
        filePrefix = self.get_variable_dict("RUN INFORMATION","LABEL")["value"]
        filePrefix = filePrefix.replace("'","")

        self.__setPrimaryHistoryParameters(filePrefix, start, stop)
        self.__setSecondaryHistoryParameters(filePrefix, start, stop)
            

class TestTIEGCM(unittest.TestCase):
    def setUp(self):
        self.TIEGCM_params = TIEGCM()

    def test__get_filename_from_date(self):
        filename = self.TIEGCM_params._TIEGCM__get_filename_from_date("test",datetime.datetime(2009,1,1,0,0,0))
        self.assertEqual(filename, "test_tie_2009-01-01T00-00-00.nc")

    def test__seconds_to_days_hours_minutes(self):
        d120 = datetime.datetime(1996, 4, 29, 16, 0)
        d140 = datetime.datetime(1996, 5, 19, 16, 0)
        dt = d140-d120
        (nDays, nHours, nMinutes) = self.TIEGCM_params._TIEGCM__seconds_to_days_hours_minutes(dt.days*24.0*(60.0*60.0) + dt.seconds)
        self.assertEqual(nDays, 20)
        self.assertEqual(nHours, 0)
        self.assertEqual(nMinutes, 0)        
        
        (nDays, nHours, nMinutes) = self.TIEGCM_params._TIEGCM__seconds_to_days_hours_minutes(4*24.0*(60.0*60.0) + 8*60.0*(60.0) + 15*60.0)
        self.assertEqual(nDays, 4)
        self.assertEqual(nHours, 8)
        self.assertEqual(nMinutes, 15)
        
        (nDays, nHours, nMinutes) = self.TIEGCM_params._TIEGCM__seconds_to_days_hours_minutes(23*24.0*(60.0*60.0) + 42*60.0*(60.0) + 108*60.0)
        self.assertEqual(nDays, 24)
        self.assertEqual(nHours, 19)
        self.assertEqual(nMinutes, 48)

    def test__getSource(self):
        d50 = datetime.datetime(1996, 2, 19, 16, 0)
        self.assertEqual('050', d50.strftime('%j'))
        (source, source_start) = self.TIEGCM_params._TIEGCM__getSource(d50)
        self.assertEqual(source, 'TGCM.tiegcm1.94.pcntr_mareqx_smin.nc')
        self.assertEqual(source_start, '80,0,0')
        
        d140 = datetime.datetime(1996, 5, 19, 16, 0)
        (source, source_start) = self.TIEGCM_params._TIEGCM__getSource(d140)
        self.assertEqual(source, 'TGCM.tiegcm1.94.pcntr_junsol_smin.nc')
        self.assertEqual(source_start, '172,0,0')

        d261 = datetime.datetime(1996, 9, 17, 16, 0)
        (source, source_start) = self.TIEGCM_params._TIEGCM__getSource(d261)
        self.assertEqual(source, 'TGCM.tiegcm1.94.pcntr_sepeqx_smin.nc')        
        self.assertEqual(source_start, '264,0,0')

        d329 = datetime.datetime(1996, 11, 24, 16, 0)
        (source, source_start) = self.TIEGCM_params._TIEGCM__getSource(d329)
        self.assertEqual(source, 'TGCM.tiegcm1.94.pcntr_decsol_smin.nc')        
        self.assertEqual(source_start, '355,0,0')


    def test__setPrimaryHistoryParmaeters(self):
        self.TIEGCM_params._TIEGCM__setPrimaryHistoryParameters('cmitTest',
                                                                datetime.datetime(2012,5,3,0,0,0),
                                                                datetime.datetime(2012,5,3,2,0,0))

        self.assertEqual("124,0,0", self.TIEGCM_params.get_variable_dict("PRIMARY HISTORY FILE PARAMETERS","START")["value"])
        self.assertEqual("124,2,0", self.TIEGCM_params.get_variable_dict("PRIMARY HISTORY FILE PARAMETERS","STOP")["value"])
        self.assertEqual("0,2,0",self.TIEGCM_params.get_variable_dict("PRIMARY HISTORY FILE PARAMETERS","HIST")["value"])
        self.assertEqual(1,self.TIEGCM_params.get_variable_dict("PRIMARY HISTORY FILE PARAMETERS","MXHIST_PRIM")["value"])
        self.assertEqual("'cmitTest_tie_2012-05-03T00-00-00.nc','cmitTest_tie_2012-05-03T02-00-00.nc'",self.TIEGCM_params.get_variable_dict("PRIMARY HISTORY FILE PARAMETERS","OUTPUT")["value"])

    def test__setSecondaryHistoryParmaeters(self):
        self.TIEGCM_params._TIEGCM__setSecondaryHistoryParameters('cmitTest',
                                                                  datetime.datetime(2012,5,3,0,0,0),
                                                                  datetime.datetime(2012,5,3,2,0,0))

        self.assertEqual("124,0,30", self.TIEGCM_params.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","SECSTART")["value"])
        self.assertEqual("124,2,0", self.TIEGCM_params.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","SECSTOP")["value"])
        self.assertEqual("0,0,30",self.TIEGCM_params.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","SECHIST")["value"])
        self.assertEqual(4,self.TIEGCM_params.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","MXHIST_SECH")["value"])
        self.assertEqual( "'cmitTest_sech_tie_2012-05-03T00-30-00_2012-05-03T02-00-00.nc'",self.TIEGCM_params.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","SECOUT")["value"])

    def test__leapYearBoundary(self):
        self.TIEGCM_params._TIEGCM__setPrimaryHistoryParameters('cmitTest',
                                                                datetime.datetime(2000,12,30,0,0,0),
                                                                datetime.datetime(2001,1,1,0,0,0))
        self.assertEqual("365,0,0", self.TIEGCM_params.get_variable_dict("PRIMARY HISTORY FILE PARAMETERS","START")["value"])
        self.assertEqual("367,0,0", self.TIEGCM_params.get_variable_dict("PRIMARY HISTORY FILE PARAMETERS","STOP")["value"])
        self.assertEqual("2,0,0",self.TIEGCM_params.get_variable_dict("PRIMARY HISTORY FILE PARAMETERS","HIST")["value"])
        self.assertEqual(1,self.TIEGCM_params.get_variable_dict("PRIMARY HISTORY FILE PARAMETERS","MXHIST_PRIM")["value"])
        self.assertEqual("'cmitTest_tie_2000-12-30T00-00-00.nc','cmitTest_tie_2001-01-01T00-00-00.nc'",self.TIEGCM_params.get_variable_dict("PRIMARY HISTORY FILE PARAMETERS","OUTPUT")["value"])

        self.TIEGCM_params._TIEGCM__setSecondaryHistoryParameters('cmitTest',
                                                                  datetime.datetime(2000,12,30,0,0,0),
                                                                  datetime.datetime(2001,1,1,0,0,0))
        self.assertEqual("365,0,30", self.TIEGCM_params.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","SECSTART")["value"])
        self.assertEqual("367,0,0", self.TIEGCM_params.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","SECSTOP")["value"])
        self.assertEqual("0,0,30",self.TIEGCM_params.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","SECHIST")["value"])
        self.assertEqual(96,self.TIEGCM_params.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","MXHIST_SECH")["value"])
        self.assertEqual( "'cmitTest_sech_tie_2000-12-30T00-30-00_2001-01-01T00-00-00.nc'",self.TIEGCM_params.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","SECOUT")["value"])

    def test__yearBoundary(self):
        self.TIEGCM_params._TIEGCM__setPrimaryHistoryParameters('cmitTest',
                                                                datetime.datetime(2001,12,30,0,0,0),
                                                                datetime.datetime(2002,1,1,0,0,0))
        self.assertEqual("364,0,0", self.TIEGCM_params.get_variable_dict("PRIMARY HISTORY FILE PARAMETERS","START")["value"])
        self.assertEqual("366,0,0", self.TIEGCM_params.get_variable_dict("PRIMARY HISTORY FILE PARAMETERS","STOP")["value"])
        self.assertEqual("2,0,0",self.TIEGCM_params.get_variable_dict("PRIMARY HISTORY FILE PARAMETERS","HIST")["value"])
        self.assertEqual(1,self.TIEGCM_params.get_variable_dict("PRIMARY HISTORY FILE PARAMETERS","MXHIST_PRIM")["value"])
        self.assertEqual("'cmitTest_tie_2001-12-30T00-00-00.nc','cmitTest_tie_2002-01-01T00-00-00.nc'",self.TIEGCM_params.get_variable_dict("PRIMARY HISTORY FILE PARAMETERS","OUTPUT")["value"])

        self.TIEGCM_params._TIEGCM__setSecondaryHistoryParameters('cmitTest',
                                                                  datetime.datetime(2001,12,30,0,0,0),
                                                                  datetime.datetime(2002,1,1,0,0,0))
        self.assertEqual("364,0,30", self.TIEGCM_params.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","SECSTART")["value"])
        self.assertEqual("366,0,0", self.TIEGCM_params.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","SECSTOP")["value"])
        self.assertEqual("0,0,30",self.TIEGCM_params.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","SECHIST")["value"])
        self.assertEqual(96,self.TIEGCM_params.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","MXHIST_SECH")["value"])
        self.assertEqual("'cmitTest_sech_tie_2001-12-30T00-30-00_2002-01-01T00-00-00.nc'",self.TIEGCM_params.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","SECOUT")["value"])


    def test__setSecondaryHistoryParmaeters_single_timestep(self):
        self.TIEGCM_params._TIEGCM__setSecondaryHistoryParameters('cmitTest',
                                                                  datetime.datetime(2012,5,3,1,30,0),
                                                                  datetime.datetime(2012,5,3,2,0,0))
        # SECSTART is 30 minutes after simulation start time because
        # the TIEGCM doesn't have secondary history variables
        # calculated at startup.
        self.assertEqual("124,2,0", self.TIEGCM_params.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","SECSTART")["value"])
        self.assertEqual("124,2,0", self.TIEGCM_params.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","SECSTOP")["value"])
        self.assertEqual("0,0,30",self.TIEGCM_params.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","SECHIST")["value"])
        self.assertEqual(1,self.TIEGCM_params.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","MXHIST_SECH")["value"])
        # Note how the filename changes:
        self.assertEqual( "'cmitTest_sech_tie_2012-05-03T02-00-00.nc'",self.TIEGCM_params.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","SECOUT")["value"])


    def test_get_spinup_times(self):
        # Test crossing the year boundary: 20 days prior to 1/1/2009. Note: 2008 is a leap year!
        (start, stop, dt) = self.TIEGCM_params.get_spinup_times(datetime.datetime(2009,1,2,0,0,0), 65536)
        self.assertEqual(len(start), 2)
        self.assertEqual(start[0], datetime.datetime(2008, 12, 13, 0, 0))
        self.assertEqual(start[1], datetime.datetime(2009, 1, 1, 0, 0))
        
        self.assertEqual(len(stop), 2)
        self.assertEqual(stop[0], datetime.datetime(2009, 1, 1, 0, 0))
        self.assertEqual(stop[1], datetime.datetime(2009, 1, 2, 0, 0))

        self.assertEqual(len(dt), 2)
        self.assertEqual(dt[0], 86400.0)
        self.assertEqual(dt[1], 86400.0)

    def test_get_spinup_coupled_model_name(self):
        self.assertEqual(self.TIEGCM_params.get_spinup_coupled_model_name("TIEGCM"), "TIEGCM")        

    def test_get_input_file(self):
        """
        Make sure the default input file matches the expected input
        """
        # Default input file as a string:
        inputFile = self.TIEGCM_params.get_input_file()
        # Expected input file as a string:
        gigantic_string="""&tgcm_input
; Auto-generated at <2011-05-05 10:19:40.908674>

; RUN INFORMATION
 LABEL = 'tiegcm1.9'
 START_YEAR = 2000
 START_DAY = 80
 CALENDAR_ADVANCE = 1
 STEP = 60
 SOURCE = 'TIEGCM_data_file.nc'
 SOURCE_START = 80,0,0

; PRIMARY HISTORY FILE PARAMETERS
 START = 80,0,0
 STOP = 80,0,0
 HIST = 0,0,30
 OUTPUT = 'myoutput_p001.nc','to','myoutput_p020.nc','by','1'
 MXHIST_PRIM = 16

; SECONDARY HISTORY FILE PARAMETERS
 SECSTART = 80,0,0
 SECSTOP = 80,0,0
 SECHIST = 0,0,30
 SECOUT = 'myoutput_p001.nc','to','myoutput_p020.nc','by','1'
 MXHIST_SECH = 16
 SECFLDS = 'TN','UN','VN','O2','O1','NO','NO_COOL','OP','TI','TE','NE','O2P','OMEGA','Z','POTEN','UI_ExB','VI_ExB','WI_ExB','TEC','QJI_TN','gpot','gpotm','geng','gflx','gzigm1','gzigm2','gnsrhs'

; MODEL PARAMETERS
 AURORA = 1
 COLFAC = 1.5

; POTENTIAL
 POTENTIAL_MODEL = 'HEELIS'
 GPI_NCFILE = '$TGCMDATA/gpi_2000001-2011059.nc'

; IMF Parameters

; GSWM
 TIDE = 0.,0.,0.,0.,0.,0.,0.,0.,0.,0.
 TIDE2 = 0.,0.
 GSWM_MI_DI_NCFILE = '$TGCMDATA/gswm_diurn_5.0d_99km.nc'
 GSWM_MI_SDI_NCFILE = '$TGCMDATA/gswm_semi_5.0d_99km.nc'
/
"""

        # The "auto-generated" date will change.  Use regex to grab the non-date parts of the string:
        import re
        r1 = re.match('([\s\S\d\D\w\W]+)\<\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d+\>([\s\S\d\D\w\W]+)',
                    inputFile)
        r2 = re.match('([\s\S\d\D\w\W]+)\<\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d+\>([\s\S\d\D\w\W]+)',
                    gigantic_string)

        #... make sure non-date parts match!
        MakeItSo.share.unittestAssertBigStrings(self, r1.groups()[0], r2.groups()[0])
        MakeItSo.share.unittestAssertBigStrings(self, r1.groups()[1], r2.groups()[1])

    def test_shared_model_parameters(self):
        self.TIEGCM_params.set_shared_model_parameters("TestCase", 3*24.0*(60.0*60.0) + 5*60.0*(60.0) + 10*60.0)

        self.assertEqual(self.TIEGCM_params.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","SECHIST")["value"], "3,5,10")

    def test_set_time_information(self):
        self.TIEGCM_params.set_time_information(datetime.datetime(2009,1,1,0,0,0), datetime.datetime(2009,1,1,4,0,0))
        self.assertEqual(self.TIEGCM_params.get_variable_dict("RUN INFORMATION","START_YEAR")["value"], 2009)
        self.assertEqual(self.TIEGCM_params.get_variable_dict("RUN INFORMATION","START_DAY")["value"], 1)
        self.assertEqual(self.TIEGCM_params.get_variable_dict("RUN INFORMATION","SOURCE")["value"], "'TGCM.tiegcm1.94.pcntr_decsol_smin.nc'")
        self.assertEqual(self.TIEGCM_params.get_variable_dict("RUN INFORMATION","SOURCE_START")["value"], "355,0,0")

        self.assertEqual(self.TIEGCM_params.get_variable_dict("PRIMARY HISTORY FILE PARAMETERS","START")["value"], "1,0,0")
        self.assertEqual(self.TIEGCM_params.get_variable_dict("PRIMARY HISTORY FILE PARAMETERS","STOP")["value"], "1,4,0")
        self.assertEqual(self.TIEGCM_params.get_variable_dict("PRIMARY HISTORY FILE PARAMETERS","HIST")["value"], "0,4,0")
        self.assertEqual(self.TIEGCM_params.get_variable_dict("PRIMARY HISTORY FILE PARAMETERS","OUTPUT")["value"], "'tiegcm1.9_tie_2009-01-01T00-00-00.nc','tiegcm1.9_tie_2009-01-01T04-00-00.nc'")
        self.assertEqual(self.TIEGCM_params.get_variable_dict("PRIMARY HISTORY FILE PARAMETERS","MXHIST_PRIM")["value"], 1)
        
        self.assertEqual(self.TIEGCM_params.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","SECSTART")["value"], "1,0,30")
        self.assertEqual(self.TIEGCM_params.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","SECSTOP")["value"], "1,4,0")
        self.assertEqual(self.TIEGCM_params.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","SECHIST")["value"], "0,0,30")
        self.assertEqual(self.TIEGCM_params.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","SECOUT")["value"], "'tiegcm1.9_sech_tie_2009-01-01T00-30-00_2009-01-01T04-00-00.nc'")
        self.assertEqual(self.TIEGCM_params.get_variable_dict("SECONDARY HISTORY FILE PARAMETERS","MXHIST_SECH")["value"], 8)

    def test_get_model_setup(self):
        model='CMIT'
        configureSpinup=True
        run_id='test'
        startDate=datetime.datetime(year=2000,month=3,day=14,hour=3,minute=23,second=int(53.60496))

        #self.assertRaises(Exception, self.TIEGCM_params.get_model_setup,  model,configureSpinup,run_id,startDate)
        #(self.TIEGCM_params._TIEGCM__sourceFilename, source_start) = self.TIEGCM_params._TIEGCM__getSource(startDate)
        (shScript, instructions) = self.TIEGCM_params.get_model_setup(model,configureSpinup,run_id,startDate)

        # start date/time is pi day (and time! 0.14159265 percentage of the day => 3:23:53.60496)

        MakeItSo.share.unittestAssertBigStrings(self, shScript, 
                                                """# TIEGCM\ncp ${LTRROOT}/build/bin/TIEGCM .\ncp ${LTRROOT}/build/bin/TIEGCM-MIX .\n\n\n\n# TIEGCM startup file\ncp ${TGCMDATA}/TGCM.tiegcm1.94.pcntr_mareqx_smin.nc .\n""")

        MakeItSo.share.unittestAssertBigStrings(self, instructions,
                                                """\tTIEGCM:\n\t   Guessed initial TIEGCM SOURCE file "TGCM.tiegcm1.94.pcntr_mareqx_smin.nc".\n\t   Make sure you have downloaded the TIEGCM data directory and set the $TGCMDATA environment variable.\n\t   See the documentation for more information.\n\t   https://wiki.ucar.edu/display/LTR/TGCMDATA""")

    def test_get_iteration_setup(self):
        self.assertEqual("", self.TIEGCM_params.get_iteration_setup("Test", 0))
                        
    def test_get_iteration_setup(self):
        self.assertEqual("", self.TIEGCM_params.get_iteration_teardown("Test", 0))

    def test_get_run_instructions(self):
        (isParallel, executable, flags, stdin, stdout_and_stderr) = self.TIEGCM_params.get_run_instructions("CMIT", 2)
        self.assertEqual(isParallel, False)
        self.assertEqual(executable, "TIEGCM-MIX")
        self.assertEqual(flags, "")
        self.assertEqual(stdin, "tiegcm1.9-002.inp")
        self.assertEqual(stdout_and_stderr, "TIEGCM-MIX-002.out")

        (isParallel, executable, flags, stdin, stdout_and_stderr) = self.TIEGCM_params.get_run_instructions("TIEGCM", 200)
        self.assertEqual(executable, "TIEGCM")
        self.assertEqual(flags, "")
        self.assertEqual(stdin, "tiegcm1.9-200.inp")
        self.assertEqual(stdout_and_stderr, "TIEGCM-200.out")

        self.assertRaises(Exception,     self.TIEGCM_params.get_run_instructions, "supercalafragalisticexpialadoshus", 2)


if __name__ == '__main__':
    import doctest
    doctest.testmod()
    unittest.main()

