from Model import Model

import datetime
import os
import unittest
import xml.dom.minidom

class XJD(Model):
    """
    InterComm XML Job Description file

    Since we only need one or two XJD files for a whole run interval,
    most functions do nothing.
    """

    def __init__(self):
        """
        Instantiate model/parameter type
        """
        filename = os.path.join(os.path.dirname(__file__), "XJD.yml")
        Model.__init__(self, filename)

        self.output_directory = ""

    def get_model_setup(self, model_name, has_spinup, run_identifier, start_datetime):
        return ('', '')
            
    def get_iteration_setup(self, model_name, iteration):
        """
        Setup environment to run a model, ends with newline when not empty
        """
        
        if (model_name == "CMIT"):
            self.write_xjd_file("lfm-mix.xjd", ["cpl","mhd"])
            self.write_xjd_file("cmit.xjd", ["cpl","mhd", "itm"])
        elif (model_name == "LFM-MIX"):
            self.write_xjd_file("lfm-mix.xjd", ["cpl","mhd"])
	elif (model_name == "LFM-RCM"):
            self.write_xjd_file("lfm-mix-rcm.xjd", ["cpl","mhd","im"])
	
        return ""

    def write_xjd_file(self, filename, components):
        """
        Write an XJD file with a particular set of components
        
        Parameters:
          filename: write XJD file to here
          components:  list of components usef for coupled models

        Example:
          write_xjd_file('LFM-MIX-RCM.xjd', ['mhd','cpl','im'])
        """

        xml_xjd = self.get_xjd_file(components)

        input_file = open(os.path.join(self.output_directory, filename), "w")

        #FIXME: Due to a bug with InterComm, do not print whitespace (" ", "\n", "\t") between xml <tag></tag>'s.
        #       ... i.e. use "xml.dom.minidom.Document::toxml" instead of "xml.dom.minidom.Document::toprettyxml"
        #input_file.write( xml_xjd.toprettyxml(indent="  ", encoding="utf-8") + "\n")
        input_file.write( xml_xjd.toxml(encoding="utf-8") + "\n")
        input_file.close()    

    def get_xjd_file(self, components):
        # Create the minidom document
        xmlDoc = xml.dom.minidom.Document()

        # <ICXJD> 
        xmlICXJD = xmlDoc.createElement("ICXJD")
        xmlDoc.appendChild(xmlICXJD)
        
        # <version>
        xmlVersion = xmlDoc.createElement("version")
        xmlVersion.appendChild(xmlDoc.createTextNode("1.5"))
        xmlICXJD.appendChild(xmlVersion)
        
        # <components>
        xmlComponents = self.__get_xml_components_(xmlDoc, components)
        xmlICXJD.appendChild(xmlComponents)

        # <connections>
        xmlConnections = self._get_xml_connections_(xmlDoc, components)
        xmlICXJD.appendChild(xmlConnections)

        return xmlDoc

    def __get_xml_components_(self, xmlDoc, components):
        """
        Returns string containing coupled model <components/> that exchange data.

        Parameters:
          xmlDoc: xml.dom.minidom.Document() object
          components:  list of components usef for coupled models

        Returns:
          xml.dom.minidom.Document() object containing XML specifications for all components.

        Example of XML returned:
        <components>
          <component>
            <id>mhd</id>
            <name>lfm</name>
            <nNode>8</nNode>
            <arrayorder>IC_COLUMN_MAJOR</arrayorder>
          </component>       
        </components>

        """
        xmlComponents = xmlDoc.createElement("components")
        # Setup each <component>
        for component in self.parameters["components"]:
            # Obtain useful information for the component
            component_name = component.keys()[0]
            component_dict = component[component_name]
            
            idValue = component_name
            nameValue = component_dict["value"]
            nNodeValue = component_dict["nNode"]
            arrayOrderValue = component_dict["arrayorder"]

            # We don't care about components that aren't part of the
            # current coupled model type.
            if ( components.count(component_name) <= 0 ):
                continue

            # Create XML to store component.
            xmlComponent = xmlDoc.createElement("component")

            xmlComponentId = xmlDoc.createElement("id")
            xmlComponentId.appendChild(xmlDoc.createTextNode( idValue ))
            xmlComponent.appendChild(xmlComponentId)
        
            xmlComponentName = xmlDoc.createElement("name")
            xmlComponentName.appendChild(xmlDoc.createTextNode( nameValue ))
            xmlComponent.appendChild(xmlComponentName)

            xmlComponentNNode = xmlDoc.createElement("nNode")
            xmlComponentNNode.appendChild(xmlDoc.createTextNode( str(nNodeValue) ))
            xmlComponent.appendChild(xmlComponentNNode)

            xmlComponentArrayOrder = xmlDoc.createElement("arrayorder")
            xmlComponentArrayOrder.appendChild(xmlDoc.createTextNode( arrayOrderValue ))
            xmlComponent.appendChild(xmlComponentArrayOrder)

            xmlComponents.appendChild(xmlComponent)
        return xmlComponents

    def _get_xml_connections_(self, xmlDoc, components):
        """
        Returns string containing one <connection>...</connection>
        containing exchange information.

        Parameters:
          xmlDoc: xml.dom.minidom.Document() object
          components:  list of components usef for coupled models

        Returns:
          xml.dom.minidom.Document() object containing XML specifications for all connections.

        Example of XML returned:

        <connections>
          <connection>
            <id>mhd-cpl_scalars</id>
            <type>double</type>
            <commtype>1xN</commtype>
            <msgtag>1000</msgtag>
            <exporter>mhd</exporter>
            <exportport>mhd-cpl_scalars</exportport>
            <importer>cpl</importer>
            <importport>mhd-cpl_scalars</importport>     
          </connection>
        </connections>
        """
        xmlConnections = xmlDoc.createElement("connections")

        # Setup each <connection>
        for connection in self.parameters["connections"]:
            # Obtain useful information for the connection
            connection_name = connection.keys()[0]
            connection_dict = connection[connection_name]
            
            idValue = connection_name
            typeValue = connection_dict["type"]
            commTypeValue = connection_dict["commtype"]
            msgTagValue = connection_dict["msgtag"]
            exporterValue = connection_dict["exporter"]
            exportPortValue = connection_name
            importerValue = connection_dict["importer"]
            importPortValue = connection_name
            
            # We only care about connections between components
            # defined for the current coupled model type
            if ( (components.count(importerValue) <= 0) |
                 (components.count(exporterValue) <= 0) ):
                continue

            # Create XML to store connection.
            xmlConnection = xmlDoc.createElement("connection")

            xmlConnectionId = xmlDoc.createElement("id")
            xmlConnectionId.appendChild(xmlDoc.createTextNode( idValue ))
            xmlConnection.appendChild(xmlConnectionId)
        
            xmlConnectionType = xmlDoc.createElement("type")
            xmlConnectionType.appendChild(xmlDoc.createTextNode( typeValue ))
            xmlConnection.appendChild(xmlConnectionType)

            xmlConnectionCommType = xmlDoc.createElement("commtype")
            xmlConnectionCommType.appendChild(xmlDoc.createTextNode( commTypeValue ))
            xmlConnection.appendChild(xmlConnectionCommType)

            xmlConnectionMsgTag = xmlDoc.createElement("msgtag")
            xmlConnectionMsgTag.appendChild(xmlDoc.createTextNode( str(msgTagValue) ))
            xmlConnection.appendChild(xmlConnectionMsgTag)

            xmlConnectionExporter = xmlDoc.createElement("exporter")
            xmlConnectionExporter.appendChild(xmlDoc.createTextNode( exporterValue ))
            xmlConnection.appendChild(xmlConnectionExporter)

            xmlConnectionExportPort = xmlDoc.createElement("exportport")
            xmlConnectionExportPort.appendChild(xmlDoc.createTextNode( exportPortValue ))
            xmlConnection.appendChild(xmlConnectionExportPort)

            xmlConnectionImporter = xmlDoc.createElement("importer")
            xmlConnectionImporter.appendChild(xmlDoc.createTextNode( importerValue ))
            xmlConnection.appendChild(xmlConnectionImporter)

            xmlConnectionImportPort = xmlDoc.createElement("importport")
            xmlConnectionImportPort.appendChild(xmlDoc.createTextNode( importPortValue ))
            xmlConnection.appendChild(xmlConnectionImportPort)

            xmlConnections.appendChild(xmlConnection)

        return xmlConnections

    
    def get_iteration_teardown(self, model_name, iteration):
        """
        Teardown environment to run a model, ends with newline when not empty
        """        
        return ""

    def get_run_instructions(self, model_name, iteration):
        return (False, "", "", "", "")

    def get_input_file(self):
        return ""

    def set_shared_parameters(self, coupled_model_name, shared_model, env = None):
        """
        Note: Required function (standardized parameters & return variables)        
        """
        if env == None:
            raise Exception, "XJD file requires an Environment object."
        # Set number of processors for the "mhd" component (typically mhd == LFM).
        cpuString = str( env.get_variable_dict("flags", "n_processors")["value"] )
        self.get_variable_dict("components", "mhd")["nNode"] = env.get_num_cpus( cpuString )

        self.output_directory = shared_model.get_variable_dict("SHARED", "Output directory")["value"]

    def set_time_information(self, start_datetime, stop_datetime):
        None

class TestXJD(unittest.TestCase):
    def setUp(self):
        self.params = XJD()

    def test_get_model_setup(self):
        (shScript, instructions) = self.params.get_model_setup('LFM-RCM',
                                                               has_spinup=True,
                                                               run_identifier='test',
                                                               start_datetime=datetime.datetime(year=2000,month=3,day=14,hour=3,minute=23,second=int(53.60496)))
        # start date/time is pi day (and time! 0.14159265 percentage of the day => 3:23:53.60496)
        self.assertEqual(shScript,'')
        self.assertEqual(instructions,'')

    def test_set_shared_parameters(self):
        # setting shared parameters requires an Environment object to get the number of processors.
        self.assertRaises(Exception,    self.params.set_shared_parameters, None)

if __name__ == '__main__':
    import doctest
    doctest.testmod()
    unittest.main()
