from MakeItSo import yaml

import datetime
import math
import os
import unittest

def getModelCoupling():
    """
    Get dictionary of queueing environments
    """
    dirname = os.path.dirname(__file__)
    models_file = open(os.path.join(dirname, "models.yml"))
    models_dict = yaml.safe_load(models_file)
    models_file.close()

    return models_dict["Models"]


def getSegmentTimes(firstYMDHMS_str, lastYMDHMS_str, dt, hoursPerSegment):
    """
    Calculates & returns two lists that represent the start & stop
    times for each segment.

    Parmaeters:
      - firstYMDHMS_str: string 'YYYY MM DD HH MM SS' to represent
          beginning time.  Compare to shared model parameter
          ['SHARED']['START_YMDHMS'].
      - lastYMDMHS_str: string 'YYYY MM DD HH MM SS' to represent
          end time.  Compare to shared model parameter
          ['SHARED']['STOP_YMDHMS'].
      - dt: number of simulation seconds between saving a file.  Compare
           to shared model parameter ['SHARED']['TIME_DUMP_INTERVAL'].
           Special case when (dt == 0.0): skip error checking on
           whether the number of segments divide evenly by dt.
           Useful for spinup runs where time synchronization
           between multiple models is unimportant (ie. LFM-MIX spinup
           for a CMIT run).           
      - hoursPerSegment: Number of simulation hours per segment.
          Compare to shared model parameter
          ['SHARED']['SIMULATION_HOURS_PER_SEGMENT'].

    Returns:
      - segment_start_times: array of datetime objects
      - segment_stop_times: array of datetime objects

    FIXME: CMIT/TIEGCM has problems crossing the year boundary.  This
    routine should probably check if the coupled model requested
    includes TIEGCM, then add en extra segment ending/starting January
    1 the next year with STOP=366,0,0 (or 367,0,0 for leap year).  The
    year boundary is okay for running the TIEGCM spinup interval (see
    TIEGCM.py::get_spinup_times).  Ultimately, I decided not to
    implement this because:
      1. We don't know what coupled model type getSegmentTimes is called for
      2. How do we propagate the TIEGCM restriction to other models (ie. LFM, MIX, etc)?
    A kludge to setup a run: configure MakeItSo for a run ending
    January 1 the next year.  Then configure another run via MakeItSo
    with no spinup interval starting January 1.
    """
    # Process inputs
    input_startUT = datetime.datetime( int( firstYMDHMS_str.split()[0] ),
                                       int( firstYMDHMS_str.split()[1] ),
                                       int( firstYMDHMS_str.split()[2] ),
                                       int( firstYMDHMS_str.split()[3] ),
                                       int( firstYMDHMS_str.split()[4] ),
                                       int( firstYMDHMS_str.split()[5] )
                                       )
    
    input_stopUT = datetime.datetime( int( lastYMDHMS_str.split()[0] ),
                                      int( lastYMDHMS_str.split()[1] ),
                                      int( lastYMDHMS_str.split()[2] ),
                                      int( lastYMDHMS_str.split()[3] ),
                                      int( lastYMDHMS_str.split()[4] ),
                                      int( lastYMDHMS_str.split()[5] )
                                      )

    input_dumpInterval = float( dt )

    input_sim_hour_dt = float( hoursPerSegment )
    
    # Process start/stop/dump date & time information
    # 1:  Check that start/stop date/time & dump interval match up
    # Need to run the LFM with 50 minutes of "blow-by":
    startUT = input_startUT
    stopUT  = input_stopUT
    start_stop_difference = stopUT - startUT

    nDays  = start_stop_difference.days
    nHours = math.floor(start_stop_difference.seconds/(60.*60.))
    nMinutes = ((start_stop_difference.seconds/(60.*60.)) - nHours) * 60.0
    

    if input_dumpInterval == 0.0:
        # We'll need to calculate the dump interval as post-processing.
        pass
    else:
        nTotalMinutes = nMinutes + nHours*60.0 + nDays*(24.0*60.0)
        nDumpMinutes = input_dumpInterval / 60.0

        remainder = nTotalMinutes % nDumpMinutes
        if ( math.fabs(remainder - 0.0) > 1e-6  ):
            errorString  = "Elapsed simulation time is not divisible by the dumpInterval!"
            errorString += "\n\tstartUT:      "+repr(input_startUT)
            errorString += "\n\tstopUT:       "+repr(input_stopUT)
            errorString += "\n\tElapsed:      "+repr(nDays)+" days "+repr(nHours)+ " hours "+repr(nMinutes)+" minutes"
            errorString += "\n\tnTotalMinutes:"+repr(nTotalMinutes)
            errorString += "\n\tnDumpMinutes: "+repr(nDumpMinutes)
            errorString += "\n\tnSegments:    "+repr(nTotalMinutes/(hoursPerSegment*60.0))
            errorString += "\n\tremainder:    "+repr(remainder)+" minutes"
        
            raise Exception, errorString

    # 2:  Figure out how many segments we need:
    dt = datetime.timedelta(hours   = input_sim_hour_dt)
    next_startUT = startUT 
    next_stopUT  = startUT
    segment_start_times = []
    segment_stop_times = []
    while (next_stopUT < stopUT):
        next_stopUT  += dt
        if next_stopUT > stopUT:
            next_stopUT = stopUT
        segment_start_times.append( next_startUT)
        segment_stop_times.append( next_stopUT)
        next_startUT += dt

    return (segment_start_times, segment_stop_times)



class TestGetSegmentTimes(unittest.TestCase):
    def test_standard(self):
        (starts,stops) = getSegmentTimes("2009 1 1 0 0 0", "2009 1 1 2 0 0", 120.0 , 0.5)
        
        self.assertEqual( len(starts), len(stops) )
        self.assertEqual( len(starts), 4)


    def test_slowCpu(self):
        (starts,stops) = getSegmentTimes("2009 1 1 0 0 0", "2009 1 1 2 0 0", 120.0 , 0.1666666667)

        self.assertEqual( len(starts), len(stops) )
        self.assertEqual( len(starts), 12 )

    def test_fastCpu(self):
        (starts,stops) = getSegmentTimes("2009 1 1 0 0 0", "2009 1 1 2 0 0", 120.0 , 2.0)

        self.assertEqual( len(starts), len(stops) )
        self.assertEqual( len(starts), 1 )

    def testInvalidDumpCadence(self):
        # 33 seconds does not divide evenly into 1 hour segments, so this should raise an exception:
        self.assertRaises(Exception, getSegmentTimes, "2012 5 3 11 0 0", "2012 5 3 12 0 0", 33.0, 1.0)

    def testArbitraryDumpCadence(self):
        (starts,stops) = getSegmentTimes("2012 5 3 11 0 0", "2012 5 3 12 0 0", 0.0, 1.0)
        self.assertEqual( len(starts), len(stops) )
        self.assertEqual( len(starts), 1 )

if __name__ == '__main__':
    import doctest
    doctest.testmod()
    unittest.main()
