"""
Functions that help get user input to model parameters.
"""
# Standard import
import os
import sys

# Custom imports
import MakeItSo.Models
from MakeItSo import share

def stdin(include_optional = False, include_expert = False):
    """
    Select a model & input its parameters.

    Models & their dependencies are defined in 'models.yml'

    Parameters:
      include_optional:  Query user to input optional variables?
      include_expert:   Query user to input expert_only variables?

    Returns:
      coupled_model_name - name of the coupled model to configure.
      models - list of model parameters
      shared_model_index - integer index into 'models' list of shared model parameters
    """

    models_dict = MakeItSo.Models.getModelCoupling()

    # Choose a (coupled) model
    options = [m.keys()[0] for m in models_dict]
    selected_coupled_model = share.stdin_select('models', options)

    # Select the chosen (coupled) model
    coupled_model_name = models_dict[selected_coupled_model].keys()[0]

    print ""
    print "Input shared model variables for " + coupled_model_name + ": "

    # Instantiate classes based on Model names
    # Could probably use getattr, but eval(...) seems to do the trick.

    models = []
    shared_model = None
    for model_name in models_dict[selected_coupled_model][coupled_model_name]:
        if model_name == "Shared":
            shared_model = eval("MakeItSo.Models.Shared")()
        else:
            models.append( eval("MakeItSo.Models."+model_name)() )

    # Input all model parameters
    stdin_input_parameters(coupled_model_name,
                           shared_model,
                           bool(include_optional),
                           bool(include_expert))
    for index, model in enumerate(models):
        stdin_input_parameters(coupled_model_name,
                               model,
                               bool(include_optional),
                               bool(include_expert))
        
    return (coupled_model_name, models, shared_model)

def stdin_input_parameters(coupled_model_name, model_object, include_optional = False, include_expert = False):
    """
    Query the user to enter all required input parameters via stdin
    input.  Note the model_object is an input & output parameter,
    since it reads & writes member data.

    Parameters:
      coupled_model_name: Name of the coupled model (CMIT, LFM-MIX,
        etc.)
      model_object: An object derived from a subclass of Model
        (e.g. LFM, MIX, etc.)
      include_optional: Should we prompt the user for optional
        variables?
      include_expert: Should we prompt the user for expert_only
        variables?

    Returns:
      None:  Has no return, but this function modifies the 'model_object' object.
    """
    is_header_type_printed = False

    print ""
    print "Input " + model_object.__class__.__name__ + " variables. (? for help)"
    print "---------------------------------"
    variable_types_sorted = model_object.parameters.keys()
    variable_types_sorted.reverse()
    for name in variable_types_sorted:
        # variables are stored as a YAML sequence (i.e. array) to preserve ordering
        for idx,variable in enumerate(model_object.parameters[name]):
            # Get a dictionary of the elements of a variable.
            variable_name = variable.keys()[0]
            variable_dict = variable[variable_name]
            # Is there a restriction on the model type?
            if (variable_dict.has_key("models") > 0):
                # If coupled_model_name is not one of the "models", skip the variable.
                if (variable_dict["models"].count(coupled_model_name) <= 0):
                    #print "Skipping variable " + variable_name + " since it is not for the model " + coupled_model_name
                    continue
                
            # user_input has three possibile strings:
            #  1.  required: Always query
            isRequired =  (variable_dict["user_input"] == "required")
            #  2.  optional: Query only when asked to input "optional" or "expert" (aka all parameters).
            isOptional = ((variable_dict["user_input"] == "optional") & (include_optional | include_expert) )
            #  3.  expert_only (i.e. dont do it unless you know what you're doing!)
            isExpert   = ((variable_dict["user_input"] == "expert_only") & include_expert)
            if ( isRequired | isOptional | isExpert ):
                if not is_header_type_printed:
                    print "\tRequired \"" + name + "\" variables (? for help):"
                    print "\t---------------------------------"
                    is_header_type_printed = True

                input = share.stdin_get_variable(variable_name, variable_dict)

                if (input == "DISABLE-THIS-VARIABLE"):
                    
                    if (not variable_dict.has_key('status') ):
                        while (input == 'DISABLE-THIS-VARIABLE'):
                            print "Skipping is not permitted (or not implemented) for this variable."
                            input = share.stdin_get_variable(variable_name, variable_dict)                    
                    elif variable_dict.has_key('status'):
                        if (variable_dict['status'] == 'mandatory'):
                            while (input == 'DISABLE-THIS-VARIABLE'):
                                print "Skipping is not permitted (or not implemented) for this variable."
                                input = share.stdin_get_variable(variable_name, variable_dict)
                        elif variable_dict['status'] != 'disabled':
                            print 'Disabling variable "' + variable_name + '"'
                            variable_dict['status'] = 'disabled'
                    else:
                        raise Exception, 'Unexpected behavior skipping this variable.'
                else:
                    variable_dict["value"] = input
                    
        if is_header_type_printed:
            print ""
            is_header_type_printed = False

if __name__ == '__main__':
    import doctest
    doctest.testmod()
