import datetime
import optparse
import math
import os
import random
import stat
import sys

import MakeItSo
from MakeItSo import Environments
from MakeItSo import Models
from MakeItSo import share

#
#FIXME: Integrate with build system to generate startup hdf files, compile executables, etc.
#

def main():
    """
    Run makeItSo:
    """
    options = parse_arguments()
    (coupled_model_name, shared_model, models, env) = getInputs(options)
    generateScripts(coupled_model_name, shared_model, models, env, options)

def parse_arguments():
    """
    Returns an optparse object of parsed command-line arguments.
    """
    # additional optparse help available at:
    # http://docs.python.org/library/optparse.html
    # http://docs.python.org/lib/optparse-generating-help.html
    parser = optparse.OptionParser(version=MakeItSo.Release.version)
    parser.add_option("-v", "--verbose", dest="verbose", default=False, action="store_true",
                      help="Run in verbose mode")
    parser.add_option("-o", "--optional", dest="optional", default=False, action="store_true",
                      help="Utilize \"optional\" variables?")
    parser.add_option("-e", "--expert", dest="expert", default=False, action="store_true", 
                      help="Utilize \"expert only\" variables?")
    parser.add_option("-c", "--configFile", dest="configFile",
                      default=None, action="store", metavar="FILE",
                      help="Path to a makeItSo configuration file.  Note: "
                      "these config files are automatically written to "
                      "the run directory whenever you run makeItSo "
                      "(look for a file ending with "
                      "\"_makeItSo.config\").")
    parser.add_option("-a", "--about", dest="about", default=False, action="store_true",
                      help="About this program.")

    #groupNotImplemented = optparse.OptionGroup(parser, "Dangerous Options",
    #                                           "The following options "
    #                                           "are not implemented or "
    #                                           "partially implemented. "
    #                                           "Use with care!")
    #groupNotImplemented.add_option("-i", "--interactive", dest="interactive",
    #                               default=True, action="store_false",
    #                               help="Run in question-and-answer form (i.e. run interactively).")
    #parser.add_option_group(groupNotImplemented)
    
    groupNotes = optparse.OptionGroup(parser, "Notes",
                                 "See README for full details on running makeItSo!")
    parser.add_option_group(groupNotes)
    
    (options, args) = parser.parse_args()

    if options.verbose:
        print "Enter \"?\" for a description of the current variable."
        print ""
        print('verbose:      ') + str(options.verbose)
        print('interactive:  ') + str(options.interactive)
        print('optional:     ') + str(options.optional)
        print('expert:       ') + str(options.expert)
        print ""

    if options.about:
        print sys.argv[0] + " version " + MakeItSo.Release.version
        print ""
        print share.word_wrap("Generates model input & job scripts specific to a given compute environment.")
        print ""
        print share.word_wrap("Want to extend this to your own environment or model(s)?  See Environments/README and Models/README for details.")
        print ""
        print share.word_wrap("Please share any questions, comments or contributions with the original author:  Peter Schmitt (schmitt <at> ucar <dot> edu)")
        print ""
        about_strings = [ "Make it so!",
                          "If the warp drive fails to activate, the results could be... unfortunate.  Very unfortunate. We will be dead.",
                          "Space... The final frontier. These are the voyages of the starship Enterprise.  " +
                          "It's continuing mission, to explore strange new worlds. To seek out new life and "+
                          "new civilizations. To boldly go where no one has gone before.",
                          "In your position it's important to ask yourself one question: what would Picard do?",
                          "Engage!",
                          "Tea, Earl Grey, hot." ]
        print share.word_wrap(random.choice(about_strings))
        sys.exit()

    return options

def getInputs(options):
    """
    Get user input (either stdin or from a config file) necessary to
    generate scripts.
    
    Parameters:
      options:  makeItSo optparse options
    Returns:
      coupled_model_name:   Name of coupled model (e.g. CMIT, LFM-MIX, etc.)
      shared_model:  Shared model object
      models: List of Model objects (e.g. LFM, MIX, TIEGCM, etc.)
      env: Environment object (e.g. SH, SGE, etc.)
    """
    # Get user input
    if options.configFile == None:
        # stdin
        (coupled_model_name, models, shared_model) = Models.inputs.stdin(options.optional, options.expert)
        env = Environments.inputs.stdin(options.optional, options.expert, shared_model.is_Coupled_Via_InterComm(coupled_model_name))
    else:
        # config file
        filename = options.configFile
        if not os.path.isfile(filename):
            raise Exception, ("Error: Config file \"" + filename + "\" does not exist.")        
        (coupled_model_name, shared_model, models, env) = MakeItSo.Configuration().readConfigurationFile(filename)

    return (coupled_model_name, shared_model, models, env)

def generateScripts(coupled_model_name, shared_model, models, env, options):
    """
    Generate the set of scripts & input files given the input parameters.
    
    Parameters:
      coupled_model_name:   Name of coupled model (e.g. CMIT, LFM-MIX, etc.)
      shared_model:  Shared model object
      models: List of Model objects (e.g. LFM, MIX, TIEGCM, etc.)
      env: Environment object (e.g. SH, SGE, etc.)
      options:  makeItSo optparse options
    """

    print ' ----------------'
    print '/  Make It So!   \\'
    print '--------------------------------------------------------------------------------'
    
    # Setup Script Generator
    output_directory = shared_model.get_variable_dict("SHARED","Output directory")["value"]
    share.setupOutputDirectory(output_directory)

    # Set any shared model parameters
    for model in models:
        model.set_shared_parameters(coupled_model_name, shared_model, env)

    # Generate "*-setup.sh" script, print instructions to STDOUT & write to a README file.
    instructions = writeSetupFile(coupled_model_name, shared_model, models, env)
    fh = open(os.path.join(output_directory, 'README'),'w')
    fh.write(instructions + '\n')
    fh.close()       

    spinup_iterations = writeSpinupScripts(coupled_model_name, shared_model, models, env )

    # Reset any shared model parameters that may've been corrupted during writeSpinupScripts
    for model in models:
        model.set_shared_parameters(coupled_model_name, shared_model, env)
    
    writeSimulationScripts(coupled_model_name, shared_model, models, env, spinup_iterations)
        
    # Write the config file so it's easy to re-run generate scripts.
    run_identifier = shared_model.get_variable_dict("SHARED","RUN_IDENTIFIER")["value"]    
    filename = os.path.join(output_directory, run_identifier + "_makeItSo.config")

    MakeItSo.Configuration().writeConfigurationFile(filename, coupled_model_name, shared_model, models, env,
                                  options.optional, options.expert)

    print instructions
    print '--------------------------------------------------------------------------------'
    

def writeSetupFile(coupled_model_name, shared_model, models, env):
    """
    Writes a `RUN_IDENTIFIER-setup.sh` script that preconfigures
    models.  Returns run instructions that should be piped to stdout.    
    """
    shScript = ['#/bin/sh','']

    run_identifier = shared_model.get_variable_dict('SHARED','RUN_IDENTIFIER')['value']    
    output_directory = shared_model.get_variable_dict('SHARED','Output directory')['value']
    filename = os.path.join(output_directory, run_identifier + '-setup.sh')

    instructions = []

    # Now configure run directory...
    #FIXME: This hard-codes the script generator to be LFM-specific!
    #       This should maybe go somewhere in the 'Models.yml' file.
    configInstructions  = 'Job scripts & input files created.  You must configure your run directory.  '
    configInstructions += 'First make sure that the $LTRROOT environment variable is set '
    configInstructions += '(see documentation* for more information).  Then '
    if os.path.dirname(filename) is not '':
        configInstructions += '`cd %s` and e' % (os.path.dirname(filename))
    else:
        configInstructions += 'E'
    configInstructions += 'xecute `./%s` to configure the model(s).' % (os.path.basename(filename))
    instructions.append(share.word_wrap(configInstructions))
    instructions.append('')
    instructions.append('Additional instructions & notes:')

    # get datetime object from string "YYYY MM DD HH MM SS":
    startDateTime = datetime.datetime(*[int(s) for s in shared_model.get_variable_dict('SHARED','START_YMDHMS')['value'].split()])        

    # Get extra instructions for each model
    for model in models:
        (spinup_start_times, spinup_stop_times, dump_interval) = model.get_spinup_times(startDateTime, shared_model.get_variable_dict('SHARED','SIMULATION_HOURS_PER_SEGMENT')['value'])

        # Some models (ie. xjd) may not be time dependent, but Model::get_model_setup expects a start time.
        if len(spinup_start_times) == 0:
            spinup_start_times = [datetime.datetime.now()]
     
        (mScript, mInstructions) = model.get_model_setup(coupled_model_name,
                                                         shared_model.get_variable_dict('SHARED', 'SPINUP')['value'],
                                                         run_identifier,
                                                         spinup_start_times[0]
                                                         )
        shScript.append(mScript)
        if (len(mInstructions) > 0):
            instructions.append(mInstructions)

    shScript.append('echo ""')                    
    shScript.append('echo "Executables copied & models prepared for run."')
    shScript.append('echo "Make sure to follow all the instructions in the README."')
    shScript.append('echo "As a reminder, here they are:"')
    shScript.append('echo ""')
    shScript.append('cat README')
    
    sh_file = open(filename, "w")
    sh_file.write( '\n'.join(shScript) )
    sh_file.close()

    # Setup script should have executable permissions.  Set 'em here:
    mode = os.stat(filename).st_mode
    os.chmod(filename, mode | stat.S_IXUSR)

    # How do you run the code?
    instructions.append('')
    instructions.append(share.word_wrap('To run the code, submit the first iteration in a sequence'))
    instructions.append('\t%s %s-001.%s' % (env.metadata['submit_via'],
                                            run_identifier,
                                            env.metadata['output_suffix']) )
    instructions.append("which should execute the entire run you've requested.")

    # Where do you learn more?
    instructions.append('')
    instructions.append('* See the LTR documentation for more information')
    instructions.append('https://wiki.ucar.edu/display/LTR')

    return '\n'.join(instructions)
    
    

def writeSpinupScripts(coupled_model_name, shared_model, models, env):
    """
    Create segments for 'spinup' times, if necessary.  These scripts
    initialize the model & its components.

    Notes:
      - Parameters are the same as generateScripts(...)
      - The shared_model parameter 'spinup' determines if any spinup
        scripts are necessary.
    """

    # Don't create spinup scripts if they're not necessary!
    if not shared_model.get_variable_dict('SHARED', 'SPINUP')['value']:
        return 0
    
    run_identifier = shared_model.get_variable_dict("SHARED","RUN_IDENTIFIER")["value"]
    output_directory = shared_model.get_variable_dict("SHARED","Output directory")["value"]

    # get datetime object from string "YYYY MM DD HH MM SS":
    startDateTime = datetime.datetime(*[int(s) for s in shared_model.get_variable_dict('SHARED','START_YMDHMS')['value'].split()])        
                                   
    # Before creating "spinup" times, record what some of the standard
    # shared model parameters are.  After doing the spin-ups, we will
    # revert parameters back
    simulation_dt_seconds = shared_model.get_variable_dict("SHARED","TIME_DUMP_INTERVAL")["value"]

    spinup_iterations = 0
    for model in models:
        (spinup_start_times, spinup_stop_times, dump_interval) = model.get_spinup_times(startDateTime, shared_model.get_variable_dict('SHARED','SIMULATION_HOURS_PER_SEGMENT')['value'])
        # Ensure data type of 'dump_interval': This should be a
        # floating-point number of seconds.  While testing, I found
        # peculiar instances where 60*60*60 seconds would be
        # interpereted as a datetime.timedelta object!
        for idx,value in enumerate(dump_interval):
            if value.__class__ == datetime.timedelta:
                dump_interval[idx] = value.seconds

        for i, startTime in enumerate(spinup_start_times):
            model_setup = []
            model_exec_commands = []
            model_teardown = []

            # Update dump interval with spinup interval
            shared_model.get_variable_dict("SHARED","TIME_DUMP_INTERVAL")["value"] = dump_interval[i]

            spinup_coupled_model_name = model.get_spinup_coupled_model_name(coupled_model_name)
            model.set_shared_parameters(spinup_coupled_model_name, shared_model)

            (setup, cmds, teardown) = configure_model(spinup_coupled_model_name, run_identifier, output_directory, model, env,
                                                      startTime, spinup_stop_times[i], 1+spinup_iterations)
            model_setup.append( setup )
            model_exec_commands.append( cmds )
            model_teardown.append( teardown )

            # Setup any coupled models required in spinup
            for model_name in model.get_spinup_coupled_models():
                for model_spinup in models:
                    if (model_name == model_spinup.metadata["model_name"]):
                        spinup_coupled_model_name = model.get_spinup_coupled_model_name(coupled_model_name)

                        model_spinup.set_shared_parameters(spinup_coupled_model_name, shared_model)

                        (setup, cmds, teardown) = configure_model(spinup_coupled_model_name, run_identifier, output_directory, model_spinup, env,
                                                                  startTime, spinup_stop_times[i], 1+spinup_iterations)
                        model_setup.append( setup )
                        model_exec_commands.append( cmds )
                        model_teardown.append( teardown )

            #FIXME: TIE-GCM should only use 8 procs
            write_run_script(run_identifier, output_directory, env, shared_model.is_Coupled_Via_InterComm(spinup_coupled_model_name),
                             "".join(model_setup), "".join(model_exec_commands), "".join(model_teardown), 1+spinup_iterations)

            spinup_iterations += 1

    # Since we're done dealing with model spinups, restore standard
    # model parameters back to simulation settings chosen by the user.
    shared_model.get_variable_dict("SHARED","TIME_DUMP_INTERVAL")["value"] = simulation_dt_seconds
    for model in models:
        model.set_shared_parameters(spinup_coupled_model_name, shared_model, env)

    return spinup_iterations

def writeSimulationScripts(coupled_model_name, shared_model, models, env, spinup_iterations):
    """
    Write the job scripts & input files necessary for the simulation.

    Notes:
      - Parameters are the same as generateScripts(...)
      - If any model spinup is required, those scripts should've been
        written in writeSpinupScripts(...).
    """
    run_identifier = shared_model.get_variable_dict("SHARED","RUN_IDENTIFIER")["value"]    
    output_directory = shared_model.get_variable_dict("SHARED","Output directory")["value"]

    # Get start/stop times for each segment
    (segment_start_times, segment_stop_times) = Models.getSegmentTimes(shared_model.get_variable_dict('SHARED','START_YMDHMS')['value'],
                                                                       shared_model.get_variable_dict('SHARED','STOP_YMDHMS')['value'],
                                                                       shared_model.get_variable_dict('SHARED','TIME_DUMP_INTERVAL')['value'],
                                                                       shared_model.get_variable_dict('SHARED','SIMULATION_HOURS_PER_SEGMENT')['value'])
    #
    # Iterate over time loop, creating segments for the requested time interval.
    #
    for i, startTime in enumerate(segment_start_times):
        model_setup = []
        model_exec_commands = []
        model_teardown = []
        for model in models:
            (setup, cmds, teardown) = configure_model(coupled_model_name, run_identifier, output_directory, model, env,
                                                      startTime, segment_stop_times[i], i+1+spinup_iterations)
            model_setup.append( setup )
            model_exec_commands.append( cmds )
            model_teardown.append( teardown )

        write_run_script(run_identifier, output_directory, env, shared_model.is_Coupled_Via_InterComm(coupled_model_name),
                         "".join(model_setup), "".join(model_exec_commands), "".join(model_teardown), i+1+spinup_iterations)
    

def configure_model(coupled_model_name, run_name, output_directory, model, environment, start, stop,  iteration):
    """
    Write the model input file and return the set of commands necessary to run a model.
    
    Parameters: 
      coupled_model_name:   Name of coupled model (e.g. CMIT, LFM-MIX, etc.)
      run_name:    Name of the run chosen by the user.
      model:       Model object (e.g. LFM, MIX, TIEGCM, etc.)
      environment: Environment object (e.g. SH, SGE, etc.)
      start:       start datetime
      stop:        stop datetime 
      iteration:   integer iteration number

    Returns:
      model_setup_str: Commands used to prepare the model for execution.
      model_exec_commands_str: Commands used to actually run a model.
      model_teardown_str: Commands used to remove temporary/intermediary files used in a model run.
    """

    # Set model time
    model.set_time_information(start, stop)

    # Get commands to setup the model to run.
    model_setup = model.get_iteration_setup(coupled_model_name, iteration)
    
    # Write input file
    input_file_str = model.get_input_file()
    if ( (input_file_str != "") & (input_file_str != None) ):

        filename = "%s-%03d.%s" % (model.metadata["output_prefix"], iteration, model.metadata["output_suffix"])

        input_file = open(os.path.join(output_directory,filename), "w")
        input_file.write(input_file_str + "\n")
        input_file.close()

    # Gather Model Instructions, such as:
    #    mpirun -np 8 ./LFM-MIX lfm-mix.xjd < /dev/null &> LFM-MIX-001.out
    #    ./ionosphere ionosphere-001.param &> IONOSPHERE-001.out
    #    ./tiegcm < tiegcm1.9-001.inp &> TIEGCM-001.out            
    (isParallel, executable, flags, stdin, stdout) = model.get_run_instructions(coupled_model_name, iteration)
    model_exec_commands = environment.get_run_command(isParallel, executable, flags, stdin, stdout)
    if (model_exec_commands != ""):
        model_exec_commands += " &\n"

    model_teardown = model.get_iteration_teardown(run_name, iteration)
    
    return (model_setup, model_exec_commands, model_teardown)



def write_run_script(run_name, output_directory, environment, uses_intercomm, setup, commands, teardown, iteration):
    """
    Write coupled model run script to file.
    
    Parameters:
      run_name:
      environment:
      uses_intercomm:
      setup:
      commands:
      teardown:
      iteration:

    Returns: None
    """
    
    # Write the job script
    run_script = []
    run_script.append( environment.get_script_header(run_name, iteration, uses_intercomm) )

    run_script.append( "# Model SetUp" )
    run_script.append( setup )
    run_script.append( "" )
    
    run_script.append( "# Run the Model(s)" )
    # Remove "&\n" for last run command:
    if commands.endswith("&\n"):
        run_script.append( commands[0:len(commands)-3] )
    else:
        run_script.append( commands )
    run_script.append( "" )
    
    run_script.append( "# Model TearDown" )
    run_script.append( teardown )
    run_script.append( "" )
    
    run_script.append( environment.get_sh_script_footer(run_name, iteration, uses_intercomm) )
    
    filename = "%s-%03d.%s" % (run_name, iteration, environment.metadata["output_suffix"])
    script_file = open(os.path.join(output_directory,filename), "w")
    script_file.write( "\n".join(run_script) )
    script_file.close()

