import Release

import copy
import os
import re
import sys

def typeCast(s, dType):
    """
    Type cast string "s" to type "t".  Tries to handle boolean values

    parameters:
     s: input string to type cast
     dType: One of < 'boolean' | 'int' | 'integer' | 'double' | 'float' | 'string' >

    returns:
      type-converted string.

    >>> type(typeCast('t', 'boolean'))
    <type 'bool'>

    >>> type(typeCast('3.1415', 'integer'))
    <type 'int'>

    >>> type(typeCast('3.1415', 'double'))
    <type 'double'>    

    >>> type(typeCast('3.1415', 'float'))
    <type 'float'>
    """
    
    if dType == 'boolean':
        return str2bool(s)
    elif (dType == 'integer') | (dType == 'int'):
        return int(float(s))
    elif dType == 'double':
        return  double(s)
    elif dType == 'float':
        return  float(s)
    elif dType == 'string':
        return s
    else:
        if Release.verboseDebug:
            print('WARNING:  Did not understand data type "' +
                  dType + '" when converting string "' + s + '".')
        return s
                

def str2bool(s):
    """
    Converts input string to a boolean value

    >>> str2bool("true")
    True

    >>> str2bool("T")
    True

    >>> str2bool("1")
    True

    >>> str2bool("fAlSe")
    False

    >>> str2bool("f")
    False

    >>> str2bool("0")
    False
    """
    s = s.upper()
    if ( (s == 'TRUE') | (s == 'T') | (s == '1') ):
        return True
    elif ( (s == 'FALSE') | (s == 'F') | (s == '0') ):
        return False
    else:
        raise Exception, "Did not understand input string. Cannot convert to boolean value"

def unittestAssertBigStrings(testObj, s1, s2):
    """
    Perform a line-by-line assertEqual on two LARGE strings for a unit test.
    
    Inputs:
      testObj: an object of a class derived from unittest.TestCase
      s1: Big string with line breaks to compare against s2.
      s2: Big string with line breaks to compare against s1.
    """
    ss1 = s1.split("\n")
    ss2 = s2.split("\n")
    for i, s in enumerate(ss1):
        testObj.assertEqual(s, ss2[i])

def setupOutputDirectory(dirpath):
    """
    Check if the directory exists, create it if it doesn't.  Print
    warnings if it's not empty.

    Parameters:
      dirpath: directory that we'll generate input files & scripts in.
    """
    if not os.path.exists(dirpath):
        print "Creating directory [" + dirpath + "] for job scripts & model inputs"
        os.makedirs(dirpath)
    elif not os.path.isdir(dirpath):
        raise Exception, "Error: A file exists at [" + dirpath + "].  Remove the file or select a different directory to generate input files & scripts"

def stdin_select(name, optionList, sorted=True):
    """
    Prompt the user via stdin input to select one item from a list.  

    Parameters: 
     name: Identify what the list represents
     optionList: List of options to select from
     sorted: Should we auto-sort the list alphabetically? 

    Returns: Integer into original selection list

    Example: 
      stdin_select('model', ['CMIT', 'LFM-MIX', 'LFM-RCM', 'LFM', 'TIEGCM'])
         Available models:
         0 CMIT
         1 LFM
         2 LFM-MIX
         3 LFM-RCM
         4 TIEGCM
         Choose a model (0-4) to configure:
    """
    if name[-1:] == 's':
        nameSingular = name[:-1]
        namePlural = name
    else:
        nameSingular = name
        namePlural = name + 's'
    
    if sorted:
        unsortedOptionList = copy.deepcopy(optionList)
        optionList.sort()

    print 'Available ' + namePlural + ':'
    for index, name in enumerate(optionList):
        print "\t" + str(index) + " " + name

    selection = -1
    while ( (selection < 0) | (selection >= len(optionList)) ):
        try:
            input_string = raw_input("Choose a " + nameSingular + " (0-" + str(len(optionList)-1) + ") to configure: ")            
        except KeyboardInterrupt:
            #KeyboardInterrupt == CTRL+C
            input_string = ''
            print 'KeyboardInterrupt.  Try CTRL+D to exit.'
        except EOFError:
            #EOFError == CTRL+D
            sys.stdin.close()
            #FIXME:  '/dev/tty' may not work in Windows:
            sys.stdin = open('/dev/tty')
            while True:
                try:
                    ins = raw_input('Exit the script generator? ([y]/n)')
                except EOFError:
                    print ''
                    sys.exit()
                ins = ins.upper()
                if (ins == 'Y') | (ins == 'YES') | (ins == ''):
                    sys.exit()
                elif (ins == 'N') | (ins == 'NO'):
                    input_string = ''
                    break

        if input_string:
            selection = int(input_string)

    if sorted:
        selection = unsortedOptionList.index( optionList[selection] )

    return selection
    

def stdin_get_variable(variable_name, variable_dict):
    """ Query the user (via stdin) to enter data for the given variable.  Return the value entered by the user.

    Parameters:
      variable_name:
      variable_dict: Dictionary that should have the following keys:
         * description: <text_description>      
         * status: < enabled | disabled | mandatory >
         * type: < boolean | integer | float | string >
         * user_input: < required | optional | expert_only | ignore | shared_model_parameter >
         * value: <variable value>
         * format_regex: <Python 're' module regular expression that value must match.>
         * format_human: <human-readable description of what the variable should contain>

    Returns: Value entered by user 
    """

    # Check if user wants to override a disabled variable_dict
    if variable_dict.has_key('status'):
        if variable_dict['status'] == 'disabled':
            enable_name = 'Enable "' + variable_name + '"?'
            enable_dict = { 'description': 'Do you want to enable this variable (yes or no)?  Full description: "' + variable_dict['description'] + '"',
                            'user_input': 'required',
                            'units': 'N/A',
                            'value': 'n',
                            "format_regex": "Y|y|YES|Yes|yEs|yeS|YEs|YeS|yES|N|n|No|nO",
                            "format_human": "Must be ether \"yes\" or \"no\"."}
            conf = stdin_get_variable(enable_name, enable_dict)
            conf = conf.upper()
            if (conf == 'Y') | (conf == 'YES'):
                variable_dict['status'] = 'enabled'
            elif (conf == 'N') | (conf == 'NO'):
                return 'DISABLE-THIS-VARIABLE'

    while True:
        try:
            input_string = raw_input("\t"+ variable_name + " ( " + str(variable_dict["value"]) + " ): ")
        except KeyboardInterrupt:
            #EOFError == CTRL+C
            print ''
            input_string = '?'
            if variable_name.count('\tIgnore variable ') <= 0:
                confirm_name = "\tIgnore variable \"" + variable_name + "\"?"
                confirm_dict = { "description": "Do you want to ignore this variable (yes or no)?",
                                 "user_input": "required",
                                 "units": "N/A",
                                 "value": "y",
                                 "format_regex": "Y|y|YES|Yes|yEs|yeS|YEs|YeS|yES|N|n|No|nO",
                                 "format_human": "Must be ether \"yes\" or \"no\"."
                                 }
                conf = stdin_get_variable(confirm_name, confirm_dict)
                conf = conf.upper()
                if (conf == 'Y') | (conf == 'YES'):
                    return "DISABLE-THIS-VARIABLE"
        except EOFError:
            #EOFError == CTRL+D
            print ''
            input_string = '?'
            if variable_name.count('\Exit script generator?') <= 0:
                confirm_name = "\tExit script generator?"
                confirm_dict = { "description": "Do you really want to exit?",
                                 "user_input": "required",
                                 "units": "N/A",
                                 "value": "y",
                                 "format_regex": "Y|y|YES|Yes|yEs|yeS|YEs|YeS|yES|N|n|No|nO",
                                 "format_human": "Must be ether \"yes\" or \"no\"."
                                 }
                sys.stdin.close()
                #FIXME:  'dev/tty/' may not work in Windows:
                sys.stdin = open('/dev/tty')
                conf = stdin_get_variable(confirm_name, confirm_dict)
                conf = conf.upper()
                if (conf == 'Y') | (conf == 'YES'):
                    sys.exit()
            
        
        if (input_string == "?"):
            # The user needs help, print additional information.
            print "\t\tDescription:"
            print word_wrap(variable_dict["description"], 72,0,0,'\t\t\t')
            
        elif (input_string != ""):
            # The user entered something.
            # Did the user enter something reasonable?
            if variable_dict.has_key("format_regex"):
                if not matches_regex(variable_dict["format_regex"], input_string):
                    print "\t\tInvalid format."
                    print word_wrap("Expected: " + variable_dict["format_human"], 72,0,0,'\t\t\t')
                    print "\t\t\tData Type: " + variable_dict["type"]                    
                    continue

            # type cast variable appropriately:
            if  variable_dict.has_key('type'):
                input_string = typeCast(input_string, variable_dict['type'])
                
            return input_string
        else:
            # The user entered nothing, assume they want the default value.
            return variable_dict["value"]

def matches_regex(regex, inString):
    """
    Does inString match regex?

    'Some people, when confronted with a problem, think `I know! I'll
    use regular expressions.` Now they have two problems.'
       -- Jamie Zawinski
    """

    # re.match returns an "re" object if it finds something, None otherwise.
    if re.match(regex, inString):
        return True
    else:
        return False

def word_wrap(inString, width=80, ind1=0, ind2=0, prefix=''):
    """ word wrapping function.
    inString: the string to wrap
    width: the column number to wrap at
    prefix: prefix each line with this string (goes before any indentation)
    ind1: number of characters to indent the first line
    ind2: number of characters to indent the rest of the lines

    Shamelessly taken from http://www.saltycrane.com/blog/2007/09/python-word-wrap-function/
    Could probably adapt from Python's textwrap module  http://docs.python.org/library/textwrap.html

    >>> word_wrap('test short string.')
    'test short string.'

    >>> word_wrap('test really_long_word_wrapping.', width=10)
    'test\\nreally_long_word_wrapping.\\n'

    >>> word_wrap('Number of processors to use.  Must be in the form nodes=N:ppn=P, where N=how many nodes you want to run on, P=how many CPUs per node.  For more information, see http://discovery.dartmouth.edu:8080/display/Discovery/Scheduling+Jobs+to+Run', 72,0,0, '\t\t\t')
    '                    Number of processors to use.  Must be in the form nodes=N:ppn=P, where\\n                    N=how many nodes you want to run on, P=how many\\n                    CPUs per node.  For more information, see\\n                    http://discovery.dartmouth.edu:8080/display/Discovery/Scheduling+Jobs+to+Run\\n'
    """
    string = copy.deepcopy(inString)
    newstring = prefix + ind1*" "
    if len(string) > width:
        while True:
            # find position of nearest whitespace char to the left of "width"
            marker = width-1
            while not string[marker].isspace():
                marker = marker - 1

                if marker == len(prefix+ind2*" "):
                    # String has no whitespace in the first (width-1)
                    # characters.  Search forward for spaces.                    
                    marker = width-1
                    while not string[marker].isspace():
                        marker += 1
                        if marker == len(string):
                            # End of string and we haven't found a space?  Let's add one.
                            string += ' '

            # remove line from original string and add it to the new string
            newline = string[0:marker] + "\n"
            newstring = newstring + newline
            string = prefix + ind2*" " + string[marker+1:]
            
            # break out of loop when finished
            if len(string) <= width:
                if len(string.strip()) == 0:
                    string=''
                break
            
    return newstring + string


if __name__ == '__main__':
    import doctest
    doctest.testmod()
    #unittest.main()
