/*
 * Virtual LFM uses InterComm's high-level XJD API (Section 6 of
 * InterComm Manual) to talk with
 *   - MIX using LFM's CPL_IC_Interface class
 *   - RCM using LFM's IM_IC_Interface class
 */

#include <cassert>
#include <iostream>
#include <iomanip>
#include <math.h>

#define PPLUSPLUS
#include <IC_EndPointSet.h>

#include "DateTime.h"
#include "Check_IC_Status.h"
#include "CPL_IC_Interface.h"
#include "IM_IC_Interface.h"

void parse_xjd(const char *, char *, char *, int &);

void readGrid(int &ni, int &nj, int &nk, doubleArray &x, doubleArray &y, doubleArray &z);

/********************************************************************//**
 * Data to keep track of model coupling times
 ************************************************************************/ 								       
struct modelCoupling {
  DateTime start;  /// starting date/time to couple
  DateTime next;   /// next date/time to exchange
  int delta_t;     /// frequency (in seconds) to exchange
  int exchangeNum; /// How many times have we exchanged?
  bool isFirstExchange;
  bool isLastExchange;
};
int calcExchangeNum(const DateTime &start, const DateTime &current, const int &delta_t);
bool isLastExchange(const DateTime &stop, const int &delta_t, const DateTime &current);

/********************************************************************//**
 * Virtual LFM:  Simulate a time loop to communicate with MIX & RCM
 ************************************************************************/ 								       
int main(int argc, char **argv)
{
  /****************** Initialize... ******************/
  int localTasks = 1;
  const char *xjdName = "LFM-MIX-RCM.xjd";
  char localName[256];
  strcpy(localName, "lfm");

  Optimization_Manager::Initialize_Virtual_Machine(localName,localTasks,argc,argv);
  int mype = Communication_Manager::My_Process_Number;

  /* Let's figure out the component name and the number of processors
     from the xjd-file */
  parse_xjd(xjdName, "mhd", localName, localTasks);
  cout << "Local component name: \"" << localName << "\"" << endl
       << "Local number of processors: " << localTasks <<  endl
       << "XJD File: " << xjdName << endl;

  // Read grid information from file.
  int ni, nj, nk;
  doubleArray x, y, z;
  readGrid(ni,nj,nk, x,y,z);
  int nip1 = ni+1; 
  int njp1 = nj+1;
  int nkp1 = nk+1;

  //FIXME:  Read these from file as done in jpara_wrapper.
  doubleArray bx(ni,nj,nk), by(ni,nj,nk), bz(ni,nj,nk);
  doubleArray rho(2,nj,nk), csound(2,nj,nk);
  doubleArray ej_pot, ek_pot, velocity;

  int num_procs = Communication_Manager::numberOfProcessors();
  Partitioning_Type all(num_procs);
  bx.setBase(1);  
  bx.partition(all);

  by.setBase(1);  
  by.partition(all);

  bz.setBase(1);
  bz.partition(all);

  rho.setBase(1);
  rho.partition(all);

  csound.setBase(1);
  csound.partition(all);

  float dipole_moment = 123.456;

  // Time vars
  DateTime currentUt(2009,1,1,0,0,0);
  DateTime stopUt(2009,1,1,0,4,0);
  int loopIterationNum = 0;

  struct modelCoupling coupler = {
    currentUt, // start date & time
    currentUt, // next exchange date & time
    30,      // delta_t in seconds
    calcExchangeNum(currentUt, currentUt, 2*60),
    true,       // isFirstExchange
    false       // isLastExchange
  };

  struct modelCoupling innerMagnetosphere = {
    currentUt, // start date & time
    currentUt, // next exchange date & time
    2*60,      // delta_t in seconds
    calcExchangeNum(currentUt, currentUt, 2*60),
    true,      // isFirstExchange
    false      // isLastExchange
  };

  int icErrorCode;
  Partitioning_Type p_zero(Range(0,0));
  doubleArray cplScalars(CPL_Interface::NUMBER_OF_SCALARS, p_zero);
  cplScalars(CPL_Interface::KILL_SIGNAL) = CPL_Interface::KILL_SIGNAL_CONTINUE;
  cplScalars(CPL_Interface::TILT) = 0;
  cplScalars(CPL_Interface::EXCHANGE_WITH_IM) = CPL_Interface::NO_IM_EXCHANGE;

  intArray imScalars(IM_Interface::NUMBER_OF_SCALARS, p_zero);
  imScalars(IM_Interface::KILL_SIGNAL) = IM_Interface::KILL_SIGNAL_CONTINUE;
  imScalars(IM_Interface::DELTA_T) = innerMagnetosphere.delta_t; // number of seconds between LFM-RCM exchange
  
  if (mype == 0)  std::cout << "Hello from the  Virtual LFM!" << std::endl;

  IC_EndPointSet *epset = new IC_EndPointSet("LFM-MIX-RCM.xjd", "lfm", icErrorCode);
  epset->printErrorMessage("     IC_EndPointSet created. Status",icErrorCode);
  CHECK_IC_EPS_STATUS("IC_EndPointSet('LFM-MIX-RCM.xjd', 'lfm", icErrorCode);

  CPL_IC_Interface *ion = new CPL_IC_Interface(epset, x, y, z, 2, nj, nk);
  IM_IC_Interface *rcm = new IM_IC_Interface(epset, 
					     x.convertTo_floatArray(), y.convertTo_floatArray(), z.convertTo_floatArray(),
					     x.getPartition(),
					     dipole_moment, 12345.6789, //FIXME:  GET_DIP_MOMENT(...)
					     innerMagnetosphere.start, innerMagnetosphere.delta_t);

  if (mype == 0)  std::cout << "Committing arrays. . . " << std::endl;
  
  // We commit the InterComm arrays here because this function commits
  // *all* regions created previously. Both CPL_IC_Interface and
  // IM_IC_Interface define arrays that need to be committed.
  epset->commitArrays(icErrorCode);
  epset->printErrorMessage("     Arrays commited. Status",icErrorCode);
  CHECK_IC_EPS_STATUS("IC_EndPointSet::commitArrays()", icErrorCode);

  if (mype == 0)  std::cout << "Comitted.  Handshaking." << std::endl;
  rcm->Initialize();

  if (mype == 0) std::cout << "Shook hands." << std::endl;

  /****************** LFM-MIX-RCM Inner Loop ******************/

  double dt = 10.0;

  Communication_Manager::Sync();

  while (currentUt <= stopUt){ 
    if (mype == 0) std::cout << "--------------------------------------------"
			     << "loop iteration " << ++loopIterationNum << std::endl;
    // Do we exchange with the Inner Magnetosphere (RCM)?
    if (currentUt >= innerMagnetosphere.next){
      if (innerMagnetosphere.isFirstExchange){
	// Tell the coupler (MIX) that it needs to export only to RCM
	cplScalars(CPL_Interface::EXCHANGE_WITH_IM) = CPL_Interface::CPL_SEND_TO_IM;
      }
      else if ( isLastExchange(stopUt, innerMagnetosphere.delta_t, innerMagnetosphere.next) ){
	innerMagnetosphere.isLastExchange = true;
	// Update kill signal.
	imScalars(IM_Interface::KILL_SIGNAL) = IM_Interface::KILL_SIGNAL_LAST_EXCHANGE;
	// Last exchange.  Tell the Coupler (MIX) that it needs import only from RCM
	cplScalars(CPL_Interface::EXCHANGE_WITH_IM) = CPL_Interface::CPL_RECV_FROM_IM;	
      }
      else{
	cplScalars(CPL_Interface::EXCHANGE_WITH_IM) = CPL_Interface::DUPLEX_IM_EXCHANGE;
      }

      imScalars(IM_Interface::YEAR)   = innerMagnetosphere.next.getYear();
      imScalars(IM_Interface::MONTH)  = innerMagnetosphere.next.getMonth();
      imScalars(IM_Interface::DAY)    = innerMagnetosphere.next.getDay();
      imScalars(IM_Interface::HOUR)   = innerMagnetosphere.next.getHour();
      imScalars(IM_Interface::MINUTE) = innerMagnetosphere.next.getMinute();
      imScalars(IM_Interface::SECOND) = int( innerMagnetosphere.next.getSecond() );
      imScalars(IM_Interface::LABEL)  = loopIterationNum;

      if (mype == 0)  std::cout << "IM exchange # " << innerMagnetosphere.exchangeNum << "\t"
				<< currentUt.getDateTimeString() << std::endl;

      rcm->SendScalars("mhd-im_scalars", imScalars);
      if (!innerMagnetosphere.isFirstExchange){
	rcm->Import();
      }
    }

    // Do we exchange with the Coupler (MIX)?
    if (currentUt >= coupler.next){
      if ( isLastExchange(stopUt, coupler.delta_t, coupler.next) ){
	coupler.isLastExchange = true;
	// Update kill signal.
	cplScalars(CPL_Interface::KILL_SIGNAL) = CPL_Interface::KILL_SIGNAL_LAST_EXCHANGE;
      }

      cplScalars(CPL_Interface::YEAR)   = coupler.next.getYear();
      cplScalars(CPL_Interface::MONTH)  = coupler.next.getMonth();
      cplScalars(CPL_Interface::DAY)    = coupler.next.getDay();
      cplScalars(CPL_Interface::HOUR)   = coupler.next.getHour();
      cplScalars(CPL_Interface::MINUTE) = coupler.next.getMinute();
      cplScalars(CPL_Interface::SECOND) = coupler.next.getSecond();
      cplScalars(CPL_Interface::LABEL)  = loopIterationNum;

      if (mype == 0)  std::cout << "CPL exchange # " << coupler.exchangeNum <<  "\t"       
				<< currentUt.getDateTimeString() << std::endl << std::endl;

      ion->sendScalars(cplScalars);
      ion->Export(bx, by, bz, rho, csound);
    }

    // Did we export to the Coupler (MIX)?  
    if (currentUt >= coupler.next){
      // ... Import the updated values.
      ion->Import(ej_pot, ek_pot, velocity);      

      // Update the time for next exchange.
      coupler.next = coupler.start;
      coupler.next.incrementSeconds(++coupler.exchangeNum*coupler.delta_t);
      coupler.isFirstExchange = false;
    }
    
    // Did we import from the inner magnetosphere (RCM)?
    if (currentUt >= innerMagnetosphere.next){
      // Do not export to RCM on the last exchange...
      if (! innerMagnetosphere.isLastExchange){
	rcm->Export();     
      }
      // Update the time for next exchange.
      innerMagnetosphere.next = innerMagnetosphere.start;
      innerMagnetosphere.next.incrementSeconds(++innerMagnetosphere.exchangeNum*innerMagnetosphere.delta_t);
      innerMagnetosphere.isFirstExchange = false;
    }

    if ( coupler.isLastExchange ){
      break;
    }

    // Assume that MIX Coupler and RCM Inner-Magnetosphere do not
    // exchange at the next LFM step.
    cplScalars(CPL_Interface::EXCHANGE_WITH_IM) = CPL_Interface::NO_IM_EXCHANGE;

    currentUt.incrementSeconds(dt);
  }

  /****************** Send kill signal ******************/
  
  cplScalars(CPL_Interface::KILL_SIGNAL) = CPL_Interface::KILL_SIGNAL_SHUTDOWN;
  ion->sendScalars(cplScalars);  

  imScalars(IM_Interface::KILL_SIGNAL) = IM_Interface::KILL_SIGNAL_SHUTDOWN;
  rcm->SendScalars("mhd-im_scalars", imScalars);
  
  /****************** Clean up & exit ******************/
  if (mype == 0)  std::cout << "-------------- Cleanup & Exit. . . . . . ;" << std::endl;

  delete rcm;
  delete ion;
  Communication_Manager::Sync();
  delete epset;
  Communication_Manager::Sync();
  Optimization_Manager::Exit_Virtual_Machine();

  return 0;
}


/********************************************************************//**
 * Calculate how many exchanges should have occured from start
 * date/time to current date/time, assuming constant exchange interval
 * delta_t
 ************************************************************************/ 								       
int calcExchangeNum(const DateTime &start, const DateTime &current, const int &delta_t)
{
  return floor( (current.getMJD() - start.getMJD()) / (double) delta_t/86400.0 );
}


/********************************************************************//**
 * Is this the final exchange that should occur (i.e. is current +
 * delta_t > stop)?
 ************************************************************************/ 								       
bool isLastExchange(const DateTime &stop, const int &delta_t, const DateTime &current)
{
  DateTime nextExchange = current;
  nextExchange.incrementSeconds(delta_t);
  if (nextExchange > stop){
    return true;
  }
  return false;
}


/********************************************************************//**
 * Read the LFM dimensions (ni,nj,nk) and grid (x,y,z) from the
 * GRIDOUT file.
 *
 * Reads the dimensions on each processor & uses P++ magic to
 * distribute the (x,y,z) grid to all tasks.
 ************************************************************************/ 
void readGrid(int &ni, int &nj, int &nk, doubleArray &x, doubleArray &y, doubleArray &z)
{
  int localTasks = Communication_Manager::Number_Of_Processors;
  int mype = Communication_Manager::My_Process_Number;

  // Read dimensions from GRIDOUT file.
  int nip1, njp1, mjin, mjwall, nkp1;
  ifstream gridFile;

  for (int i=0; i < localTasks; i++){
    // Note: We do serial file I/O operations one at a time to avoid slamming the disk.
    //
    // A more efficient way of doing this would be to read data on a
    // single process & then broadcast data out.  
    
    if ( i == mype){
      gridFile.open("GRIDOUT", ios::in);
      if (!gridFile){
	cerr << "Could not open the file " << "GRIDOUT" << endl;
	exit(1);
      }
      gridFile >> nip1 >> njp1 >> mjin >> mjwall >> nkp1;
      gridFile.close();
    }
    Communication_Manager::Sync();
  }

  ni = nip1-1;
  nj = njp1-1;
  nk = nkp1-1;
  
  // Now each procesor has the dimensions.  Prepare to do I/O on the head node.
   
  Partitioning_Type ioProcess(Range(0,0));
  
  doubleArray phi(nkp1);
  phi.partition(ioProcess);
  
  doubleArray x2(nip1,njp1);
  x2.partition(ioProcess);

  doubleArray y2(nip1,njp1);
  y2.partition(ioProcess);
  
  x.partition(ioProcess);
  x.redim(nip1, njp1, nkp1);
  
  y.partition(ioProcess);
  y.redim(nip1,njp1,nkp1);

  z.partition(ioProcess);
  z.redim(nip1,njp1,nkp1);

  Optimization_Manager::setOptimizedScalarIndexing(On);

  double earthRadiusInCm = 6.38e8;
  // I/O on the head node: Read & set the (x,y,z) grid.
  if (mype == 0){
    cout << "Reading grid (" << nip1 << ", " << njp1 << ", " << nkp1 << ")." << endl;
    gridFile.open("GRIDOUT",ios::in);
    if (!gridFile) {
      cerr << "Could not open the file" <<endl;
      exit(1);
    }
    gridFile >> nip1 >> njp1 >> mjin >> mjwall >> nkp1;

    for (int k=0; k < nkp1; k++){
      gridFile >> phi(k);
    }
    
    for (int j=0; j < njp1; j++){
      for (int i = 0; i < nip1; i++){
	gridFile >> x2(i,j);
	x2(i,j) *= earthRadiusInCm; // convert from Re to cm.
      }
    }

    for (int j=0; j < njp1; j++){
      for (int i = 0; i < nip1; i++){
	gridFile >> y2(i,j);
	y2(i,j) *= earthRadiusInCm; // convert from Re to cm.
      }
    }
    
    gridFile.close();
    
    for (int k = 0; k < nkp1; k++) {
      for (int j = 0; j < njp1; j++) {
	for (int i = 0; i < nip1; i++) {
	  x(i,j,k) = x2(i,j);
	  y(i,j,k) = y2(i,j)*cos(PI/180.*phi(k));
	  z(i,j,k) = y2(i,j)*sin(PI/180.*phi(k));
	}
      }
    }
  }
	
  Optimization_Manager::setOptimizedScalarIndexing(Off);
  Communication_Manager::Sync();
  
  // All the data is on the head node.  Distribute to all processes.
  Partitioning_Type all(localTasks);
  x.setBase(1);
  x.partition(all); 
  
  y.setBase(1);
  y.partition(all); 
  
  z.setBase(1);
  z.partition(all);
}
