/*
 * Virtual MIX uses InterComm's low-level XJD API (Section 4 of
 * InterComm Manual) to talk with
 *   - LFM using MIX's MHD_IC_Interface class
 *   - RCM using MIX's IM_IC_Interface class
 */

// MIX includes
#include "IM_IC_Interface.h"
#include "IM_Interface.h"
#include "MHD_IC_Interface.h"
#include "MHD_Interface.h"
#include "Params.h"
#include "Solver.h"
#include "Grid.h"
#include "functions.h"

// Common includes
#include "DateTime.h"
#include "Check_IC_Status.h"

bool receiveMHDscalars(MHD_Interface *mhd,  DateTime &currentDateTime, real &tilt, long &mhdTimestep, bool &isLastStep, int &exchangeWithIm);

int main(int argc, char **argv)
{
  /****************** Initialize... ******************/
  Overture::start(argc,argv);  // initialize Overture

  char* paramFileName = argv[1];
  Params params(paramFileName);
  
  realArray ionScalars;

  int iStatus;
  IC_XJD *xjd = IC_Initialize("mix",0,params.xjdName,&iStatus);
  CHECK_IC_STATUS(xjd, "IC_Initialize", iStatus);

  MHD_Interface *mhd = new MHD_IC_Interface(&params, xjd);
  // Get the second shell MHD ionospheric grid from the MHD Interface
  CompositeGrid cg = mhd->getSecondShellIonCartesianGrid();
  realCompositeGridFunction secondShellMHDFunctions(cg,all,all,all,8);

  // Get the first shell MHD ionospheric grid from the MHD interface
  CompositeGrid cg1 = mhd->getFirstShellIonCartesianGrid();
  realCompositeGridFunction firstShellMHDFunctions(cg1,all,all,all,2);

  IM_Interface *rcm = new IM_IC_Interface(xjd);

  // Create a representation of MIX's solver functions
  Solver ionSolver(params.mixGrid.getNLAT(),
		   params.mixGrid.getLATITUDE(),
		   params.mixGrid.getSOLVER_LON_GRID_STEP());
  CompositeGrid sgrid = ionSolver.generateCartesian2DGrid();
  realCompositeGridFunction solverFunctions(sgrid,all,all,all,18);
  solverFunctions.setName("Solver grid functions");
  solverFunctions.setName("Potential North [V]",SOLVER_POTENTIAL_NORTH);
  solverFunctions.setName("Potential South [V]",SOLVER_POTENTIAL_SOUTH);

  std::cout << "Interface defined.  Comitting regions." << std::endl;
  IC_Commit_region(xjd,&iStatus);
  CHECK_IC_STATUS(xjd, "IC_Commit_region", iStatus);

  std::cout << "Comitted.  Handshaking." << std::endl;
  rcm->Initialize();

  std::cout << "Shook hands." << std::endl;

  /****************** LFM-MIX-RCM Inner Loop ******************/


  DateTime currentUT;
  real tilt;
  long mhdTimeStep;
  bool isLastStep = false;
  int exchangeWithIm = 0;

  int imExchangeNum = 0;
  int mhdExchangeNum = 0;
  while ( receiveMHDscalars(mhd, currentUT, tilt, mhdTimeStep, isLastStep, exchangeWithIm) ){
    std::cout << "MHD Exchange # " << mhdExchangeNum++ << "\t"
	      << currentUT.getDateTimeString() << std::endl;

    std::cout << "1.  RCM might export to LFM... see VirtualLFM.C" << std::endl;

    std::cout << "2-a.  LFM -> MIX Scalars" << std::endl;
    // already happened in receiveMHDscalars(...)

    std::cout << "2-b.  LFM -> MIX" << std::endl;
    mhd->Import(secondShellMHDFunctions);

    // MIX (RCM) does not import (export) on the first exchange, as
    // instructed by the LFM via scalar at index
    // MHD_Interface::EXCHANGE_WITH_IM
    if ( (exchangeWithIm == MHD_Interface::CPL_RECV_FROM_IM) ||
	 (exchangeWithIm == MHD_Interface::DUPLEX_IM_EXCHANGE) ){
      std::cout << "IM Exchange # " << imExchangeNum << "\t"
		<< currentUT.getDateTimeString() << std::endl;
      imExchangeNum++;
      std::cout << "3.  RCM -> MIX" << std::endl;
      rcm->Import();
    }
    
    std::cout << "4-a.  Solve Potential. . . " << std::endl;
    std::cout << "4-b.  MIX -> LFM" << std::endl;
    mhd->Export(firstShellMHDFunctions,secondShellMHDFunctions);

    std::cout << "5.  LFM exports to RCM... see VirtualLFM.C" << std::endl;

    // MIX (RCM) does not export (import) on the last exchange, as
    // instructed by the LFM via scalar at index
    // MHD_Interface::EXCHANGE_WITH_IM
    if ( (exchangeWithIm == MHD_Interface::CPL_SEND_TO_IM) || 
	 (exchangeWithIm == MHD_Interface::DUPLEX_IM_EXCHANGE) ){
      std::cout << "6.  MIX -> RCM" << std::endl;
      rcm->Export(solverFunctions);
    }

    std::cout << std::endl;
  
    if (isLastStep){
      break;
    }
  
  }

  /****************** Clean up & exit ******************/

  IC_Finalize(xjd, &iStatus);
  CHECK_IC_STATUS(xjd, "IC_Finalize", iStatus);

  delete mhd;
#if 0
  delete rcm;
#endif
}


/********************************************************************//**
 * Receive MHD scalars that describe the state of LFM-MIX exchange
 * 
 * \param[in] mhd
 * \param[out] currentDateTime
 * \param[out] tilt
 * \param[out] mhdTimeStep
 * \param[out] isLastStep
 * \param[out] exchangeWithIm - one of the 4 values defined in MHD_Interface::MHD_exchange_with_IM
 *
 * \return elapsed CPU time (in seconds)
 *
 * \note taken from MIX/src/ionosphere_functions at r901.
 ************************************************************************/
bool receiveMHDscalars(MHD_Interface *mhd, DateTime &currentDateTime, real &tilt, long &mhdTimestep, bool &isLastStep, int &exchangeWithIm)
{
#ifdef DEBUG_MODE_ON
  cout << "MIX/src/ionosphere_functions.h::receiveMHDscalars(...)" << endl;
#endif
  realArray scalars = mhd->getScalars();

  isLastStep = false; // must initialize to false!!!

  if ( fabs(scalars(MHD_Interface::KILL_SIGNAL) - MHD_Interface::KILL_SIGNAL_LAST_EXCHANGE) < 1e-5 ){
    cout << "Last LFM-MIX exchange" << endl;
    isLastStep = true;
  }   
  else if ( fabs(scalars(MHD_Interface::KILL_SIGNAL) - MHD_Interface::KILL_SIGNAL_SHUTDOWN) < 1e-5 ){  
    cout << "MIX received kill signal from the MHD code" << endl;
    //  exit the code
    return false;
  }
  
  currentDateTime = DateTime(scalars(MHD_Interface::YEAR),
			     scalars(MHD_Interface::MONTH),
			     scalars(MHD_Interface::DAY), 
			     scalars(MHD_Interface::HOUR), 
			     scalars(MHD_Interface::MINUTE), 
			     (double) scalars(MHD_Interface::SECOND));
  tilt = scalars(MHD_Interface::TILT);
  mhdTimestep = (long)(scalars(MHD_Interface::LABEL));

  // Should be one of the integer values defined in MHD_Interface::MHD_exchange_with_IM
  exchangeWithIm = (int) scalars(MHD_Interface::EXCHANGE_WITH_IM);

  // If we made it this far, MIX should keep running.
  return true;
}
